
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ImageMatrix.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Functions of image algebra.
//
// ==========================================================================


#ifndef BoxySoft_ImageMatrix
#define BoxySoft_ImageMatrix


#include "HIL_ImageFor.h"

using namespace PHIL;

/////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function returns a new matrix which is 
// result of multiplication of the operands a and b.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			a - reference to the first matrix
//			b - reference to the first matrix
//		
// OUTPUT:
//			a * b if possible
//			0 otherwise
//		
// REMARKS:
//			Template parameter T stands for pixel type;
//			Template parameter D stands for intermediate
//				results accumulator;
//
//			The caller is responsible of disposing
//			the returned object!!!
//		
template< typename T, typename D/* = double*/ >	
TImageFor< T > * Orphan_Mult_Matrix( const TImageFor< T > & a, const TImageFor< T > & b )
{
	const int kCols_a = a.GetCol();
	const int kRows_a = a.GetRow();

	const int kCols_b = b.GetCol();

	if( kCols_a != b.GetRow() )
		return 0;					// cannot multiply, dimensions mismatch


	// Create the output matrix-image
	TImageFor< T > * outImage = new TImageFor< T >( kCols_b, kRows_a );

	register T * a_data_ptr = a.GetImageData();
	register T * b_data_ptr = b.GetImageData();
	register T * out_data_ptr = outImage->GetImageData();

	register int i, j, k;

	register T * loc_a_data_ptr = 0;
	register T * loc_b_data_ptr = 0;

	for( i = 0; i < kRows_a; ++ i )
	{
		for( j = 0; j < kCols_b; ++ j )
		{
			D theSum( 0 );

			loc_a_data_ptr = a_data_ptr;
			loc_b_data_ptr = b_data_ptr + j;

			for( k = 0; k < kCols_a; ++ k )
			{
				theSum += * loc_a_data_ptr ++ * * loc_b_data_ptr;
				loc_b_data_ptr += kCols_b;
			}

			* out_data_ptr ++ = (T)theSum;
		}

		a_data_ptr += kCols_a;
	}

	return outImage;
}



///////////////////////////////////////////////////////////
// This function returns a new matrix which is 
// inverted matrix of the input one (if possible).
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in_data - reference to the first matrix
//		
// OUTPUT:
//			inverse matrix of in_data, if possible
//			0 otherwise
//		
// REMARKS:
//			Template parameter T stands for pixel type;
//			Template parameter D stands for intermediate
//				results accumulator;
//
//			The caller is responsible of disposing
//			the returned object!!!
//		
template < class T, class D /*= double*/ >
TImageFor< T > * Orphan_Inv_Matrix( const TImageFor< T > & in_data )
{
	int size = in_data.GetCol();
	if( size != in_data.GetRow() )
		return 0;					// the matrix is not square, cannot use this function

	int i, j, k;
	
	// Prepare the working buffer
	TImageFor< T > data( in_data );			// create a temporary buffer with copied data

	// Set the unity matrix - finally, this will become an output
	TImageFor< T > * outBuf = new TImageFor< T >( size, size, 0.0 );			// create a temporary buffer with copied data
	for( i = 0; i < size; ++ i )
		outBuf->SetPixel( i, i, 1.0 );

	D theFactor;

	// An improvement would be to find the biggest element at first, anyway...
	// For each row
	for( i = 0; i < size; ++ i )
	{
		theFactor = data.GetPixel( i, i );		// it is always the diagonal element 

		// --------------------------------------------------------------------------------------
		// Find the greatest factor in the following column (pivoting for the numerical stability)
		D tmp;
		j = i;
		for( k = i + 1; k < size; ++ k )
		{
			if( fabs( tmp = data.GetPixel( i, k ) ) > fabs( theFactor ) )
			{
				j = k;				// store that position
				theFactor = tmp;	// this is our new and better factor
			}
		}

		if( theFactor == 0.0 )
		{
			delete outBuf;
			return 0;				// cannot invert this matrix
		}

		// The factor is ok, so exchange the rows: i-th and j-th
		if( j > i )
		{
			// Exchange the j-th and i-th rows
			for( k = 0; k < size; ++ k )
			{
				tmp = data.GetPixel( k, i );
				data.SetPixel( k, i, data.GetPixel( k, j ) );
				data.SetPixel( k, j, tmp );
			}
			for( k = 0; k < size; ++ k )
			{
				tmp = outBuf->GetPixel( k, i );
				outBuf->SetPixel( k, i, outBuf->GetPixel( k, j ) );
				outBuf->SetPixel( k, j, tmp );
			}
		}
		// --------------------------------------------------------------------------------------



		for( k = 0; k < size; ++ k )
		{
			data.GetRefPixel( k, i )	/= theFactor;
			outBuf->GetRefPixel( k, i ) /= theFactor;
		}


		// Traverse the rest and zeroth the i-th column below, starting from the row index i-th down
		for( j = i + 1; j < size; ++ j )		// the rest rows
		{
			REQUIRE( data.GetPixel( i, i ) == 1.0 );
			theFactor = data.GetPixel( i, j );	// the same for this row
			for( k = 0; k < size; ++ k )		// all the columns
			{
				data.GetRefPixel( k, j ) -= theFactor * data.GetPixel( k, i );
				outBuf->GetRefPixel( k, j ) -= theFactor * outBuf->GetPixel( k, i );
			}

			REQUIRE( data.GetPixel( i, j ) == 0.0 );
		}	
	}

	REQUIRE( data.GetPixel( size - 1, size - 1 ) == 1.0 );

	// At this stage we should have a triangle matrix. What is left is to convert it
	// into canonical ones' matrix.
	// This time we traverse from bottom up
	for( i = size - 1; i > 0; -- i )
	{
		for( j = i - 1; j >= 0; -- j )		// the rest rows up
		{
			theFactor = data.GetPixel( i, j );	// the same for this row
			for( k = 0; k < size; ++ k )		// all the columns
				outBuf->GetRefPixel( k, j ) -= theFactor * outBuf->GetPixel( k, i );

		}	
	}

	return outBuf;
}



///////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function returns the conjugate (i.e. transposed)
// matrix.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in_matrix - the input matrix (= an image);
//				its value is only read
//		
// OUTPUT:
//			a pointer to the orphaned transposed matrix if possible,
//			0 otherwise
//		
// REMARKS:
//			The returned object is to be disposed
//			by a caller (i.e. it is orphaned)
//
//			Actually it is not a proper conjugate, only transposed.
//		
template < class T, class D /*= double*/ >
TImageFor< T > * Orphan_Conjugate_Matrix( const TImageFor< T > & in_data )
{
	const int kCols = in_data.GetCol();
	const int kRows = in_data.GetRow();

	REQUIRE( kCols != 0 );
	REQUIRE( kRows != 0 );

	REQUIRE( kCols < 100000 );	// sanity check
	REQUIRE( kRows < 100000 );

	TImageFor< T > * outImage	= new TImageFor< T >( kRows, kCols );	// just opposite

	register T * data_ptr		= in_data.GetImageData();

	register T * out_ptr		= outImage->GetImageData();
	register T * loc_out_ptr	= out_ptr;
	register T * loc_out_ptr2	= out_ptr;

	register int i, j;

	for( i = 0; i < kRows; ++ i )
	{
		loc_out_ptr = loc_out_ptr2;

		for( j = 0; j < kCols; ++ j )
		{

			* loc_out_ptr = * data_ptr ++;
			loc_out_ptr += kRows;			// kRows denotes rows but in the ORIGINAL matrix, for this one it is columns
		}

		++ loc_out_ptr2;
	}

	return outImage;
}

///////////////////////////////////////////////////////////
// This function returns the Penrose-Moore pseudo-inverse
// matrix (a matrix is an image):
//
//		A~ = ( A* x A )-1 x A*
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in_matrix - the input matrix (= an image);
//				its value is only read
//		
// OUTPUT:
//			a pointer to the orphaned pseudo-inverse matrix if possible,
//			0 otherwise
//		
// REMARKS:
//			The returned object is to be disposed
//			by a caller (i.e. it is orphaned)
//		
template < class T, class D /*= double*/ >
TImageFor< T > * Orphan_PseudoInv_Matrix( const TImageFor< T > & in_matrix )
{
	REQUIRE( in_matrix.GetCol() != in_matrix.GetRow() );	// although possible, in the opposite use simple inverse
	
	REQUIRE( in_matrix.GetCol() < 1000 );	// sanity check
	REQUIRE( in_matrix.GetRow() < 1000 );

	TImageFor< T > * theConjugate = Orphan_Conjugate_Matrix< T, D >( in_matrix );
	REQUIRE( theConjugate != 0 );

	TImageFor< T > * theConjMultIn = Orphan_Mult_Matrix< T, D >( * theConjugate, in_matrix );
	REQUIRE( theConjMultIn != 0 );

	TImageFor< T > * invertedMatrix = Orphan_Inv_Matrix< T, D >( * theConjMultIn );

	delete theConjMultIn;

	TImageFor< T > * retMatrix = invertedMatrix != 0 ? Orphan_Mult_Matrix< T, D >( * invertedMatrix, * theConjugate ) : 0;
	
	delete invertedMatrix;
	delete theConjugate;

	return retMatrix;
}


///////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function returns the solution to the linear
// system of linear equations in the form:
//
//				Ax=B
//
// if such exists.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			A - the input matrix 
//			B - the input matrix (a vector) 
//		
// OUTPUT:
//			a pointer to the orphaned solution vector x,
//			0 otherwise
//		
// REMARKS:
//			The function implements partial pivoting.
//
//			The returned object is to be disposed
//			by a caller (i.e. it is orphaned)
//		
template < class T, class D /*= double*/ >
TImageFor< T > * Orphan_Linear_Solution( const TImageFor< T > & A, const TImageFor< T > & B )
{
	int size = A.GetCol();
	if( size != A.GetRow() )
		return 0;					// the matrix is not square, cannot use this function
	if( size != B.GetRow() || B.GetCol() != 1 )
		return 0;					// the matrix is not a vector of a proper size, cannot use this function

	int i, j, k;
	
	// Prepare the working buffer
	int size_1 = size + 1;
	TImageFor< T > data( size_1, size );			// create a temporary buffer with copied data

	// Copy data to the new local buffer
	for( i = 0; i < size; ++ i )				// through cols
		for( j = 0; j < size; ++ j )			// through rows
			data.SetPixel( i, j, A.GetPixel( i, j ) );


	for( j = 0; j < size; ++ j )				// through rows
		data.SetPixel( size, j, B.GetPixel( 0, j ) );


	D theFactor;

	// An improvement would be to find the biggest element at first, anyway...
	// For each row
	for( i = 0; i < size; ++ i )
	{
		theFactor = data.GetPixel( i, i );		// it is always the diagonal element 

		// --------------------------------------------------------------------------------------
		// Find the greatest factor in the following column (partial-pivoting for the numerical stability)
		D tmp;
		j = i;
		for( k = i + 1; k < size; ++ k )
		{
			if( fabs( tmp = data.GetPixel( i, k ) ) > fabs( theFactor ) )
			{
				j = k;				// store that position
				theFactor = tmp;	// this is our new and better factor
			}
		}

		if( theFactor == 0.0 )
			return 0;				// cannot invert this matrix

		// The factor is ok, so exchange the rows: i-th and j-th
		if( j > i )
		{
			// Exchange the j-th and i-th rows
			for( k = 0; k < size_1; ++ k )
			{
				tmp = data.GetPixel( k, i );
				data.SetPixel( k, i, data.GetPixel( k, j ) );
				data.SetPixel( k, j, tmp );
			}
		}
		// --------------------------------------------------------------------------------------

		for( k = 0; k < size_1; ++ k )
			data.GetRefPixel( k, i )	/= theFactor;


		// Traverse the rest and zeroth the i-th column below, starting from the row index i-th down
		for( j = i + 1; j < size; ++ j )		// the rest rows
		{
			REQUIRE( data.GetPixel( i, i ) == 1.0 );
			theFactor = data.GetPixel( i, j );	// the same for this row
			for( k = 0; k < size_1; ++ k )		// all the columns
				data.GetRefPixel( k, j ) -= theFactor * data.GetPixel( k, i );

			REQUIRE( data.GetPixel( i, j ) == 0.0 );
		}	
	}


	REQUIRE( data.GetPixel( size - 1, size - 1 ) == 1.0 );

	TImageFor< T > * x_vec = new TImageFor< T >( 1, size, 0.0 );

	// At this stage we should have a triangle matrix. What is left is to compute unknown.
	// This time we traverse from bottom up
	for( i = size - 1; i >= 0; -- i )
	{
		D tmp = data.GetPixel( size, i );		// the right-most column

		for( j = size - 1; j > i; -- j )		
			tmp -= data.GetPixel( j, i ) * x_vec->GetPixel( 0, j );

		x_vec->GetRefPixel( 0, i ) = tmp;
	}

	return x_vec;
}

///////////////////////////////////////////////////////////////////////////////////////




#endif BoxySoft_ImageMatrix




