
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ColorSpaceConverters.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the color converter classes.
//
// ==========================================================================



#ifndef BoxySoft_ColorSpaceConverters
#define BoxySoft_ColorSpaceConverters


#include "HIL_ImageFor.h"
#include "HIL_AutoPtrDefs.h"
#include "FixedNumber.h"
#include "MathLibraryForFixed.h"
#include <math.h>



using namespace PHIL;

//
// The base class for color space conversions.
//
// REMARK: These conversions use the virtual mechanism
// which can slow down a little bit for massive data.
//
template < class PIX >
class ColorImageConverters
{
	protected:

		///////////////////////////////////////////////////////////
		// This function converts a pixel from one color space
		// to the complementary pixel in the output color space.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			out - reference to the out pixel
		//			in - reference to the input pixel
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		virtual void operator()( MMultiPixelFor< PIX > & out, const MMultiPixelFor< PIX > & in ) = 0;

	public:

		///////////////////////////////////////////////////////////
		// This function converts one color model based image
		// into another color image in-situ.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			dest_image - a reference to the input image
		//			src_image - an output image (in another color
		//				space)
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		virtual void operator()( TImageFor< MMultiPixelFor< PIX > > & dest_image, const TImageFor< MMultiPixelFor< PIX > > & src_image )
		{
			register int data_num = src_image.GetElems();

			REQUIRE( dest_image.GetElems() == data_num );
			if( dest_image.GetElems() != data_num )
				return;		// or throw !!!

			register MMultiPixelFor< PIX > * src_data_ptr	= src_image.GetImageData();
			register MMultiPixelFor< PIX > * dst_data_ptr	= dest_image.GetImageData();

			if( src_data_ptr != dst_data_ptr )
				while( data_num -- != 0 )
					operator()( * dst_data_ptr ++, * src_data_ptr ++ );
			else
				while( data_num -- != 0 )
					operator()( * src_data_ptr, * src_data_ptr ), src_data_ptr ++;

		}

		virtual void operator()( TImageFor< MMultiPixelFor< PIX > > & image )
		{
			register int data_num				= image.GetElems();
			register MMultiPixelFor< PIX > * data_ptr	= image.GetImageData();

			while( data_num -- != 0 )
				operator()( * data_ptr, * data_ptr ), data_ptr ++;
		}


};




// The template parameter T defines type of data for pixel computations
template< class PIX = unsigned char, class T = double >
class RGB_2_HSI : public ColorImageConverters< PIX >
{

public:

		///////////////////////////////////////////////////////////
		// This function converts an RGB pixel to HSI representation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theOutPixel - reference to the output pixel
		//			theInPixel - reference to the input pixel
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The two references can be the same.
		//		
		virtual void operator()( MMultiPixelFor< PIX > & theOutPixel, const MMultiPixelFor< PIX > & theInPixel )
		{
			T red, green, blue;

			T red_diff_green;
			T red_diff_blue;
			T green_diff_blue;

			T z, n, delta, sum, hue, satur, intens;

			const T kMaxVal( 255 );

			const T kMaxVal_k2Pi( 255 / k2Pi );

			red		= R( theInPixel );
			green	= G( theInPixel );
			blue	= B( theInPixel );

			REQUIRE( red >= T(0) && red <= kMaxVal );
			REQUIRE( green >= T(0) && green <= kMaxVal );
			REQUIRE( blue >= T(0) && blue <= kMaxVal );

			red_diff_green	= red - green;
			red_diff_blue	= red - blue;	
			green_diff_blue	= green - blue;	

			z = ( red_diff_green + red_diff_blue ) / (T)2.0;
			
			n = sqrt( red_diff_green * red_diff_green + red_diff_blue * green_diff_blue );

			if( n != (T)0 )
			{
				delta = acos( z / n );			// in radians 0 <= delta < pi
				REQUIRE( delta >= (T)0 && delta <= (T)kPi );
			}
			else
			{
				delta = (T)0;
				//REQUIRE( 0 );					// hue undef - check it!
			}

			// i.e., blue <= green
			hue = green_diff_blue >= (T)0 ? delta : (T)k2Pi - delta;

			// adjust to the integer value
			REQUIRE( hue >= (T)0 && hue <= (T)k2Pi );
			H( theOutPixel ) = hue * kMaxVal_k2Pi/*kMaxVal / (T)k2Pi*/;		// store hue [in radians!] in the red field

			sum = red + green + blue;

			if( sum != (T)0 )
			{
				satur = (T)1.0 - (T)3.0 * min( min( red, green ), blue ) / sum;
			}
			else
			{
				satur = (T)0;
				//REQUIRE( 0 );							// saturation undef - check it!
			}

			// adjust to the integer value
			REQUIRE( satur >= (T)0 && satur <= (T)1.0 );
			S( theOutPixel )= satur * kMaxVal;	// store saturation in the green field

			// compute intensity
			intens = sum / (T)3.0;
			REQUIRE( intens >= (T)0 && intens <= kMaxVal );
			I( theOutPixel ) = intens;				// store intensity in the blue field
		}


};



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// The template parameter T defines type of data for pixel computing
template< class PIX = unsigned char, class T = double >
class HSI_2_RGB : public ColorImageConverters< PIX >
{


public:

		///////////////////////////////////////////////////////////
		// This function converts an RGB pixel to HSI representation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theOutPixel - reference to the output pixel
		//			theInPixel - reference to the input pixel
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The two references can be the same.
		//		
		virtual void operator()( MMultiPixelFor< PIX > & theOutPixel, const MMultiPixelFor< PIX > & theInPixel )
		{
			T	r, g, b, h;

			const T rt3( 0.577350269 ); // = 1.0 / sqrt( 3.0 )
			
			const T	kMaxVal( 255 );
			const T _360_kMaxVal( 360.0 / 255 );

			const T k2Pi_360( k2Pi / 360.0 );
			

			T u1	= (T)H( theInPixel ) * _360_kMaxVal;	// when using fixed format we must be carefull with order of computations to avoid overflow!!!
			T u2	= (T)S( theInPixel ) / kMaxVal;
			T u3	= (T)I( theInPixel ) / kMaxVal;

			REQUIRE( u1 >= (T)0 && u1 <= (T)360.0 );
			REQUIRE( u2 >= (T)0 && u2 <= (T)1.0 );
			REQUIRE( u3 >= (T)0 && u3 <= (T)1.0 );

			if( u2 == (T)0 )
			{
				r = g = b = u3;				// gray tone
			}
			else
			{

				if( u1 < (T)120 )
				{
					// B is minimum
					b = ( (T)1.0 - u2 ) * u3;
					h = rt3 * (T)tan( (u1 - (T)60) * k2Pi_360 );	// tan takes radians!
					g = ( (T)1.5 + (T)1.5 * h ) * u3 - ( (T)0.5 + (T)1.5 * h ) * b;
					r = (T)3.0 * u3 - g - b;
					
				}
				else
				{
					if( u1 >= (T)120 && u1 < (T)240 )
					{
						// R is minimum
						r = ( (T)1.0 - u2 ) * u3;
						h = rt3 * (T)tan( (u1 - (T)180) * k2Pi_360 );	// tan takes radians!
						b = ( (T)1.5 + (T)1.5 * h ) * u3 - ( (T)0.5 + (T)1.5 * h ) * r;
						g = (T)3.0 * u3 - b - r;

					}
					else
					{
						// G is minimum
						g = ( (T)1.0 - u2 ) * u3;
						h = rt3 * (T)tan( (u1 - (T)300) * k2Pi_360 );	// tan takes radians!
						r = ( (T)1.5 + (T)1.5 * h ) * u3 - ( (T)0.5 + (T)1.5 * h ) * g;
						b = (T)3.0 * u3 - r - g;

					}

				}
			}

			if( r < (T)0.0 )
				r = (T)0.0;
			else
				if( r > (T)1.0 )
					r = (T)1.0;

			R( theOutPixel )= r * kMaxVal;


			if( g < (T)0.0 )
				g = (T)0.0;
			else
				if( g > (T)1.0 )
					g = (T)1.0;

			G( theOutPixel ) = g * kMaxVal;


			if( b < (T)0.0 )
				b = (T)0.0;
			else
				if( b > (T)1.0 )
					b = (T)1.0;

			B( theOutPixel ) = b * kMaxVal;

		}


};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// The template parameter T defines type of data for pixel computing
template< class PIX = unsigned char, class T = double >
class RGB_2_YCrCb : public ColorImageConverters< PIX >
{

public:

		///////////////////////////////////////////////////////////
		// This function converts an RGB pixel to HSI representation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theOutPixel - reference to the output pixel
		//			theInPixel - reference to the input pixel
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The two references can be the same.
		//		
		virtual void operator()( MMultiPixelFor< PIX > & out, const MMultiPixelFor< PIX > & in )
		{
			T r = R( in );
			T g = G( in );
			T b = B( in );

			Y( out )  = 0.2990000 * r + 0.587000 * g + 0.114000 * b;
			Cb( out ) = -0.168736 * r - 0.331264 * g + 0.500000 * b;
			Cr( out ) = 0.5000000 * r - 0.418668 * g - 0.081312 * b;

			REQUIRE( Y( out ) >= 0 && Cb( out ) <= 255 );		// check the range
			REQUIRE( Cb( out ) >= -128 && Cb( out ) <= 127 );		
			REQUIRE( Cr( out ) >= -128 && Cr( out ) <= 127 );
		}

};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


// The template parameter T defines type of data for pixel computing
template< class PIX = unsigned char, class T = double >
class YCrCb_2_RGB : public ColorImageConverters< PIX >
{

public:

		///////////////////////////////////////////////////////////
		// This function converts an RGB pixel to HSI representation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theOutPixel - reference to the output pixel
		//			theInPixel - reference to the input pixel
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The two references can be the same.
		//		
		virtual void operator()( MMultiPixelFor< PIX > & out, const MMultiPixelFor< PIX > & in )
		{
			T y	=			 Y( in );		// this should be always positive in the full range
			T cb	= (char)Cb( in );		// this can be negative or positive in the half range 
			T cr	= (char)Cr( in );

			REQUIRE( y >= 0.0 );

			T r = y - 0.00092640 * cb + 1.40168676 * cr;
			T g = y - 0.34369538 * cb - 0.71416904 * cr;
			T b = y + 1.77216042 * cb + 0.00099022 * cr;


			// We need to offset the results
			if( r < 0.0 )
				r = -r;
			if( g < 0.0 )
				g = -g;
			if( b < 0.0 )
				b = -b;
			
			if( r > 255.0 )
				r = 255.0;
			if( g > 255.0 )
				g = 255.0;
			if( b > 255.0 )
				b = 255.0;


			R( out ) = (unsigned char) r;		// red
			G( out ) = (unsigned char) g;		// green
			B( out ) = (unsigned char) b;		// blue

		}

};



///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// Helper functions
template< class PIX >
inline void RGB_2_HSI_Convert( TImageFor< MMultiPixelFor< PIX > > & dest_image, const TImageFor< MMultiPixelFor< PIX > > & src_image )
{
	RGB_2_HSI< PIX, double > theConverter;
//	(*(ColorImageConverters< PIX >*)&theConverter)( dest_image, src_image );
	theConverter.ColorImageConverters< PIX >::operator()( dest_image, src_image );
}

template< class PIX >
inline void HSI_2_RGB_Convert( TImageFor< MMultiPixelFor< PIX > > & dest_image, const TImageFor< MMultiPixelFor< PIX > > & src_image )
{
	HSI_2_RGB< PIX, double > theConverter;
	theConverter.ColorImageConverters< PIX >::operator()( dest_image, src_image );
}

template< class PIX >
inline void RGB_2_YCrCb_Convert( TImageFor< MMultiPixelFor< PIX > > & dest_image, const TImageFor< MMultiPixelFor< PIX > > & src_image )
{
	RGB_2_YCrCb< PIX, double > theConverter;
	theConverter.ColorImageConverters< PIX >::operator()( dest_image, src_image );
}

template< class PIX >
inline void YCrCb_2_RGB_Convert( TImageFor< MMultiPixelFor< PIX > > & dest_image, const TImageFor< MMultiPixelFor< PIX > > & src_image )
{
	YCrCb_2_RGB< PIX, double > theConverter;
	theConverter.ColorImageConverters< PIX >::operator()( dest_image, src_image );
}


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////



#endif BoxySoft_ColorSpaceConverters




