
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: BinomialFilter.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Classes for bionomial filters.	
//
// ==========================================================================


#include <math.h>
#include "BinomialFilter.h"
#include "GraphicLibrary.h"
#include <functional>
#include <fstream>



///////////////////////////////////////////////////////////
// This function is a class constructor
///////////////////////////////////////////////////////////
//		
// INPUT:
//			mask_width - width of the mask; preferrable
//				odd numbers	
//		
// OUTPUT:
//			constructed object
//		
// REMARKS:
//			This member initializes necessary binomial
//			coefficients, so they are available immidiately
//		
TBinomialFilter::TBinomialFilter( int mask_width )		// class constructor
: fN( mask_width )
{
	// Get binomial coefficients.
	FillFilterCoefficients( fN, fBinomialCoeff );
}


///////////////////////////////////////////////////////////
// This function is a virtual destructor.
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TBinomialFilter::~TBinomialFilter()		// class destructor
{
}


///////////////////////////////////////////////////////////
// This function fills binomial coefficients based on 
// mask size.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			n - width of the binomial mask, should be
//				greater than 0
//
//						PASCAL TRIANGLE
//			----------------------------------------
//			n |         binomial coefficients
//			----------------------------------------
//			1 |                  1
//			2 |                1 2 1
//			3 |               1 3 3 1
//          4 |              1 4 6 4 1
//			5 |             1 5 10 10 5 1
//          .........................................
//
//
//			binoCoefVect - reference to the input/output
//				vector: on input it contains k'th row of
//				the Pascal traingle, on output it contains
//				the k+1'th row.
//		
// OUTPUT:
//			binoCoefVect contains the k+1'th row of
//				the Pascal triangle
//		
// REMARKS:
//			On the first run it is only required that size
//			of the binoCoefVect is exactly 0.
//		
void TBinomialFilter::FillFilterCoefficients( int n, vector< double > & binoCoefVect )
{
	REQUIRE( n > 0 );

	vector< unsigned long > bino;

	REQUIRE( bino.size() == 0 );		// Required by the GetNextPascalTriangleRow()


	double multiCoeff = pow( 2.0, n - 1 );	

	while( n -- )
		GetNextPascalTriangleRow( bino );

	// Fill out vector
	REQUIRE( binoCoefVect.size() == 0 );

	transform( bino.begin(), bino.end(), back_inserter( binoCoefVect ), bind2nd( divides< double >(), /*(double)*/multiCoeff ) );	// add elements to the end of binoCoefVect

#if _DEBUG
	ofstream test_file( "bino.txt" );
	ostream_iterator<double> iter( test_file, ", " );
	copy( binoCoefVect.begin(), binoCoefVect.end(), iter );
#endif


	REQUIRE( binoCoefVect.size() == bino.size() );
}


///////////////////////////////////////////////////////////
// This function returns next Pascal triangle row.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			bino - a reference to the k-1 row of the Pascal triangle.
//				At the first call it is only required that size
//				of the bino is exactly 0; all the next iterations
//				will fill next values appropriately.
//		
// OUTPUT:
//			bino contains the k'th row of the Pascal triangle	
//		
// REMARKS:
//		
//		
void TBinomialFilter::GetNextPascalTriangleRow( vector< unsigned long > & bino )
{
	int len = bino.size();

	REQUIRE( len >= 0 );

	register unsigned long prev_val_1 = len > 0 ? bino[ 0 ] : 1;
	register unsigned long prev_val_2;

	register int i;

	REQUIRE( prev_val_1 == 1 );


	// Max value for "i" will be len-2
	for( i = 0; i < len - 1; i ++ )
	{
		prev_val_2 = bino[ i + 1 ];		// save for next iteration

		bino[ i + 1 ] += prev_val_1;

		prev_val_1 = prev_val_2;		// save for next iteration
	}


	bino.push_back( 1 );	// all rows end with "1"
}


///////////////////////////////////////////////////////////
// This function does binomial filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - const reference to the input image
//			mask_width - width of the binomial filter
//			Out - reference to the output image
//		
// OUTPUT:
//			none
//		
// REMARKS:
//			See also overloaded versions of this function.
//		
void TBinomialFilter::operator() ( const TRealImage & In, TRealImage & Out ) const
{
	REQUIRE( fBinomialCoeff.size() > 0 );

	TRealImage * outHolder = new TRealImage( In.GetCol(), In.GetRow() );

	Horz1DConvolve( In, fBinomialCoeff, * outHolder );
	Vert1DConvolve( * outHolder, fBinomialCoeff, Out );

	delete outHolder;
}


///////////////////////////////////////////////////////////
// This function does binomial filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - const reference to the input image
//		
// OUTPUT:
//			ORPHANED filtered image.
//		
// REMARKS:
//			The returned image is orphaned, i.e. the caller
//			is responsible for its further management and
//			its final memory de-allocation !!!
//
//			See also overloaded versions of this function.
//		
TRealImage * TBinomialFilter::operator() ( const TRealImage & In ) const
{
	TRealImage * out = new TRealImage( In.GetCol(), In.GetRow() );

	operator() ( In, * out );

	return out;
}

///////////////////////////////////////////////////////////
// This function does binomial filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - reference to the input image
//		
// OUTPUT:
//			none
//		
// REMARKS:
//			The input image is CHANGED!
//
//			See also overloaded versions of this function.
//		
void TBinomialFilter::operator() ( TRealImage & In ) const
{
	operator() ( In, In );
}








