
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: BinomialFilter.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Classes for bionomial filters.	
//
// ==========================================================================




#ifndef BoxySoft_BinomialFilter
#define BoxySoft_BinomialFilter


#include "RealLinearFilter.h"


///////////////////////////////////////////////////////////
// This class implements binomial filter. For more information
// on this class of filters see:
// 1. Jaehne B.:	Digital Image Processing; pg. 297-304,
//					Springer 1997.
// 2. Klette R., Zamperoni P.:	Handbook of Image Processing
//					Operators; pg. 203-206, Wiley, 1996.
///////////////////////////////////////////////////////////
class TBinomialFilter : public TRealLinearFilter
{
	private:

		// Class inherent variables

		// Contains mask size "n" value
		int	fN;

		// Contains binomial coefficients for the given "n",
		// i.e. mask size. They are computed according to the rule:
		//
		//                        | n-1 |
		//	fBinomialCoeff[ i ] = |     |, where 0 <= i < n
		//                        |  i  |
		//
		vector< double >	fBinomialCoeff;

	public:

		const vector< double > & GetBinomialCoefficients( void ) const { return fBinomialCoeff; }

	protected:

		///////////////////////////////////////////////////////////
		// This function returns next Pascal triangle row.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			bino - a reference to the k-1 row of the Pascal triangle.
		//				At the first call it is only required that size
		//				of the bino is exactly 0; all the next iterations
		//				will fill next values appropriately.
		//		
		// OUTPUT:
		//			bino contains the k'th row of the Pascal triangle	
		//		
		// REMARKS:
		//		
		//		
		void GetNextPascalTriangleRow( vector< unsigned long > & bino );

		
		///////////////////////////////////////////////////////////
		// This function fills binomial coefficients based on 
		// mask size.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			n - width of the binomial mask, should be
		//				greater than 0
		//
		//						PASCAL TRIANGLE
		//			----------------------------------------
		//			n |         binomial coefficients
		//			----------------------------------------
		//			1 |                  1
		//			2 |                1 2 1
		//			3 |               1 3 3 1
		//          4 |              1 4 6 4 1
		//			5 |             1 5 10 10 5 1
		//          .........................................
		//
		//
		//			binoCoefVect - reference to the input/output
		//				vector: on input it contains k'th row of
		//				the Pascal traingle, on output it contains
		//				the k+1'th row.
		//		
		// OUTPUT:
		//			binoCoefVect contains the k+1'th row of
		//				the Pascal traingle
		//		
		// REMARKS:
		//			On the first run it is only required that size
		//			of the binoCoefVect is exactly 0.
		//		
		void FillFilterCoefficients( int n, vector< double > & binoCoefVect );

	public:

		// ===================================================
		TBinomialFilter( int mask_width );		// class default constructor
		virtual ~TBinomialFilter();				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - const reference to the input image
		//			Out - reference to the output image
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			See also overloaded versions of this function.
		//		
		virtual void operator() ( const TRealImage & In, TRealImage & Out ) const;

		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - const reference to the input image
		//		
		// OUTPUT:
		//			ORPHANED filtered image.
		//		
		// REMARKS:
		//			The returned image is orphaned, i.e. the caller
		//			is responsible for its further management and
		//			its final memory de-allocation !!!
		//
		//			See also overloaded versions of this function.
		//		
		virtual TRealImage * operator() ( const TRealImage & In ) const;

		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - reference to the input image
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The input image is CHANGED!
		//
		//			See also overloaded versions of this function.
		//		
		virtual void operator() ( TRealImage & In ) const;

};





#endif BoxySoft_BinomialFilter




