
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: GaussianFilter.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the Gaussian filters.
//
// ==========================================================================






#include <math.h>
#include <functional>
#include <fstream>
#include "GaussianFilter.h"
#include "GraphicLibrary.h"




///////////////////////////////////////////////////////////
// This function is a class constructor
///////////////////////////////////////////////////////////
//		
// INPUT:
//			mask_width - width of the mask; preferrable
//				odd numbers; We assume window length
//				(mask_width here) equal to: 3*sigma	
//		
// OUTPUT:
//			constructed object
//		
// REMARKS:
//			This member initializes necessary binomial
//			coefficients, so they are available immidiately
//		
TGaussianFilter::TGaussianFilter( int mask_width )		// class constructor
: fN( mask_width )
{
	// Get Gaussian coefficients.
	FillFilterCoefficients( fN, fGaussianCoeff );
}

///////////////////////////////////////////////////////////
// This function is a class constructor
///////////////////////////////////////////////////////////
//		
// INPUT:
//			mask_width - width of the mask; preferrable
//				odd numbers	
//		
// OUTPUT:
//			constructed object
//		
// REMARKS:
//			This member initializes necessary binomial
//			coefficients, so they are available immidiately
//		
TGaussianFilter::TGaussianFilter( int mask_width, double sigma )		// class constructor
: fN( mask_width )
{
	// Get Gaussian coefficients.
	FillFilterCoefficients( fN, sigma, fGaussianCoeff );
}

///////////////////////////////////////////////////////////
// This function is a virtual destructor.
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TGaussianFilter::~TGaussianFilter()		// class destructor
{
}


///////////////////////////////////////////////////////////
// This function fills the Gaussian coefficients based on 
// mask size.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			n - width of the binomial mask, should be
//				greater than 0
//			gaussCoefVect - reference to the input/output
//				vector.
//		
// OUTPUT:
//			gaussCoefVect contains gaussian coefficients
//		
// REMARKS:
//		
void TGaussianFilter::FillFilterCoefficients( int n, double sigma, vector< double > & gaussCoefVect )
{
	REQUIRE( n > 0 );

	// The mask should be symetric, so check it out.
	if( ( n & 0x01 ) == 0 )
		n ++;		// make it odd

	// We assume window length (n here) equal 3*sigma, so
	double _2_sigma_square = 2.0 * sigma * sigma;

	const int kHalfCoefNum = n / 2;

	// Make room for n coefficients
	gaussCoefVect.clear();
	gaussCoefVect.insert( gaussCoefVect.begin(), n, 0.0 );

	double sum = 1.0;
	gaussCoefVect[ kHalfCoefNum ] = 1.0;

	for( int i = 1; i <= kHalfCoefNum; i ++ )
	{
		double c = exp( - ((double)i * i) / _2_sigma_square );
		gaussCoefVect[ kHalfCoefNum - i ] = gaussCoefVect[ kHalfCoefNum + i ] = c;		
		sum += c + c;
	}

	// Normalize all coefficients such they sum up to 1.0
	transform( gaussCoefVect.begin(), gaussCoefVect.end(), gaussCoefVect.begin(), bind2nd( divides< double >(), sum ) );

#if _DEBUG
	ofstream test_file( "gauss.txt" );
	ostream_iterator<double> iter( test_file, ", " );
	copy( gaussCoefVect.begin(), gaussCoefVect.end(), iter );
#endif

}

void TGaussianFilter::FillFilterCoefficients( int n, vector< double > & gaussCoefVect )
{
	double assumed_sigma = (double) n / /*3.0*/6.0;		// ==> n is NOT a radius, it is a diameter
	FillFilterCoefficients( n, assumed_sigma, gaussCoefVect );
}



///////////////////////////////////////////////////////////
// This function does gaussian filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - const reference to the input image
//			mask_width - width of the binomial filter
//			Out - reference to the output image
//		
// OUTPUT:
//			none
//		
// REMARKS:
//			See also overloaded versions of this function.
//		
void TGaussianFilter::operator() ( const TRealImage & In, TRealImage & Out ) const
{
	REQUIRE( fGaussianCoeff.size() > 0 );

	TRealImage * outHolder = new TRealImage( In.GetCol(), In.GetRow() );

	Horz1DConvolve( In, fGaussianCoeff, * outHolder );
	Vert1DConvolve( * outHolder, fGaussianCoeff, Out );

	delete outHolder;
}


///////////////////////////////////////////////////////////
// This function does gaussian filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - const reference to the input image
//		
// OUTPUT:
//			ORPHANED filtered image.
//		
// REMARKS:
//			The returned image is orphaned, i.e. the caller
//			is responsible for its further management and
//			its final memory de-allocation !!!
//
//			See also overloaded versions of this function.
//		
TRealImage * TGaussianFilter::operator() ( const TRealImage & In ) const
{
	TRealImage * out = new TRealImage( In.GetCol(), In.GetRow() );

	operator() ( In, * out );

	return out;
}

///////////////////////////////////////////////////////////
// This function does gaussian filtering of the input image. 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - reference to the input image
//		
// OUTPUT:
//			none
//		
// REMARKS:
//			The input image is CHANGED!
//
//			See also overloaded versions of this function.
//		
void TGaussianFilter::operator() ( TRealImage & In ) const
{
	operator() ( In, In );
}


////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
DOG_Filter::DOG_Filter( int mask_width, double sigma_1, double sigma_2 )		// class constructor
: TGaussianFilter( mask_width, sigma_1 )
{
	// The first part of the Gaussian coefficients is already filled by
	// the base constructor. 
	// What remains is the second set of Gaussian coefficients and their
	// subtraction.

	vector< double > secondSetOfGaussianCoeff;

	FillFilterCoefficients( mask_width, sigma_2, secondSetOfGaussianCoeff );

	REQUIRE( fGaussianCoeff.size() == secondSetOfGaussianCoeff.size() );

	// Do subtractions (soemthing like 	fGaussianCoeff -= secondSetOfGaussianCoeff; )
	transform(	fGaussianCoeff.begin(), fGaussianCoeff.end(),							// first source
				secondSetOfGaussianCoeff.begin(),										// second source
				fGaussianCoeff.begin(),													// destination
				minus< double >() );													// what to do...

#if _DEBUG
	ofstream test_file( "DoG_Mask.txt" );
	ostream_iterator< double > iter( test_file, ", " );
	copy( fGaussianCoeff.begin(), fGaussianCoeff.end(), iter );
#endif

}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
DOG_Filter::~DOG_Filter()		// class destructor
{
}




////////////////////////////////////////////////////////////////////////////////////////////////











