
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: GaussianFilter.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the Gaussian filters.
//
// ==========================================================================



#ifndef BoxySoft_GaussianFilter
#define BoxySoft_GaussianFilter


#include "RealLinearFilter.h"


///////////////////////////////////////////////////////////
// This class implements the Gaussian filter. For more information
// on this class of filters see:
///////////////////////////////////////////////////////////
class TGaussianFilter : public TRealLinearFilter
{
	private:
	protected:

		// Class inherent variables

		// Contains mask size "n" value
		int	fN;

		// Contains Gaussian coefficients for the given "n",
		vector< double >	fGaussianCoeff;

	public:

		const vector< double > & GetGaussianCoefficients( void ) const { return fGaussianCoeff; }

	protected:

		
		///////////////////////////////////////////////////////////
		// This function fills the Gaussian coefficients based on 
		// mask size.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			n - width of the binomial mask, should be
		//				greater than 0
		//			gaussCoefVect - reference to the input/output
		//				vector.
		//		
		// OUTPUT:
		//			gaussCoefVect contains gaussian coefficients
		//		
		// REMARKS:
		//			The version without explicit sigma assumes
		//				sigma = n / 3
		//
		void FillFilterCoefficients( int n, vector< double > & GaussCoefVect );
		void FillFilterCoefficients( int n, double sigma, vector< double > & GaussCoefVect );

	public:

		// ===================================================
		///////////////////////////////////////////////////////////
		// This function is a class constructor
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			mask_width - width of the mask; preferrable
		//				odd numbers; We assume window length
		//				(mask_width here) equal to: 3*sigma	
		//		
		// OUTPUT:
		//			constructed object
		//		
		// REMARKS:
		//			This member initializes necessary binomial
		//			coefficients, so they are available immidiately
		//		
		//			The version without explicit sigma assumes
		//				sigma = mask_width / 3

		TGaussianFilter( int mask_width );						// class constructor
		TGaussianFilter( int mask_width, double sigma );		// class constructor
		virtual ~TGaussianFilter();								// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - const reference to the input image
		//			Out - reference to the output image
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			See also overloaded versions of this function.
		//		
		virtual void operator() ( const TRealImage & In, TRealImage & Out ) const;

		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - const reference to the input image
		//		
		// OUTPUT:
		//			ORPHANED filtered image.
		//		
		// REMARKS:
		//			The returned image is orphaned, i.e. the caller
		//			is responsible for its further management and
		//			its final memory de-allocation !!!
		//
		//			See also overloaded versions of this function.
		//		
		virtual TRealImage * operator() ( const TRealImage & In ) const;

		///////////////////////////////////////////////////////////
		// This function does binomial filtering of the input image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - reference to the input image
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The input image is CHANGED!
		//
		//			See also overloaded versions of this function.
		//		
		virtual void operator() ( TRealImage & In ) const;

};

//////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This class implements the DOG (Difference of Gaussians)
// filter.
///////////////////////////////////////////////////////////
class DOG_Filter : public TGaussianFilter
{
	private:

		// Class inherent variables

	protected:

	public:

		// ===================================================
		DOG_Filter( int mask_width, double sigma_1, double sigma_2 );		// class constructor
		virtual ~DOG_Filter();												// class virtual destructor
		// ===================================================

	public:



};




//////////////////////////////////////////////////////////////////////////////////////


#endif BoxySoft_GaussianFilter




