
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: GraphicLibrary.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				The file contains functions for different simple actions
//				for image processing.
//
// ==========================================================================





#ifndef GraphicLibrary_h
#define GraphicLibrary_h


#include "HIL_ImageFor.h"
#include "AccumulationTraits.h"
#include <algorithm>


using namespace PHIL;




///////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This class performs 1D horizontal convolution 
//
// The type "T" defines what is a type of a single pixel.
// The type "ATrait" defines what is a type for the
// accumulator used during convolution.
//
// We create a class here instead of the template function.
// The problem is that the template functions don't allow
// for default parameters. In this fashion a user would
// be forced to provied ATrait all the time. Therefore 
// we have this class.
//
// For the actual convolution use the helper functions.
//
///////////////////////////////////////////////////////////
template < typename T, typename ATrait = Accumulation_Trait< T > > 
class _Convolve
{
	public:


		///////////////////////////////////////////////////////////
		// This function does the convolution Out = In * Mask. 
		// The convolution is performed symmetrically, i.e. at a 
		// given point of the image convolution summation goes
		// symmetrically in all directions. If value of a image 
		// or mask outside their boundaries is necessary it is 
		// assumed that this value is 0.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - a reference to the input image
		//			Mask - a reference to the convolution kernel image
		//			Out - a reference to the outcome image	
		//		
		// OUTPUT:
		//			Out = In * Mask	
		//		
		// REMARKS:
		//			Although it is not necessary, the best if mask size
		//			is odd in both directions -- in such a case the whole
		//			procedure can be simplified of the conditions.
		//		
		static void Convolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
		{
			register long i, j, m, n, vertMaskIndex, horzMaskIndex, imTempCol, imTempRow;

			typename ATrait::TheAccumulatorType conv_sum;

			const long im_col = In.GetCol();
			const long im_row = In.GetRow();

			const long mask_col = Mask.GetCol();
			const long mask_row = Mask.GetRow();

			long horzHalfMask = mask_col / 2;
			long vertHalfMask = mask_row / 2; 


			// Require mask sizes to be odd - for real filters
			//REQUIRE( 2 * vertHalfMask + 1 == mask_row );
			//REQUIRE( 2 * horzHalfMask + 1 == mask_col );


			// Go through all the points of the input picture
			for( i = 0; i < im_row; i ++ )
			{
				for( j = 0; j < im_col; j ++ )
				{
					// Count the sum for the whole mask at each point of the input image.
					conv_sum = ATrait::GetZero();

					for( m = - vertHalfMask; m <= + vertHalfMask; m ++ )
					{
						vertMaskIndex = m + vertHalfMask;

						for( n = - horzHalfMask; n <= + horzHalfMask; n ++ )
						{
							imTempCol = j - n;
							imTempRow = i - m;

							horzMaskIndex = n + horzHalfMask;

							if( imTempCol >= 0 && imTempCol < im_col && imTempRow >= 0 && imTempRow < im_row 		
								// >>> This part of AND is necessary only if size of the mask is even <<<
								&& horzMaskIndex < mask_col && vertMaskIndex < mask_row )
								conv_sum += Mask.GetPixel( horzMaskIndex, vertMaskIndex ) * In.GetPixel( imTempCol, imTempRow );
						
						}
					}

					Out.SetPixel( j, i, (T)conv_sum );

				}

			}

		}

};

// The helpers:
#if MODERN_COMPILER
template < typename T, typename Traits >
inline void Convolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Convolve< T, Traits >::Convolve( In, Mask, Out );
}
#endif MODERN_COMPILER

template < typename T >
inline void Convolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Convolve< T >::Convolve( In, Mask, Out );
}




///////////////////////////////////////////////////////////
// This class performs 1D horizontal convolution 
//
// The type "T" defines what is a type of a single pixel.
// The type "ATrait" defines what is a type for the
// accumulator used during convolution.
//
// We create a class here instead of the template function.
// The problem is that the template functions don't allow
// for default parameters. In this fashion a user would
// be forced to provied ATrait all the time. Therefore 
// we have this class.
//
// For the actual convolution use the helper functions.
//
///////////////////////////////////////////////////////////
template < typename T, typename ATrait = Accumulation_Trait< T > > 
class _Horz1DConvolve
{
	public:

		///////////////////////////////////////////////////////////
		// This function performs 1D convolution of each row of
		// a given image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - a reference to the input image
		//			Mask - a reference to the convolution kernel vector
		//			Out - a reference to the output image (In and Out
		//				should have the same size)	
		//		
		// OUTPUT:
		//			Out set with convolution outcome	
		//		
		// REMARKS:
		//		
		//		
		static void Convolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
		{
			register long i, j, k;

			typename ATrait::TheAccumulatorType conv_sum;

			const long row = In.GetRow();
			const long col = In.GetCol();

			const int start = Mask.size() / 2;
			REQUIRE( start != 0 );				// empty mask supplied, probably an error
			if( start == 0 )
				return;

			// Go through all the rows of the input picture
			for( i = 0; i < row; i ++ )
			{
				// Get each pixel in the row
				for( j = 0; j < col; j ++ )
				{
					conv_sum = ATrait::GetZero();

					for( k = - start; k <= + start; k ++ )
						if( j - k >= 0 && j - k < col )
							conv_sum += Mask[ k + start ] * In.GetPixel( j - k, i );

					Out.SetPixel( j, i, (T)conv_sum );
				}
			}
		}

		// Overloaded version - the mask is in a form of an 1D image (i.e. a single row)
		static void Convolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
		{
			register long i, j, k;

			typename ATrait::TheAccumulatorType conv_sum;

			const long row = In.GetRow();
			const long col = In.GetCol();

			const int kMaskSize = Mask.GetCol();
			REQUIRE( Mask.GetRow() == 1 );			// since this should be an 1D image with only one row
			
			const int start = kMaskSize / 2;
			REQUIRE( start != 0 );					// empty mask supplied, probably an error
			if( start == 0 )
				return;


			// Go through all the rows of the input picture
			for( i = 0; i < row; i ++ )
			{
				// Get each pixel in the row
				for( j = 0; j < col; j ++ )
				{
					conv_sum = ATrait::GetZero();

					for( k = - start; k <= + start; k ++ )
						if( j - k >= 0 && j - k < col )
							conv_sum += Mask.GetPixel( k + start, 0 ) * In.GetPixel( j - k, i );

					Out.SetPixel( j, i, (T)conv_sum );
				}
			}
		}

};

// The helpers:
#if MODERN_COMPILER
template < typename T, typename Traits >
inline void Horz1DConvolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
{
	_Horz1DConvolve< T, Traits >::Convolve( In, Mask, Out );
}
template < typename T, typename Traits >
inline void Horz1DConvolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Horz1DConvolve< T, Traits >::Convolve( In, Mask, Out );
}
#endif MODERN_COMPILER

template < typename T >
inline void Horz1DConvolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
{
	_Horz1DConvolve< T >::Convolve( In, Mask, Out );
}

template < typename T >
inline void Horz1DConvolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Horz1DConvolve< T >::Convolve( In, Mask, Out );
}

///////////////////////////////////////////////////////////
// This class performs 1D vertical convolution 
//
// The type "T" defines what is a type of a single pixel.
// The type "ATrait" defines what is a type for the
// accumulator used during convolution.
//
// We create a class here instead of the template function.
// The problem is that the template functions don't allow
// for default parameters. In this fashion a user would
// be forced to provied ATrait all the time. Therefore 
// we have this class.
//
// For the actual convolution use the helper functions.
//
///////////////////////////////////////////////////////////
template < typename T, typename ATrait = Accumulation_Trait< T > > 
class _Vert1DConvolve
{
	public:

		///////////////////////////////////////////////////////////
		// This function performs 1D convolution of each column of
		// a given image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			In - a reference to the input image
		//			Mask - a reference to the convolution kernel vector
		//			Out - a reference to the output image (In and Out
		//				should have the same size)	
		//		
		// OUTPUT:
		//			Out set with convolution outcome	
		//		
		// REMARKS:
		//		
		//		
		static void Convolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
		{
			register long i, j, k;

			typename ATrait::TheAccumulatorType conv_sum;

			const long row = In.GetRow();
			const long col = In.GetCol();
			
			const int start = Mask.size() / 2;
			REQUIRE( start != 0 );					// empty mask supplied, probably an error
			if( start == 0 )
				return;


			// Go through all the rows of the input picture
			for( i = 0; i < col; i ++ )
			{

				// Get each pixel in the row
				for( j = 0; j < row; j ++ )
				{
					conv_sum = ATrait::GetZero();

					for( k = - start; k <= + start; k ++ )
						if( j - k >= 0 && j - k < row )
							conv_sum += Mask[ k + start ] * In.GetPixel( i, j - k );

					Out.SetPixel( i, j, (T)conv_sum );
				}
			}
		}

		// Overloaded version - the mask is in a form of an 1D image (i.e. a single row)
		static void Convolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
		{
			register long i, j, k;

			typename ATrait::TheAccumulatorType conv_sum;

			const long row = In.GetRow();
			const long col = In.GetCol();
			
			const int kMaskSize = Mask.GetCol();
			REQUIRE( Mask.GetRow() == 1 );			// since this should be an 1D image with only one row
			
			const int start = kMaskSize / 2;
			REQUIRE( start != 0 );					// empty mask supplied, probably an error
			if( start == 0 )
				return;


			// Go through all the rows of the input picture
			for( i = 0; i < col; i ++ )
			{

				// Get each pixel in the row
				for( j = 0; j < row; j ++ )
				{
					conv_sum = ATrait::GetZero();

					for( k = - start; k <= + start; k ++ )
						if( j - k >= 0 && j - k < row )
							conv_sum += Mask.GetPixel( k + start, 0 ) * In.GetPixel( i, j - k );

					Out.SetPixel( i, j, (T)conv_sum );
				}
			}
		}


};

// The helpers:
#if MODERN_COMPILER
template < typename T, typename Traits >
inline void Vert1DConvolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
{
	_Vert1DConvolve< T, Traits >::Convolve( In, Mask, Out );
}
template < typename T, typename Traits >
inline void Vert1DConvolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Vert1DConvolve< T, Traits >::Convolve( In, Mask, Out );
}
#endif MODERN_COMPILER

template < typename T >
inline void Vert1DConvolve( const TImageFor< T > & In, const vector< T > & Mask, TImageFor< T > & Out )
{
	_Vert1DConvolve< T >::Convolve( In, Mask, Out );
}

template < typename T >
inline void Vert1DConvolve( const TImageFor< T > & In, const TImageFor< T > & Mask, TImageFor< T > & Out )
{
	_Vert1DConvolve< T >::Convolve( In, Mask, Out );
}


// Computes the multi-graylevel thresholded output image.
// The output gray value is one of the three ranges:
//
// I(x,y) = GrayVal1 <==> origVal < Thresh1
// I(x,y) = GrayVal2 <==> Thresh1 <= origVal < Thresh2
// I(x,y) = GrayVal3 <==> origVal >= Thresh2
//
// All threshold levels are assumed to be UnSIGNED!
//
template < class T > 
void MultiThresholding( const TImageFor< T > & In, 
						T Thresh1, T Thresh2, 
						TImageFor< T > & Out,
						T GrayVal1,
						T GrayVal2,
						T GrayVal3
						)
{
	REQUIRE( In.GetCol() == Out.GetCol() );
	REQUIRE( In.GetRow() == Out.GetRow() );

	register T * src_data_ptr = In.GetImageData();
	register T * dst_data_ptr = Out.GetImageData();
	register int data_num = In.GetElems();

	T pixelVal;

	while( data_num -- != 0 )
	{
		pixelVal = * src_data_ptr ++;
		if( pixelVal < Thresh1 )
			* dst_data_ptr ++ = GrayVal1;
		else
			if( pixelVal < Thresh2 )
				* dst_data_ptr ++ = GrayVal2;
			else
				* dst_data_ptr ++ = GrayVal3;

	}

}

// Computes the multi-graylevel thresholded output image.
// The output gray value is one of the three ranges:
//
// I(x,y) = GrayVal1 <==> origVal < Thresh1
// I(x,y) = GrayVal2 <==> Thresh1 <= origVal
//
// All threshold levels are assumed to be UnSIGNED!
//
template < class T > 
void MultiThresholding( const TImageFor< T > & In, 
						T Thresh1, 
						TImageFor< T > & Out,
						T GrayVal1,
						T GrayVal2
					)
{
	REQUIRE( In.GetCol() == Out.GetCol() );
	REQUIRE( In.GetRow() == Out.GetRow() );

	register T * src_data_ptr = In.GetImageData();
	register T * dst_data_ptr = Out.GetImageData();
	register int data_num = In.GetElems();

	while( data_num -- != 0 )
		* dst_data_ptr ++ = * src_data_ptr ++ < Thresh1 ? GrayVal1 : GrayVal2;
}


///////////////////////////////////////////////////////////
// This function returns a minimum value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image	
//		
// OUTPUT:
//			the smallest value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
T GetImageMinValue( const TImageFor< T > & in )
{
	register T * src_data_ptr = in.GetImageData();
	register int data_num = in.GetElems();

	T retVal = in.GetPixel( 0, 0 );	
	T pixVal;
	while( data_num -- != 0 )
		if( ( pixVal = * src_data_ptr ++ ) < retVal )
			retVal = pixVal;

	return retVal;
}

///////////////////////////////////////////////////////////
// This function returns a minimum value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image	
//			position - on exit contains a position of
//				a pixel with minimal value
//		
// OUTPUT:
//			the smallest value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
T GetImageMinValue( const TImageFor< T > & in, Index & position )
{
	register T * src_data_ptr = in.GetImageData();
	register int data_num = in.GetElems();

	REQUIRE( data_num > 0 );
	if( in.GetCol() == 0 || in.GetRow() == 0 )
		//throw "Wrong image dimensions.";
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );

	T retVal = in.GetPixel( 0, 0 );	
	T pixVal;
	int offset( 0 );
	while( data_num -- != 0 )
		if( ( pixVal = * src_data_ptr ++ ) < retVal )
			retVal = pixVal, offset = data_num;

	offset = in.GetElems() - ( offset + 1 );				// now offset is from the beginning
	position.fRow = offset / in.GetCol();
	position.fCol = offset - position.fRow * in.GetCol();

	return retVal;
}


///////////////////////////////////////////////////////////
// This function returns a maximum value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image	
//		
// OUTPUT:
//			the biggest value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
T GetImageMaxValue( const TImageFor< T > & in )
{
	register T * src_data_ptr = in.GetImageData();
	register int data_num = in.GetElems();

	T retVal = in.GetPixel( 0, 0 );	
	T pixVal;
	while( data_num -- != 0 )
		if( ( pixVal = * src_data_ptr ++ ) > retVal )
			retVal = pixVal;

	return retVal;
}

///////////////////////////////////////////////////////////
// This function returns a maximum value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image	
//			position - on exit contains a position of
//				a pixel with maximal value
//		
// OUTPUT:
//			the greatest value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
T GetImageMaxValue( const TImageFor< T > & in, Index & position )
{
	register T * src_data_ptr = in.GetImageData();
	register int data_num = in.GetElems();

	REQUIRE( data_num > 0 );
	if( in.GetCol() == 0 || in.GetRow() == 0 )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );

	T retVal = in.GetPixel( 0, 0 );	
	T pixVal;
	int offset( 0 );
	while( data_num -- != 0 )
		if( ( pixVal = * src_data_ptr ++ ) > retVal )
			retVal = pixVal, offset = data_num;

	offset = in.GetElems() - ( offset + 1 );				// now offset is from the beginning
	position.fRow = offset / in.GetCol();
	position.fCol = offset - position.fRow * in.GetCol();

	return retVal;
}

///////////////////////////////////////////////////////////
// This function returns a mean value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image	
//		
// OUTPUT:
//			the mean value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
double GetImageMeanValue( const TImageFor< T > & in )
{
	register T * src_data_ptr = in.GetImageData();
	register int data_num = in.GetElems();

	REQUIRE( data_num > 0 );
	if( data_num <= 0 )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );

	double sum = 0.0;
	while( data_num -- != 0 )
		sum += (double)* src_data_ptr ++;

	return sum / (double)in.GetElems();
}

///////////////////////////////////////////////////////////
// This function returns a mean value of a pixel from image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			binaryMask - an image with binary values
//				defining places that have to be counted
//				(denoted with consider_pixels_value),
//				all remaining will be omitted from calculations
//			consider_pixels_value - only pixels with this value
//				are taken into consideration
//		
// OUTPUT:
//			the mean value from all pixel values of an image	
//		
// REMARKS:
//		
//		
template< class T >
double GetImageMeanValue( const TImageFor< T > & in, const MonochromeImage & binaryMask, const MonoPixel_Type consider_pixels_value, unsigned long & consider_pixels_counter )
{
	const int row = in.GetRow();
	const int col = in.GetCol();

	REQUIRE( binaryMask.GetRow() == row && binaryMask.GetCol() == col );
	if( ! ( binaryMask.GetRow() == row && binaryMask.GetCol() == col ) )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );

	REQUIRE( row != 0 && col != 0 )
	if( row == 0 || col == 0 )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );

	
	register T * src_data_ptr = in.GetImageData();
	register MonochromeImage::PixelType * mask_data_ptr = binaryMask.GetImageData();;
	register int data_num = in.GetElems();

	double sum = 0.0;
	consider_pixels_counter = 0;

	while( data_num -- != 0 )
		if( * mask_data_ptr ++ == consider_pixels_value )
			sum += (double)* src_data_ptr ++, consider_pixels_counter ++;

	return consider_pixels_counter == 0 ? 0.0 : sum / (double)consider_pixels_counter;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This function takes two images, input and output, and
// fills the latter with the former median filtered.  
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - a reference to the input image
//			Out - a reference to the output image that will
//					receive an input image median filtered
//			mask_horz_size - horizontal dimension of the median
//					mask (assumed to be odd)
//			mask_vert_size - vertical dimension of the median
//					mask (assumed to be odd)
//		
// OUTPUT:
//			the Out image is filled with the filtered In image	
//		
// REMARKS:
//			If either of the mask size is even it is automatically
//			exchanged to the nearest odd value, i.e. given x and x
//			is even then x += 1. 	
//
//			For this filter we assume a Moore neighborhood of pixels.
//		
template < class T > 
void MedianFilter( const TImageFor< T > & In, Dimension mask_horz_size, Dimension mask_vert_size, TImageFor< T > & Out )
{
	// Invariant at first
	REQUIRE( In.GetCol() <= Out.GetCol() );
	REQUIRE( In.GetRow() <= Out.GetRow() );

	REQUIRE( mask_horz_size <= In.GetCol() );
	REQUIRE( mask_vert_size <= In.GetRow() );

	register Dimension i, j, k, l, m;
	

	const Dimension kCol = In.GetCol();
	const Dimension kRow = In.GetRow();

	const Dimension dispX = mask_horz_size / 2;
	const Dimension dispY = mask_vert_size / 2;

	// If supplied with an even number for a mask, then it will be extanded to 
	// the nearest odd value, e.g. if mask_horz_size is 4 then the actual mask
	// window will have 5 pixels of horizontal extend.
	const Dimension kNeighbourElems = ( 2 * dispX + 1 ) * ( 2 * dispY + 1 );
	const Dimension kHalfElems = kNeighbourElems / 2;


	vector< T >	dataGatherVec( kNeighbourElems );
	REQUIRE( dataGatherVec.size() == kNeighbourElems );

	// Go through all the points of the input picture
	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			// Now, for a given central pixel, gather all its neighbour pixels

			Dimension col_start_from = j > dispX ? j - dispX : 0;
			Dimension col_end_to = j + dispX < kCol ? j + dispX : kCol - 1;

			Dimension row_start_from = i > dispY ? i - dispY : 0;
			Dimension row_end_to = i + dispY < kRow ? i + dispY : kRow - 1;

			m = 0;		// restart the element index

			for( k = col_start_from; k <= col_end_to; k ++ )
			{
				for( l = row_start_from; l <= row_end_to; l ++ )
				{
					dataGatherVec[ m ++ ] = In.GetPixel( k, l );
				}

			}
			
			REQUIRE( dataGatherVec.size() == kNeighbourElems );

			// Find the median (or other value, such as max or min)
			nth_element( dataGatherVec.begin(), dataGatherVec.begin() + kHalfElems, dataGatherVec.end() );

/*			REQUIRE(	kNeighbourElems <= 1 ||
						dataGatherVec[ kHalfElems ] >= dataGatherVec[ kHalfElems - 1 ] &&
						dataGatherVec[ kHalfElems ] <= dataGatherVec[ kHalfElems + 1 ]
				);
*/
			Out.SetPixel( j, i, dataGatherVec[ kHalfElems ] );

		}

	}

}

// In the following there is a specialization of the 
// median filter for vector valued pixels
#include "VectorRankFilter.h"



#if ! MODERN_COMPILER		// the following overrides are used because of the flawed compiler
							// that cannot create a partially specialized template (in Median_Filter_OperationFor)

// An overriden version
template < class T > 
void MedianFilter( const TImageFor< T > & In, Index mask_size, TImageFor< T > & Out )
{
	// We have to cast, otherwise the scalar version will be chosen since template match
	// is preferred over the casting.
	MedianFilter( In, (int)mask_size.fCol, (int)mask_size.fRow, Out );
}


#endif MODERN_COMPILER

// An overriden version
template < class T, class U > 
void MedianFilter( const TImageFor< T > & In, const TImageFor< U > & mask_size, TImageFor< T > & Out )
{
	MedianFilter( In, mask_size.GetPixel( 0, 0 ), mask_size.GetPixel( 0, 1 ), Out );
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function changes all image values to the corresponding
// absolute values.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - a reference to the input image		
//		
// OUTPUT:
//			the input image has been "absolute valued"		
//		
// REMARKS:
//		
//		
template < class T > 
void ImageAbsoluteValue( TImageFor< T > & image )
{
	register T * src_data_ptr					= image.GetImageData();
	register T * one_after_last_src_data_ptr	= src_data_ptr + image.GetElems();

	for( ; src_data_ptr < one_after_last_src_data_ptr; src_data_ptr ++ )
	{
		T thePixel = * src_data_ptr;
		if( thePixel < 0 )
			* src_data_ptr = - thePixel;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function does mapping of the input image, which pixel
// values fall in range [a,b] into an image which pixel
// values fall in a given range [lowerBound, upperBound].
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - a reference to the image to be changed
//			lowerBound - first value of the new image range
//			upperBound - second value of the new image range
//			(no relation is assumed between these two values)	
//		
// OUTPUT:
//			in image has changed its pixels value range.		
//		
// REMARKS:
//		
//		
template< class T >
void ChangeImageRange( TImageFor< T > & in, T lowerBound, T upperBound )
{
	double minImageVal = (double)GetImageMinValue( in );
	double maxImageVal = (double)GetImageMaxValue( in );

	// Make the right order.
	if( lowerBound > upperBound )
	{
		T temp = upperBound;
		upperBound = lowerBound;
		lowerBound = temp;
	}

	double delta = upperBound - lowerBound;

	if( minImageVal == maxImageVal )
	{
		in.SetAll( (T)((double)lowerBound + delta / 2.0) );
		return;
	}

	double scaleCoeff = delta / ( maxImageVal - minImageVal );

	register T * src_data_ptr = in.GetImageData();
	for( register int data_num = in.GetElems(); data_num -- != 0; ++ src_data_ptr )
		* src_data_ptr = lowerBound + ( (double)* src_data_ptr - minImageVal ) * scaleCoeff;
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////


#endif GraphicLibrary_h


