
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: Morphology.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of image morphology.
//
// ==========================================================================



#ifndef BoxySoft_Morphology
#define BoxySoft_Morphology


#include "HIL_MaskedImage.h"
#include "HIL_AutoPtrDefs.h"



using namespace PHIL;




///////////////////////////////////////////////////////////
//
// The class for morphological operations on different
// type images. 
//
// This is a general implementation that takes into account
// the grey valued definition of the morphological 
// operations. This means that the presented functions
// can operate with almost any types of pixels for which
// the order relation has been defined. However, this
// means also that the implementation might not be optimal
// in many cases (e.g. for pure binary images).
//
// The structural element is always defined as the masked
// image. This means that a shape of the structural element
// is defined by a binary mask inherent to the masked image, 
// whereas the pixel values of the element are taken into
// computations of the morphological functions. Certainly,
// if the pixels are set to 0 then only a shape of the 
// structural element is taken into consideration.
//
// Optimization issues: the most important optimization can
// be achieved by virtue of SEPARABLE structural elements
// (the concept similar to the separable convolution masks).
// For more information see e.g. the book by Soille.
//
///////////////////////////////////////////////////////////
template < class T >
class MorphologyFor
{
	public:

		typedef std::auto_ptr< TImageFor< T > > TheImageAutoPtr;

	public:

		// ===================================================
		MorphologyFor( void ) {}			// class default constructor
		virtual ~MorphologyFor() {}			// class virtual destructor
		// ===================================================

	public:


		///////////////////////////////////////////////////////////
		// This is morphological dilation on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains a dilated version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr Dilate( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			register long i, j, m, n, vertMaskIndex, horzMaskIndex, imTempCol, imTempRow;

			const long im_col = theImage.GetCol();
			const long im_row = theImage.GetRow();

			const long mask_col = theStructureElement.GetCol();
			const long mask_row = theStructureElement.GetRow();

			// We don't assume any specific size or center of the structural element here,
			// although most often than not it is symmetrical with geometrical center 
			// being a pivoting point.

			long horzHalfMask = mask_col / 2;
			long vertHalfMask = mask_row / 2; 

			// This is an output image in its initial state (all pixels 0)
			TheImageAutoPtr outImage( new TImageFor< T >( im_col, im_row ) );
			
			bool row_is_ok, mask_index_in_range;

			// Go through all the points of the input picture
			for( i = 0; i < im_row; i ++ )
			{
				for( j = 0; j < im_col; j ++ )
				{
					T maxVal( theImage.GetPixel( j, i ) );

					for( m = - vertHalfMask; m <= + vertHalfMask; m ++ )
					{
						vertMaskIndex = m + vertHalfMask;

						imTempRow = i - m;

						row_is_ok = imTempRow >= 0 && imTempRow < im_row;
						mask_index_in_range = vertMaskIndex < mask_row;

						for( n = - horzHalfMask; n <= + horzHalfMask; n ++ )
						{
							imTempCol = j - n;

							horzMaskIndex = n + horzHalfMask;

							if( imTempCol >= 0 && imTempCol < im_col && row_is_ok/*imTempRow >= 0 && imTempRow < im_row*/ 		
								// >>> This part of AND is necessary only if size of the mask is even <<<
								&& horzMaskIndex < mask_col && mask_index_in_range/*vertMaskIndex < mask_row*/ )
							{
								bool is_in_structure;
								T pixel = theStructureElement.GetPixel( horzMaskIndex, vertMaskIndex, is_in_structure );

								if( is_in_structure == true )
								{
									pixel += theImage.GetPixel( imTempCol, imTempRow );
									if( pixel > maxVal )
										maxVal = pixel;			// store the maximal value as so far
								}
							}

						}
					}

					outImage->SetPixel( j, i, maxVal );
				}

			}

			return outImage;
		}


		///////////////////////////////////////////////////////////
		// This is morphological erosion on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains an eroded version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr Erode( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			register long i, j, m, n, vertMaskIndex, horzMaskIndex, imTempCol, imTempRow;

			const long im_col = theImage.GetCol();
			const long im_row = theImage.GetRow();

			const long mask_col = theStructureElement.GetCol();
			const long mask_row = theStructureElement.GetRow();

			// We don't assume any specific size or center of the structural element here,
			// although most often than not it is symmetrical with geometrical center 
			// being a pivoting point.

			long horzHalfMask = mask_col / 2;
			long vertHalfMask = mask_row / 2; 

			// This is an output image in its initial state (all pixels 0)
			TheImageAutoPtr outImage( new TImageFor< T >( im_col, im_row ) );		

			bool row_is_ok, mask_index_in_range;

			// Go through all the points of the input picture
			for( i = 0; i < im_row; i ++ )
			{
				for( j = 0; j < im_col; j ++ )
				{
					T minVal( theImage.GetPixel( j, i ) );

					for( m = - vertHalfMask; m <= + vertHalfMask; m ++ )
					{
						vertMaskIndex = m + vertHalfMask;

						imTempRow = i + m;

						row_is_ok = imTempRow >= 0 && imTempRow < im_row;

						mask_index_in_range = vertMaskIndex < mask_row;

						for( n = - horzHalfMask; n <= + horzHalfMask; n ++ )
						{
							imTempCol = j + n;

							horzMaskIndex = n + horzHalfMask;

							if( imTempCol >= 0 && imTempCol < im_col && row_is_ok/*imTempRow >= 0 && imTempRow < im_row*/ 		
								// >>> This part of AND is necessary only if size of the mask is even <<<
								&& horzMaskIndex < mask_col && mask_index_in_range/*vertMaskIndex < mask_row*/ )
							{
								bool is_in_structure;
								T pixel = theStructureElement.GetPixel( horzMaskIndex, vertMaskIndex, is_in_structure );

								if( is_in_structure == true )
								{
									pixel = theImage.GetPixel( imTempCol, imTempRow ) - pixel;
									if( pixel < minVal )
										minVal = pixel;			// store the min value as so far
								}
							}

						}
					}

					outImage->SetPixel( j, i, minVal );
				}

			}

			return outImage;
		}


		///////////////////////////////////////////////////////////
		// This is morphological open on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains an opened version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr Open( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			// The morpological opening is erosion followed by dilation with the same structural element
			TheImageAutoPtr erodedImage( Erode( theImage, theStructureElement ) );
			return Dilate( * erodedImage, theStructureElement );
		}


		///////////////////////////////////////////////////////////
		// This is morphological close on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains a close version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr Close( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			// The morpological opening is erosion followed by dilation with the same structural element
			TheImageAutoPtr dilatedImage( Dilate( theImage, theStructureElement ) );
			return Erode( * dilatedImage, theStructureElement );
		}

		///////////////////////////////////////////////////////////
		// This is morphological gradient on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains a gradient version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr Gradient( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			// The morpological opening is erosion followed by dilation with the same structural element
			TheImageAutoPtr dilatedImage( Dilate( theImage, theStructureElement ) );
			TheImageAutoPtr erodedImage( Erode( theImage, theStructureElement ) );

			* dilatedImage -= * erodedImage;
			return dilatedImage;
		}

		///////////////////////////////////////////////////////////
		// This is morphological top-hat on any pixel type input
		// image. The structural element is in the form of a masked
		// image where the mask denotes which pixel belongs to the 
		// that element and the values are taken to the computation.
		// The geometrical center of the sctructural element is taken
		// as the central point.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theImage - the input image with pixels of type T
		//			theStructureElement - the structural element with
		//				pixels of type T and binary mask defining
		//				shape of this element
		//		
		// OUTPUT:
		//			Auto ptr to the returned (orphaned) object which
		//			contains a top-hat version of the input image
		//		
		// REMARKS:
		//			The returned object is orphaned!
		//		
		TheImageAutoPtr TopHat( const TImageFor< T > & theImage, const TMaskedImageFor< T > & theStructureElement )
		{
			// The morpological opening is erosion followed by dilation with the same structural element
			TheImageAutoPtr openImage( Open( theImage, theStructureElement ) );
			TheImageAutoPtr retImage( new TImageFor< T >( theImage ) );

			* retImage -= * openImage;
			return retImage;
		}


};





#endif BoxySoft_Morphology




