
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: NonParametricTransformations.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of non-parametric transformations of images.
//
// ==========================================================================



#ifndef BoxySoft_NonParametricTransformations
#define BoxySoft_NonParametricTransformations


#include "HIL_ImageFor.h"
#include <bitset>


using namespace PHIL;




// ==========================================================================
template < typename PIX, int BITS_PER_ENCODING = 3 >
class Census_BitEncodingPolicy
{

	public:


		///////////////////////////////////////////////////////////
		// This function returns non parametric bit encoding
		// for a given pair of pixels.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			pixel - a given pixel to be encoded (a neighbour)
		//			ref_pixel - a reference pixel
		//		
		// OUTPUT:
		//			Bit encoding up to a length of "unsigned long"
		//			which is usually 32 bits.
		//
		//			In this case it is bit "1" if pixel value i greater
		//			than the reference or "0" otherwise.
		//		
		// REMARKS:
		//		
		//		
		unsigned long GetBitEncoding( const PIX pixel, const PIX ref_pixel )
		{
			return pixel > ref_pixel ? 1 : 0;
		}

};

template < typename PIX, int BITS_PER_ENCODING = 3 >
class Precise3BitCensus_BitEncodingPolicy
{

	public:


		///////////////////////////////////////////////////////////
		// This function returns non parametric bit encoding
		// for a given pair of pixels.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			pixel - a given pixel to be encoded (a neighbour)
		//			ref_pixel - a reference pixel
		//		
		// OUTPUT:
		//			Bit encoding up to a length of "unsigned long"
		//			which is usually 32 bits.
		//
		// REMARKS:
		//		
		//			In this case we have special 3-bits "fuzzy" encoding
		//			as follows:
		//
		//
		//				|	Relation type S(k)		|	Bit encoding E(k)
		//			=========================================
		//			1	| strongly 		smaller		|		011
		//			2	|				smaller		|		010
		//			3	| little 		smaller		|		001
		//			4	|				equal		|		000
		//			5	| little		greater		|		101
		//			6	|				greater		|		110
		//			7	| strongly		greater		|		111
		//
		//
		//				Bit number	|		2		|	  1		|	  0
		//				============================================================================
		//				Bit meaning	|      sign	   Difference category, bit 1	Difference category, bit 0
		//
		//				If |R|  in  [0,3] then S(k) = "equal". 
		//				If |R|  in  (3, 15] then S(k) = "little smaller/greater", 
		//				If |R|  in  (15, 63] then S(k) = "smaller/greater", 
		//				If |R|  in  64 then we classify as S(k) = "strongly smaller/greater". 
		//
		//		
		//		
		unsigned long GetBitEncoding( const PIX pixel, const PIX ref_pixel )
		{
			long R = pixel - ref_pixel;
			unsigned long E = 0x00;

			if( R < 0 ) 
			{
			  R = -R;
			  E = 0x01 << BITS_PER_ENCODING - 1;
			}

			const int kRightShift = 2;			// just for this algorithm (see my paper)
			for( int i = 0; i < BITS_PER_ENCODING; ++ i )
			  if( ( R >>= kRightShift ) != 0 )
				++ E;
			  else
				break;

			return E;
		}

};


// ==========================================================================



template < typename PIX = unsigned char, int WINDOW_SIZE = 3, int BITS_PER_ENCODING = 3, class BitEncodingPolicy = Census_BitEncodingPolicy< PIX, BITS_PER_ENCODING > >
class TNeighborRelations_NonParam : public BitEncodingPolicy
{
	protected:

		enum { kNonParamWindow = WINDOW_SIZE, kNonParamBits = kNonParamWindow * kNonParamWindow };
		enum { kBitsPerSingleEncoding = BITS_PER_ENCODING };

	

	public:

		typedef std::bitset< ( kNonParamBits - 1 ) * kBitsPerSingleEncoding > _NonParam;	// - 1 because the central pixel value is not counted
		typedef TImageFor< _NonParam > _NonParamImage;

		typedef TImageFor< PIX > _InputImageType;

	protected:

		const _InputImageType & fInImage;

	public:

		TNeighborRelations_NonParam( const _InputImageType & inImage )
			: fInImage( inImage ) {}

	public:

		_NonParamImage * OrphanNonParamImage( const TImageFor< PIX > & image )
		{
 			register Dimension i, j;

			Dimension row = image.GetRow();
			Dimension col = image.GetCol();

			_NonParamImage * nonParamImage = new _NonParamImage( col, row );
			nonParamImage -> SetAll( 0 );


			const int win_swing = kNonParamWindow / 2;

			int h_from, h_to, v_from, v_to, bit_pos;
			register int m, n; 

			unsigned char central_pixel;

 			// For each pixel set its census
			for( i = 0; i < row; i ++ )
			{
				v_from = ( i >= win_swing ? i - win_swing : 0 );
				v_to = ( i + win_swing >= row ? row - 1 : i + win_swing );
				
				for( j = 0; j < col; j ++ )
				{

					// Now move in the census window
					h_from = ( j >= win_swing ? j - win_swing : 0 );
					h_to = ( j + win_swing >= col ? col - 1 : j + win_swing );

					_NonParam & pixelNonParam = nonParamImage->GetRefPixel( j, i );
					bit_pos = 0;

					central_pixel = image.GetPixel( j, i );

					for( m = h_from; m <= h_to; m ++ )
					{
						for( n = v_from; n <= v_to; n ++ )
						{
							if( m == j && n == i )
								continue;	// skip the central pixel from the census window

							// THE BIGGEST DIFFERENCE IS HERE: we assing '0' if the pixels
							// are the same. The key idea behing this change is that 
							// the Tanimoto distance prefers matching of '1s'. That is, having
							// two matches, one with zeros, the second wiht ones, the second
							// will be given higher score in a sense of the Tanimoto distance.
//							if( image.GetPixel( m, n ) >/*=*/ central_pixel )
//								pixelNonParam.set( bit_pos ++ );	// set bit if beyond the threshold
//							else
//								pixelNonParam.reset( bit_pos ++ );

				
							unsigned long theEncoding = GetBitEncoding( image.GetPixel( m, n ) , central_pixel );
							for( int k = 0; k < BITS_PER_ENCODING; ++ k, theEncoding >>= 1 )
							{
								if( ( theEncoding & 0x01 ) != 0 )
									pixelNonParam.set( bit_pos ++ );
								else
									pixelNonParam.reset( bit_pos ++ );
							}
						
						}
					}
				
					REQUIRE( bit_pos <= ( kNonParamBits - 1 ) * kBitsPerSingleEncoding );
				}

			}

			return nonParamImage;
		}


		// Helper function
		template < typename T >
		_NonParamImage * OrphanNonParamImage( const TImageFor< PIX > & image, TImageFor< T > & out_image )
		{
			// We require that the output monochrome pixel is able to store required num of bits per pixel
			REQUIRE( ( kNonParamBits - 1 ) * kBitsPerSingleEncoding <= 8 * sizeof( MonoPixel_Type ) );

			REQUIRE( out_image.GetCol() == image.GetCol() && out_image.GetRow() == image.GetRow() );
			if( ! ( out_image.GetCol() == image.GetCol() && out_image.GetRow() == image.GetRow() ) )
				return 0;			// wron input parameters


			_NonParamImage * nonParamImage = OrphanNonParamImage( image );


			Dimension row = image.GetRow();
			Dimension col = image.GetCol();


			register _NonParam * src_data_ptr					= nonParamImage->GetImageData();
			register _NonParam * one_after_last_src_data_ptr	= src_data_ptr + nonParamImage->GetElems();

			register T * dst_data_ptr = out_image.GetImageData();

			while( src_data_ptr < one_after_last_src_data_ptr )
				* dst_data_ptr ++ = (* src_data_ptr ++).to_ulong();

			return nonParamImage;
		}


		template < typename T >
		void Convert( TImageFor< T > & out_image )
		{
			// We require that the output monochrome pixel is able to store required num of bits per pixel
			REQUIRE( ( kNonParamBits - 1 ) * kBitsPerSingleEncoding <= 8 * sizeof( MonoPixel_Type ) );

			REQUIRE( out_image.GetCol() == fInImage.GetCol() && out_image.GetRow() == fInImage.GetRow() );
			if( ! ( out_image.GetCol() == fInImage.GetCol() && out_image.GetRow() == fInImage.GetRow() ) )
				return;			// wrong input parameters


			Dimension row = fInImage.GetRow();
			Dimension col = fInImage.GetCol();

			_NonParamImage * nonParamImage = OrphanNonParamImage( fInImage );


			register _NonParam * src_data_ptr					= nonParamImage->GetImageData();
			register _NonParam * one_after_last_src_data_ptr	= src_data_ptr + nonParamImage->GetElems();

			register T * dst_data_ptr = out_image.GetImageData();

			while( src_data_ptr < one_after_last_src_data_ptr )
				* dst_data_ptr ++ = (* src_data_ptr ++).to_ulong();

		}


		template <>
		void Convert( TImageFor< _NonParam > & out_image )
		{
			_NonParamImage * nonParamImage = OrphanNonParamImage( fInImage );
			if( nonParamImage != 0 )
				out_image = * nonParamImage;
			delete nonParamImage;
		}


};

// ==========================================================================

typedef TNeighborRelations_NonParam< MonoPixel_Type, 3, 1 > Census_3x3;
typedef TNeighborRelations_NonParam< MonoPixel_Type, 5, 1 > Census_5x5;
typedef TNeighborRelations_NonParam< MonoPixel_Type, 7, 1 > Census_7x7;

typedef TNeighborRelations_NonParam< MonoPixel_Type, 3, 3, Precise3BitCensus_BitEncodingPolicy< MonoPixel_Type, 3 > > PrecCensus_3x3;
typedef TNeighborRelations_NonParam< MonoPixel_Type, 5, 3, Precise3BitCensus_BitEncodingPolicy< MonoPixel_Type, 3 > > PrecCensus_5x5;
typedef TNeighborRelations_NonParam< MonoPixel_Type, 7, 3, Precise3BitCensus_BitEncodingPolicy< MonoPixel_Type, 3 > > PrecCensus_7x7;

// ==========================================================================





#endif BoxySoft_NonParametricTransformations




