
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: SavitzkyGolayFilter.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Savitzky-Golay filters.	
//
// ==========================================================================



#ifndef BoxySoft_SavitzkyGolayFilter
#define BoxySoft_SavitzkyGolayFilter




#include "ImageMatrix.h"



// =========================================================================================

// Implementation of the generator
// of coefficients of the Savitzky-Golay filter.
// Once computed, the coefficients can be used
// in the convolution operator.

class TSavitzkyGolayFilter
{



	public:


		///////////////////////////////////////////////////////////
		// This function computes a power of integer values: x^a
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			x - argument value
		//			a - power value
		//		
		// OUTPUT:
		//			x ^ a
		//		
		// REMARKS:
		//		
		//		
		long Power( long x, long a )
		{
			register long retVal = 1;
			while( a -- > 0 )
				retVal *= x;
			return retVal;
		}


		///////////////////////////////////////////////////////////
		// This function orphanes a coordinate matrix for the
		// Savitzky-Golay filter. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			kN - degree of the interpolating polynomial
		//			kIndexFrom - number of samples to the left 
		//				from the central one
		//			kIndexTo - number of samples from to the right 
		//				from the central one
		//		
		// OUTPUT:
		//			Orphaned image - a matrix P
		//		
		// REMARKS:
		//			The returned object has to be deleted by 
		//			a calling part.
		//		
		LongImage * Generate_SavGol_2D_Coordinate_Matrix( const long kN, const long kIndexFrom, const long kIndexTo )
		{		
			const long kElemsInRow = ( ( kN + 1 ) * ( kN + 2 ) ) / 2;
			const long kNumOfRows = ( kIndexTo - kIndexFrom + 1 ) * ( kIndexTo - kIndexFrom + 1 );
			register long i, j, a, b;

			//                                           cols         rows
			LongImage * indexMatrix = new LongImage( kElemsInRow, kNumOfRows, 0.0 );

			int row_counter = 0;

			//////////////////////////////////////////////////////////////////////
			// These loops run through all possible indexes in the sample window W
			// i runs horizontally
			// j runs vertically
			for( j = kIndexFrom; j <= kIndexTo; ++ j )
			{
				for( i = kIndexFrom; i <= kIndexTo; ++ i )
				{
					///////////////////////////////////////////
					// These two loops generate a single row Pk
					int col_counter = 0;

					for( a = 0; a <= kN; ++ a )
					{
						for( b = 0; b <= kN; ++ b )
						{
							if( a + b <= kN )
							{
								long theElement = Power( i, a ) * Power( j, b );
								indexMatrix->SetPixel( col_counter ++, row_counter, theElement );
							}
						}
					}

					REQUIRE( col_counter == kElemsInRow );
					///////////////////////////////////////////

					++ row_counter;		// changing any value of i or j moves us to a new row
				}

			}
			//////////////////////////////////////////////////////////////////////

			REQUIRE( row_counter == kNumOfRows );

			return indexMatrix;
		}



		///////////////////////////////////////////////////////////
		// This function orphanes coefficients of the Savitzky-Golay 
		// filter. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			kN - degree of the interpolating polynomial
		//			kIndexFrom - number of samples to the left 
		//				from the central one
		//			kIndexTo - number of samples from to the right 
		//				from the central one
		//		
		// OUTPUT:
		//			Orphaned image - a matrix ~P
		//		
		// REMARKS:
		//			The returned object has to be deleted by 
		//			a calling part.
		//		
		TRealImage * Compute_SavGol_Filer( const long kN, const long kIndexFrom, const long kIndexTo )
		{
			LongImage * SG_CoordMatrix = Generate_SavGol_2D_Coordinate_Matrix( kN, kIndexFrom, kIndexTo );

			REQUIRE( SG_CoordMatrix != 0 );

			TRealImage tmp_SG_CoordMatrix( * SG_CoordMatrix );

			Save_ASCII_Image( * SG_CoordMatrix, "Matrix_P.txt", ", " );
			delete SG_CoordMatrix;

			TRealImage * SG_FilterCoeffs = Orphan_PseudoInv_Matrix< double, double >( tmp_SG_CoordMatrix );

			if( SG_FilterCoeffs != 0 )
				Save_ASCII_Image( * SG_FilterCoeffs, "Matrix_PseudoInv_P.txt", ", " );

			return SG_FilterCoeffs;
		}


};


// =========================================================================================



#endif BoxySoft_SavitzkyGolayFilter




