
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: VectorRankFilter.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Vector rank filters.	
//				Implemented in accordance with:
//
//					Cyganek B.: Computational Framework for Family of Order Statistic 
//					Filters for Tensor Valued Data. 
//					Lecture Notes in Computer Science, LNCS 4141. 
//					Springer, 2006, pp. 156-162.
//
// ==========================================================================



#ifndef BoxySoft_VectorRankFilter
#define BoxySoft_VectorRankFilter



#include "HIL_ImageFor.h"
#include "HIL_MultiPixelFor.h"
#include <math.h>


using namespace PHIL;



// =================================================================================================

// The squared Frobenius norm
template < typename T, int C >	
double Metric( MMultiPixelFor< T > & p1, MMultiPixelFor< T > & p2 )
{
	double tmp = 0.0, theMetric = 0.0;
	for( int i = 0; i < C; ++ i )
	{
		tmp = p1.fData[ i ] - p2.fData[ i ];
		theMetric += tmp * tmp;
	}

	return theMetric;
}


// =================================================================================================

///////////////////////////////////////////////////////////
// This function takes two images, input and output, and
// fills the latter with the former median filtered.  
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - a reference to the input image
//			Out - a reference to the output image that will
//					receive an input image median filtered
//			mask_horz_size - horizontal dimension of the median
//					mask (assumed to be odd)
//			mask_vert_size - vertical dimension of the median
//					mask (assumed to be odd)
//		
// OUTPUT:
//			the Out image is filled with the filtered In image	
//		
// REMARKS:
//			If either of the mask size is even it is automatically
//			exchanged to the nearest odd value, i.e. given x and x
//			is even then x += 1. 		
//	
template < typename T, int C >	

#if MODERN_COMPILER		// the older compilers cannot deal with partial specialization

	void MedianFilter( const TImageFor< MMultiPixelFor< T, C > > & In, Dimension mask_horz_size, Dimension mask_vert_size, TImageFor< MMultiPixelFor< T, C > > & Out )

#else MODERN_COMPILER

	void MedianFilter( const TImageFor< MMultiPixelFor< T, C > > & In, int mask_horz_size, int mask_vert_size, TImageFor< MMultiPixelFor< T, C > > & Out )

#endif MODERN_COMPILER
{ 
	// Invariant at first
	REQUIRE( In.GetCol() <= Out.GetCol() );
	REQUIRE( In.GetRow() <= Out.GetRow() );

	REQUIRE( mask_horz_size < In.GetCol() );
	REQUIRE( mask_vert_size < In.GetRow() );

	typedef MMultiPixelFor< T, C > MPIX;

	register Dimension i, j, k, l, current_elems = 0, prev_current_elems = 0;
	

	const Dimension kCol = In.GetCol();
	const Dimension kRow = In.GetRow();

	const Dimension dispX = mask_horz_size / 2;
	const Dimension dispY = mask_vert_size / 2;

	// If supplied with an even number for a mask, then it will be extanded to 
	// the nearest odd value, e.g. if mask_horz_size is 4 then the actual mask
	// window will have 5 pixels of horizontal extend.

	const Dimension kMaxMutualDistMatrix_Size_X = dispX * 2 + 1;
	const Dimension kMaxMutualDistMatrix_Size_Y = dispY * 2 + 1;
	const Dimension kMaxMutualDistMatrix_Size = kMaxMutualDistMatrix_Size_X * kMaxMutualDistMatrix_Size_Y;

	TRealImage bi_metrics( kMaxMutualDistMatrix_Size, kMaxMutualDistMatrix_Size, 0.0 );


	vector< MPIX >	dataGatherVec;



	// Go through all the points of the input picture
	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			// Now, for a given central pixel, gather all its neighbour pixels

			Dimension col_start_from = j > dispX ? j - dispX : 0;
			Dimension col_end_to = j + dispX < kCol ? j + dispX : kCol - 1;

			Dimension row_start_from = i > dispY ? i - dispY : 0;
			Dimension row_end_to = i + dispY < kRow ? i + dispY : kRow - 1;


			Dimension one_column_elems = row_end_to - row_start_from + 1;		
			Dimension one_row_elems = col_end_to - col_start_from + 1;		
			current_elems = one_row_elems * one_column_elems;


			// Now we compute all the 1/2 * N * (N-1) comparisons of pairs of vectors.
			// From the definition of a metric it follows that d( x, x ) = 0
			// We also use the metric feature: d( x, y ) = d( y, x )

			// If the "false &&" is ON, then there is no acceleration (i.e. brute force run)
			if( /*false && */prev_current_elems == kMaxMutualDistMatrix_Size && current_elems == kMaxMutualDistMatrix_Size )
			{
				Dimension p, q;

				// In this case we can reuse the distances already computed in the previous step

				// one_column_elems is an amount of pixels that we have to refill
				REQUIRE( current_elems == dataGatherVec.size() );
				//copy( dataGatherVec.begin() + one_column_elems, dataGatherVec.end(), dataGatherVec.begin() );
				p = 0;
				for( k = one_column_elems; k < current_elems; ++ k )
					dataGatherVec[ p ++ ] = dataGatherVec[ k ];
				REQUIRE( p == current_elems - one_column_elems );
				/////

				//p = current_elems - one_column_elems;
				for( l = row_start_from; l <= row_end_to; l ++ )
					dataGatherVec[ p ++ ] = In.GetPixel( col_end_to, l );


				// Now special copy the mutual distance matrix
				for( k = one_column_elems, p = 0; k < current_elems - 1; ++ k, ++ p )
				{
					for( l = k + 1, q = p + 1; l < current_elems; ++ l, ++ q )
					{
						bi_metrics.SetPixel( q, p, bi_metrics.GetPixel( l, k ) );	
					}
				}

				// Fill the new positions in the mutual distance matrix
				// It is convenient to go by rows
				for( l = current_elems - 1, p = 0; p < one_column_elems ; ++ p, -- l )
				{
					MPIX theReferencePixel = dataGatherVec[ l ];
					for( k = 0; k < current_elems - 1 - p; ++ k )
					{
						double theDistance = Metric< T, 3 >( theReferencePixel, dataGatherVec[ k ] );
						bi_metrics.SetPixel( l, k, theDistance );	
					}
				}


			}
			else
			{
				dataGatherVec.clear();			// clear all data
				REQUIRE( dataGatherVec.size() == 0 );

				for( k = col_start_from; k <= col_end_to; k ++ )
					for( l = row_start_from; l <= row_end_to; l ++ )
						dataGatherVec.push_back( In.GetPixel( k, l ) );
				
				REQUIRE( current_elems == dataGatherVec.size() );

				// The triangle matrix - diagonal and lower triangle set to 0.0
				bi_metrics.SetAll( 0.0 );


				for( k = 0; k < current_elems - 1; ++ k )
				{
					MPIX theReferencePixel = dataGatherVec[ k ];
					for( l = k + 1; l < current_elems; ++ l )
					{
						double theDistance = Metric< T, 3 >( theReferencePixel, dataGatherVec[ l ] );
						bi_metrics.SetPixel( l, k, theDistance );	
					}
				}
			}

			/////////////////////////


			// At this point we'll have computed all necessary comparisons among vector pairs.
			vector< double >	partial_sum_vec( current_elems );

			// Precompute the partial sums
			partial_sum_vec[ 0 ] = 0.0;
			for( k = 1; k < current_elems; ++ k )
			{
				double p_sum = 0.0;
				for( l = 0; l < k; ++ l )
					p_sum += bi_metrics.GetPixel( k, l );

				partial_sum_vec[ k ] = p_sum;
			}


			// Let's select a winner.
			Dimension theWinnerIndex = 0;
			double prev_cumulation = 100000000.0;				// something to start from

			for( k = 0; k < current_elems; ++ k )
			{
				double cumulation = partial_sum_vec[ k ];		// take the precomputed triangular value
				
				for( l = k + 1; l < current_elems; ++ l )
					cumulation += bi_metrics.GetPixel( l, k );
			
				if( cumulation < prev_cumulation )
				{
					prev_cumulation = cumulation;
					theWinnerIndex = k;
				}
			}


			REQUIRE( theWinnerIndex < dataGatherVec.size() );
			Out.SetPixel( j, i, dataGatherVec[ theWinnerIndex ] );

			prev_current_elems = current_elems;		// save for the next iteration
		}

	}

}



// =================================================================================================


///////////////////////////////////////////////////////////
// This function takes two images, input and output, 
// as well as a vector with weights, and then
// fills the latter image with the former median filtered.  
///////////////////////////////////////////////////////////
//		
// INPUT:
//			In - a reference to the input image
//			Out - a reference to the output image that will
//					receive an input image median filtered
//			mask_horz_size - horizontal dimension of the median
//					mask (assumed to be odd)
//			mask_vert_size - vertical dimension of the median
//					mask (assumed to be odd)
//			theWeights - vector of the weights for median
//				filtering
//		
// OUTPUT:
//			the Out image is filled with the filtered In image	
//		
// REMARKS:
//			If either of the mask size is even it is automatically
//			exchanged to the nearest odd value, i.e. given x and x
//			is even then x += 1. 		
//	
template < typename T, int C >	

#if MODERN_COMPILER		// the older compilers cannot deal with partial specialization

void MedianFilter(	const TImageFor< MMultiPixelFor< T, C > > & In, Dimension mask_horz_size, Dimension mask_vert_size, TImageFor< MMultiPixelFor< T, C > > & Out,
							vector< double > & theWeights )


#else MODERN_COMPILER

void MedianFilter(	const TImageFor< MMultiPixelFor< T, C > > & In, long mask_horz_size, long mask_vert_size, TImageFor< MMultiPixelFor< T, C > > & Out,
							vector< double > & theWeights )


#endif MODERN_COMPILER


{
	// Invariant at first
	REQUIRE( In.GetCol() <= Out.GetCol() );
	REQUIRE( In.GetRow() <= Out.GetRow() );

	REQUIRE( mask_horz_size < In.GetCol() );
	REQUIRE( mask_vert_size < In.GetRow() );

	typedef MMultiPixelFor< T, C > MPIX;

	register Dimension i, j, k, l, current_elems = 0, prev_current_elems = 0;
	

	const Dimension kCol = In.GetCol();
	const Dimension kRow = In.GetRow();

	const Dimension dispX = mask_horz_size / 2;
	const Dimension dispY = mask_vert_size / 2;

	// If supplied with an even number for a mask, then it will be extanded to 
	// the nearest odd value, e.g. if mask_horz_size is 4 then the actual mask
	// window will have 5 pixels of horizontal extend.

	const Dimension kMaxMutualDistMatrix_Size_X = dispX * 2 + 1;
	const Dimension kMaxMutualDistMatrix_Size_Y = dispY * 2 + 1;
	const Dimension kMaxMutualDistMatrix_Size = kMaxMutualDistMatrix_Size_X * kMaxMutualDistMatrix_Size_Y;

	TRealImage bi_metrics( kMaxMutualDistMatrix_Size, kMaxMutualDistMatrix_Size, 0.0 );


	vector< MPIX >	dataGatherVec;

	REQUIRE( theWeights.size() >= kMaxMutualDistMatrix_Size_X );

	// Go through all the points of the input picture
	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			// Now, for a given central pixel, gather all its neighbour pixels

			Dimension col_start_from = j > dispX ? j - dispX : 0;
			Dimension col_end_to = j + dispX < kCol ? j + dispX : kCol - 1;

			Dimension row_start_from = i > dispY ? i - dispY : 0;
			Dimension row_end_to = i + dispY < kRow ? i + dispY : kRow - 1;


			Dimension one_column_elems = row_end_to - row_start_from + 1;		
			Dimension one_row_elems = col_end_to - col_start_from + 1;		
			current_elems = one_row_elems * one_column_elems;


			// Now we compute all the 1/2 * N * (N-1) comparisons of pairs of vectors.
			// From the definition of a metric it follows that d( x, x ) = 0
			// We also use the metric feature: d( x, y ) = d( y, x )

			if( prev_current_elems == kMaxMutualDistMatrix_Size && current_elems == kMaxMutualDistMatrix_Size )
			{
				Dimension p, q;

				// In this case we can reuse the distances already computed in the previous step

				// one_column_elems is an amount of pixels that we have to refill
				REQUIRE( current_elems == dataGatherVec.size() );
				p = 0;
				for( k = one_column_elems; k < current_elems; ++ k )
					dataGatherVec[ p ++ ] = dataGatherVec[ k ];
				REQUIRE( p == current_elems - one_column_elems );
				/////

				//p = current_elems - one_column_elems;
				for( l = row_start_from; l <= row_end_to; l ++ )
					dataGatherVec[ p ++ ] = In.GetPixel( col_end_to, l );


				// Now special copy the mutual distance matrix
				for( k = one_column_elems, p = 0; k < current_elems - 1; ++ k, ++ p )
				{
					for( l = k + 1, q = p + 1; l < current_elems; ++ l, ++ q )
					{
						double tmp = bi_metrics.GetPixel( l, k );
						bi_metrics.SetPixel( q, p, tmp );	
						bi_metrics.SetPixel( p, q, tmp );	
					}
				}

				// Fill the new positions in the mutual distance matrix
				// It is convenient to go by rows
				for( l = current_elems - 1, p = 0; p < one_column_elems ; ++ p, -- l )
				{
					MPIX theReferencePixel = dataGatherVec[ l ];
					for( k = 0; k < current_elems - 1 - p; ++ k )
					{
						double theDistance = Metric< T, 3 >( theReferencePixel, dataGatherVec[ k ] );
						bi_metrics.SetPixel( l, k, theDistance );	
						
						bi_metrics.SetPixel( k, l, theDistance );	// This is new with respect to the non-weighted version,
																	// since all elements now must be filled.
					}
				}

			}
			else
			{
				dataGatherVec.clear();			// clear all data
				REQUIRE( dataGatherVec.size() == 0 );

				for( k = col_start_from; k <= col_end_to; k ++ )
					for( l = row_start_from; l <= row_end_to; l ++ )
						dataGatherVec.push_back( In.GetPixel( k, l ) );
				
				REQUIRE( current_elems == dataGatherVec.size() );

				// The triangle matrix - diagonal and lower triangle set to 0.0
				bi_metrics.SetAll( 0.0 );


				for( k = 0; k < current_elems - 1; ++ k )
				{
					MPIX theReferencePixel = dataGatherVec[ k ];
					for( l = k + 1; l < current_elems; ++ l )
					{
						double theDistance = Metric< T, 3 >( theReferencePixel, dataGatherVec[ l ] );
						bi_metrics.SetPixel( l, k, theDistance );	

						bi_metrics.SetPixel( k, l, theDistance );	// This is new with respect to the non-weighted version,
																	// since all elements now must be filled.

					}
				}

			}

			/////////////////////////


			// At this point we'll have computed all necessary comparisons among vector pairs.
			// Let's select a winner.
			Dimension theWinnerIndex = 0;
			double prev_cumulation = 100000000.0;				// something to start from

			
			// We traverse the mutual-distances-matrix (which is current_elems x current_elems)
			// and compute the weighted sums from each row.
			for( k = 0; k < current_elems; ++ k )
			{
				double cumulation = 0.0;
				
				for( l = 0; l < current_elems; ++ l )
					cumulation += theWeights[ l ] * bi_metrics.GetPixel( l, k );

				if( cumulation < prev_cumulation )
				{
					prev_cumulation = cumulation;
					theWinnerIndex = k;
				}
			}

			REQUIRE( theWinnerIndex < dataGatherVec.size() );
			Out.SetPixel( j, i, dataGatherVec[ theWinnerIndex ] );

			prev_current_elems = current_elems;		// save for the next iteration
		}

	}

}


// =================================================================================================


#endif BoxySoft_VectorRankFilter




