
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: FixedNumber.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A class for definition of the fixed representation of numbers.
//
// ==========================================================================



#ifndef BoxySoft_FixedNumber
#define BoxySoft_FixedNumber



#include "ImageDeclarations.h"
#include "HIL_Exceptions.h"
#include <iostream>


// If this is set to 1 then the class can throw an
// exception (DivideByZero_Exception) on divide by zero
#define THROW_EXCEPTIONS 0

// If set to 1 then a saturated arithmetic is implemented
#define SATURATED_ARITHMETIC 1

// If set to 1 then, depending on other options, the static
// flags are compiled that store status of some operations
// such as: saturation, divide by zero, etc.
#define USE_STATIC_FLAGS 1


using namespace PHIL;



///////////////////////////////////////////////////////////
// This class represents the (missing) fixed data type.
// This class is not intended to be derived from (no
// virtual destructor, etc.).
///////////////////////////////////////////////////////////
template< typename DATA_TYPE, int PRECISION >
class FixedFor
{
	public:

//#if THROW_EXCEPTIONS
//		class FixedFor_Exception {};
//		class DivideByZero_Exception : public FixedFor_Exception {};
//#endif

	private:

		// Class inherent variables
		DATA_TYPE	fValue;

	public:

		DATA_TYPE GetInternalValue( void ) { return fValue; }
		
		void SetInternalValue( DATA_TYPE new_val ) { fValue = new_val; }


	public:

		enum { kPrecision = PRECISION };

		enum { kSignMask = 1 << 8 * sizeof( DATA_TYPE ) - 1 };

		typedef FixedFor< DATA_TYPE, PRECISION > ThisFixedFor_Type;

	public:

		// This flags can be examined to see results of some operations
		// Be careful, however, when using in the MULTITHREADED environment!!!
		// You are also responsible for clearing these flags
#if USE_STATIC_FLAGS
	#if SATURATED_ARITHMETIC
			static bool gSaturation_Flag;
			static void ClearSaturationFlag( void ) { gSaturation_Flag = false; }
			static bool GetSaturationFlag( void )	{ return gSaturation_Flag; }
	#endif
	#if ! THROW_EXCEPTIONS
			static bool gDivideByZero_Flag;
			static void ClearDivideByZeroFlag( void )	{ gDivideByZero_Flag = false; }
			static bool GetDivideByZeroFlag( void )		{ return gDivideByZero_Flag; }
	#endif
#endif

		
		// Here define the longest integer type available on your machine
		typedef __int64 ACC_TYPE;			// This is Microsoft specific
		//typedef unsigned long ACC_TYPE;	// for all other system you can use this one


		static const DATA_TYPE kMaxSaturatedMagnitude;		// = -1 & ~kSignMask;
		static const ACC_TYPE kSaturationShift;				// = sizeof( DATA_TYPE ) * 8 - 1;
		// Here we simply check whether theAcc > max_val
		///////////////////////////////////////////////////////////
		// This function checks whether a given value
		// will saturate this representation. It happens
		// if number of bits in the input value exceeds 
		// the allowable number which is equal to 
		// the kSaturationShift ( = sizeof( DATA_TYPE ) * 8 - 1 )
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			theAcc - the value to be checked
		//		
		// OUTPUT:
		//			true - if the input value is greater
		//				than the maximum allowable value
		//				for this fixed representation
		//			false - otherwise (i.e. a value fits nicely)
		//		
		// REMARKS:
		//			This function assumes that the input value
		//			is compliant with the fixed point representation.
		//			Put it otherwise it cannot be simply applied to the
		//			integer (not shifted) values!
		//		
		bool IsSaturated( ACC_TYPE theAcc ) const { return theAcc >> kSaturationShift > 0 ? true : false; }

		bool DoesIntegerSaturate( ACC_TYPE theAcc ) const 
		{ 
			#if USE_STATIC_FLAGS
				return gSaturation_Flag = theAcc >> ( kSaturationShift - kPrecision ) > 0 ? true : false;
			#else
				return theAcc >> ( kSaturationShift - kPrecision ) > 0 ? true : false; 
			#endif
		}

	public:

		// Some raw data manipulators
		DATA_TYPE DataValueAbs( void ) const { return fValue & ~kSignMask; }	// Return the absolute value
		// This always returns positive value 
		DATA_TYPE GetMagnitudeOfIntegerPart( void ) const { return DataValueAbs() >> kPrecision; }
		DATA_TYPE GetFractionalPart( void ) const { return fValue & ( ( 1 << kPrecision ) - 1 ); }


	public:

		// ===================================================
		FixedFor( void ) : fValue( 0 ) 					// class default constructor
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
		}

		FixedFor( char x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = x < 0 ? ( DoesIntegerSaturate( -x ) ? kMaxSaturatedMagnitude : -x ) << kPrecision | kSignMask : ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x < 0 ? -x << kPrecision | kSignMask : x << kPrecision;
			#endif
		}

		FixedFor( short x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = x < 0 ? ( DoesIntegerSaturate( -x ) ? kMaxSaturatedMagnitude : -x ) << kPrecision | kSignMask : ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x < 0 ? -x << kPrecision | kSignMask : x << kPrecision;
			#endif
		}

		FixedFor( int x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = x < 0 ? ( DoesIntegerSaturate( -x ) ? kMaxSaturatedMagnitude : -x ) << kPrecision | kSignMask : ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x < 0 ? -x << kPrecision | kSignMask : x << kPrecision;
			#endif
		}

		FixedFor( long x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = x < 0 ? ( DoesIntegerSaturate( -x ) ? kMaxSaturatedMagnitude : -x ) << kPrecision | kSignMask : ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x < 0 ? -x << kPrecision | kSignMask : x << kPrecision;
			#endif
		}

		FixedFor( unsigned char x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x << kPrecision;
			#endif
		}

		FixedFor( unsigned short x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x << kPrecision;
			#endif
		}

		FixedFor( unsigned int x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x << kPrecision;
			#endif
		}

		FixedFor( unsigned long x )
		{
			REQUIRE( sizeof( DATA_TYPE ) * 8 > kPrecision );
			#if SATURATED_ARITHMETIC
				fValue = ( DoesIntegerSaturate( x ) ? kMaxSaturatedMagnitude : x ) << kPrecision;	
			#else
				fValue = x << kPrecision;
			#endif
		}



		FixedFor( float x );							// class constructor
		FixedFor( double x );							// class constructor
		
		
		template< class D, int P >
		FixedFor( const FixedFor< D, P > & f )			// mixed copy constructor
		{
			// At first copy the integer part not to loose the MSB
			fValue = f.GetMagnitudeOfIntegerPart() << kPrecision;
			// Then copy the most significant part of the fractional part
			fValue |= f.kPrecision > kPrecision ? f.GetFractionalPart() >> f.kPrecision - kPrecision : f.GetFractionalPart() << kPrecision - f.kPrecision;
			// Finally set the sign
			if( f.IsNegative() )
				MakeNegative();
		}

		// ===================================================
	
		template< class D, int P >
		ThisFixedFor_Type & operator = ( const FixedFor< D, P > & f )	// mixed assignement
		{
			// For different types it can cause lost of data
			if( (void*)& f != (void*)this )
			{
				// At first copy the integer part not to loose the MSB
				fValue = f.GetMagnitudeOfIntegerPart() << kPrecision;
				// Then copy the most significant part of the fractional part
				fValue |= f.kPrecision > kPrecision ? f.GetFractionalPart() >> f.kPrecision - kPrecision : f.GetFractionalPart() << kPrecision - f.kPrecision;
				// Finally set the sign
				if( f.IsNegative() )
					MakeNegative();
			}

			return * this;
		}

		// ===================================================
		// This group of assignements was added to remedy weird
		// behaviour of the Microsoft Visual C++ 6.0 when
		// performing an assignements like this: 
		//		double d = -13.17;
		//		FixedFor< short, 8 > f;
		//		f = d;	==> this line was compiled but then... skipped???
		//
		// In Stroustrup it is written (pg. 284): "By default, a single argument
		// constructor also defines an implicit conversion. For some types, that
		// is ideal. For example 
		//		complex z = 2; // initialize z with complex(2)
		// ". 
		// Apparently,this does not work for some reason under Visual C++ 6.0.
		//
		ThisFixedFor_Type & operator = ( double d )	
		{
			// It can cause lost of data
			* this = ThisFixedFor_Type( d );
			return * this;
		}
		ThisFixedFor_Type & operator = ( long d )	
		{
			// It can cause lost of data
			* this = ThisFixedFor_Type( d );
			return * this;
		}
		ThisFixedFor_Type & operator = ( int d )	
		{
			// It can cause lost of data
			* this = ThisFixedFor_Type( d );
			return * this;
		}
		ThisFixedFor_Type & operator = ( char d )	
		{
			// It can cause lost of data
			* this = ThisFixedFor_Type( d );
			return * this;
		}

		// ===================================================
		
		~FixedFor() {}			// class destructor (not virtual since we don't wish to derive from this class)

		// ===================================================

		// Conversion operators:

		operator char()				const;
		operator short()			const;
		operator int()				const;
		operator long()				const;
		operator unsigned char()	const;
		operator unsigned short()	const;
		operator unsigned int()		const;
		operator unsigned long()	const;
		operator float()			const;
		operator double()			const;

		// ===================================================
		// Helpers:

		bool IsNegative( void ) const { return ( fValue & kSignMask ) == 0 ? false : true; }
		bool IsPositive( void ) const { return ( fValue & kSignMask ) != 0 ? false : true; }	// 0 is also positive

		void ChangeSign( void ) { fValue ^= kSignMask; }				// Simple XOR will do this nicely
		void MakeNegative( void ) { REQUIRE( fValue != 0 ); fValue |= kSignMask; }				// Turn the sign bit ON
		void MakePositive( void ) { fValue &= ~kSignMask; }				// Turn the sign bit OFF
		void MakeAbs( void ) { MakePositive(); }						// Get rid of the sign, i.e. make it positive
		
		ThisFixedFor_Type GetAbs( void ) const { ThisFixedFor_Type tmp( * this ); tmp.MakeAbs(); return tmp; }	// Return the absolute value

		// This always returns positive value 
		static ThisFixedFor_Type GetMinimumMagnitude( void ) { ThisFixedFor_Type x(0); x.fValue |= 0x01; return x; }
		static ThisFixedFor_Type GetMaximumMagnitude( void ) { ThisFixedFor_Type x(0); x.fValue = ~x.fValue; x.MakeAbs(); return x; }
		
		// ===================================================
		// Basic library of operations:

		ThisFixedFor_Type operator + ( ThisFixedFor_Type f ) const;
		ThisFixedFor_Type operator - ( ThisFixedFor_Type f ) const;
		ThisFixedFor_Type operator * ( ThisFixedFor_Type f ) const;
		ThisFixedFor_Type operator / ( ThisFixedFor_Type f ) const;

		ThisFixedFor_Type & operator += ( ThisFixedFor_Type f );	
		ThisFixedFor_Type & operator -= ( ThisFixedFor_Type f );
		ThisFixedFor_Type & operator *= ( ThisFixedFor_Type f );
		ThisFixedFor_Type & operator /= ( ThisFixedFor_Type f );

		// ===================================================
		
		ThisFixedFor_Type operator & ( ThisFixedFor_Type f ) const;
		ThisFixedFor_Type operator | ( ThisFixedFor_Type f ) const;
		ThisFixedFor_Type operator ^ ( ThisFixedFor_Type f ) const;

		ThisFixedFor_Type & operator &= ( ThisFixedFor_Type f );
		ThisFixedFor_Type & operator |= ( ThisFixedFor_Type f );
		ThisFixedFor_Type & operator ^= ( ThisFixedFor_Type f );

		ThisFixedFor_Type & operator ~ ( void );

		// ===================================================

		ThisFixedFor_Type operator ++ ( int );		// postfix
		ThisFixedFor_Type & operator ++ ();			// prefix

		ThisFixedFor_Type operator -- ( int );		// postfix
		ThisFixedFor_Type & operator -- ();			// prefix

		// ===================================================

		ThisFixedFor_Type operator << ( int shift ) const;
		ThisFixedFor_Type operator >> ( int shift ) const;

		ThisFixedFor_Type & operator <<= ( int shift );// { fValue <<= shift; return * this; }
		ThisFixedFor_Type & operator >>= ( int shift );// { fValue >>= shift; return * this; }

		// ===================================================
		// Comparators:

		bool operator ==	( ThisFixedFor_Type f ) const;
		bool operator !=	( ThisFixedFor_Type f ) const;
		bool operator <		( ThisFixedFor_Type f ) const;
		bool operator <=	( ThisFixedFor_Type f ) const;
		bool operator >		( ThisFixedFor_Type f ) const;
		bool operator >=	( ThisFixedFor_Type f ) const;

		// ===================================================

};

///////////////////////////////////////////////////////////







// ===================================================
// The constructors:


template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::FixedFor( double x ) : fValue( 0 )
{			
	bool minus_sign;
	if( x < 0.0 )
	{
		minus_sign = true;		// remember the sign,
		x = -x;					// then take the absolute value
	}
	else
	{
		minus_sign = false;
	}

	long int_part = x;			// here we use the standard conversion from double to long

#if SATURATED_ARITHMETIC
	if( int_part > kMaxSaturatedMagnitude >> kPrecision )	
	{
		#if USE_STATIC_FLAGS
			gSaturation_Flag = true; 
		#endif
		fValue = kMaxSaturatedMagnitude | ( minus_sign ? kSignMask : 0 );	
		return;					// nothing more to do
	}
#endif

	double fract_part = x - int_part;

	fValue = int_part;
	fValue <<= kPrecision;		// put integral part into its place

	// store the fractional part
	DATA_TYPE fraction = 0;
	for( register int i = 0; i < kPrecision; ++ i )
	{
		fraction <<= 1;			// at each iteration shift to the left to make room
								// for next bits
		
		fract_part *= 2.0;
		if( fract_part >= 1.0 )	// soft check !
		{
			// set the bit
			fraction |= 0x01;	// set this bit
			fract_part -= 1.0;
		}

	}

	fValue |= fraction;			// join the two together

	if( minus_sign == true && fValue != 0 )		// bug fix thxs to Herman
		MakeNegative();			// finally consider the sign
}			

// We repeat the same code as for the "double" constructor to
// avoid implicit conversion from "float" to "double" if a additional 
// function were used.
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::FixedFor( float x ) : fValue( 0 )
{			
	bool minus_sign;
	if( x < 0.0 )
	{
		minus_sign = true;		// remember the sign,
		x = -x;					// then take the absolute value
	}
	else
	{
		minus_sign = false;
	}

	long int_part = x;			// here we use the standard conversion from double to long

#if SATURATED_ARITHMETIC
	if( int_part > kMaxSaturatedMagnitude >> kPrecision )	
	{
		#if USE_STATIC_FLAGS
			gSaturation_Flag = true; 
		#endif
		fValue = kMaxSaturatedMagnitude | ( minus_sign ? kSignMask : 0 );	
		return;					// nothing more to do
	}
#endif

	double fract_part = x - int_part;

	fValue = int_part;
	fValue <<= kPrecision;		// put integral part into its place

	// store the fractional part
	DATA_TYPE fraction = 0;
	for( register int i = 0; i < kPrecision; ++ i )
	{
		fraction <<= 1;			// at each iteration shift to the left to make room
								// for next bits
		
		fract_part *= 2.0;
		if( fract_part >= 1.0 )	// soft check !
		{
			// set the bit
			fraction |= 0x01;	// set this bit
			fract_part -= 1.0;
		}

	}

	fValue |= fraction;			// join the two together

	if( minus_sign == true && fValue != 0 )		// bug fix thxs to Herman
		MakeNegative();			// finally consider the sign
}			

// ===================================================

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator float() const
{ 
	double retVal = 0.0;

	DATA_TYPE fraction = fValue & ( ( 1 << kPrecision ) - 1 );
	for( register int i = 0; i < kPrecision; ++ i )
	{
		if( ( fraction & 0x01 ) != 0 )
			retVal += 1.0;

		fraction >>= 1;					// shift to the right 
		retVal /= 2.0;
	}

	retVal += DataValueAbs() >> kPrecision;	// now add the integer part;
	return IsNegative() ? - retVal : retVal;
}

// We repeat the same code as for the "double" conversion to
// avoid implicit conversion from "float" to "double" if a additional 
// function were used.
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator double() const
{ 
	double retVal = 0.0;

	DATA_TYPE fraction = fValue & ( ( 1 << kPrecision ) - 1 );
	for( register int i = 0; i < kPrecision; ++ i )
	{
		if( ( fraction & 0x01 ) != 0 )
			retVal += 1.0;

		fraction >>= 1;					// shift to the right 
		retVal /= 2.0;
	}

	retVal += DataValueAbs() >> kPrecision;	// now add the integer part;
	return IsNegative() ? - retVal : retVal;
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator char() const
{ 
	const char kMaxOutMagnitude = -1 & ~( 1	<< sizeof( char ) * 8 - 1 );
	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	// and the magnitude value is less than the maximum for the output type (otherwise
	// there would be an overflow in the output value)
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 || DataValueAbs() >= kMaxOutMagnitude ? 0 : 1;
	return IsPositive() ? ( fValue >> kPrecision ) + fractionalRemainder : - ( ( DataValueAbs() >> kPrecision ) + fractionalRemainder );
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator short() const
{ 
	const char kMaxOutMagnitude = -1 & ~( 1	<< sizeof( short ) * 8 - 1 );
	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	// and the magnitude value is less than the maximum for the output type (otherwise
	// there would be an overflow in the output value)
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 || DataValueAbs() >= kMaxOutMagnitude ? 0 : 1;
	return IsPositive() ? ( fValue >> kPrecision ) + fractionalRemainder : - ( ( DataValueAbs() >> kPrecision ) + fractionalRemainder );
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator int() const
{ 
	const char kMaxOutMagnitude = -1 & ~( 1	<< sizeof( int ) * 8 - 1 );
	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	// and the magnitude value is less than the maximum for the output type (otherwise
	// there would be an overflow in the output value)
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 || DataValueAbs() >= kMaxOutMagnitude ? 0 : 1;
	return IsPositive() ? ( fValue >> kPrecision ) + fractionalRemainder : - ( ( DataValueAbs() >> kPrecision ) + fractionalRemainder );
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator long() const
{ 
	const char kMaxOutMagnitude = -1 & ~( 1	<< sizeof( long ) * 8 - 1 );
	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	// and the magnitude value is less than the maximum for the output type (otherwise
	// there would be an overflow in the output value)
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 || DataValueAbs() >= kMaxOutMagnitude ? 0 : 1;
	return IsPositive() ? ( fValue >> kPrecision ) + fractionalRemainder : - ( ( DataValueAbs() >> kPrecision ) + fractionalRemainder );
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator unsigned char() const
{ 
//	return (unsigned char)operator char();
	return fValue >> kPrecision;
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator unsigned short() const
{ 
	if( IsNegative() )
		return 0;			// when converting to an unsigned representation, 0 is the lowest value

	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 ? 0 : 1;
	return ( fValue >> kPrecision ) + fractionalRemainder;
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator unsigned int() const
{ 
	if( IsNegative() )
		return 0;			// when converting to an unsigned representation, 0 is the lowest value

	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 ? 0 : 1;
	return ( fValue >> kPrecision ) + fractionalRemainder;
}

template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION >::operator unsigned long() const
{ 
	if( IsNegative() )
		return 0;			// when converting to an unsigned representation, 0 is the lowest value

	// The fractionalRemainder takes on 1 if the fractional part is greater or equal to 0.5
	DATA_TYPE fractionalRemainder = ( fValue & ( 1 << kPrecision - 1 ) ) == 0 ? 0 : 1;
	return ( fValue >> kPrecision ) + fractionalRemainder;
}

// ===================================================

// prefix
template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator ++ ()
{
	( * this ) += 1;
	return * this;
}
template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator -- ()
{
	( * this ) -= 1;
	return * this;
}

// postfix
template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator ++ ( int )
{
	ThisFixedFor_Type tmp( * this );	// this unfortunately requires a temporary copy
	( * this ) += 1;
	return tmp;
}
template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator -- ( int )
{
	ThisFixedFor_Type tmp( * this );	// this unfortunately requires a temporary copy
	( * this ) -= 1;
	return tmp;
}



// ===================================================

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator + ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp += f;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator - ( ThisFixedFor_Type f ) const
{
	f.ChangeSign();
	return operator + ( f );
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator * ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp *= f;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator / ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp /= f;
}


// ===================================================

//
// C++ standard stands that: The RIGHT shift is quaranteed to be logical (0-fill)
// if the shifted value has an unsigned type or if it has a nonnegative value;
// otherwise the result is implementation dependent.
// The LEFT (...) vacated bits are 0-filled.
//
template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator <<= ( int shift )
{
	IsPositive() ? fValue <<= shift : fValue <<= shift, MakeNegative();
	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator >>= ( int shift )
{
	IsPositive() ? fValue >>= shift : ( fValue &= ~kSignMask ) >>= shift, MakeNegative();
	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator << ( int shift ) const
{
	ThisFixedFor_Type tmp( * this );
	tmp.fValue <<= shift;
	return tmp;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator >> ( int shift ) const
{
	ThisFixedFor_Type tmp( * this );
	tmp.fValue >>= shift;
	return tmp;
}

// ===================================================

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator += ( ThisFixedFor_Type f )
{
	// We perform the sign-magnitude arithmetic
	bool first_is_negative	= IsNegative();
	bool second_is_negative = f.IsNegative();

	if( first_is_negative )
	{
		if( second_is_negative )
		{
			// Both are negative

			MakeAbs();				// Take the absolute value of the first operand
			f.MakeAbs();			// Take the absolute value of the second operand

			
#if SATURATED_ARITHMETIC
			ACC_TYPE theAccumulator = fValue;			// The first component goes here
			theAccumulator += f.fValue;					// We add the bare magnitudes;
														// In the FPGA we can avoid "abs" by constraining
														// only to magnitude part, i.e. disregarding the MSB bit
			fValue = IsSaturated( theAccumulator ) ? kMaxSaturatedMagnitude : theAccumulator;	// Check for saturation
			
			// We can also turn-on a saturation flag here, if necessary
			#if USE_STATIC_FLAGS
				gSaturation_Flag = true;
			#endif
#else

			fValue += f.fValue;		// Add the bare magnitudes;
									// In the FPGA we can avoid "abs" by constraining
									// only to magnitude part, i.e. disregarding the MSB bit
#endif

			MakeNegative();			// And negate the result
		}
		else
		{
			// The first is negative, the second is positive

			MakeAbs();		// in the FPGA implementation simply operate on magnitue (disregarding the sign bit)
			
			// Here we compare only the magnitudes; The second is positive, so there is no need to get its abs
			if( fValue <= f.fValue )		// "soft" to make 0 positive		
			{
				fValue = f.fValue - fValue;	// and we do NOT change the sign
			}
			else
			{
				fValue -= f.fValue;
				REQUIRE( IsPositive() );
				if( fValue != 0 )
					MakeNegative();					// and negate the result
			}
		}
	}
	else
	{
		if( second_is_negative )
		{
			// The first is positive, the second is negative

			f.MakeAbs();		// in the FPGA implementation simply operate on magnitue (disregarding the sign bit)
			
			// Here we compare only the magnitudes; The first is positive, so there is no need to get its abs
			if( fValue >= f.fValue )	// "soft" to make 0 positive		
			{
				fValue -= f.fValue;		// and we do NOT change the sign
			}
			else
			{
				fValue = f.fValue - fValue;
				REQUIRE( IsPositive() );
				if( fValue != 0 )
					MakeNegative();			// and negate the result
			}
		}
		else
		{
			// Both are positive
#if SATURATED_ARITHMETIC
			ACC_TYPE theAccumulator = fValue;			// The first component goes here
			theAccumulator += f.fValue;					// We add the bare magnitudes;
														// In the FPGA we can avoid "abs" by constraining
														// only to magnitude part, i.e. disregarding the MSB bit
			fValue = IsSaturated( theAccumulator ) ? kMaxSaturatedMagnitude : theAccumulator;	// Check for saturation
			
			// We can also turn-on a saturation flag here, if necessary
			#if USE_STATIC_FLAGS
				gSaturation_Flag = true;
			#endif
#else
			fValue += f.fValue;
#endif
		}
	}

	return * this;
}


template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator -= ( ThisFixedFor_Type f )
{
	f.ChangeSign();
	return operator += ( f );
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator *= ( ThisFixedFor_Type f )
{
	// We perform the sign-magnitude arithmetic
	bool first_is_negative	= IsNegative();
	bool second_is_negative = f.IsNegative();

	// At first we need an accumulator that is long enough to store the result.
	ACC_TYPE theAccumulator = DataValueAbs();		// we load here a value which is then shifted, e.g. if we
													// load a fixed 1.0 then this value in acc is NOT a unit 
													// but a value shifted to the left by kPrecision
	REQUIRE( sizeof( theAccumulator ) >= 2 * sizeof( DATA_TYPE ) );

	theAccumulator *= f.DataValueAbs();		// multiply the magnitudes - we employ here the available
											// standard multiplication of POSITIVE integers
											// For the FPGA we need to apply a multiplicator here.
												

	// Now we need to shift the result right, twice the PRECISION
	theAccumulator >>= kPrecision;

	// Copy the shifted accumulator to the return object ==> possible lost of precision
#if SATURATED_ARITHMETIC
	fValue = IsSaturated( theAccumulator ) ? kMaxSaturatedMagnitude : theAccumulator;	// Check for saturation
	// We can also turn-on a saturation flag here, if necessary
	#if USE_STATIC_FLAGS
		gSaturation_Flag = true;
	#endif
#else
	fValue = theAccumulator;
#endif

	// Finally set the sign of the result
	if( ( first_is_negative ^ second_is_negative ) && fValue != 0 )
		MakeNegative();

	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator /= ( ThisFixedFor_Type f )
{
	if( f.fValue == 0 )
	{
#if THROW_EXCEPTIONS
//		throw DivideByZero_Exception();
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kDivideByZero );
#else
		#if USE_STATIC_FLAGS
			gDivideByZero_Flag = true;
		#endif
#endif
		return * this;							// error, trying to divide by zero! (set an error flag?)
	}

	// We perform the sign-magnitude arithmetic
	bool first_is_negative	= IsNegative();
	bool second_is_negative = f.IsNegative();

	// At first we need an accumulator that is long enough to store the result.
	ACC_TYPE theAccumulator = DataValueAbs();	// a divident goes here
	REQUIRE( sizeof( theAccumulator ) >= 2 * sizeof( DATA_TYPE ) );


	theAccumulator <<= kPrecision;				// we need to shift the first operand
												// to accomodate for the integer division used 
												// in the next statement

	theAccumulator /= f.DataValueAbs();			// divide the magnitudes- we employ here the available
												// standard multiplication of POSITIVE integers
												// For the FPGA we need to apply a multiplicator here.

	fValue = theAccumulator;					// copy back the result

	// Finally set the sign of the result
	if( ( first_is_negative ^ second_is_negative ) && fValue != 0 )
		MakeNegative();

	return * this;
}

///////////////////////////////////////////////////////////
// Logical operators:


template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator & ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp &= f;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator | ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp |= f;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type FixedFor< DATA_TYPE, PRECISION >::operator ^ ( ThisFixedFor_Type f ) const
{
	ThisFixedFor_Type tmp( * this );
	return tmp ^= f;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator ~ ( void )
{
	fValue = ~ fValue;
	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator &= ( ThisFixedFor_Type f )
{
	fValue &= f.fValue;
	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator |= ( ThisFixedFor_Type f )
{
	fValue |= f.fValue;
	return * this;
}

template< typename DATA_TYPE, int PRECISION >
typename FixedFor< DATA_TYPE, PRECISION >::ThisFixedFor_Type & FixedFor< DATA_TYPE, PRECISION >::operator ^= ( ThisFixedFor_Type f )
{
	fValue ^= f.fValue;
	return * this;
}






///////////////////////////////////////////////////////////
// Comparison operators:

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator == ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	return fValue == f.fValue;
}

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator != ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	return fValue != f.fValue;
}

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator < ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	if( IsNegative() == true )														// we compare ABS only if the same sign
		return f.IsNegative() == true ? DataValueAbs() > f.DataValueAbs() : true;	// an opposite condition if both negative
	else
		return f.IsNegative() == true ? false : DataValueAbs() < f.DataValueAbs();
}

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator <= ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	if( IsNegative() == true )														// we compare ABS only if the same sign
		return f.IsNegative() == true ? DataValueAbs() >= f.DataValueAbs() : true;	// an opposite condition if both negative
	else
		return f.IsNegative() == true ? false : DataValueAbs() <= f.DataValueAbs();
}

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator > ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	if( IsNegative() == true )														// we compare ABS only if the same sign
		return f.IsNegative() == true ? DataValueAbs() < f.DataValueAbs() : false;	// an opposite condition if both negative
	else
		return f.IsNegative() == true ? true : DataValueAbs() > f.DataValueAbs();
}

template< typename DATA_TYPE, int PRECISION >
bool FixedFor< DATA_TYPE, PRECISION >::operator >= ( ThisFixedFor_Type f ) const
{
	REQUIRE( DataValueAbs() != 0 || IsPositive() );
	REQUIRE( f.DataValueAbs() != 0 || f.IsPositive() );
	if( IsNegative() == true )														// we compare ABS only if the same sign
		return f.IsNegative() == true ? DataValueAbs() <= f.DataValueAbs() : false;	// an opposite condition if both negative
	else
		return f.IsNegative() == true ? true : DataValueAbs() >= f.DataValueAbs();
}

///////////////////////////////////////////////////////////

template< typename D, int P >
std::ostream & operator << ( std::ostream & o, const FixedFor< D, P > f )
{
	o << (double)f;
	return o;
}

template< typename D, int P >
std::istream & operator >> ( std::istream & i, FixedFor< D, P > & f )
{
	double d;
	i >> d;
	f = d;
	return i;
}


///////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function returns difference in representations
// between double and FixedFor<> 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			d - an input double value
//			f - exemplary FixedFor<> to be checked
//		
// OUTPUT:
//			difference (as a double value) between pure double
//			and fixed representation of this number
//		
// REMARKS:
//		
//		
template< typename D, int P >
double GetDifferenceInRepresentations( double d, FixedFor< D, P > f )
{
	f = d;					// convert double to FixedFor<>
	return d - (double)f;	// convert back to double and return a difference
}

///////////////////////////////////////////////////////////
// Static definitions:

template< typename DATA_TYPE, int PRECISION >
const DATA_TYPE FixedFor< DATA_TYPE, PRECISION >::kMaxSaturatedMagnitude = -1 & ~FixedFor< DATA_TYPE, PRECISION >::kSignMask;

template< typename DATA_TYPE, int PRECISION >
typename const FixedFor< DATA_TYPE, PRECISION >::ACC_TYPE FixedFor< DATA_TYPE, PRECISION >::kSaturationShift = sizeof( DATA_TYPE ) * 8 - 1;

#if USE_STATIC_FLAGS
	#if SATURATED_ARITHMETIC
			template< typename DATA_TYPE, int PRECISION >
			bool FixedFor< DATA_TYPE, PRECISION >::gSaturation_Flag = false;
	#endif
	#if ! THROW_EXCEPTIONS
			template< typename DATA_TYPE, int PRECISION >
			bool FixedFor< DATA_TYPE, PRECISION >::gDivideByZero_Flag = false;
	#endif
#endif



///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

// Here we define the specific data types. Be careful, however,
// to check what is size of built-in types on your machine!
typedef FixedFor< short, 8 >		FIXED_8_8;
typedef FixedFor< long, 16 >		FIXED_16_16;
typedef FixedFor< long, 8 >			FIXED_24_8;
typedef FixedFor< long, 24 >		FIXED_8_24;

// The following types are for normalized numbers
// for which their magnitude is less than 1.0.
typedef FixedFor< short, 15 >		FIXED_PURE_FRACT_15;
typedef FixedFor< long, 31 >		FIXED_PURE_FRACT_31;

// This is specific for Microsoft (i.e. the __int64 type)
// It can fire REQUIRE() for multiplication and division
//typedef FixedFor< __int64, 32 >		FIXED_32_32;

///////////////////////////////////////////////////////////




#endif BoxySoft_FixedNumber




