
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: MathLibraryForFixed.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Some math functions for fixed point data.
//
// ==========================================================================



#ifndef BoxySoft_MathLibraryForFixed
#define BoxySoft_MathLibraryForFixed


#include "FixedNumber.h"



///////////////////////////////////////////////////////////
// This function computes the sinus function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//				(representing radians)
//		
// OUTPUT:
//			sinus in the fixed-point format	
//		
// REMARKS:
//			We use the first four components of
//			the Taylor expansion series:
//
//			sin(x) = SUM(0,k,inf,) [ (-1)^k	* x^(2k+1) / (2k+1)! ]
//		
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Sin_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	FX tmp_3 = x * x * x;
	FX tmp_5 = tmp_3 * x * x;
	FX tmp_7 = tmp_5 * x * x;
	return x - tmp_3 / (FX)6.0 + tmp_5 / (FX)120.0 - tmp_7 / (FX)5040.0;
}

///////////////////////////////////////////////////////////
// This function computes the cosinus function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//				(representing radians)
//		
// OUTPUT:
//			cosinus in the fixed-point format	
//		
// REMARKS:
//			We use the first four components of
//			the Taylor expansion series:
//
//			cos(x) = SUM(0,k,inf,) [ (-1)^k	* x^(2k) / (2k)! ]
//		
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Cos_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	FX tmp_2 = x * x;
	FX tmp_4 = tmp_2 * tmp_2;
	FX tmp_6 = tmp_4 * tmp_2;
	return (FX)1.0 - tmp_2 / (FX)2.0 + tmp_4 / (FX)24.0 - tmp_6 / (FX)720.0;
}


///////////////////////////////////////////////////////////
// This function computes the tangent function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//				(representing radians)
//		
// OUTPUT:
//			tangent in the fixed-point format	
//		
// REMARKS:
//			We use the first four components of
//			the Taylor expansion series:
//
//			tan(x) = x + x^3/3 + 2x^5/15 + 17x^7/315 + ... + F(Bn(x)) + ... for |x|<pi/2
//
//			Can throw an exception if tangent +/- infinity (e.g. for x = pi/2, etc.)
//
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Tan_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	const FX kPi_Half( 1.570796327 );

	if( x.GetAbs() < kPi_Half )
	{
		FX tmp_3 = x * x * x;
		FX tmp_5 = tmp_3 * x * x;
		FX tmp_7 = tmp_5 * x * x;
		return x + tmp_3 / (FX)3.0 + tmp_5 / (FX)7.5 + tmp_7 / (FX)(315.0/17.0);
	}
	else
	{
		// Here we use the simple formula: tan(x) = sin(x) / cos(x), if cos(x)<>0
		FX cos_x = Cos_ForFixed( x );
		if( cos_x == (FX)0.0 )
			throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kResultInfinity );

		return Sin_ForFixed( x ) / cos_x;
	}
}

///////////////////////////////////////////////////////////
// This function computes the ArcTan function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//		
// OUTPUT:
//			arcus tan in the fixed-point format	
//				(representing radians)
//		
// REMARKS:
//			We use the first components of
//			the Taylor expansion series
//
//		
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Atan_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	if( x.GetAbs() <= (FX)1.0 )
	{
		FX tmp_3 = x * x * x;
		FX tmp_4 = tmp_3 * x;
		FX tmp_5 = tmp_4 * x;
		return x - tmp_3 / (FX)3.0 + tmp_5 / (FX)5.0 - ( tmp_3 * tmp_4 ) / (FX)7.0 + ( tmp_4 * tmp_5 ) / (FX)9.0;
	}
	else
	{
		const FX kPi_Half( 1.570796327 );
		FX tmp = (FX)1.0 / x;
		FX tmp_3 = tmp * tmp * tmp;
		FX tmp_5 = tmp_3 * tmp * tmp;
		return kPi_Half - tmp + tmp_3 / (FX)3.0 - tmp_5 / (FX)5.0 + tmp_5 * tmp * tmp / (FX)7.0;
	}
}


///////////////////////////////////////////////////////////
// This function computes the e^x function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//		
// OUTPUT:
//			e^x in the fixed-point format	
//		
// REMARKS:
//			We use the first components of
//			the Taylor expansion series
//
//		
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Exp_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	FX tmp_2 = x * x;			// x^2
	FX tmp_3 = tmp_2 * x;		// x^3
	FX tmp_5 = tmp_2 * tmp_3;	// x^5
	return (FX)1.0 + x + tmp_2 / (FX)2.0 + tmp_3 / (FX)6.0 + tmp_2 * tmp_2 / (FX)24.0 + tmp_5 / (FX)120.0 + tmp_5 * x / (FX)720.0;
}

///////////////////////////////////////////////////////////
// This function computes the ln(x) function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//		
// OUTPUT:
//			ln(x) in the fixed-point format	
//		
// REMARKS:
//			We use the first components of
//			the Taylor expansion series.
//			This series can converge very slowly.
//
//			Can throw an exception if argument negative
//
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Ln_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	REQUIRE( x.IsPositive() == true );		// otherwise undefined
	if( x.IsPositive() != true )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kWrongInputParam );

	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	FX tmp = ( x - (FX)1.0 ) / ( x + (FX)1.0 );	
	FX tmp2 = tmp * tmp;
	FX tmp3 = tmp2 * tmp;
	tmp = tmp + tmp3 / (FX)3.0 + tmp2 * tmp3 / (FX)5.0 + tmp3 * tmp3 * tmp / (FX)7.0;
	return (FX)2.0 * tmp;
}


///////////////////////////////////////////////////////////
// This function computes the power function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format;
//				x must be positive
//			pow - power value
//		
// OUTPUT:
//			x^pow in the fixed-point format	
//		
// REMARKS:
//
//		
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Power_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x, const FixedFor< DATA_TYPE, PRECISION > pow )
{
	REQUIRE( x.IsPositive() == true );		// otherwise undefined
	if( x.IsPositive() != true )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kWrongInputParam );

	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	return Exp_ForFixed( pow * Ln_ForFixed( x ) );
}

///////////////////////////////////////////////////////////
// This function computes the square root function.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			x - the input value in the fixed-point format
//		
// OUTPUT:
//			sqrt in the fixed-point format	
//		
// REMARKS:
//			We use the first four components of
//			the Taylor expansion series:
//
//			Can throw an exception if argument negative
//
template< typename DATA_TYPE, int PRECISION >
FixedFor< DATA_TYPE, PRECISION > Sqrt_ForFixed( const FixedFor< DATA_TYPE, PRECISION > x )
{
	REQUIRE( x.IsPositive() == true );		// otherwise undefined
	if( x.IsPositive() != true )
		throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kWrongInputParam );

	typedef FixedFor< DATA_TYPE, PRECISION > FX;
	return Power_ForFixed( x, (FX)0.5 );
}

/////////////////////////////////////////////////////////////////////////////////////////////////

// Helpers:


template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > sin( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Sin_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > cos( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Cos_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > tan( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Tan_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > atan( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Atan_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > exp( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Exp_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > log( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Ln_ForFixed( x );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > pow( FixedFor< DATA_TYPE, PRECISION > x, FixedFor< DATA_TYPE, PRECISION > pow )
{
	return Power_ForFixed( x, pow );
}

template< typename DATA_TYPE, int PRECISION >
inline FixedFor< DATA_TYPE, PRECISION > sqrt( FixedFor< DATA_TYPE, PRECISION > x )
{
	return Sqrt_ForFixed( x );
}





/////////////////////////////////////////////////////////////////////////////////////////////////


#endif BoxySoft_MathLibraryForFixed




