
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: IntelLib_ImageConverters.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Auxiliary file for those who cooperate with Intel's Open CV.
//
// ==========================================================================



#ifndef BoxySoft_IntelLib_ImageConverters
#define BoxySoft_IntelLib_ImageConverters


#define CONNECT_2_OPEN_CV 0

#if CONNECT_2_OPEN_CV

#include <highgui.h>
#include "HIL_ImageFor.h"
#include "HIL_MultiChannelImageFor.h"
#include "HIL_AutoPtrDefs.h"


//////////////////////////////////////////////////////////////////////////////////////////



namespace PHIL		
{


///////////////////////////////////////////////////////////
// Copy from non-interlaced multi-channel image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. destination IplImage
//				has different size)
//		
// REMARKS:
//		
//	
template< typename T >
bool MixedCopy( IplImage & dest, const TMultiChannelImageFor< T > & source ) 
{
	REQUIRE( dest.nChannels == 3 );
	REQUIRE( dest.depth == IPL_DEPTH_8U );

	register int i, j;

	const int kRow	= dest.height;
	const int kCol	= dest.width;

	if( source.GetRow() != 1 || source.GetCol() != dest.nChannels )
		return false;			// wrong format of the destination file

	TImageFor< T > & theRedChannel		= source.GetRefPixel( 0, 0 );
	TImageFor< T > & theGreenChannel	= source.GetRefPixel( 1, 0 );
	TImageFor< T > & theBlueChannel		= source.GetRefPixel( 2, 0 );

	if( theRedChannel.GetRow() != kRow || theRedChannel.GetCol() != kCol )
		return false;			// cannot copy these images because different size of channels

	register unsigned char * dest_data_ptr = (unsigned char *)dest.imageData;

	// IPLs are interlaced
 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			* dest_data_ptr ++ = theRedChannel.GetPixel( j, i );
			* dest_data_ptr ++ = theGreenChannel.GetPixel( j, i );
			* dest_data_ptr ++ = theBlueChannel.GetPixel( j, i );
		}
	}

	return true;
}

///////////////////////////////////////////////////////////
// Copy to non-interlaced multi-channel image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. destination IplImage
//				has different size)
//		
// REMARKS:
//		
//	
template< typename T >
bool MixedCopy( TMultiChannelImageFor< T > & dest, const IplImage & source ) 
{
	REQUIRE( source.nChannels == 3 );
	REQUIRE( source.depth == IPL_DEPTH_8U );

	register int i, j;

	const int kRow	= source.height;
	const int kCol	= source.width;

	if( dest.GetRow() != 1 || dest.GetCol() != source.nChannels )
		return false;			// wrong format of the destination file

	TImageFor< T > & theRedChannel		= dest.GetRefPixel( 0, 0 );
	TImageFor< T > & theGreenChannel	= dest.GetRefPixel( 1, 0 );
	TImageFor< T > & theBlueChannel		= dest.GetRefPixel( 2, 0 );

	if( theRedChannel.GetRow() != kRow || theRedChannel.GetCol() != kCol )
		return false;			// cannot copy these images because different size of channels

	register unsigned char * source_data_ptr = (unsigned char *)source.imageData;

 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			theRedChannel.SetPixel( j, i, * source_data_ptr ++ );
			theGreenChannel.SetPixel( j, i, * source_data_ptr ++ );
			theBlueChannel.SetPixel( j, i, * source_data_ptr ++ );
		}
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function does CONVERT-and-copy from TImageFor<T>
// to certain IplImages.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. destination IplImage
//				has pixels that are not supported here)
//		
// REMARKS:
//		
//			Some copyings can lead to lost of precision. 
//	
template< typename T >
bool MixedCopy( IplImage & dest, const TImageFor< T > & source ) 
{
 	register int i, j;
	const int kRow	= dest.height;
	const int kCol	= dest.width;
	
	if( dest.nChannels != 1 )
		return false;	// wrong destination format - only one channel allowed

	if( kRow != source.GetRow() || kCol != source.GetCol() )
		return false;	// different sizes, cannot copy

	// dest.widthStep contains the size of aligned image kRow in bytes;
	// It is given in bytes, therefore we need to divide 
	const Dimension step = dest.widthStep / ( dest.depth / IPL_DEPTH_8U );	// now "step" is in pixel-wide-words

	// Copy data - however, not always it is a byte by byte copying
	switch( dest.depth )
	{
		case IPL_DEPTH_8U:

			{
 				register unsigned char * dest_data_ptr = (unsigned char *)dest.imageData;
				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest_data_ptr[ j + i * step ] = (unsigned char)source.GetPixel( j, i ); // we must call the right conversion here
			}
			break;

		case IPL_DEPTH_8S:

			{
 				register char * dest_data_ptr = (char *)dest.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest_data_ptr[ j + i * step ] = (char)source.GetPixel( j, i ); // we must call the right conversion here
			}
			break;

		case IPL_DEPTH_16U:

 			{
 				register unsigned short * dest_data_ptr = (unsigned short *)dest.imageData;
				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest_data_ptr[ j + i * step ] = (unsigned short)source.GetPixel( j, i ); // we must call the right conversion here
			}
			break;

		case IPL_DEPTH_16S:

			{
 				register short * dest_data_ptr = (short *)dest.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest_data_ptr[ j + i * step ] = (short)source.GetPixel( j, i ); // we must call the right conversion here
			}
			break;

		case IPL_DEPTH_32F:

			{
 				register float * dest_data_ptr = (float *)dest.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest_data_ptr[ j + i * step ] = (float)source.GetPixel( j, i ); // we must call the right conversion here
			}
			break;

		default :
			REQUIRE( false );		// sorry, we don't know how to deal with this format
			break;
	}


	return true;		// data copied successfully
}

//////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function does CONVERT-and-copy from certain IplImages
// to TImageFor<T>.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. the source IplImage
//				has pixels that are not supported here)
//		
// REMARKS:
//		
//			Some copyings can lead to lost of precision. 
//	
template< typename T >
bool MixedCopy( TImageFor< T > & dest, const IplImage & source ) 
{
 	register int i, j;
	const int kRow	= source.height;
	const int kCol	= source.width;

	if( source.nChannels != 1 )
		return false;	// wrong destination format - only one channel allowed

	if( kRow != dest.GetRow() || kCol != dest.GetCol() )
		return false;	// different sizes, cannot copy

	// Copy data 
	// dest.widthStep is given in bytes, therefore we need to divide 
	const Dimension step = source.widthStep / ( source.depth / IPL_DEPTH_8U );	// now "step" is in pixel-wide-words


	// Copy data - however, not always it is a byte by byte copying
	switch( source.depth )
	{
		case IPL_DEPTH_8U:

			{
 				register unsigned char * source_data_ptr = (unsigned char *)source.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest.SetPixel( j, i, (T)source_data_ptr[ j + i * step ] );
			}
			break;

		case IPL_DEPTH_8S:

			{
 				register char * source_data_ptr = (char *)source.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest.SetPixel( j, i, (T)source_data_ptr[ j + i * step ] );
			}
			break;

		case IPL_DEPTH_16U:

 			{
 				register unsigned short * source_data_ptr = (unsigned short *)source.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest.SetPixel( j, i, (T)source_data_ptr[ j + i * step ] );
			}
			break;

		case IPL_DEPTH_16S:

			{
 				register short * source_data_ptr = (short *)source.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest.SetPixel( j, i, (T)source_data_ptr[ j + i * step ] );
			}
			break;

		case IPL_DEPTH_32F:

			{
 				register float * source_data_ptr = (float *)source.imageData;
 				for( i = 0; i < kRow; i ++ )
					for( j = 0; j < kCol; j ++ )
						dest.SetPixel( j, i, (T)source_data_ptr[ j + i * step ] );
			}
			break;

		default :
			REQUIRE( false );		// sorry, we don't know how to deal with this format
			break;
	}

	return true;		// data copied successfully
}

//////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// Specialized template --
// Copy from interlaced color image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. destination IplImage
//				has different size)
//		
// REMARKS:
//		
//		It is required that the IplImage has exactly three channels
//		and its pixels are IPL_DEPTH_8U
//		Sizes of both images must be the same!
//	
template<>
inline bool MixedCopy( IplImage & dest, const TImageFor< MMultiPixelFor< unsigned char, 3 > > & source ) 
{
	REQUIRE( dest.nChannels == 3 );
	REQUIRE( dest.depth == IPL_DEPTH_8U );

	register int i, j;

	const int kRow	= dest.height;
	const int kCol	= dest.width;

	if( source.GetRow() != kRow || source.GetCol() != kCol )
		return false;			// wrong format of the destination file

	register unsigned char * dest_data_ptr = (unsigned char *)dest.imageData;

	// IPLs are interlaced
 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			MMultiPixelFor< unsigned char, 3 > thePixel = source.GetPixel( j, i );

			* dest_data_ptr ++ = R( thePixel );	// red
			* dest_data_ptr ++ = G( thePixel );	// green
			* dest_data_ptr ++ = B( thePixel );	// blue
		}
	}

	return true;
}

///////////////////////////////////////////////////////////
// Specialized template --
// Copy to interlaced color image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. source IplImage
//				has different size)
//		
// REMARKS:
//		
//		It is required that the IplImage has exactly three channels
//		and its pixels are IPL_DEPTH_8U.
//		Sizes of both images must be the same!
//	
template<>
inline bool MixedCopy( TImageFor< MMultiPixelFor< unsigned char, 3 > > & dest, const IplImage & source ) 
{
	REQUIRE( source.nChannels == 3 );
	REQUIRE( source.depth == IPL_DEPTH_8U );

	register int i, j;

	const int kRow	= source.height;
	const int kCol	= source.width;


	if( dest.GetRow() != kRow || dest.GetCol() != kCol )
		return false;			// cannot copy these images because different size

	register unsigned char * source_data_ptr = (unsigned char *)source.imageData;

 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			MMultiPixelFor< unsigned char, 3 > thePixel;

			R( thePixel ) = * source_data_ptr ++;		// red
			G( thePixel ) = * source_data_ptr ++;		// green
			B( thePixel ) = * source_data_ptr ++;		// blue
			
			dest.SetPixel( j, i, thePixel );
		}
	}

	return true;
}



//////////////////////////////////////////////////////////////////////////////////////////

}	// end of the PHIL namespace


//////////////////////////////////////////////////////////////////////////////////////////


#endif CONNECT_2_OPEN_CV










#endif BoxySoft_IntelLib_ImageConverters




