
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_Exceptions.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A suite of exception definitions for the HIL.
//
// ==========================================================================



#ifndef BoxySoft_HIL_Exceptions
#define BoxySoft_HIL_Exceptions


#include "ImageDeclarations.h"
#include <iostream>


using namespace std;


// ==========================================================================
namespace PHIL		
{
// ==========================================================================



////////////////////////////////////////////////////////////////////////////////////////////////
// We make it derived from std::exception so the user  
// can catch all types of errors with one "catch" block.
class T_HIL_Exception : public std::exception
{
	public:
		
		enum EErrorCode{ kUnknown = 0 };


	public:

		///////////////////////////////////////////////////////////
		// This function returns an error code.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			error code of this object
		//		
		// REMARKS:
		//			Each T_HIL_Exception object has
		//			associated EXACTLY ONE error code,
		//			supplied during construction.
		//		
		int Get_Error_Code( void ) const { return fErrorCode; }
		
		
		///////////////////////////////////////////////////////////
		// This function returns a string with description
		// of an exception.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			error explanation message for this object
		//		
		// REMARKS:
		//			This function will NOT throw any exceptions.
		//		
		virtual const char * what() const throw() = 0;

	protected:
		
		///////////////////////////////////////////////////////////
		// Class (default) constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			errorCode - error code for this exception-object
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		T_HIL_Exception( int errorCode = kUnknown )
			: fErrorCode( errorCode ) {}

		virtual ~T_HIL_Exception() {}

	private:

		int fErrorCode;		// interpretation of this value depends on a derived class (i.e. data abstraction)
};

////////////////////////////////////////////////////////////////////////////////////////////////


class T_Standard_HIL_Exception : public T_HIL_Exception
{

private:

	const static char * HIL_Standard_ErrorHandlerMessage[];	// a table of message strings

public:

	// When changing this enum remember to appropriately change
	// the HIL_Standard_ErrorHandlerMessage (in the HIL_Exceptions.cpp file)
	enum HIL_Standard__ErrorHandlerError{	kUnknown = 0,
		
											kDivideByZero,
											kWrongInputParam,
											kWrongOutputParam,
											kImageDimensionsMismatch,
											kWrongPixelType,
											kResultInfinity,

											////////////////////////////////////////
											k_SENTINEL			// keep it always at the end 
								};
public:


		///////////////////////////////////////////////////////////
		// This function returns a string with description
		// of an exception.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			error explanation message for this object
		//		
		// REMARKS:
		//			This function will NOT throw any exceptions.
		//		
		virtual const char * what() const throw();


public:

		///////////////////////////////////////////////////////////
		// Class (default) constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			errorCode - error code for this exception-object;
		//				Must be in a range [ kUnknown, k_SENTINE )
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		T_Standard_HIL_Exception( int error )
			: T_HIL_Exception( error ) { REQUIRE( error >= kUnknown && error < k_SENTINEL ); }

		virtual ~T_Standard_HIL_Exception() {}

};

////////////////////////////////////////////////////////////////////////////////////////////////

// Helper stream out operator. Allows streaming of T_HIL_Exception the objects.
inline ostream & operator << ( ostream & o, const T_HIL_Exception & exc )
{
	o << "Exc. no. " << exc.Get_Error_Code() << ": ";
	o << exc.what();
	return o;
}

////////////////////////////////////////////////////////////////////////////////////////////////

// ==========================================================================
}	// end of the PHIL namespace
// ==========================================================================


#endif BoxySoft_HIL_Exceptions




