
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: Histogram.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the image histogram.
//
// ==========================================================================



#include "Histogram.h"





/////////////////////////////////////////////////////////////////////////////
// Static helpers:
int THistogram::FindMedian( const MonochromeImage & image )
{
	THistogram histogram;
	return histogram.GetMedian( image );
}

// Helpers - might not be as efficient if used more than once since they 
// compute histogram by themself each time they are invoked.
int THistogram::FindPercentile( const MonochromeImage & image, const double quantil )
{
	THistogram histogram;
	return histogram.GetPercentile( image, quantil );
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
THistogram::THistogram( void )		// class constructor
: kHistogramElements( 256 )//, fHistogramVector( new HistoVector( kHistogramElements, 0 ) )
{
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
THistogram::~THistogram()		// class destructor
{

}


///////////////////////////////////////////////////////////
// This function returns auto pointer to the histogram
// vector.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - reference to the monochrome image object
//			maxValPtr - optional pointer to the variable
//				that stores maximum encountered value in 
//				the histogram (the most frequent one)	
//		
// OUTPUT:
//			auto ptr to the histogram table	
//		
// REMARKS:
//		
//		
THistogram::HistoVector_AP THistogram::FindHistogram( const MonochromeImage & image, int * maxValPtr /*= 0*/ )
{
	HistoVector_AP histogramVector( new HistoVector( kHistogramElements, 0 ) );

	int val;
	int max_val = 0;

	register MonochromeImage::PixelType * data_ptr = image.GetImageData();
	register int data_num = image.GetElems();

	MonochromeImage::PixelType pixVal;

	while( data_num -- != 0 )
	{
		pixVal = * data_ptr ++;
		
		REQUIRE( pixVal >= 0 && pixVal < kHistogramElements );

		val = ++ (*histogramVector)[ pixVal ];
		
		if( val > max_val )
			max_val = val;			// find and store the maximum value encountered
	}



	if( maxValPtr != 0 )
		* maxValPtr = max_val;

	return histogramVector;
}


///////////////////////////////////////////////////////////
// This function finds the median (or other percentile)
// It is simple: fMedianVal represents an intensity value present in the picture,
// and the sum of all occurences of all intensity values up to the fMedianVal
// doesn't exceed the HALF (or other fraction for different percentile) of all
// possible occurences (an occurency is just one histogram value).
// Putting this in probability language - P( I <= fMedianVal ) >= 1/2 and
// P( I >= fMedianVal ) >= 1/2.
// Intensity is a random value with probability equal to the number of its 
// occurences in the picture to all the occurences, i.e. to the number of all
// pixels in this image. 
// Value of the quantil parameter is as follows:	0.5 for median
//													0.25 for lower quantil
//													0.1, 0.2, ..., 0.9 for decils
// INPUT:
//			histogram - auto ptr of the histogram table
//			quantil - quantil value (see description above); should
//				not exceed 1.0
//			totalNumOfPixelsInImage - total number of pixels in the
//				image (simply cols * rows of the input image)
//		
// OUTPUT:
//			percentile value, i.e. we return x0, whereas P( x < x0 ) = quantil	
//		
// REMARKS:
//		
//		
int THistogram::GetPercentile( const HistoVector_AP histogram, const double quantil, const unsigned long totalNumOfPixelsInImage )
{
	REQUIRE( quantil <= 1.0 );


	double medianVal = quantil * (double)( totalNumOfPixelsInImage );
	double sum = 0.0;
	for( int m = 0; m < histogram->size(); m ++ )
	{
		sum += (*histogram)[ m ];
		if( sum >= medianVal )
			return m;
	}

	REQUIRE( 0 );		// should never reach this point unless quantil greater than 1
	return 0;
}





