
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: Histogram.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the image histogram.
//
// ==========================================================================


#ifndef BoxySoft_Histogram
#define BoxySoft_Histogram


#include "HIL_ImageFor.h"


using namespace PHIL;



///////////////////////////////////////////////////////////
// This class implements histogram computation from the
// monochrome images.
///////////////////////////////////////////////////////////
class THistogram
{
	private:

		// Contains number of histogram elements, thus size of the fHistogramVector
		const unsigned int kHistogramElements;

		// Class inherent variables

	public:

		typedef vector< unsigned int >			HistoVector;
		typedef std::auto_ptr< HistoVector >	HistoVector_AP;		


	protected:

	public:

		// ===================================================
		THistogram( void );			// class default constructor
		virtual ~THistogram();		// class virtual destructor
		// ===================================================

		
		///////////////////////////////////////////////////////////
		// This function returns auto pointer to the histogram
		// vector.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			image - reference to the monochrome image object
		//			maxValPtr - optional pointer to the variable
		//				that stores maximum encountered value in 
		//				the histogram (the most frequent one)	
		//		
		// OUTPUT:
		//			auto ptr to the histogram table	
		//		
		// REMARKS:
		//		
		//		
		HistoVector_AP FindHistogram( const MonochromeImage & image, int * maxValPtr = NULL );

		///////////////////////////////////////////////////////////
		// This function finds the median (or other percentile)
		// It is simple: fMedianVal represents an intensity value present in the picture,
		// and the sum of all occurences of all intensity values up to the fMedianVal
		// doesn't exceed the HALF (or other fraction for different percentile) of all
		// possible occurences (an occurency is just one histogram value).
		// Putting this in probability language - P( I <= fMedianVal ) >= 1/2 and
		// P( I >= fMedianVal ) >= 1/2.
		// Intensity is a random value with probability equal to the number of its 
		// occurences in the picture to all the occurences, i.e. to the number of all
		// pixels in this image. 
		// Value of the quantil parameter is as follows:	0.5 for median
		//													0.25 for lower quantil
		//													0.1, 0.2, ..., 0.9 for decils
		// INPUT:
		//			histogram - auto ptr of the histogram table
		//			quantil - quantil value (see description above); should
		//				not exceed 1.0
		//			totalNumOfPixelsInImage - total number of pixels in the
		//				image (simply cols * rows of the input image)
		//		
		// OUTPUT:
		//			percentile value, i.e. we return x0, whereas P( x < x0 ) = quantil	
		//		
		// REMARKS:
		//		
		//		
		int GetPercentile( const HistoVector_AP histogram, const double quantil, const unsigned long numOfPixels );

		// Helpers - might not be as efficient if used more than once since they 
		// compute histogram by themself each time they are invoked.
		int GetPercentile( const MonochromeImage & image, const double quantil )
		{
			return GetPercentile( FindHistogram( image ), quantil, image.GetCol()*image.GetRow() );
		}

		// Just to get median
		int GetMedian( HistoVector_AP histogram, const unsigned long numOfPixels )
		{
			return GetPercentile( histogram, 0.5, numOfPixels );
		}
		int GetMedian( const MonochromeImage & image )
		{
			return GetMedian( FindHistogram( image ), image.GetCol()*image.GetRow() );
		}

		/////////////////////////////////////////////////////////////////////////////
		// Static helpers:
		static int FindMedian( const MonochromeImage & image );

		// Helpers - might not be as efficient if used more than once since they 
		// compute histogram by themself each time they are invoked.
		static int FindPercentile( const MonochromeImage & image, const double quantil );

};




#endif BoxySoft_Histogram




