
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ImageErrorMeasurementLibrary.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Suite of functions for comparisons of image contents.
//
// ==========================================================================



#ifndef BoxySoft_ImageErrorMeasurementLibrary
#define BoxySoft_ImageErrorMeasurementLibrary



#include "HIL_ImageFor.h"


using namespace PHIL;


///////////////////////////////////////////////////////////
// This function computes MSE (mean squared error ) from 
// the two images of the same type:
//
//	MSE = Sum( [ in_1( i, j ) - in_2( i, j ) ]^2 ) / num_of_pixels
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in_1 - a const reference to the first input image
//			in_2 - a const reference to the second input image
//		
// OUTPUT:
//			MSE value (can be 0.0)
//							
// REMARKS:
//		
//			The two input images should be of the same size
//			however this is not a must. If the images are different
//			then MSE is computed only where the two images overlap.
//
template < class T, class S >
double Get_MSE( const TImageFor< T > & in_1, const TImageFor< S > & in_2 )
{
	const int kCols_1 = in_1.GetCol();
	const int kRows_1 = in_1.GetRow();

	const int kCols_2 = in_2.GetCol();
	const int kRows_2 = in_2.GetRow();

	const int kCols = min( kCols_1, kCols_2 );
	const int kRows = min( kRows_1, kRows_2 );

	if( kCols == 0 || kRows == 0 )
		return 0.0;

	double retVal = 0.0;

	for( register int i = 0; i < kRows; ++ i )
	{
		for( register int j = 0; j < kCols; ++ j )
		{
			double diff = (double) in_2.GetPixel( j, i ) - (double) in_1.GetPixel( j, i );
			retVal += diff * diff;
		}
	}

	retVal /= kCols * kRows;
	return retVal;
}


///////////////////////////////////////////////////////////
// This function computes PSNR measure.
// (peak signal-to-reconstructed image measure)
//
//	PSNR = 20log10( 255/ RMSE )
//
//	where:
//
//	RMSE = sqrt( MSE );
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in_1 - a const reference to the first input image
//			in_2 - a const reference to the second input image
//		
// OUTPUT:
//			PSNR value in dB
//			or DBL_MAX if MSE is 0.0
//							
// REMARKS:
//		
//			The two input images should be of the same size
//			however this is not a must. If the images are different
//			then MSE is computed only where the two images overlapp.

//			If MSE is exactly 0.0 then returned PSNR is DBL_MAX
//
template < class T, class S >
double Get_PSNR( const TImageFor< T > & in_1, const TImageFor< S > & in_2 )
{
	double MSE = Get_MSE( in_1, in_2 );
	return MSE == 0.0 ? DBL_MAX : 20.0 * log10( 255.0 / sqrt( MSE ) );
}






#endif BoxySoft_ImageErrorMeasurementLibrary




