
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: AffineMoments.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the affine moment invariants.	
//
// ==========================================================================



#ifndef BoxySoft_AffineMoments
#define BoxySoft_AffineMoments



#include "Moments.h"


using namespace PHIL;







///////////////////////////////////////////////////////////
// This function computes the set of the four affine 
// moment invariants as defined by Flusser & Suk
//
// This function assumes that the center of gravity
// is already given (e.g. from the CenterOfGravity(...))
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes).
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			theCentroid - the center of gravity point,
//				if already known
//			theInvariants - the output data structure,
//				a vector that, if returned true, contains
//				the four affine invariants.
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise	
//		
// REMARKS:
//				
template< class T >
bool AffineMomentInvariants( const TMaskedImageFor< T > & in, const Real_2D_Point theCentroid, vector< double > & theInvariants )
{
	if( theCentroid == kWrongPoint )
		return false;		// error

	theInvariants.clear();

	// Ok, the centroid found, so let's compute the requested central moment
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double m_centr_00 = 0.0;

	double m_centr_20 = 0.0;
	double m_centr_02 = 0.0;
	double m_centr_11 = 0.0;

	double m_centr_30 = 0.0;
	double m_centr_03 = 0.0;
	double m_centr_12 = 0.0;
	double m_centr_21 = 0.0;

	double x_first, x_second, x_third;
	double y_first, y_second, y_third;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				m_centr_00 += thePixel;

				// Compute the auxiliary variables with powers of the relative coordinates
				x_first		= j - x( theCentroid );//.x;
				x_second	= x_first * x_first;
				x_third		= x_second * x_first;

				y_first		= i - y( theCentroid );//.y;
				y_second	= y_first * y_first;
				y_third		= y_second * y_first;

				// Find the central moments
				m_centr_20 += x_second					* thePixel;
				m_centr_02 +=				y_second	* thePixel;

				m_centr_11 += x_first *		y_first		* thePixel;

				m_centr_30 += x_third					* thePixel;
				m_centr_03 +=				y_third		* thePixel;

				m_centr_12 += x_first *		y_second	* thePixel;
				m_centr_21 += x_second *	y_first		* thePixel;
			}
		}
	}

	if( m_centr_00 == 0.0 )
		return false;		// cannot compute the normalized moments

	double m_centr_00_pow_2		= m_centr_00		* m_centr_00;
	double m_centr_00_pow_4		= m_centr_00_pow_2	* m_centr_00_pow_2;
	double m_centr_00_pow_7		= m_centr_00_pow_4	* m_centr_00_pow_2	* m_centr_00;
	double m_centr_00_pow_10	= m_centr_00_pow_7	* m_centr_00_pow_2	* m_centr_00;
	double m_centr_00_pow_11	= m_centr_00_pow_10	* m_centr_00;

	double m_centr_12_pow_2		= m_centr_12		* m_centr_12;
	double m_centr_12_pow_3		= m_centr_12_pow_2	* m_centr_12;

	double m_centr_21_pow_2		= m_centr_21		* m_centr_21;
	double m_centr_21_pow_3		= m_centr_21_pow_2	* m_centr_21;

	double m_centr_02_pow_2		= m_centr_02		* m_centr_02;
	double m_centr_02_pow_3		= m_centr_02_pow_2	* m_centr_02;

	double m_centr_20_pow_2		= m_centr_20		* m_centr_20;
	double m_centr_20_pow_3		= m_centr_20_pow_2	* m_centr_20;

	double m_centr_03_pow_2		= m_centr_03 * m_centr_03;
	double m_centr_30_pow_2		= m_centr_30 * m_centr_30;

	double m_centr_11_pow_2		= m_centr_11 * m_centr_11;
	double m_centr_11_pow_3		= m_centr_11_pow_2 * m_centr_11;



	///////////////////////////////////////////////////////////
	// Compute the 4 premier affine moments (see Flusser & Suk)
	double m_affine_1, m_affine_2, m_affine_3, m_affine_4;

	m_affine_1 = ( m_centr_20 * m_centr_02 - m_centr_11_pow_2 ) / m_centr_00_pow_4;

	m_affine_2 = m_centr_30_pow_2 * m_centr_03_pow_2;
	m_affine_2 -= 6.0 * m_centr_30 * m_centr_21 * m_centr_12 * m_centr_03;
	m_affine_2 += 4.0 * m_centr_30 * m_centr_12_pow_3;
	m_affine_2 += 4.0 * m_centr_03 * m_centr_21_pow_3;
	m_affine_2 -= 3.0 * m_centr_21_pow_2 * m_centr_12_pow_2;
	m_affine_2 /= m_centr_00_pow_10;

	m_affine_3 = m_centr_20 * ( m_centr_21 * m_centr_03 - m_centr_12_pow_2 );
	m_affine_3 -= m_centr_11 * ( m_centr_30 * m_centr_03 - m_centr_21 * m_centr_12 );
	m_affine_3 += m_centr_02 * ( m_centr_30 * m_centr_12 - m_centr_21_pow_2 );
	m_affine_3 /= m_centr_00_pow_7;

	m_affine_4 = m_centr_20_pow_3 * m_centr_03_pow_2;
	m_affine_4 -= 6.0 * m_centr_20_pow_2 * m_centr_11 * m_centr_12 * m_centr_03;
	m_affine_4 -= 6.0 * m_centr_20_pow_2 * m_centr_02 * m_centr_21 * m_centr_03;
	m_affine_4 += 9.0 * m_centr_20_pow_2 * m_centr_02 * m_centr_12_pow_2;
	m_affine_4 += 12.0 * m_centr_20 * m_centr_11_pow_2 * m_centr_21 * m_centr_03;
	m_affine_4 += 6.0 * m_centr_20 * m_centr_11 * m_centr_02 * m_centr_30 * m_centr_03; 
	m_affine_4 -= 18.0 * m_centr_20 * m_centr_11 * m_centr_02 * m_centr_21 * m_centr_12;
	m_affine_4 -= 8.0 * m_centr_11_pow_3 * m_centr_30 * m_centr_03;
	m_affine_4 -= 6.0 * m_centr_20 * m_centr_02_pow_2 * m_centr_30 * m_centr_12;
	m_affine_4 += 9.0 * m_centr_20 * m_centr_02_pow_2 * m_centr_21_pow_2;
	m_affine_4 += 12.0 * m_centr_11_pow_2 * m_centr_02 * m_centr_30 * m_centr_12;
	m_affine_4 -= 6.0 * m_centr_11 * m_centr_02_pow_2 * m_centr_30 * m_centr_21;
	m_affine_4 += m_centr_02_pow_3 * m_centr_30_pow_2;
	m_affine_4 /= m_centr_00_pow_11;


	theInvariants.push_back( m_affine_1 );
	theInvariants.push_back( m_affine_2 );
	theInvariants.push_back( m_affine_3 );
	theInvariants.push_back( m_affine_4 );


	return true;
}








#endif BoxySoft_AffineMoments




