
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: Moments.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Functions for computation of image moments.
//
// ==========================================================================




#ifndef BoxySoft_Moments
#define BoxySoft_Moments


#include "HIL_ImageFor.h"
#include "HIL_MaskedImage.h"
#include "HIL_ConvolveOperators.h"
#include <math.h>


using namespace PHIL;




const static Real_2D_Point	kWrongPoint;			// Use this constant to express a situation
													// when it is not possible to get the centroid

const static double			kWrongValue( -1.0 );	// Use it as a return value if computations not possible


///////////////////////////////////////////////////////////
// This function computes a 2D moment of the rank:
// x_moment_rank, y_moment_rank
//
// The formula for 2D moments is as follows:
//
//		m_ij( I ) = SUM{ x^i * y^j * I(x,y) }
//
// where: i == x_moment_rank, j == y_moment_rank
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes)
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			x_moment_rank - the first rank of the moment
//			y_moment_rank - the second rank of the moment
//		
// OUTPUT:
//				
//		
// REMARKS:
//			See e.g. Klette: "Digital Geometry" pg. 541.	
//		
template< class T >
double Moment( const TMaskedImageFor< T > & in, const double x_moment_rank, const double y_moment_rank )
{
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double theMoment = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				theMoment += ::pow( j, x_moment_rank ) * ::pow( i, y_moment_rank ) * thePixel;
			}
		}
	}

	return theMoment;
}


///////////////////////////////////////////////////////////
// This function computes an inertia moment m_0.
//
// The formula for 2D moments is as follows:
//
// m_0( I ) = SUM{ x^2 * y^2 * I(x,y) } = m_02( I ) + m_20( I )
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes)
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//		
// OUTPUT:
//				
//		
// REMARKS:
//		
//		
template< class T >
double InertiaMoment( const TMaskedImageFor< T > & in )
{
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double theMoment = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				theMoment += ( j * j + i * i ) * thePixel;
			}
		}
	}

	return theMoment;
}

///////////////////////////////////////////////////////////
// This function computes the controid (i.e. the center of
// gravity):
//
//	x = m_10 / m_00,   y = m_01 / m_00
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes)
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			ret_m00 - the optional pointer to the double
//				that upon return obtains the m_00 value
//			ret_m01 - the optional pointer to the double
//				that upon return obtains the m_01 value
//			ret_m10 - the optional pointer to the double
//				that upon return obtains the m_10 value
//
// OUTPUT:
//				
//		
// REMARKS:
//		
//		
template< class T >
Real_2D_Point CenterOfGravity(	const TMaskedImageFor< T > & in, 
								double * ret_m00 = 0, double * ret_m01 = 0, double * ret_m10 = 0 )
{
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double m_00 = 0.0, m_10 = 0.0, m_01 = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				m_00 += thePixel;
				m_10 += j * thePixel;
				m_01 += i * thePixel;
			}
		}
	}

	REQUIRE( fabs( m_00 ) > 1e-8 );				// otherwise your data is probably normalized - check it!!!

	// Write out the side products if requried
	if( ret_m00 != 0 )
		* ret_m00 = m_00;
	if( ret_m01 != 0 )
		* ret_m01 = m_01;
	if( ret_m10 != 0 )
		* ret_m10 = m_10;

	return m_00 == 0 ? kWrongPoint : Real_2D_Point( m_10 / m_00, m_01 / m_00 );
}

// Specialization for the window of an image
template< class T >
Real_2D_Point CenterOfGravity(	const TProxyImageFor< T > & in, 
								double * ret_m00 = 0, double * ret_m01 = 0, double * ret_m10 = 0 )
{
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double m_00 = 0.0, m_10 = 0.0, m_01 = 0.0;

	register double i, j;
	register double thePixel;

	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			thePixel = in.GetPixel( j, i );

			// Check whether that pixel belongs to the "object"
			m_00 += thePixel;
			m_10 += j * thePixel;
			m_01 += i * thePixel;
		}
	}

	// Write out the side products if requried
	if( ret_m00 != 0 )
		* ret_m00 = m_00;
	if( ret_m01 != 0 )
		* ret_m01 = m_01;
	if( ret_m10 != 0 )
		* ret_m10 = m_10;

	return m_00 == 0 ? kWrongPoint : Real_2D_Point( m_10 / m_00, m_01 / m_00 );
}




///////////////////////////////////////////////////////////
// This function computes a 2D central moment of the rank:
// x_moment_rank, y_moment_rank
//
// This function assumes that the center of gravity
// is already given (e.g. from the CenterOfGravity(...))
//
// The formula for 2D moments is as follows:
//
//		m_ij( I ) = SUM{ (x-ox)^i * (y-oy)^j * I(x,y) }
//
// where: i == x_moment_rank, j == y_moment_rank
// and a point (ox, oy) is a center of gravity (centroid)
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes).
//
// The central moments are computed the same was as moments
// if we use a coordinate system that has its origin
// at the center of gravity.
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			theCentroid - the center of gravity point,
//				if already known
//			x_moment_rank - the first rank of the moment
//			y_moment_rank - the second rank of the moment
//		
// OUTPUT:
//			the central moment of order (x_moment_rank, y_moment_rank)	
//		
// REMARKS:
//				
template< class T >
double CentralMoment( const TMaskedImageFor< T > & in, const Real_2D_Point theCentroid, const double x_moment_rank, const double y_moment_rank )
{
	if( theCentroid == kWrongPoint )
		return kWrongValue;		// error

	// Ok, the centroid found, so let's compute the requested central moment
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double theMoment = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				theMoment += ::pow( j - theCentroid.x, x_moment_rank ) * ::pow( i - theCentroid.y, y_moment_rank ) * thePixel;
			}
		}
	}

	return theMoment;
}



///////////////////////////////////////////////////////////
// This function computes a 2D central normalized moment 
// of the rank:
// x_moment_rank, y_moment_rank
//
// This function assumes that the center of gravity
// is already given (e.g. from the CenterOfGravity(...))
//
// The formula for 2D moments is as follows:
//
//		m_centr_norm_ij( I ) = m_centr_ij / m_centr_00 ^ gamma
//		gamma = ( i + j ) / 2 + 1
//
// where: i == x_moment_rank, j == y_moment_rank
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes).
//
// The central moments are computed the same was as moments
// if we use a coordinate system that has its origin
// at the center of gravity.
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			theCentroid - the center of gravity point,
//				if already known
//			x_moment_rank - the first rank of the moment
//			y_moment_rank - the second rank of the moment
//		
// OUTPUT:
//			the central moment of order (x_moment_rank, y_moment_rank)	
//		
// REMARKS:
//				
template< class T >
double NormalizedCentralMoment( const TMaskedImageFor< T > & in, const Real_2D_Point theCentroid, const double x_moment_rank, const double y_moment_rank )
{
	if( theCentroid == kWrongPoint )
		return kWrongValue;		// error

	// Ok, the centroid found, so let's compute the requested central moment
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double theMoment = 0.0, m_centr_00 = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				m_centr_00 += thePixel;
				theMoment += ::pow( j - theCentroid.x, x_moment_rank ) * ::pow( i - theCentroid.y, y_moment_rank ) * thePixel;
			}
		}
	}



	return m_centr_00 == 0.0 ? kWrongValue : theMoment / ::pow( m_centr_00, ( ( x_moment_rank + y_moment_rank ) / 2.0 + 1.0 ) );
}


///////////////////////////////////////////////////////////
// This function computes the set of moment invariants
// as defined by M. Hu in "Visual pattern recognition
// by moment invariants" IRE Transactions on Information
// Theory, IT-8 (1962) 179-187
//
// This function assumes that the center of gravity
// is already given (e.g. from the CenterOfGravity(...))
//
// The "object" is defined by a mask associated
// with the original image (thus we use TMaskedImageFor<>
// classes).
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			theCentroid - the center of gravity point,
//				if already known
//			theInvariants - the output data structure,
//				a vector that, if returned true, contains
//				the Hu invariants.
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise	
//		
// REMARKS:
//				
template< class T >
bool MomentInvariants( const TMaskedImageFor< T > & in, const Real_2D_Point theCentroid, vector< double > & theInvariants )
{
	if( theCentroid == kWrongPoint )
		return false;		// error

	theInvariants.clear();

	// Ok, the centroid found, so let's compute the requested central moment
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double m_centr_00 = 0.0;

	double m_centr_20 = 0.0;
	double m_centr_02 = 0.0;
	double m_centr_11 = 0.0;

	double m_centr_30 = 0.0;
	double m_centr_03 = 0.0;
	double m_centr_12 = 0.0;
	double m_centr_21 = 0.0;

	double x_first, x_second, x_third;
	double y_first, y_second, y_third;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				m_centr_00 += thePixel;

				// Compute the auxiliary variables with powers of the relative coordinates
				x_first		= j - theCentroid.x;
				x_second	= x_first * x_first;
				x_third		= x_second * x_first;

				y_first		= i - theCentroid.y;
				y_second	= y_first * y_first;
				y_third		= y_second * y_first;

				// Find the central moments
				m_centr_20 += x_second					* thePixel;
				m_centr_02 +=				y_second	* thePixel;

				m_centr_11 += x_first *		y_first		* thePixel;

				m_centr_30 += x_third					* thePixel;
				m_centr_03 +=				y_third		* thePixel;

				m_centr_12 += x_first *		y_second	* thePixel;
				m_centr_21 += x_second *	y_first		* thePixel;
			}
		}
	}

	if( m_centr_00 == 0.0 )
		return false;		// cannot compute the normalized moments


	// Now, from the central moments, create the normalized central moments (see Ritter, pg. 295)
	const double m_centr_00_norm_s2 = ::pow( m_centr_00, ( 2.0 + 0.0 ) / 2.0 + 1.0 );	// p + q = 2
	const double m_centr_00_norm_s3 = m_centr_00_norm_s2 * sqrt( m_centr_00 );	// p + q = 3

	m_centr_20 /= m_centr_00_norm_s2;
	m_centr_02 /= m_centr_00_norm_s2;

	m_centr_11 /= m_centr_00_norm_s2;

	m_centr_30 /= m_centr_00_norm_s3;
	m_centr_03 /= m_centr_00_norm_s3;

	m_centr_12 /= m_centr_00_norm_s3;
	m_centr_21 /= m_centr_00_norm_s3;

	// Finally, find the Hu's invariants (see Ritter, pg. 295)

	// o1:
	theInvariants.push_back( m_centr_20 + m_centr_02 );

	// o2:
	theInvariants.push_back( ::pow( m_centr_20 - m_centr_02, 2.0 ) + 4.0 * m_centr_11 * m_centr_11 );

	// o3:
	theInvariants.push_back( ::pow( m_centr_30 - 3.0 * m_centr_12, 2.0 ) + ::pow( 3.0 * m_centr_21 - m_centr_03, 2.0 ) );

	// o4:
	theInvariants.push_back( ::pow( m_centr_30 + m_centr_12, 2.0 ) + ::pow( m_centr_21 + m_centr_03, 2.0 ) );

	// o5:
	theInvariants.push_back(	( m_centr_30 - 3.0 * m_centr_12 ) * ( m_centr_30 + m_centr_12 ) * ( ::pow( m_centr_30 + m_centr_12, 2.0 ) - 3.0 * ::pow( m_centr_21 + m_centr_03, 2.0 ) ) +
								( 3.0 * m_centr_21 - m_centr_03 ) * ( m_centr_21 + m_centr_03 ) * ( 3.0 * ::pow( m_centr_30 + m_centr_12, 2.0 ) - ::pow( m_centr_21 + m_centr_03, 2.0 ) ) );

	// o6:
	theInvariants.push_back(	( m_centr_20 - m_centr_02 ) * ( ::pow( m_centr_30 + m_centr_12, 2.0 ) - ::pow( m_centr_21 + m_centr_03, 2.0 ) ) +
								4.0 * m_centr_11 * ( m_centr_30 + m_centr_12 ) * ( m_centr_21 + m_centr_03 ) );

	// o7:
	theInvariants.push_back(	( 3.0 * m_centr_21 - m_centr_03 ) * ( m_centr_30 + m_centr_12 ) * ( ::pow( m_centr_30 + m_centr_12, 2.0 ) - 3.0 * ::pow( m_centr_21 + m_centr_03, 2.0 ) ) +
								( 3.0 * m_centr_12 - m_centr_30 ) * ( m_centr_21 + m_centr_03 ) * ( 3.0 * ::pow( m_centr_30 + m_centr_12, 2.0 ) - ::pow( m_centr_21 + m_centr_03, 2.0 ) ) );


	return true;
}


static double GetInertiaMomentAroundAxis( double theta, double m_centr_20, double m_centr_11, double m_centr_02 )
{
	double sin_theta = sin( theta );
	double cos_theta = cos( theta );
	return m_centr_20 * sin_theta * sin_theta - 2.0 * m_centr_11 * sin_theta * cos_theta + m_centr_02 * cos_theta * cos_theta;
}

///////////////////////////////////////////////////////////
// This function computes parameters of the principal axis.
// This is a line that passes trough the centroid 
// and has a slope as follows:
//
//		tan( 2*theta ) = 2 * m_centr_11 / ( m_centr_20 - m_centr_02 )
//
// This function assumes that the center of gravity
// is already given (e.g. from the CenterOfGravity(...))
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - reference to the image
//			theCentroid - the center of gravity point,
//				if already known
//		
// OUTPUT:
//				
//		
// REMARKS:
//				
template< class T >
bool Find_ThePrincipalAxis_Slope( const TMaskedImageFor< T > & in, const Real_2D_Point theCentroid, double & theta_major, double & theta_minor, double & elongation_ratio, double & length, double & width )
{
	if( theCentroid == kWrongPoint )
		return false;		// error

	// Ok, the centroid found, so let's compute the requested central moment
	const double row = in.GetRow();
	const double col = in.GetCol();

	REQUIRE( row > 0 );
	REQUIRE( col > 0 );

	double m_centr_11 = 0.0, m_centr_20 = 0.0, m_centr_02 = 0.0;

	register double i, j;
	for( i = 0; i < row; i ++ )
	{
		for( j = 0; j < col; j ++ )
		{
			bool theMaskVal;
			double thePixel = in.GetPixel( j, i, theMaskVal );

			// Check whether that pixel belongs to the "object"
			if( theMaskVal == true )
			{
				double x_shifted_coord = j - theCentroid.x;
				double y_shifted_coord = i - theCentroid.y;

				m_centr_11 += x_shifted_coord * y_shifted_coord * thePixel;
				m_centr_20 += x_shifted_coord * x_shifted_coord * thePixel;
				m_centr_02 += y_shifted_coord * y_shifted_coord * thePixel;
			}
		}
	}

	theta_major = m_centr_20 == m_centr_02 ? kPiHalf / 2.0 : atan( 2.0 * m_centr_11 / ( m_centr_20 - m_centr_02 ) ) / 2.0;
	if( theta_major < 0.0 )
		theta_major += kPi;
	theta_minor = theta_major + kPiHalf;

	// Check which one maximizes the moment of intertia

	double m_1 = GetInertiaMomentAroundAxis( theta_major, m_centr_20, m_centr_11, m_centr_02 );
	double m_2 = GetInertiaMomentAroundAxis( theta_minor, m_centr_20, m_centr_11, m_centr_02 );

	elongation_ratio = m_1 == 0.0 ? 0.0 : m_2 / m_1;

	if( m_1 < m_2 )
	{
		// Exchange the values
		double temp = theta_major;
		theta_major = theta_minor;
		theta_minor = temp;

		if( elongation_ratio != 0.0 )
			elongation_ratio = 1.0 / elongation_ratio;
	}

	double sec_moment_diff = m_centr_20 - m_centr_02;
	double sec_moment_sum = m_centr_20 + m_centr_02;
	double common = sqrt( 4.0 * m_centr_11 * m_centr_11 + sec_moment_diff * sec_moment_diff );

	length = 0.5 * ( sec_moment_sum + common );
	width = 0.5 * ( sec_moment_sum - common );

	return true;
}




#endif BoxySoft_Moments




