
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: GaussianImagePyramids.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the image pyramids.
//
// ==========================================================================



#include "GaussianImagePyramids.h"
#include "GaussianFilter.h"
#include "ImageReSamplers.h"


const TGaussianFilter_Factory	TGaussianFilter_Factory::gkGaussianFilter_Factory_Object;
const TBinomialFilter_Factory	TBinomialFilter_Factory::gkBinomialFilter_Factory_Object;




///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TGaussianImagePyramids::TGaussianImagePyramids( int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory /*= TGaussianFilter_Factory::gkGaussianFilter_Factory_Object*/ )		// class constructor
: fMaskWidth( maskWidth ), fSigma( sigma ), fRealLinearFilter_Factory( theFactory ) 
{
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TGaussianImagePyramids::~TGaussianImagePyramids()		// class destructor
{
}

///////////////////////////////////////////////////////////
// This function creates an image pyramid from the input
// image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			theInputImage - reference to the input image
//			theOutputPyramid - vector of output images
//			requiredPyramidLevels - required number of 
//				the pyramid levels
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise
//		
// REMARKS:
//		
//		
bool TGaussianImagePyramids::operator () ( const TRealImage & theInputImage, RealImage_Vector & theOutputPyramid, int requiredPyramidLevels )
{
	// At first we need a Gaussian filter.

	// Get the smoothing filter through the current factory.
	RealLinFilter_AP theSmoothingFilter( fRealLinearFilter_Factory( fMaskWidth, fSigma ) );

	TImage_SimpleByTwo_DownSampler theDecimator;

	REQUIRE( theOutputPyramid.size() == 0 );	// require no prior images in the output vector


	theOutputPyramid.push_back( theInputImage );						// save the first image of the pyramid (which is the original, unchanged image)

	RIAP smoothedImageAtLevel( ( * theSmoothingFilter )( theInputImage ) );	// prepare the starting smoothed version

	// Go through all remaining levels
	for( int theLevel = 1; theLevel < requiredPyramidLevels; theLevel ++ )
	{
		RIAP decimatedImage( theDecimator( * smoothedImageAtLevel ) );	// decimate the smoothed version
		
		theOutputPyramid.push_back( * decimatedImage );					// save that level of images

		smoothedImageAtLevel = RIAP( ( * theSmoothingFilter )( * (const TRealImage *)decimatedImage.get() ) );	// the starting image for the next level
	}


	return true;
}



///////////////////////////////////////////////////////////
// This function creates an image pyramid from the input
// image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			theInputImage - reference to the input image
//			theOutputPyramid - vector of output images
//			requiredPyramidLevels - required number of 
//				the pyramid levels
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise
//		
// REMARKS:
//		
//		
bool TGaussianImagePyramids::operator () ( const TRealImage & theInputImage, Real_Template_Image & theOutputPyramid, int requiredPyramidLevels )
{
	// At first we need a Gaussian filter.

	// Get the smoothing filter through the current factory.
	RealLinFilter_AP theSmoothingFilter( fRealLinearFilter_Factory( fMaskWidth, fSigma ) );

	TImage_SimpleByTwo_DownSampler theDecimator;

	// We store images of the pyramid in a single row template image.
	REQUIRE( theOutputPyramid.GetCol() >= requiredPyramidLevels );
	REQUIRE( theOutputPyramid.GetRow() >= 1 );

	theOutputPyramid.SetRefPixel( 0, 0, theInputImage );

	RIAP smoothedImageAtLevel( ( * theSmoothingFilter )( theInputImage ) );	// prepare the starting smoothed version

	// Go through all remaining levels
	for( int theLevel = 1; theLevel < requiredPyramidLevels; theLevel ++ )
	{
		RIAP decimatedImage( theDecimator( * smoothedImageAtLevel ) );	// decimate the smoothed version
		
		theOutputPyramid.SetRefPixel( theLevel, 0, * decimatedImage );

		smoothedImageAtLevel = RIAP( ( * theSmoothingFilter )( * (const TRealImage *)decimatedImage.get() ) );	// the starting image for the next level
	}


	return true;
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TDOGImagePyramids::TDOGImagePyramids( int maskWidth, double sigma_1, double sigma_2, const TRealLinearFilter_Factory & theFactory /*= TGaussianFilter_Factory::gkGaussianFilter_Factory_Object*/ )		// class constructor
: TGaussianImagePyramids( maskWidth, sigma_1 ), fSigma_2( sigma_2 ) 
{
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TDOGImagePyramids::~TDOGImagePyramids()		// class destructor
{
}

///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//
bool TDOGImagePyramids::operator () ( const TRealImage & theInputImage, RealImage_Vector & theOutputPyramid, int requiredPyramidLevels )
{
	// At first we need a Gaussian filter.

	DOG_Filter theSmoothingFilter( fMaskWidth, fSigma, fSigma_2 );

	TImage_SimpleByTwo_DownSampler theDecimator;

	REQUIRE( theOutputPyramid.size() == 0 );	// require no prior images in the output vector


	theOutputPyramid.push_back( theInputImage );						// save the first image of the pyramid (which is the original, unchanged image)

	RIAP smoothedImageAtLevel( theSmoothingFilter( theInputImage ) );	// prepare the starting smoothed version

	// Go through all remaining levels
	for( int theLevel = 1; theLevel < requiredPyramidLevels; theLevel ++ )
	{
		RIAP decimatedImage( theDecimator( * smoothedImageAtLevel ) );	// decimate the smoothed version
		
		theOutputPyramid.push_back( * decimatedImage );					// save that level of images

		smoothedImageAtLevel = RIAP( theSmoothingFilter( * (const TRealImage *)decimatedImage.get() ) );	// the starting image for the next level
	}


	return true;
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//
bool TDOGImagePyramids::operator () ( const TRealImage & theInputImage, Real_Template_Image & theOutputPyramid, int requiredPyramidLevels )
{
	// At first we need a Gaussian filter.

	DOG_Filter theSmoothingFilter( fMaskWidth, fSigma, fSigma_2 );

	TImage_SimpleByTwo_DownSampler theDecimator;

	REQUIRE( theOutputPyramid.GetCol() >= requiredPyramidLevels );
	REQUIRE( theOutputPyramid.GetRow() >= 1 );


	theOutputPyramid.SetRefPixel( 0, 0, theInputImage );

	RIAP smoothedImageAtLevel( theSmoothingFilter( theInputImage ) );	// prepare the starting smoothed version

	// Go through all remaining levels
	for( int theLevel = 1; theLevel < requiredPyramidLevels; theLevel ++ )
	{
		RIAP decimatedImage( theDecimator( * smoothedImageAtLevel ) );	// decimate the smoothed version
		
		theOutputPyramid.SetRefPixel( theLevel, 0, * decimatedImage );

		smoothedImageAtLevel = RIAP( theSmoothingFilter( * (const TRealImage *)decimatedImage.get() ) );	// the starting image for the next level
	}


	return true;
}
//////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TLaplacianImagePyramids::TLaplacianImagePyramids( int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory /*= TGaussianFilter_Factory::gkGaussianFilter_Factory_Object*/ )		// class constructor
: TGaussianImagePyramids( maskWidth, sigma, theFactory )
{
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TLaplacianImagePyramids::~TLaplacianImagePyramids()		// class destructor
{
}

// VERSION: DOWN, UP, SUBTRACT
bool TLaplacianImagePyramids::operator () ( const TRealImage & theInputImage, RealImage_Vector & theOutputPyramid, int requiredPyramidLevels )
{
	// At first create the pure Gaussian pyramid
	RealImage_Vector thePureGaussianPyramid;
	if( TGaussianImagePyramids::operator ()( theInputImage, thePureGaussianPyramid, requiredPyramidLevels + 1 ) == false )
		return false;

	// Now take two neighbouring images from the pure Gaussian pyramid, up-sample the smaller one,
	// and subtract the two. The result put into the output pyramid.

	TImage_TwoNeighborInterpol_UpSampler		theImageUpSampler;

	// Go through all remaining levels
	REQUIRE( thePureGaussianPyramid.size() == requiredPyramidLevels + 1 );
	int theLevel;
	for( theLevel = 0; theLevel < requiredPyramidLevels - 1; theLevel ++ )
	{
		REQUIRE( theLevel + 1 < thePureGaussianPyramid.size() );
		RIAP upSampledImage( theImageUpSampler( thePureGaussianPyramid[ theLevel + 1 ] ) );			// interpolate the smoothed and down-sampled version

		REQUIRE( upSampledImage.get() != 0 );
		REQUIRE( thePureGaussianPyramid[ theLevel ].GetCol() >= upSampledImage->GetCol() );
		REQUIRE( thePureGaussianPyramid[ theLevel ].GetRow() >= upSampledImage->GetRow() );

		// It can happen that the two images we wish to subtract can have different dimensions.
		// This happens because when down sampling, then up sampling, image dimensions are 
		// adjusted to the nearest even value. At the same time the second image, the one
		// from the Gaussian pyramid that has not been downsampled, can have an odd dimension.
		// Thus a necessity to adjust image dimensions before subtracting.

		RIAP adjustedGaussianImage( thePureGaussianPyramid[ theLevel /*- 1*/ ].OrphanBitBlit( 0, 0, upSampledImage->GetCol(), upSampledImage->GetRow() ) );

		theOutputPyramid.push_back( * adjustedGaussianImage - * upSampledImage );
	}

	// The last image in the Laplacian pyramid is a low-pass filtered
	// Gaussian image containing only the coarsest structures. 
	REQUIRE( theLevel == requiredPyramidLevels - 1 );
	theOutputPyramid.push_back( thePureGaussianPyramid[ theLevel ] );

	return true;
}

bool TLaplacianImagePyramids::operator () ( const TRealImage & theInputImage, Real_Template_Image & theOutputPyramid, int requiredPyramidLevels )
{
	// At first create the pure Gaussian pyramid
	RealImage_Vector thePureGaussianPyramid;
	if( TGaussianImagePyramids::operator ()( theInputImage, thePureGaussianPyramid, requiredPyramidLevels + 1 ) == false )
		return false;

	REQUIRE( theOutputPyramid.GetCol() >= requiredPyramidLevels );
	REQUIRE( theOutputPyramid.GetRow() >= 1 );

	// Now take two neighbouring images from the pure Gaussian pyramid, up-sample the smaller one,
	// and subtract the two. The result put into the output pyramid.

	TImage_TwoNeighborInterpol_UpSampler		theImageUpSampler;

	// Go through all remaining levels
	REQUIRE( thePureGaussianPyramid.size() == requiredPyramidLevels + 1 );
	int theLevel;
	for( theLevel = 0; theLevel < requiredPyramidLevels - 1; theLevel ++ )
	{
		REQUIRE( theLevel + 1 < thePureGaussianPyramid.size() );
		RIAP upSampledImage( theImageUpSampler( thePureGaussianPyramid[ theLevel + 1 ] ) );			// interpolate the smoothed and down-sampled version

		REQUIRE( upSampledImage.get() != 0 );
		REQUIRE( thePureGaussianPyramid[ theLevel ].GetCol() >= upSampledImage->GetCol() );
		REQUIRE( thePureGaussianPyramid[ theLevel ].GetRow() >= upSampledImage->GetRow() );

		// It can happen that the two images we wish to subtract can have different dimensions.
		// This happens because when down sampling, then up sampling, image dimensions are 
		// adjusted to the nearest even value. At the same time the second image, the one
		// from the Gaussian pyramid that has not been downsampled, can have an odd dimension.
		// Thus a necessity to adjust image dimensions before subtracting.

		RIAP adjustedGaussianImage( thePureGaussianPyramid[ theLevel /*- 1*/ ].OrphanBitBlit( 0, 0, upSampledImage->GetCol(), upSampledImage->GetRow() ) );

		theOutputPyramid.SetRefPixel( theLevel, 0, * adjustedGaussianImage - * upSampledImage );
	}

	// The last image in the Laplacian pyramid is a low-pass filtered
	// Gaussian image containing only the coarsest structures. 
	REQUIRE( theLevel == requiredPyramidLevels - 1 );
	theOutputPyramid.SetRefPixel( theLevel, 0, thePureGaussianPyramid[ theLevel ] );

	return true;
}


///////////////////////////////////////////////////////////
// This function recovers original images from the Laplacian
// pyramid.
// The Laplacian pyramid allows for image reconstruction
// quickly from the sequence of images in the Laplacian
// pyramid by recursively expanding the images and summing
// them up:
//
//			G[ p - 1 ] = L[ p - 1 ] + F2[ G[ p ] ]
//
// where: the finest level is at p = 0, L is the Laplace
// pyramid, G is Gaussian pyramid, F2 is up-sampler-by-two.
//
// Even if the interpolation algorithms required to expand
// the image contain errors, they effect only the Laplacian
// pyramid because the same algorithm is used twice.
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		This function relies on the Gaussian image pyramid
//		stored in the internal data structure:
//		fPureGaussianPyramid
//		
bool TLaplacianImagePyramids::RecoverGaussian ( const RealImage_Vector & theInputLaplacePyramid, RealImage_Vector & theOutputPyramid )
{
	int pyramidLevels = theInputLaplacePyramid.size();

	if( pyramidLevels < 1 )
		return false;

	// Proceed according to the following scheme:
	// up-sample Laplacian and ADD it to the previous file

	TImage_TwoNeighborInterpol_UpSampler theImageUpSampler;


	// Get the coarsest level from the Laplacian pyramid
	TRealImage prevLaplaceImage( theInputLaplacePyramid[ pyramidLevels - 1 ] );	
	
	for( int theLevel = pyramidLevels - 2; theLevel >= 0; theLevel -- )
	{
		// Upsample the current PREVIOUS (i.e. one level closer to the coarsest) image from the Laplacian pyramid
		RIAP upSampledImage( theImageUpSampler( prevLaplaceImage ) );

		// Add the two but be carefull on image size; e.g. if the original (Laplace) pyramid
		// is like this: 420 => 210 => (105) 104 and back, from 104 we get 208 but not 210!
		// Therefore, when going up (i.e. from the coarsest level to the finest) we have to
		// check dimensions of the two files and sometimes we have to adjust their dimensions.
		REQUIRE( upSampledImage->GetCol() <= theInputLaplacePyramid[ theLevel ].GetCol() );
		REQUIRE( upSampledImage->GetRow() <= theInputLaplacePyramid[ theLevel ].GetRow() );

		RIAP theCurrentLayer_LaplaceImage( theInputLaplacePyramid[ theLevel ].OrphanBitBlit( 0, 0, upSampledImage->GetCol(), upSampledImage->GetRow() ) );


		* theCurrentLayer_LaplaceImage += * upSampledImage;

		theOutputPyramid.push_back( * theCurrentLayer_LaplaceImage );

		// Store the WHOLE image for the next iteration
		prevLaplaceImage = * theCurrentLayer_LaplaceImage;
	}


	return true;
}


