
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ImageReSamplers.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the image by 2 re-samplers.
//
// ==========================================================================


#include "ImageReSamplers.h"


///////////////////////////////////////////////////////////
// This function down samples an input image by factor
// set in kDownStep (two)
///////////////////////////////////////////////////////////
//		
// INPUT:
//			in - a reference to the original real image (unchanged)
//		
// OUTPUT:
//			auto ptr to the down sampled copy of the original
//		
// REMARKS:
//		
// I  Create a new image with dimensions equal to the
//    original image dimensions divided by kDownStep 
// II For each output pixel ( m, n ) set its value the same
//    as can be found in the original image at location:
//    ( m * kDownStep, n * kDownStep )
//
RIAP TImage_SimpleByTwo_DownSampler::operator() ( const TRealImage & in )
{
	const int kInCol = in.GetCol();
	const int kInRow = in.GetRow();
	int outCol = kInCol / kDownStep;
	int outRow = kInRow / kDownStep;

	RIAP outImage( new TRealImage( outCol, outRow ) );

	register int i, j, inputRowIndex;

	for( j = 0; j < outRow; j ++ )
	{
		inputRowIndex = j * kDownStep;
		for( i = 0; i < outCol; i ++ )
		{
			outImage->SetPixel( i, j, in.GetPixel( i * kDownStep, inputRowIndex ) );
		}
	}

	return outImage;
}



/////////////////////////////////////////////////
RIAP TImage_PixelCopy_UpSampler::operator() ( const TRealImage & in )
{
	const int kUpStep = 2;
	const int kInCol = in.GetCol();
	const int kInRow = in.GetRow();
	int outCol = kInCol * kUpStep;
	int outRow = kInRow * kUpStep;

	RIAP outImage( new TRealImage( outCol, outRow ) );

	register int i, j, inputRowIndex;

	for( j = 0; j < outRow; j ++ )
	{
		inputRowIndex = j / kUpStep;
		REQUIRE( inputRowIndex < kInRow );
		for( i = 0; i < outCol; i ++ )
		{
			REQUIRE( i / kUpStep < kInCol );
			outImage->SetPixel( i, j, in.GetPixel( i / kUpStep, inputRowIndex ) );
		}
	}

	return outImage;
}



RIAP TImage_TwoNeighborInterpol_UpSampler::operator() ( const TRealImage & in )
{
	const int kInCol = in.GetCol();
	const int kInRow = in.GetRow();
	int outCol = kInCol * kUpStep;
	int outRow = kInRow * kUpStep;

	RIAP outImage( new TRealImage( outCol, outRow ) );

	register int i, j;

	// Fill each column: take 2 pixels from the input image,
	// then to the output set the first of these two and in the second their avrage.
	// Do this at each second row in the output image
	for( j = 0; j < kInRow; j ++ )
	{
		int out_image_row = j * kUpStep;

		double prev_pix = in.GetPixel( 0, j );
		double this_pix;

		for( i = 0; i < kInCol; i ++ )
		{

			this_pix = in.GetPixel( i, j );
	
			int out_image_column = kUpStep * i;

			outImage->SetPixel( out_image_column,		out_image_row,  ( this_pix + prev_pix ) / 2.0 );	// set a pixel being an average of the two pixels
			outImage->SetPixel( out_image_column + 1,	out_image_row,  this_pix );							// copy pixel
			

			prev_pix = this_pix;
			
		}
	
	}

	// Now interpolate missing rows. For this we do not need the original 
	// image anymore.
	for( j = 1; j < outRow; j += 2 )
	{
		for( i = 0; i < outCol; i ++ )
		{
			double pix_1 = outImage->GetPixel( i, j - 1 );
			double pix_2 = j + 1 < outRow ? outImage->GetPixel( i, j + 1 ) : pix_1;

			outImage->SetPixel( i, j, ( pix_1 + pix_2 ) / 2.0 );	// set a pixel being an average of the two pixels
		
		}

	}

	return outImage;
}
//////////////////////////////////////////////////////////////////////////////////////////////////////////

