
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ImageReSamplers.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the image by 2 re-samplers.
//
// ==========================================================================



#ifndef ImageReSamplers_h
#define ImageReSamplers_h


#include "HIL_AutoPtrDefs.h"

using namespace PHIL;



///////////////////////////////////////////////////////////
// This class defines an abstract hierarchy of image
// re-samplers. These can be down or up samplers.
///////////////////////////////////////////////////////////
class TImageReSampler
{
	private:

		// Class inherent variables

	protected:

	public:

		// ===================================================
		TImageReSampler( void ) {}				// class default constructor
		virtual ~TImageReSampler() {}			// class virtual destructor
		// ===================================================


		virtual RIAP operator() ( const TRealImage & in ) = 0;
};


////////////////////////////////////////////////////////////////
// This class implements a simple skip-by-two down sampling
// scheme. It must be stressed here that such a simple 
// down sampling can be used with properly low pass filtered
// images e.g. with Gaussain filter with appropirate mask size.
// Otherwise, the aliasing effect can be introduced into
// the output.
// The algorithm is quite easy and can be summarized as follows:
// 1) Low-pass filter the input image (outside this class)
// 2) Create a new image with dimensions equal to the
//    original image dimensions divided by the down-step factor
//    denoted as : kDownStep 
// 3) For each output pixel ( m, n ) set its value the same
//    as can be found in the original image at location:
//    ( m * kDownStep, n * kDownStep )
// 
class TImage_SimpleByTwo_DownSampler : public TImageReSampler
{
	private:

		// Class inherent variables

	protected:

		const int kDownStep;


	public:

		// ===================================================
		TImage_SimpleByTwo_DownSampler( void ) : kDownStep( 2 ) {}	// class default constructor
		virtual ~TImage_SimpleByTwo_DownSampler() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function down samples an input image by factor
		// set in kDownStep (two)
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - a reference to the original real image (unchanged)
		//		
		// OUTPUT:
		//			auto ptr to the down sampled copy of the original
		//		
		// REMARKS:
		//		
		// I  Create a new image with dimensions equal to the
		//    original image dimensions divided by kDownStep 
		// II For each output pixel ( m, n ) set its value the same
		//    as can be found in the original image at location:
		//    ( m * kDownStep, n * kDownStep )
		//
		RIAP operator() ( const TRealImage & in );
};






////////////////////////////////////////////////////////////////
// This class implements a simple the simplest up-sampler. It
// simply copies a neighbour pixel when necessary. 
// 
class TImage_PixelCopy_UpSampler : public TImageReSampler
{
	private:

		// Class inherent variables

	protected:

		const int kUpStep;

	public:

		// ===================================================
		TImage_PixelCopy_UpSampler( void ) : kUpStep( 2 ) {}	// class default constructor
		virtual ~TImage_PixelCopy_UpSampler() {}				// class virtual destructor
		// ===================================================


		RIAP operator() ( const TRealImage & in );
};


////////////////////////////////////////////////////////////////
// This class implements a simple up sampler that operates by
// means of missing pixels interpolation based on its nearest
// two neighbourhoods. The interpolation is done separately
// in horizontal, then vertical directions. 
//
// 
class TImage_TwoNeighborInterpol_UpSampler : public TImageReSampler
{
	private:

		// Class inherent variables

	protected:

		const int kUpStep;

	public:

		// ===================================================
		TImage_TwoNeighborInterpol_UpSampler( void ) : kUpStep( 2 ) {}	// class default constructor
		virtual ~TImage_TwoNeighborInterpol_UpSampler() {}				// class virtual destructor
		// ===================================================


		RIAP operator() ( const TRealImage & in );
};


#endif ImageReSamplers_h

