
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: StereoMatcherFor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A suite of stereo matching hierarchies.
//
// ==========================================================================



#ifndef BoxySoft_HIL_StereoMatcherFor
#define BoxySoft_HIL_StereoMatcherFor



#include "HIL_ImageFor.h"
#include "GraphicLibrary.h"




using namespace PHIL;




////////////////////////////////////////////////////////////////////////////////////////////////////////////////


template < typename T, typename U = long >
class Pixel_SAD_Metric
{
	public:

		typedef T PixelType;
		typedef U MatchType;

	public:

		///////////////////////////////////////////////////////////
		// This function returns a match between two points a and b
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			a - point a
		//			b - point b
		//		
		// OUTPUT:
		//			match measure between points a and b
		//		
		// REMARKS:
		//		
		//		
		U operator () ( const T a, const T b );


};

// Specialization
template <>
class Pixel_SAD_Metric< unsigned char, long >
{
	public:

		typedef unsigned char	PixelType;
		typedef long			MatchType;

	public:

		inline long operator () ( const unsigned char a, const unsigned char b )
		{
			return abs( a - b );
		}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


template < typename T, typename U = long >
class Pixel_SSD_Metric
{
	public:

		typedef T PixelType;
		typedef U MatchType;

	public:

		///////////////////////////////////////////////////////////
		// This function returns a match between two points a and b
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			a - point a
		//			b - point b
		//		
		// OUTPUT:
		//			match measure between points a and b
		//		
		// REMARKS:
		//		
		//		
		U operator () ( const T a, const T b );


};

// Specialization
template <>
class Pixel_SSD_Metric< unsigned char, long >
{
	public:

		typedef unsigned char	PixelType;
		typedef long			MatchType;

	public:

		inline long operator () ( const unsigned char a, const unsigned char b )
		{
			unsigned long dp = (unsigned long)a - (unsigned long)b;
			return dp * dp;
		}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


template < typename T, typename U = long >
class Pixel_SCP_Metric
{
	public:

		typedef T PixelType;
		typedef U MatchType;

	public:

		///////////////////////////////////////////////////////////
		// This function returns a match between two points a and b
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			a - point a
		//			b - point b
		//		
		// OUTPUT:
		//			match measure between points a and b
		//		
		// REMARKS:
		//		
		//		
		U operator () ( const T a, const T b );


};

// Specialization
template <>
class Pixel_SCP_Metric< double, double >
{
	public:

		typedef double PixelType;
		typedef double MatchType;

	public:

		inline double operator () ( const double a, const double b )
		{
			return a * b;
		}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Base template class for matching of two images (stereo)
//
// PIX_TYPE is a template parameter that defines type of a pixel of input images
// _DisparityType a template parameter that defines type of a single disparity
//
template < typename PIX_TYPE, typename _DisparityType >
class TStereoMatcher
{
	protected:

		const TImageFor< PIX_TYPE > & fLeftImage;
		const TImageFor< PIX_TYPE > & fRightImage;


	protected:

		typedef PIX_TYPE						PixelType;
		
		typedef _DisparityType					DisparityType;
		typedef TImageFor< DisparityType >		DisparityMap;

		DisparityMap * fDisparityMap;

	public:


		DisparityMap * GetDisparityMap( void ) const { return fDisparityMap; }

	public:


		TStereoMatcher( const TImageFor< PIX_TYPE > & leftImage, const TImageFor< PIX_TYPE > & rightImage )
			: fLeftImage( leftImage ), fRightImage( rightImage ), fDisparityMap( 0 ) {}

		virtual ~TStereoMatcher()
		{
			delete fDisparityMap;
		}


	public:

		// Match action
		virtual void operator () ( void ) = 0;

};

template < typename PIX_TYPE, typename DisparityType >
class TAreaBased_StereoMatcher : public TStereoMatcher< PIX_TYPE, DisparityType >
{
	protected:

		const int fMatch_area_cols;
		const int fMatch_area_rows;
		const int fMax_disp;

		const bool fDispSearchDirection;	// if true then forward search direction

	public:


		TAreaBased_StereoMatcher(	const TImageFor< PIX_TYPE > & leftImage, const TImageFor< PIX_TYPE > & rightImage,
									const int match_area_cols, const int match_area_rows, 
									const int max_disp, const bool dispSearchDirection )
			:	TStereoMatcher< PIX_TYPE, DisparityType >( leftImage, rightImage ),
				fMatch_area_cols( match_area_cols ), fMatch_area_rows( match_area_rows ), fMax_disp( max_disp ),
				fDispSearchDirection( dispSearchDirection )
		{}

		virtual ~TAreaBased_StereoMatcher() {}

	public:

		// Match action
		virtual void operator () ( void ) = 0;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template < typename T, typename _PixelMatchMeasure_Policy, typename DisparityType = long >
class TPointOriented_AreaBased_Matcher : public TAreaBased_StereoMatcher< T, DisparityType >
{
	protected:

		_PixelMatchMeasure_Policy	fPixelMatchMeasure_Policy;

		typedef typename _PixelMatchMeasure_Policy::MatchType MatchType;

	public:

		///////////////////////////////////////////////////////////
		// This function 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			leftImage - a reference to the left image of a
		//				stereo pair
		//			rightImage - a reference to the right image of a
		//				stereo pair
		//			match_area_cols - a horizontal size of the 
		//				matching window
		//			match_area_rows - a vertical size of the 
		//				matching window
		//			max_disp - expected maximum horizontal
		//				disparity value
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TPointOriented_AreaBased_Matcher(	const TImageFor< T > & leftImage, const TImageFor< T > & rightImage,
											const int match_area_cols, const int match_area_rows, 
											const int max_disp, const bool dispSearchDirection )
		: TAreaBased_StereoMatcher< T, DisparityType >( leftImage, rightImage, match_area_cols, match_area_rows, max_disp, dispSearchDirection )
		{}

		virtual ~TPointOriented_AreaBased_Matcher() {}

		///////////////////////////////////////////////////////////
		// This function computes a disparity map from the two
		// monochrome input images. The point-oriented (local)
		// algorithm is used. It is assumed that the input pair
		// is already rectified, so only horizontal scanlines
		// are searched for the matches.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			none 
		//		
		virtual void operator () ( void )
		{
			const int kTotalCols = fLeftImage.GetCol();
			const int kTotalRows = fLeftImage.GetRow();

			delete fDisparityMap;
			fDisparityMap = new DisparityMap( kTotalCols, kTotalRows, 0 );

			register int col, row, disp;

			// Traverse each row
			if( fDispSearchDirection == true )
			{
				for( row = 0; row < kTotalRows; ++ row )
				{
					// Traverse each column
					for( col = 0; col < kTotalCols; ++ col )
					{
						MatchType prev_match = (MatchType)LONG_MAX;		// get a large value
						DisparityType best_disp = (DisparityType)/*-1*/0;

						// Traverse each possible disparity
						for( disp = 0; disp < fMax_disp; ++ disp )
						{
							MatchType area_match_value;

							// Compute match value for a given area and disparity
							if( AreaMatch( col, row, disp, area_match_value ) == true )
							{
								// Compare this match to select the best one
								if( area_match_value < prev_match )
								{
									prev_match = area_match_value;	// save the best values
									best_disp = disp;		
								}
							}
						}

						// Store found the best disparity
						fDisparityMap->SetPixel( col, row, best_disp );
					}
				}
			}
			else
			{
				for( row = 0; row < kTotalRows; ++ row )
				{
					// Traverse each column
					for( col = 0; col < kTotalCols; ++ col )
					{
						MatchType prev_match = (MatchType)LONG_MAX;		// get a large value
						DisparityType best_disp = (DisparityType)/*-1*/0;

						// Traverse each possible disparity
						for( disp = 0; disp < fMax_disp; ++ disp )
						{
							MatchType area_match_value;

							// Compute match value for a given area and disparity
							if( AreaMatch( col, row, - disp, area_match_value ) == true )
							{
								// Compare this match to select the best one
								if( area_match_value < prev_match )
								{
									prev_match = area_match_value;	// save the best values
									best_disp = disp;		
								}
							}
						}

						// Store found the best disparity
						fDisparityMap->SetPixel( col, row, best_disp );
					}
				}
			}

		}


		///////////////////////////////////////////////////////////
		// This function computes an area match
		// in a window which is set around a pixel at coordinates
		// (col,row) in the reference image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			col - a horizontal coordinate of a reference 
		//				pixel in the left image
		//			row - a vertical coordinate of a reference 
		//				pixel in the left image
		//			disp - a disparity value between two images
		//			match_value - a reference to the variable
		//				that upon successful return contains
		//				the computed match value between two
		//				matched areas
		//		
		// OUTPUT:
		//			true - if success (match_value contains match value)
		//			false - otherwise 
		//		
		virtual bool AreaMatch(	const int col, const int row, const int disp, MatchType & match_value )
		{
			const int kTotalCols = fLeftImage.GetCol();
			const int kTotalRows = fLeftImage.GetRow();

			const int kColSwing = fMatch_area_cols / 2;
			const int kRowSwing = fMatch_area_rows / 2;

			// Find border values for column index
			int col_from = col - kColSwing;	// left
			if( col_from < 0 )
				col_from = 0;
			int col_to = col + kColSwing;	// right
			if( col_to > kTotalCols )
				col_to = kTotalCols;

			// Find border values for row index
			int row_from = row - kRowSwing;	// top
			if( row_from < 0 )
				row_from = 0;
			int row_to = row + kRowSwing;	// bottom
			if( row_to > kTotalRows )
				row_to = kTotalRows;

			match_value = 0.0;	// starting match value

			register int r, c;
			for( c = col_from; c < col_to; ++ c )
			{
				int right_c = c - disp;	// find column index
				if( right_c < 0 || right_c >= kTotalCols )		// for the right image
					return false;		// exit - cannot compute this match

				for( r = row_from; r < row_to; ++ r )
				{
					// update match value for each pair of pixels
					match_value += fPixelMatchMeasure_Policy( fLeftImage.GetPixel( c, r ), fRightImage.GetPixel( right_c, r ) );
				}
			}

			return true;	// successful computation of a match value
		}

};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template < typename T, typename _PixelMatchMeasure_Policy, typename DisparityType = double >
class TPointOriented_ZeroMeanAreaBased_Matcher : public TPointOriented_AreaBased_Matcher< T, _PixelMatchMeasure_Policy, DisparityType >
{

	public:

		///////////////////////////////////////////////////////////
		// This function 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			leftImage - a reference to the left image of a
		//				stereo pair
		//			rightImage - a reference to the right image of a
		//				stereo pair
		//			match_area_cols - a horizontal size of the 
		//				matching window
		//			match_area_rows - a vertical size of the 
		//				matching window
		//			max_disp - expected maximum horizontal
		//				disparity value
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TPointOriented_ZeroMeanAreaBased_Matcher(	const TImageFor< T > & leftImage, const TImageFor< T > & rightImage,
													const int match_area_cols, const int match_area_rows, 
													const int max_disp, const bool dispSearchDirection )
		: TPointOriented_AreaBased_Matcher< T, _PixelMatchMeasure_Policy, DisparityType >( leftImage, rightImage, match_area_cols, match_area_rows, max_disp, dispSearchDirection )
		{}

		virtual ~TPointOriented_ZeroMeanAreaBased_Matcher() {}



		///////////////////////////////////////////////////////////
		// This function computes an area match
		// in a window which is set around a pixel at coordinates
		// (col,row) in the reference image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			col - a horizontal coordinate of a reference 
		//				pixel in the left image
		//			row - a vertical coordinate of a reference 
		//				pixel in the left image
		//			disp - a disparity value between two images
		//			match_value - a reference to the variable
		//				that upon successful return contains
		//				the computed match value between two
		//				matched areas
		//		
		// OUTPUT:
		//			true - if success (match_value contains match value)
		//			false - otherwise 
		//		
		virtual bool AreaMatch(	const int col, const int row, const int disp, MatchType & match_value )
		{
			const int kTotalCols = fLeftImage.GetCol();
			const int kTotalRows = fLeftImage.GetRow();

			const int kColSwing = fMatch_area_cols / 2;
			const int kRowSwing = fMatch_area_rows / 2;

			// Find border values for column index
			int col_from = col - kColSwing;	// left
			if( col_from < 0 )
				col_from = 0;
			int col_to = col + kColSwing;	// right
			if( col_to > kTotalCols )
				col_to = kTotalCols;

			// Find border values for row index
			int row_from = row - kRowSwing;	// top
			if( row_from < 0 )
				row_from = 0;
			int row_to = row + kRowSwing;	// bottom
			if( row_to > kTotalRows )
				row_to = kTotalRows;

			register int r, c;

			// Simple version
			MatchType left_mean = 0;
			MatchType right_mean = 0;

			MatchType pixel_counter = 0;

			for( c = col_from; c < col_to; ++ c )
			{
				int right_c = c - disp;	// find column index
				if( right_c < 0 || right_c >= kTotalCols )		// for the right image
					return false;		// exit - cannot compute this match

				for( r = row_from; r < row_to; ++ r )
				{
					left_mean += fLeftImage.GetPixel( c, r );
					right_mean += fRightImage.GetPixel( right_c, r );

					++ pixel_counter;
				}
			}

			if( pixel_counter == 0 )
				return false;				// cannot compute this measure

			left_mean /= pixel_counter;		// compute the mean value in the left image
			right_mean /= pixel_counter;	// compute the mean value in the right image

			match_value = 0;	// starting match value

			MatchType pix_diff_1 = 0, pix_diff_2 = 0;
			MatchType sum_1 = 0, sum_2 = 0;

			for( c = col_from; c < col_to; ++ c )
			{
				int right_c = c - disp;	// find column index

				for( r = row_from; r < row_to; ++ r )
				{
					pix_diff_1 = fLeftImage.GetPixel( c, r ) - left_mean;
					pix_diff_2 = fRightImage.GetPixel( right_c, r ) - right_mean;

					// update match value for each pair of pixels
					match_value += fPixelMatchMeasure_Policy( pix_diff_1, pix_diff_2 );
				
					sum_1 += pix_diff_1 * pix_diff_1;
					sum_2 += pix_diff_2 * pix_diff_2;
				}
			}

			if( sum_1 == 0 || sum_2 == 0 )
				return false;			// cannot compute this measure

			match_value /= - sqrt( sum_1 * sum_2 );

			return true;	// successful computation of a match value
		}

};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template < typename T, typename PixelMatchMeasure_Policy, typename DisparityType = double >
class TDisparityOriented_AreaBased_Matcher : public TAreaBased_StereoMatcher< T, DisparityType >
{

	protected:

		void DeleteDisparitySpace( void )
		{
			if( fDisparitySpace != 0 )
			{
				for( int disp = 0; disp < fDisparitySpace->size(); ++ disp )
					delete ( * fDisparitySpace )[ disp ];

				delete fDisparitySpace;
			}
		}

	protected:

		PixelMatchMeasure_Policy	fPixelMatchMeasure_Policy;

		typedef typename PixelMatchMeasure_Policy::MatchType MatchType;
		typedef typename TImageFor< MatchType >				DisparitySlice;
		typedef typename vector< DisparitySlice * >			DisparitySpace;

		DisparitySpace * fDisparitySpace;	// this is our disparity space
	
	protected:

		// The scaling factor for computation of disparity
		// with sub-pixel accuracy. For floating-point disparity
		// type it can be set to 1.0. For integer values of 
		// disparity it should be chosen as follows:
		//
		//		kScalingFactor * max_disparity <= max_value_for_disparity_type
		//	
		// For example, if max_disparity ~ 20, disparity is set on unsigned char, then:
		//		kScalingFactor <= 255 / 20 ~ 12
		// 
		// Thanks to this scaling factor we extend range of disparities
		// so the full dynamics of data is used and there is "room" for
		// sub-pixel values.
		//
		const MatchType kScalingFactor;	

	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			leftImage - a reference to the left image of a
		//				stereo pair
		//			rightImage - a reference to the right image of a
		//				stereo pair
		//			match_area_cols - a horizontal size of the 
		//				matching window
		//			match_area_rows - a vertical size of the 
		//				matching window
		//			max_disp - expected maximum horizontal
		//				disparity value
		//			scalingFactor - a multiplication factor for
		//				output disparities used for computation
		//				with sub-pixel accuracy on integer data
		//				format of the disparity maps
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TDisparityOriented_AreaBased_Matcher(	const TImageFor< T > & leftImage, const TImageFor< T > & rightImage,
												const int match_area_cols, const int match_area_rows, 
												const int max_disp, const bool dispSearchDirection,
												const MatchType scalingFactor = 100 )
		:	TAreaBased_StereoMatcher< T, DisparityType >( leftImage, rightImage, match_area_cols, match_area_rows, max_disp, dispSearchDirection ),
			fDisparitySpace( 0 ), kScalingFactor( scalingFactor )
		{}


		virtual ~TDisparityOriented_AreaBased_Matcher()
		{
			DeleteDisparitySpace();
		}

		///////////////////////////////////////////////////////////
		// This function computes a disparity map from the two
		// monochrome input images. The disparity-oriented
		// algorithm is used. It is assumed that the input pair
		// is already rectified, so only horizontal scanlines
		// are searched for the matches.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none (set in the constructor)
		// OUTPUT:
		//			none 
		//		
		// REMARKS:
		//		
		//		
		virtual void operator () ( void )
		{
			const int kTotalCols = fLeftImage.GetCol();
			const int kTotalRows = fRightImage.GetRow();

			register int col, row, disp;

			DeleteDisparitySpace();
			fDisparitySpace = new DisparitySpace( fMax_disp );

			// Traverse each possible disparity
			for( disp = 0; disp < fMax_disp; ++ disp )
			{
				// Create a single disparity "plane" for disp disparity
				( * fDisparitySpace )[ disp ] = new DisparitySlice( kTotalCols, kTotalRows, (MatchType)0 );

				if( fDispSearchDirection == true )
				{
					// Traverse each column
					for( col = 0; col < kTotalCols; ++ col )
					{
						int right_col = col - disp;
						if( right_col < 0 )
							continue;

						// Traverse each row
						for( row = 0; row < kTotalRows; ++ row )
						{
							MatchType pixel_match_value = fPixelMatchMeasure_Policy( fLeftImage.GetPixel( col, row ), fRightImage.GetPixel( right_col, row ) );
							( * fDisparitySpace )[ disp ]->SetPixel( col, row, pixel_match_value );
						}
					}
				}
				else
				{
					// Traverse each column
					for( col = 0; col < kTotalCols; ++ col )
					{
						int right_col = col + disp;
						if( right_col >= kTotalCols )
							continue;

						// Traverse each row
						for( row = 0; row < kTotalRows; ++ row )
						{
							MatchType pixel_match_value = fPixelMatchMeasure_Policy( fLeftImage.GetPixel( col, row ), fRightImage.GetPixel( right_col, row ) );
							( * fDisparitySpace )[ disp ]->SetPixel( col, row, pixel_match_value );
						}
					}
				}

			}

			// At this point we have created the disparity space.
			// In disparity space we can find a GLOBAL optimum
			DisparityFromDisparitySpace();

		}

		///////////////////////////////////////////////////////////
		// This function computes a disparity map from the disparity
		// space.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			disparityMap - a reference to an image that
		//				upon return contains the disparity map;
		//			disparitySpace - a reference to the disparity
		//				space (a disparity vector); Number of 
		//				elements in this vector is the same
		//				as the maximum horizontal disparity
		//			match_area_cols - a horizontal size of the 
		//				matching window
		//			match_area_rows - a vertical size of the 
		//				matching window
		//		
		// OUTPUT:
		//			none 
		//		
		// REMARKS:
		//		
		//		
		virtual bool DisparityFromDisparitySpace( void )
		{
			if( fDisparitySpace == 0 )
				return false;

			REQUIRE( fDisparitySpace->size() > 0 );
			REQUIRE( ( * fDisparitySpace )[ 0 ] != 0 );

			const int kTotalCols = ( * fDisparitySpace )[ 0 ]->GetCol();
			const int kTotalRows = ( * fDisparitySpace )[ 0 ]->GetRow();

			register int col, row, disp;

			vector< MatchType > theAggregMask_horz( fMatch_area_cols, (MatchType) 1 );
			vector< MatchType > theAggregMask_vert( fMatch_area_rows, (MatchType) 1 );

			DisparitySlice tmpImage( kTotalCols, kTotalRows );
			// Smooth the disparity space
			for( disp = 0; disp < fMax_disp; ++ disp )
			{
				Horz1DConvolve( * ( * fDisparitySpace )[ disp ], theAggregMask_horz, tmpImage );
				Vert1DConvolve( tmpImage, theAggregMask_vert, * ( * fDisparitySpace )[ disp ] );
			}


			delete fDisparityMap;
			fDisparityMap = new DisparityMap( kTotalCols, kTotalRows );

			// Traverse each row
			for( row = 0; row < kTotalRows; ++ row )
			{
				// Traverse each column
				for( col = 0; col < kTotalCols; ++ col )
				{
					MatchType		prev_match = LONG_MAX;		// get a large value
					DisparityType	best_disp = (DisparityType)/*-1*/0;

					// Traverse each possible disparity
					for( disp = 0; disp < fMax_disp; ++ disp )
					{
						MatchType area_match_value = ( * fDisparitySpace )[ disp ]->GetPixel( col, row );
						if( area_match_value < prev_match )
						{
							// save the best values
							prev_match	= area_match_value;	
							best_disp	= disp;		
						}
					}

					///////////////////////////////////////////////////////////////////////////
					// Sub-pixel estimation:
					MatchType sub_pix_disparity = kScalingFactor * best_disp;
					if( best_disp > 0 && best_disp < fMax_disp - 1 ) 
					{
						MatchType m_im1	= ( * fDisparitySpace )[ best_disp - 1 ]->GetPixel( col, row );
						MatchType m_i	= ( * fDisparitySpace )[ best_disp     ]->GetPixel( col, row );
						MatchType m_ip1	= ( * fDisparitySpace )[ best_disp + 1 ]->GetPixel( col, row );

						MatchType denom	= m_im1 - m_i - m_i + m_ip1;
						if( denom != 0.0 )
							sub_pix_disparity += ( kScalingFactor * ( m_im1 - m_ip1 ) ) / ( denom + denom );
					}
					///////////////////////////////////////////////////////////////////////////

					// Save found disparity
					fDisparityMap->SetPixel( col, row, sub_pix_disparity );
				}
			}

			return true;
		}


};


////////////////////////////////////////////////////////////////////////////////////////////////////////////////


//
// Interesingly enough, the cross checking is also kind of 
// two image matching, so we derive it from the common
// stereo framework.
//
// PIX_TYPE is actually disparity type from other
// matching methods.
//
// DisparityType - type of an output disparity map;
// Usually we set: DisparityType == PIX_TYPE
//
template < typename PIX_TYPE, typename DisparityType >
class TDisparityMap_CrossCheck_Matcher : public TStereoMatcher< PIX_TYPE, DisparityType >
{
	protected:

		const int kDisparitySimilarityThresh;

		int fRejectedPoints_Counter;

		DisparityMap * fRejectedPointsMap;

		enum { kGoodPointMarker = 0, kRejectedPointMarker = 1 };

	public:

		DisparityMap *	GetRejectedPointsMap( void ) const { return fRejectedPointsMap; }
		int				GetRejectedPoints_Counter( void ) const { return fRejectedPoints_Counter; }

	public:


		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			leftRigh_DispMap - the first disparity map
		//				with left image set as a reference
		//			rightLeft_DispMap - the second disparity map
		//				with right image kept as a reference
		//			disparitySimilarityThresh - threshold value
		//				of allowable dissimilarity between 
		//				the pairs of disparities 
		//		
		// REMARKS:
		//			Disparities should be true disparities, 
		//			i.e. they should not be scaled.
		//		
		TDisparityMap_CrossCheck_Matcher(	const TImageFor< PIX_TYPE > & leftRigh_DispMap, const TImageFor< PIX_TYPE > & rightLeft_DispMap,
											const int disparitySimilarityThresh = 1 )
			:	TStereoMatcher< PIX_TYPE, DisparityType >( leftRigh_DispMap, rightLeft_DispMap ),
				kDisparitySimilarityThresh( disparitySimilarityThresh ), fRejectedPoints_Counter( 0 ), fRejectedPointsMap( 0 )
		{}

		virtual ~TDisparityMap_CrossCheck_Matcher() 
		{
			delete fRejectedPointsMap;
		}

	public:

		///////////////////////////////////////////////////////////
		// This function does stereo cross-checking.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			Input parameters are supplied during
		//				construction of this object
		//		
		// OUTPUT:
		//			The following members are set upon completion:
		//			fDisparityMap - an output cross-checked 
		//				disparity map
		//			fRejectedPointsMap - an image of rejected
		//				points (denoted by kRejectedPointMarker)
		//			fRejectedPoints_Counter - contains a number
		//				of rejected points (this can be obtained
		//				from the fRejectedPointsMap, as well)
		//		
		// REMARKS:
		//		
		//		
		virtual void operator () ( void )
		{
			fRejectedPoints_Counter = 0;

			const int kCols = fLeftImage.GetCol();
			const int kRows = fRightImage.GetRow();

			REQUIRE( kCols == fRightImage.GetCol() );
			REQUIRE( kRows == fRightImage.GetRow() );

			// Create an output disparity map (i.e. the one after cross checking)
			delete fDisparityMap;
			fDisparityMap = new DisparityMap( kCols, kRows, (DisparityType)0 );

			// Create a map of rejected points
			delete fRejectedPointsMap;
			fRejectedPointsMap = new DisparityMap( kCols, kRows, kGoodPointMarker );


			register int i, j, j_off;

			PixelType left_disp, right_disp;

			for( i = 0; i < kRows; i ++ )
			{
				for( j = 0; j < kCols; j ++ )
				{
					left_disp = fLeftImage.GetPixel( j, i );

					j_off = j /*+*/- left_disp;
					if( j_off >= kCols || j_off < 0 )
						continue;

					right_disp = fRightImage.GetPixel( j_off, i );

					if( abs( left_disp - right_disp ) <= kDisparitySimilarityThresh )
					{
						// ok, match validated ok
						fDisparityMap->SetPixel( j, i, left_disp );
					}
					else
					{
						// unfortunately, this is an outlier, i.e. a rejected point
						++ fRejectedPoints_Counter;
						fRejectedPointsMap->SetPixel( j, i, kRejectedPointMarker );
					}

				}

			}
			
		}


};



////////////////////////////////////////////////////////////////////////////////////////////////////////////////



#endif BoxySoft_HIL_StereoMatcherFor




