
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: LogPolarSpace.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the log-polar space converters.
//
// ==========================================================================


#include "LogPolarSpace.h"



///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TLogPolar_Space::TLogPolar_Space( void )		// class constructor
{
}


///////////////////////////////////////////////////////////
// This function 
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
TLogPolar_Space::~TLogPolar_Space()		// class destructor
{
}


///////////////////////////////////////////////////////////
// This function determines the best base of the logarithm
// (a natural logarithm) such as the log-polar transformation
// will cover exactly the range of the new image given
// by its cols-rows dimensions and an anchor point.
// This is to solve a problem how inscribe the best 
// a circle within a rectangle.
//
//	dist = sqrt( ( x - xc ) ^ 2 + ( y - yc ) ^ 2 )
//	r = log_B( dist ) 
//
// Thus:
//
//	B = exp( log( dist ) / (2*r_max) )
//
///////////////////////////////////////////////////////////
//		
// INPUT:
//			anchorPointInTheNewImage - the central point
//				of the log-polar transformation (an anchor
//				point, an origin point, etc.)
//			image_Cols - num of cols of the input (cartesian 
//				representation) image (i.e. a rectangle)
//			image_Rows - num of rows of the input (cartesian 
//				representation) image (i.e. a rectangle)
//			r_max - an expected maximum radius of the log-polar
//				representation (usually this is a number of columns,
//				or rows in an image that holds the log-polar representation
//				of the input cartesian image)
//		
// OUTPUT:
//			The BASE of the natural logarithm (ln) that allows 
//			the best coverage when transforming from the cartesian
//			to the log-polar space.
//		
// REMARKS:
//		
//		
double TLogPolar_Space::GetLogBaseToCoverTheSpace( const Real_2D_Point const & anchorPointInTheNewImage, int image_Cols, int image_Rows, double r_max, bool diagonal )
{
	double dx_max, dy_max, d_max, B;
	if( diagonal == true )
	{
		dx_max = max( x( anchorPointInTheNewImage ), image_Cols - x( anchorPointInTheNewImage ) );
		dy_max = max( y( anchorPointInTheNewImage ), image_Rows - y( anchorPointInTheNewImage ) );
		d_max = dx_max * dx_max + dy_max * dy_max;
		B = exp( log( d_max ) / ( 2.0 * r_max ) );
	}
	else
	{
		dx_max	= min( x( anchorPointInTheNewImage ), image_Cols - x( anchorPointInTheNewImage ) );
		dy_max	= min( y( anchorPointInTheNewImage ), image_Rows - y( anchorPointInTheNewImage ) );
		d_max	= min( dx_max, dy_max );
		B		= exp( log( d_max ) / r_max );
	}

	return B;
}

///////////////////////////////////////////////////////////
// This function prepares the generic (i.e. a look-up table)
// mapping (transformation) from the cartesian to the log-polar 
// coordinate space.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			kCartesianImage_Cols - horz size of the input
//				cartesian space
//			kCartesianImage_Rows - vert size of the input
//				cartesian space
//			centralPointInCartesianSubSpaceRect - the central
//				(anchor) point in the input spcace; usually
//				it is: (kCartesianImage_Cols/2, kCartesianImage_Rows/2)
//			logPolarImage_Cols - horz size of the output log-polar space
//			logPolarImage_Rows - vert size of the output log-polar space
//			diagonal - if true then maximal radius is computed
//				from the central point to the furthest corner
//				(thus a maximal circle is taken)
//				If false, then the minimal distance to one of
//				the sides is taken (thus a minimal but totally
//				inner circle is taken).//		
// OUTPUT:
//			An auto-ptr containing the generic coordinate 
//			transformation from the cartesian to the log-polar spaces.
//		
// REMARKS:
//		
//		
TLogPolar_Space::TGenericTransformEngine_AP TLogPolar_Space::Orphan_LogPolar_GenericTransformationEngine( 
							const int kCartesianImage_Cols, const int kCartesianImage_Rows, const Real_2D_Point const & centralPointInCartesianSubSpaceRect,
							const int logPolarImage_Cols, const int logPolarImage_Rows, bool diagonal  )
{
	TGenericTransformEngine_AP retVal( new TGenericTransformEngine ( logPolarImage_Cols, logPolarImage_Rows ) );

	// In the log-polar image the horizontal axis denotes "r", whereas the vertical "phi".
	double r_max = logPolarImage_Cols;
	double theLnBase = GetLogBaseToCoverTheSpace( centralPointInCartesianSubSpaceRect, kCartesianImage_Cols, kCartesianImage_Rows, r_max, diagonal );

	double * power_LUT = new double[ logPolarImage_Cols ];

	register int i, j;

	for( i = 0; i < logPolarImage_Cols; ++ i )				// B^r
		power_LUT[ i ] = pow( theLnBase, (double)i );

	const double kFromAngle = 0;
	const double kToAngle = k2Pi;
	const double kStep = ( kToAngle - kFromAngle ) / (double)logPolarImage_Rows;

	// Now we are ready to fill up the log-polar image
	double c1 = x( centralPointInCartesianSubSpaceRect);
	double c2 = y( centralPointInCartesianSubSpaceRect);

	double x1, x2;
	double alpha;
	double abs_cos_alpha, cos_alpha, sin_alpha;
	double sgn_cos_alpha;
	double power_value;

	// Go through angles (and rows)
	for( alpha = kFromAngle, i = 0; i < logPolarImage_Rows; alpha += kStep, ++ i )		// cos(alpha) and sgn(.)
	{
		cos_alpha = cos( alpha );
		sin_alpha = sin( alpha );

		if( cos_alpha >= 0.0 )
		{
			abs_cos_alpha = cos_alpha;
			sgn_cos_alpha = 1.0;				// true for positive
		}
		else
		{
			abs_cos_alpha = - cos_alpha;
			sgn_cos_alpha = -1.0;	
		}

		// Go through radius (and columns)
		for( j = 0; j < logPolarImage_Cols; ++ j )	// r
		{	
			power_value = power_LUT[ j ];

			x1 = cos_alpha * power_value + c1;
			x2 = sin_alpha * power_value + c2;

			retVal->Get_ImageOf_Real_2D_Point().SetPixel( j, i, Real_2D_Point( x1, x2 ) );			
		}
	}

	delete [] power_LUT;

	return retVal;
}

///////////////////////////////////////////////////////////
// This function orphans a real image being a log-polar
// representation of the input image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inImage - the input cartesian space image
//			coordMappingEngine - the proper coordinates mapping
//				engine that fits the coord space of the input
//				image (actually it defines a bacward transformation
//				from the log-polar to the cartesian input space
//				for the inverse warping scheme)
//		
// OUTPUT:
//			the real image with the log-polar representation
//			of the input image, if success;
//			otherwise a null auto ptr;
//		
// REMARKS:
//		
//		
RIAP TLogPolar_Space::Orphan_LogPolarRepresentationOfImage( const TRealImage const & inImage, TGenericTransformEngine_AP & coordMappingEngine )
{
	REQUIRE( coordMappingEngine.get() != 0 );
	TSimpleImageBackwardWarp< double > simpleWarpEngine( inImage, * coordMappingEngine );
	return simpleWarpEngine();
}



///////////////////////////////////////////////////////////
// This function orphans a real image being a log-polar
// representation of the input image patch
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inImage - the input cartesian space image
//			s_col - upper left coordinate of the new image (expressed
//				in the mother coordinate system)
//			s_row - top coordinate of the new image
//			width - width of a patch in the input image
//			height - height of a patch in the input image
//			coordMappingEngine - the proper coordinates mapping
//				engine that fits the coord space of the input
//				image (actually it defines a bacward transformation
//				from the log-polar to the cartesian input space
//				for the inverse warping scheme)
//		
// OUTPUT:
//			the real image with the log-polar representation
//			of the input image, if success;
//			otherwise a null auto ptr;
//		
// REMARKS:
//		
//		
RIAP TLogPolar_Space::Orphan_LogPolarRepresentationOfImage( MonochromeImage & inImage, int s_col, int s_row, int width, int height, TGenericTransformEngine_AP & coordMappingEngine )
{
	MonochromeImageProxy theProxy( inImage, s_col, s_row, width, height );
	TRealImage tmp( width, height );
	MixedCopy( tmp, theProxy );
	return Orphan_LogPolarRepresentationOfImage( tmp, coordMappingEngine );
}


RIAP TLogPolar_Space::Orphan_LogPolarRepresentationOfImage( MonochromeImage & inImage, int s_col, int s_row, int width, int height, bool center_of_gravity, bool diagonal )
{
	Real_2D_Point centralPoint4LogPolar( width / 2, height / 2 );		// let's assume a geometrical center at first
	if( center_of_gravity == true )
	{
		MonochromeImageProxy paneImage( inImage, s_col, s_row, width, height );
		Real_2D_Point theCentroid = CenterOfGravity( paneImage );
		centralPoint4LogPolar = theCentroid;
		REQUIRE( x( theCentroid ) >= 0 );
		REQUIRE( x( theCentroid ) < width );
		REQUIRE( y( theCentroid ) >= 0 );
		REQUIRE( y( theCentroid ) < height );
	}

	TGenericTransformEngine_AP coordMappingEngine( Orphan_LogPolar_GenericTransformationEngine( width, height, centralPoint4LogPolar, width, height, diagonal ) );

	REQUIRE( coordMappingEngine.get() != 0 );
	if( coordMappingEngine.get() == 0 )
		return RIAP( 0 );
	
	return Orphan_LogPolarRepresentationOfImage( inImage, s_col, s_row, width, height, coordMappingEngine );
}





RIAP TLogPolar_Space::Orphan_LogPolarRepresentationOfImage( TRealImage & inImage, int s_col, int s_row, int width, int height, TGenericTransformEngine_AP & coordMappingEngine )
{
	RIAP tmp( inImage.OrphanBitBlit( s_col, s_row, width, height ) );
	return Orphan_LogPolarRepresentationOfImage( * tmp, coordMappingEngine );
}



