
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: LogPolarSpace.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the log-polar space converters.
//
// ==========================================================================





#ifndef BoxySoft_LogPolarSpace
#define BoxySoft_LogPolarSpace



#include "Warping.h"
#include "Moments.h"




///////////////////////////////////////////////////////////
// This class transforms images into the log-polar
// representation.
///////////////////////////////////////////////////////////
class TLogPolar_Space
{
	public:

		typedef std::auto_ptr< TGenericTransformEngine > TGenericTransformEngine_AP;

	private:

		// Class inherent variables

	protected:


		///////////////////////////////////////////////////////////
		// This function determines the best base of the logarithm
		// (a natural logarithm) such as the log-polar transformation
		// will cover exactly the range of the new image given
		// by its cols-rows dimensions and an anchor point.
		// This is to solve a problem how inscribe the best 
		// a circle within a rectangle.
		//
		//	dist = sqrt( ( x - xc ) ^ 2 + ( y - yc ) ^ 2 )
		//	r = log_B( dist ) 
		//
		// Thus:
		//
		//	B = exp( log( dist ) / (2*r_max) )
		//
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			anchorPointInTheNewImage - the central point
		//				of the log-polar transformation (an anchor
		//				point, an origin point, etc.)
		//			image_Cols - num of cols of the input (cartesian 
		//				representation) image (i.e. a rectangle)
		//			image_Rows - num of rows of the input (cartesian 
		//				representation) image (i.e. a rectangle)
		//			r_max - an expected maximum radius of the log-polar
		//				representation (usually this is a number of columns,
		//				or rows in an image that holds the log-polar representation
		//				of the input cartesian image)
		//			diagonal - if true then maximal radius is computed
		//				from the central point to the furthest corner
		//				(thus a maximal circle is taken)
		//				If false, then the minimal distance to one of
		//				the sides is taken (thus a minimal but totally
		//				inner circle is taken).
		//		
		// OUTPUT:
		//			The BASE of the natural logarithm (ln) that allows 
		//			the best coverage when transforming from the cartesian
		//			to the log-polar space.
		//		
		// REMARKS:
		//		
		//		
		inline double GetLogBaseToCoverTheSpace( const Real_2D_Point const & anchorPointInTheNewImage, int image_Cols, int image_Rows, double r_max, bool diagonal );


	public:

		// ===================================================
		TLogPolar_Space( void );				// class default constructor
		virtual ~TLogPolar_Space();				// class virtual destructor
		// ===================================================


	public:


		///////////////////////////////////////////////////////////
		// This function prepares the generic (i.e. a look-up table)
		// mapping (transformation) from the cartesian to the log-polar 
		// coordinate space.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			kCartesianImage_Cols - horz size of the input
		//				cartesian space
		//			kCartesianImage_Rows - vert size of the input
		//				cartesian space
		//			centralPointInCartesianSubSpaceRect - the central
		//				(anchor) point in the input spcace; usually
		//				it is: (kCartesianImage_Cols/2, kCartesianImage_Rows/2)
		//			logPolarImage_Cols - horz size of the output log-polar space
		//			logPolarImage_Rows - vert size of the output log-polar space
		//			diagonal - if true then maximal radius is computed
		//				from the central point to the furthest corner
		//				(thus a maximal circle is taken)
		//				If false, then the minimal distance to one of
		//				the sides is taken (thus a minimal but totally
		//				inner circle is taken).
		//		
		// OUTPUT:
		//			An auto-ptr containing the generic coordinate 
		//			transformation from the cartesian to the log-polar spaces.
		//		
		// REMARKS:
		//		
		//		
		TGenericTransformEngine_AP Orphan_LogPolar_GenericTransformationEngine( 
									const int kCartesianImage_Cols, const int kCartesianImage_Rows, const Real_2D_Point const & centralPointInCartesianSubSpaceRect,
									const int logPolarImage_Cols, const int logPolarImage_Rows, bool diagonal );


		///////////////////////////////////////////////////////////
		// This function orphans a real image being a log-polar
		// representation of the input image
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inImage - the input cartesian space image
		//			coordMappingEngine - the proper coordinates mapping
		//				engine that fits the coord space of the input
		//				image (actually it defines a bacward transformation
		//				from the log-polar to the cartesian input space
		//				for the inverse warping scheme)
		//		
		// OUTPUT:
		//			the real image with the log-polar representation
		//			of the input image, if success;
		//			otherwise a null auto ptr;
		//		
		// REMARKS:
		//		
		//		
		inline RIAP Orphan_LogPolarRepresentationOfImage( const TRealImage const & inImage, TGenericTransformEngine_AP & coordMappingEngine );

		///////////////////////////////////////////////////////////
		// This function orphans a real image being a log-polar
		// representation of the input image patch
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inImage - the input cartesian space image
		//			s_col - upper left coordinate of the new image (expressed
		//				in the mother coordinate system)
		//			s_row - top coordinate of the new image
		//			width - width of a patch in the input image
		//			height - height of a patch in the input image
		//			coordMappingEngine - the proper coordinates mapping
		//				engine that fits the coord space of the input
		//				image (actually it defines a bacward transformation
		//				from the log-polar to the cartesian input space
		//				for the inverse warping scheme)
		//		
		// OUTPUT:
		//			the real image with the log-polar representation
		//			of the input image, if success;
		//			otherwise a null auto ptr;
		//		
		// REMARKS:
		//		
		//		
		RIAP Orphan_LogPolarRepresentationOfImage( MonochromeImage & inImage, int s_col, int s_row, int width, int height, TGenericTransformEngine_AP & coordMappingEngine );
		RIAP Orphan_LogPolarRepresentationOfImage( TRealImage & inImage, int s_col, int s_row, int width, int height, TGenericTransformEngine_AP & coordMappingEngine );
		// Input and output images are of the same size
		// If center_of_gravity == true then the center of gravity will be computed and taken as a central point for LP
		RIAP Orphan_LogPolarRepresentationOfImage( MonochromeImage & inImage, int s_col, int s_row, int width, int height, bool center_of_gravity, bool diagonal );

};


#endif BoxySoft_LogPolarSpace




