
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: PolyWarpingParameterCalculator.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the polynomial warping classes.
//
// ==========================================================================



#include "PolyWarpingParameterCalculator.h"
#include "ImageMatrix.h"




///////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This function does 2nd order polynomial warping on
// the supplied mono image.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inMonoImage - reference to the input image
//				which transformed (warped) copy will
//				be returned upon success
//			origImagePoints - a collection of the points
//				selected in the original image space
//			mappedImagePoints - a collection of the points
//				defining new positions (the mapped space)
//		
// OUTPUT:
//			warped image if success
//			0 otherwise
//		
// REMARKS:
//			The returned object is an auto pointer object
//			that contains orphaned object on success
//		
MIAP OrphanPolyWarpedImage( const MonochromeImage & inMonoImage, Int_2D_Point_Vector & origImagePoints, Int_2D_Point_Vector & mappedImagePoints )
{
	T_2ndOrder_Polynomial_TransformEngine						the2ndOrderPolyTransformEngine;
	
	T_2ndOrder_Polynomial_TransformEngine::Order2ndPolyMatrix	thePolyMatrix;	
	if( DeterminePolynomialTransformParameters( origImagePoints, mappedImagePoints, thePolyMatrix ) == false )
		return MIAP( 0 );

	the2ndOrderPolyTransformEngine.Set2ndPolyMatrix( thePolyMatrix );

	TRealImage inRealImage( inMonoImage.GetCol(), inMonoImage.GetRow() );
	MixedCopy( inRealImage, inMonoImage );

	TSimpleImageBackwardWarp< double > simpleWarpEngine( inRealImage, the2ndOrderPolyTransformEngine );

	RIAP outRealImage( simpleWarpEngine() );
	if( outRealImage.get() == 0 )
		return MIAP( 0 );

	MIAP outMonoImage( new MonochromeImage( inMonoImage.GetCol(), inMonoImage.GetRow() ) );
	MixedCopy( * outMonoImage.get(), * outRealImage );

	return outMonoImage;
}




///////////////////////////////////////////////////////////
// This function finds coefficients of the 2nd order
// polynomial warping transformation based on the 
// supplied two sets of corresponding points.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			origImagePoints - a collection of the points
//				selected in the original image space
//			mappedImagePoints - a collection of the points
//				defining new positions (the mapped space)
//			thePolyMatrix - the transformation matrix
//				to be filled
//		
// OUTPUT:
//			true if success
//			false otherwise
//		
// REMARKS:
//		
bool DeterminePolynomialTransformParameters( Int_2D_Point_Vector & origImagePoints, Int_2D_Point_Vector & mappedImagePoints, T_2ndOrder_Polynomial_TransformEngine::Order2ndPolyMatrix & thePolyMatrix )
{
	// At first built the A matrix from the origImagePoints point set

	const int kNumOfPoints = origImagePoints.size();

	if( kNumOfPoints < 6 )
		return false;			// too few input points

	if( kNumOfPoints != mappedImagePoints.size() )
		return false;			// the two sets should contain the same number of points


	const int kACols = 6;				// this is constant
	const int kARows = kNumOfPoints;	// this is not

	TRealImage theAMatrix( kACols, kARows );		// anyway, the guts will be created by new (i.e. on the heap)
	// Init "A"
	register int i;

	for( i = 0; i < kARows; ++ i )
	{
		// Get the points
		double x = origImagePoints[ i ].fData[ 0 ];
		double y = origImagePoints[ i ].fData[ 1 ];
		
		// Set six values
		theAMatrix.SetPixel( 0, i, 1.0 );
		theAMatrix.SetPixel( 1, i, x );
		theAMatrix.SetPixel( 2, i, y );
		theAMatrix.SetPixel( 3, i, x * x );
		theAMatrix.SetPixel( 4, i, x * y );
		theAMatrix.SetPixel( 5, i, y * y );
	}

	// Create the two column matrices
	TRealImage theUMatrix( 1, kARows );
	TRealImage theVMatrix( 1, kARows );

	// Init them
	for( i = 0; i < kARows; ++ i )
	{
		// Get the points
		theUMatrix.SetPixel( 0, i, (double)mappedImagePoints[ i ].fData[ 0 ] );
		theVMatrix.SetPixel( 0, i, (double)mappedImagePoints[ i ].fData[ 1 ] );
	}


	// Find the pseudoinverse of the A matrix
	TRealImage * the_PseudoInv_A_Matrix = Orphan_PseudoInv_Matrix< double, double >( theAMatrix );

	if( the_PseudoInv_A_Matrix == 0 )
		return false;			// degenerated case of points, cannot invert the matrix

	TRealImage * a_coeffs = Orphan_Mult_Matrix< double, double >( * the_PseudoInv_A_Matrix, theUMatrix );
	REQUIRE( a_coeffs != 0 );
	REQUIRE( a_coeffs->GetCol() == 1 && a_coeffs->GetRow() == T_2ndOrder_Polynomial_TransformEngine::k2ndPolyMatrix_Cols );
	TRealImage * b_coeffs = Orphan_Mult_Matrix< double, double >( * the_PseudoInv_A_Matrix, theVMatrix );
	REQUIRE( b_coeffs != 0 );
	REQUIRE( b_coeffs->GetCol() == 1 && b_coeffs->GetRow() == T_2ndOrder_Polynomial_TransformEngine::k2ndPolyMatrix_Cols );

	// The inverse is not used anymore
	delete the_PseudoInv_A_Matrix;

#if _DEBUG
	Save_ASCII_Image( * a_coeffs, "a_coeffs.txt" ); 
	Save_ASCII_Image( * b_coeffs, "b_coeffs.txt" );
#endif _DEBUG


	// The last thing is to fill the poly matrix
	for( i = 0; i < T_2ndOrder_Polynomial_TransformEngine::k2ndPolyMatrix_Cols; ++ i )
	{
		thePolyMatrix[ 0 ][ i ] = a_coeffs->GetPixel( 0, i );
		thePolyMatrix[ 1 ][ i ] = b_coeffs->GetPixel( 0, i );
	}

	delete a_coeffs;
	delete b_coeffs;

	return true;			// operation ok
}

///////////////////////////////////////////////////////////////////////////////////////////////////////

