
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: Warping.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the warping classes.
//
// ==========================================================================



#include "Warping.h"
#include <math.h>
#include <float.h>			// for DBL_MAX



// ==========================================================================


// Use it as an output whenever it is not possible to determine the transformation
const Real_2D_Point TCoordTransformEngine::kImpossiblePoint( -1, -1 );

// ==========================================================================

TLinearTransformEngine::TLinearTransformEngine( void ) 
{
	//memset( fLinMatrix, 0, sizeof( fLinMatrix ) );		// set the ZERO matrix
	ClearTheLinMatrix();
}


Real_2D_Point TLinearTransformEngine::operator () ( const Real_2D_Point & in )
{
	Real_3D_Point homo_in( x(in), y(in), 1.0 );		// convert the input point into the homogenous coordinates
	Real_3D_Point homo_out;

	// Multiply
	homo_out.fData[ 0 ] = fLinMatrix[ 0 ][ 0 ] * homo_in.fData[ 0 ] + fLinMatrix[ 0 ][ 1 ] * homo_in.fData[ 1 ] + fLinMatrix[ 0 ][ 2 ] * homo_in.fData[ 2 ];
	homo_out.fData[ 1 ] = fLinMatrix[ 1 ][ 0 ] * homo_in.fData[ 0 ] + fLinMatrix[ 1 ][ 1 ] * homo_in.fData[ 1 ] + fLinMatrix[ 1 ][ 2 ] * homo_in.fData[ 2 ];
	homo_out.fData[ 2 ] = fLinMatrix[ 2 ][ 0 ] * homo_in.fData[ 0 ] + fLinMatrix[ 2 ][ 1 ] * homo_in.fData[ 1 ] + fLinMatrix[ 2 ][ 2 ] * homo_in.fData[ 2 ];

	REQUIRE( homo_out.fData[ 2 ] == 1.0 );
	return Real_2D_Point( homo_out.fData[ 0 ], homo_out.fData[ 1 ] );
}


// ==========================================================================

const Real_2D_Point TLinearTransformEngine::kPointAtInfinity( DBL_MAX, DBL_MAX );


TCoordRotationEngine::TCoordRotationEngine( double theta, const Real_2D_Point & rotationAxisPoint /*= Real_2D_Point( 0, 0 )*/ )
{
	double c = cos( theta );			// theta in radians
	double s = sin( theta );			// theta in radians

	//         row  col
	fLinMatrix[ 0 ][ 0 ] = c;
	fLinMatrix[ 0 ][ 1 ] = s;
	fLinMatrix[ 0 ][ 2 ] = - rotationAxisPoint.fData[ 0 ] * c - rotationAxisPoint.fData[ 1 ] * s + rotationAxisPoint.fData[ 0 ];
	
	fLinMatrix[ 1 ][ 0 ] = - s;
	fLinMatrix[ 1 ][ 1 ] = c;
	fLinMatrix[ 1 ][ 2 ] = rotationAxisPoint.fData[ 0 ] * s - rotationAxisPoint.fData[ 1 ] * c + rotationAxisPoint.fData[ 1 ];

	fLinMatrix[ 2 ][ 0 ] = 0.0;
	fLinMatrix[ 2 ][ 1 ] = 0.0;
	fLinMatrix[ 2 ][ 2 ] = 1.0;
}

// ==========================================================================

TCoordTranslationEngine::TCoordTranslationEngine( const Real_2D_Point & translationVector )
{

	//         row  col
	fLinMatrix[ 0 ][ 0 ] = 1.0;
	fLinMatrix[ 0 ][ 1 ] = 0.0;
	fLinMatrix[ 0 ][ 2 ] = - translationVector.fData[ 0 ];
	
	fLinMatrix[ 1 ][ 0 ] = 0.0;
	fLinMatrix[ 1 ][ 1 ] = 1.0;
	fLinMatrix[ 1 ][ 2 ] = - translationVector.fData[ 1 ];

	fLinMatrix[ 2 ][ 0 ] = 0.0;
	fLinMatrix[ 2 ][ 1 ] = 0.0;
	fLinMatrix[ 2 ][ 2 ] = 1.0;
}

// ==========================================================================

TCoordScalingEngine::TCoordScalingEngine( const Real_2D_Point & scalingVector )
{
	if( scalingVector.fData[ 0 ] != 0.0 && scalingVector.fData[ 1 ] != 0.0 )
	{
		//         row  col
		fLinMatrix[ 0 ][ 0 ] = 1.0 / scalingVector.fData[ 0 ];
		fLinMatrix[ 0 ][ 1 ] = 0.0;
		fLinMatrix[ 0 ][ 2 ] = 0.0;
		
		fLinMatrix[ 1 ][ 0 ] = 0.0;
		fLinMatrix[ 1 ][ 1 ] = 1.0 / scalingVector.fData[ 1 ];
		fLinMatrix[ 1 ][ 2 ] = 0.0;

		fLinMatrix[ 2 ][ 0 ] = 0.0;
		fLinMatrix[ 2 ][ 1 ] = 0.0;
		fLinMatrix[ 2 ][ 2 ] = 1.0 /*/ ( scalingVector.fData[ 0 ] * scalingVector.fData[ 1 ] )*/;
	}
}

// ==========================================================================

TCompound_LinearTransformEngine::TCompound_LinearTransformEngine( void )
{
	fLinMatrix[ 0 ][ 0 ] = 1.0;
	fLinMatrix[ 1 ][ 1 ] = 1.0;
	fLinMatrix[ 2 ][ 2 ] = 1.0;
}


///////////////////////////////////////////////////////////
// This function adds a new linear transformation to the
// one common linear transformation.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			t - ref to the new lin transformation
//		
// OUTPUT:
//			none
//		
// REMARKS:
//		
//		
void TCompound_LinearTransformEngine::AddNewTransformation( TLinearTransformEngine & t )
{
	int i, j, k;

	int a_rows = kLinMatrix_Rows;
	int a_cols = kLinMatrix_Cols;
	int b_cols = kLinMatrix_Cols;

	LinMatrix a;
	memcpy( a, fLinMatrix, sizeof( fLinMatrix ) );		// a will store the "old" matrix

	LinMatrix & b = t.GetTheLinMatrix();


	// A general matrix multiplication procedure
	for( i = 0; i < a_rows; ++ i )
	{
		for( j = 0; j < b_cols; ++ j )
		{
			double theSum = 0.0; 

			for( k = 0; k < a_cols; ++ k )
				theSum += a[ i ][ k ] * b[ k ][ j ];

			fLinMatrix[ i ][ j ] = theSum;
		}
	}

}



// ==========================================================================



Real_2D_Point T_2ndOrder_Polynomial_TransformEngine::operator () ( const Real_2D_Point & in )
{
	double theInPoint[ k2ndPolyMatrix_Cols ];

	theInPoint[ 0 ] = 1.0;
	theInPoint[ 1 ] = in.fData[ 0 ];
	theInPoint[ 2 ] = in.fData[ 1 ];
	theInPoint[ 3 ] = in.fData[ 0 ] * in.fData[ 0 ];
	theInPoint[ 4 ] = in.fData[ 0 ] * in.fData[ 1 ];
	theInPoint[ 5 ] = in.fData[ 1 ] * in.fData[ 1 ];

	double theOutPoint[ 2 ];
	for( int i = 0; i < k2ndPolyMatrix_Rows; ++ i )
	{
		double theSum = 0.0;
		for( int j = 0; j < k2ndPolyMatrix_Cols; ++ j )
			theSum += theInPoint[ j ] * f2ndPolyMatrix[ i ][ j ];

		theOutPoint[ i ] = theSum;
	}

	return Real_2D_Point( theOutPoint[ 0 ], theOutPoint[ 1 ] );
}

// ==========================================================================


Real_2D_Point TGenericTransformEngine::operator () ( const Real_2D_Point & in )
{
	int c = in.fData[ 0 ];
	int r = in.fData[ 1 ];
	if( c < 0.0 || c > fLookUpTable.GetCol() || r < 0.0 || r > fLookUpTable.GetRow() )
		return kImpossiblePoint;
	else
		return fLookUpTable.GetPixel( c, r );
}

// ==========================================================================











