
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: LogPolarSpace.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Image warping classes.
//
// ==========================================================================


#ifndef BoxySoft_Warping_h
#define BoxySoft_Warping_h


#include "HIL_ImageFor.h"
#include "HIL_AutoPtrDefs.h"
#include "HIL_MultiPixelFor.h"
#include <math.h>

using namespace PHIL;

// ===========================================================================================


///////////////////////////////////////////////////////////
// This class implements the hierarchy of the coordinate
// transformation classes.
///////////////////////////////////////////////////////////
class TCoordTransformEngine
{
	private:

		// Class inherent variables

	protected:

	public:

		static const Real_2D_Point kImpossiblePoint;	// use it as an output whenever it is not 
														// possible to determine the transformation

	public:

		// ===================================================
		TCoordTransformEngine( void ) {}
			
		virtual ~TCoordTransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function converts a given point based on some
		// external parameters (set in derived classes)
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - the input point
		//		
		// OUTPUT:
		//			the output (converted) point
		//		
		// REMARKS:
		//		
		//		
		virtual Real_2D_Point operator () ( const Real_2D_Point & in ) = 0;


};


// ===========================================================================================



class TNonLinearTransformEngine : public TCoordTransformEngine
{
	private:

		// Class inherent variables

	public:

		// ===================================================
		TNonLinearTransformEngine( void ) {}
		virtual ~TNonLinearTransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function converts a given point based on the
		// linear transformation described by the 3x3 matrix
		// fLinMatrix.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - the input point
		//		
		// OUTPUT:
		//			the output (converted) point
		//		
		// REMARKS:
		//		
		//		
		virtual Real_2D_Point operator () ( const Real_2D_Point & in ) = 0;


};


//
// The second-order (and higher) address mappings are performed
// to compensate for spatial distortion caused by a physical
// imagining system.
//
class T_2ndOrder_Polynomial_TransformEngine : public TNonLinearTransformEngine
{
	private:

		// Class inherent variables


	public:

		enum { k2ndPolyMatrix_Cols = 6, k2ndPolyMatrix_Rows = 2 };


		typedef double Order2ndPolyMatrix[ k2ndPolyMatrix_Rows ][ k2ndPolyMatrix_Cols ];

	protected:

		// Here we have two rows of six coefficients, 
		// [ a0 a1 a2 a3 a4 a5 ] and
		// [ b0 b1 b2 b3 b4 b5 ],	
		// each corresponding to the following vector:
		// [ 1 x y x2 xy y2 ]
		Order2ndPolyMatrix	f2ndPolyMatrix;

	public:

		// ===================================================
	public:

		Order2ndPolyMatrix & Get2ndPolyMatrix( void ) { return f2ndPolyMatrix; }

		void Set2ndPolyMatrix( Order2ndPolyMatrix & matrix ) { memcpy( f2ndPolyMatrix, matrix, sizeof( f2ndPolyMatrix ) ); }


	public:

		// ===================================================
		T_2ndOrder_Polynomial_TransformEngine( void ) { f2ndPolyMatrix, 0, sizeof( f2ndPolyMatrix ); }
		T_2ndOrder_Polynomial_TransformEngine( Order2ndPolyMatrix & theMatrix ) { Set2ndPolyMatrix( theMatrix ); }
		T_2ndOrder_Polynomial_TransformEngine( T_2ndOrder_Polynomial_TransformEngine & ref ) { Set2ndPolyMatrix( ref.Get2ndPolyMatrix() ); }
		
		T_2ndOrder_Polynomial_TransformEngine & operator = ( T_2ndOrder_Polynomial_TransformEngine & ref ) 
		{ 
			if( this != & ref )
				Set2ndPolyMatrix( ref.Get2ndPolyMatrix() );
			return * this;
		}
		
		virtual ~T_2ndOrder_Polynomial_TransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function converts a given point based on the
		// linear transformation described by the 3x3 matrix
		// fLinMatrix.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - the input point
		//		
		// OUTPUT:
		//			the output (converted) point
		//		
		// REMARKS:
		//		
		//		
		virtual Real_2D_Point operator () ( const Real_2D_Point & in );


};


// ===========================================================================================


// This class originates a new hierarchy for the generic
// transformations of coordinates.
class TGenericTransformEngine : public TCoordTransformEngine
{
	private:

		// Class inherent variables


	public:

		typedef TImageFor< Real_2D_Point > ImageOf_Real_2D_Point;


	protected:

		// This table contains pre-defined new coordinates
		ImageOf_Real_2D_Point fLookUpTable;

	public:

		ImageOf_Real_2D_Point & Get_ImageOf_Real_2D_Point( void ) { return fLookUpTable; }

	public:

		// ===================================================
		TGenericTransformEngine( int lookUpCols, int lookUpRows ) : fLookUpTable( lookUpCols, lookUpRows ) {}
		virtual ~TGenericTransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function converts a given point based on the
		// linear transformation described by the 3x3 matrix
		// fLinMatrix.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - the input point
		//		
		// OUTPUT:
		//			the output (converted) point
		//		
		// REMARKS:
		//		
		//		
		virtual Real_2D_Point operator () ( const Real_2D_Point & in );


};


// ===========================================================================================

class TLinearTransformEngine : public TCoordTransformEngine
{
	private:

		// Class inherent variables

	public:

		enum { kLinMatrix_Cols = 3, kLinMatrix_Rows = 3 };

		typedef double LinMatrix[ kLinMatrix_Rows ][ kLinMatrix_Cols ];

		static const Real_2D_Point kPointAtInfinity;
	
	protected:

		LinMatrix	fLinMatrix;


	public:

		LinMatrix & GetTheLinMatrix( void ) { return fLinMatrix; }

		void SetTheLinMatrix( LinMatrix & matrix ) { memcpy( fLinMatrix, matrix, sizeof( fLinMatrix ) ); }

		void ClearTheLinMatrix( void ) { memset( fLinMatrix, 0, sizeof( fLinMatrix ) ); }		// set the ZERO matrix

	public:

		// ===================================================
		TLinearTransformEngine( void );
		TLinearTransformEngine( LinMatrix & theMatrix ) { SetTheLinMatrix( theMatrix ); }
		TLinearTransformEngine( TLinearTransformEngine & ref ) { SetTheLinMatrix( ref.GetTheLinMatrix() ); }
		
		TLinearTransformEngine & operator = ( TLinearTransformEngine & ref ) 
		{ 
			if( this != & ref )
				SetTheLinMatrix( ref.GetTheLinMatrix() );
			return * this;
		}
			
		virtual ~TLinearTransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function converts a given point based on the
		// linear transformation described by the 3x3 matrix
		// fLinMatrix.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			in - the input point
		//		
		// OUTPUT:
		//			the output (converted) point
		//		
		// REMARKS:
		//		
		//		
		virtual Real_2D_Point operator () ( const Real_2D_Point & in );


};


class TCoordRotationEngine : public TLinearTransformEngine
{
	private:

		// Class inherent variables

	protected:


	public:

		// ===================================================
		TCoordRotationEngine( double theta, const Real_2D_Point & rotationAxisPoint = Real_2D_Point( 0, 0 ) );
			
		virtual ~TCoordRotationEngine() {}				// class virtual destructor
		// ===================================================


};


class TCoordTranslationEngine : public TLinearTransformEngine
{
	private:

		// Class inherent variables

	public:

		// ===================================================
		TCoordTranslationEngine( const Real_2D_Point & translationVector );
			
		virtual ~TCoordTranslationEngine() {}				// class virtual destructor
		// ===================================================

};


class TCoordScalingEngine : public TLinearTransformEngine
{
	private:

		// Class inherent variables

	protected:
		

	public:

		// ===================================================
		TCoordScalingEngine( const Real_2D_Point & scalingVector );
			
		virtual ~TCoordScalingEngine() {}				// class virtual destructor
		// ===================================================

};


// A composite of linear transformations
class TCompound_LinearTransformEngine : public TLinearTransformEngine
{
	private:

		// Class inherent variables

	protected:
		

	public:

		// ===================================================
		TCompound_LinearTransformEngine( void );
		virtual ~TCompound_LinearTransformEngine() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function adds a new linear transformation to the
		// one common linear transformation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			t - ref to the new lin transformation
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		void AddNewTransformation( TLinearTransformEngine & t );


};

//////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This class is a base for pixel interpolation
///////////////////////////////////////////////////////////
template < class T = double >
class TPixelInterpolation
{
	private:

		// Class inherent variables

	public:
		
		TPixelInterpolation( void ) {}

		virtual ~TPixelInterpolation() {}

	public:

		///////////////////////////////////////////////////////////
		// This function interpolates a pixel value
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inImage - the image that serves pixel values
		//				for interpolation
		//			pointCoords - a point to be interpolated (it
		//				can be further processed in derived classes
		//				e.g. for the inverse warping)
		//			theInterpolatedValue - the output interpolated
		//				value of a pixel
		//		
		// OUTPUT:
		//			true if operation successful,
		//			false otherwise
		//		
		// REMARKS:
		//		
		//		
		virtual bool operator()( const TImageFor< T > & inImage, const Real_2D_Point & inCoord, T & theInterpolatedVal ) = 0;

};



template < class T = double >
class TBilinearInterpolation : public TPixelInterpolation< T >
{
	private:

		// Class inherent variables

	public:

		typedef vector< Real_2D_Point > Real_2D_Point_Vec;

	//protected:
	public:

		///////////////////////////////////////////////////////////
		// This function finds the four nearest neighbour points
		// that lie on the integer grid in the input image 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inCoord - coordinates of the input point
		//				for which the nearest points on integer
		//				grid (its neighbours) will be searched for
		//			neighbourIntPoints - the vector of four output 
		//				points; upon successfull return contains 
		//				the nearest points on the regular grid
		//				It is cleared upon each call.
		//		
		// OUTPUT:
		//			true - if operation successful
		//			false - otherwise
		//
		// REMARKS:
		//		
		//		
		// Order of the points in the neighbourIntPoints is as follows:
		//
		// [3] ------- [2]
		//  |           |
		//  |           |
		//  |      x    |
		//  |           |
		// [0] ------- [1]
		//
		//	x - denotes a position of the input point given by the "inCoord"
		//  parameter
		//
		bool FindTheNearestIntegerCoords( const Real_2D_Point & inCoord, Real_2D_Point_Vec & neighbourIntPoints );


		///////////////////////////////////////////////////////////
		// This function find the horizontal and vertical distances
		// of a point to be interpolated (any position) to its 
		// nearest neighbours that are on a regular grid.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inCoord - a point to be interpolated (it
		//				can be further processed in derived classes
		//				e.g. for the inverse warping)
		//			neighbourIntPoints - a vector that upon successfull
		//				return contains the four neighbours of
		//				a point of interpolation (see the scheme below)
		//				It is cleared upon each call.
		//			horz_disp - upons success it contains the horizontal
		//				displacement withing the grid of four 
		//				neighbours (a value in the range of 0 to 1)
		//			vert_disp - upons success it contains the vertical
		//				displacement withing the grid of four 
		//				neighbours (a value in the range of 0 to 1)
		//			dx - a horizontal span between adjacent points (usually
		//				should be 1)
		//			dy - a vertical span between adjacent points (usually
		//				should be 1)
		//				
		//		
		// OUTPUT:
		//		
		//			true if operation successful,
		//			false otherwise
		//		
		// REMARKS:
		//		
		// Order of the points in the neighbourIntPoints is as follows:
		//
		//  |<--  dx -->|
		// [3] ------- [2]--
		//  |   |       |  |
		//  |   |       | dy
		//  |---*       |
		//  |           |  |
		// [0] ------- [1]--
		//
		//		
		bool FindLocalDisplacements(	const Real_2D_Point & inCoord,
										Real_2D_Point_Vec & neighbourIntPoints, double & horz_disp, double & vert_disp,
										double & dx, double & dy );
			
	
	public:
		
		TBilinearInterpolation( void ) {}

		virtual ~TBilinearInterpolation() {}

	public:

		///////////////////////////////////////////////////////////
		// This function interpolates a pixel value
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inImage - the image that serves pixel values
		//				for interpolation
		//			inCoord - a point to be interpolated (it
		//				can be further processed in derived classes
		//				e.g. for the inverse warping)
		//			theInterpolatedValue - the output interpolated
		//				value of a pixel
		//		
		// OUTPUT:
		//			true if operation successful,
		//			false otherwise
		//		
		// REMARKS:
		//		
		//		
		virtual bool operator()( const TImageFor< T > & inImage, const Real_2D_Point & inCoord, T & theInterpolatedVal );

	public:

		static TBilinearInterpolation< T > gkBilinearInterpolation;
};


//////////////////////////////////////////////////////////////////////////////////////////////////////


template < class T >
TBilinearInterpolation< T > TBilinearInterpolation< T >::gkBilinearInterpolation;


//////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This function finds the four nearest neighbour points
// that lie on the integer grid in the input image 
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inCoord - coordinates of the input point
//				for which the nearest points on integer
//				grid (its neighbours) will be searched for
//			neighbourIntPoints - the vector of four output 
//				points; upon successfull return contains 
//				the nearest points on the regular grid;
//				It is cleared upon each call.
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise
//
// REMARKS:
//		
//		
// Order of the points in the neighbourIntPoints is as follows:
//
// [3] ------- [2]
//  |           |
//  |           |
//  |      x    |
//  |           |
// [0] ------- [1]
//
//	x - denotes a position of the input point given by the "inCoord"
//  parameter
//
template < class T >
bool TBilinearInterpolation< T >::FindTheNearestIntegerCoords( const Real_2D_Point & inCoord, Real_2D_Point_Vec & neighbourIntPoints )
{
	// What to do with a negative coordinates? 
	double xx = x( inCoord );
	double yy = y( inCoord );

	if( xx < 0.0 || yy < 0.0 )
		return false;				// coordinates out of range

	double x_f = floor( xx );
	double x_c = ceil( xx );
	if( x_f == x_c )
		x_c ++;
	double y_f = floor( yy );
	double y_c = ceil( yy );
	if( y_f == y_c )
		y_c ++;

	neighbourIntPoints.clear();

	neighbourIntPoints.push_back( Real_2D_Point( x_f, y_c ) );			// bottom left
	neighbourIntPoints.push_back( Real_2D_Point( x_c, y_c ) );			// bottom right
	neighbourIntPoints.push_back( Real_2D_Point( x_c, y_f ) );			// top right
	neighbourIntPoints.push_back( Real_2D_Point( x_f, y_f ) );			// top left

	return true;
}

///////////////////////////////////////////////////////////
// This function find the horizontal and vertical distances
// of a point to be interpolated (any position) to its 
// nearest neighbours that are on a regular grid.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inCoord - a point to be interpolated (it
//				can be further processed in derived classes
//				e.g. for the inverse warping)
//			neighbourIntPoints - a vector that upon successfull
//				return contains the four neighbours of
//				a point of interpolation (see the scheme below)
//				It is cleared upon each call.
//			horz_disp - upons success it contains the horizontal
//				displacement withing the grid of four 
//				neighbours (a value in the range of 0 to 1)
//			vert_disp - upons success it contains the vertical
//				displacement withing the grid of four 
//				neighbours (a value in the range of 0 to 1)
//			dx - a horizontal span between adjacent points (usually
//				should be 1)
//			dy - a vertical span between adjacent points (usually
//				should be 1)
//				
//		
// OUTPUT:
//		
//			true if operation successful,
//			false otherwise
//		
// REMARKS:
//		
// Order of the points in the neighbourIntPoints is as follows:
//
//  |<--  dx -->|
// [3] ------- [2]--
//  |   |       |  |
//  |   |       | dy
//  |---*       |
//  |           |  |
// [0] ------- [1]--
//
//		
template < class T >
bool TBilinearInterpolation< T >::FindLocalDisplacements(	const Real_2D_Point & inCoord,
															Real_2D_Point_Vec & neighbourIntPoints, double & horz_disp, double & vert_disp,
															double & dx, double & dy )
{
	// For the given point in the input-image-space find the neaighbouring pixels 
	// that can be used to interpolate the value
	if( FindTheNearestIntegerCoords( inCoord, neighbourIntPoints ) == false )
		return false;
	
	dx = x( neighbourIntPoints[ 1 ] ) - x( neighbourIntPoints[ 0 ] );
	dy = y( neighbourIntPoints[ 1 ] ) - y( neighbourIntPoints[ 2 ] );

	if( dx <= 0.0 || dy <= 0.0 )
		return false;				// degenerated case

	horz_disp = ( x( inCoord ) - x( neighbourIntPoints[ 0 ] ) ) / dx;
	vert_disp = ( y( inCoord ) - y( neighbourIntPoints[ 2 ] ) ) / dy; 

	return true;
}

///////////////////////////////////////////////////////////
// This function interpolates a pixel value
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inImage - the image that serves pixel values
//				for interpolation
//			inCoord - a point to be interpolated (it
//				can be further processed in derived classes
//				e.g. for the inverse warping)
//			theInterpolatedValue - the output interpolated
//				value of a pixel
//		
// OUTPUT:
//			true if operation successful,
//			false otherwise
//		
// REMARKS:
//		
//		
template < class T >
bool TBilinearInterpolation< T >::operator ()( const TImageFor< T > & inImage, const Real_2D_Point & inCoord, T & theInterpolatedVal )
{
	Real_2D_Point_Vec	neighbourIntPoints;
	double				horz_disp, vert_disp, dx, dy;

	if( FindLocalDisplacements( inCoord, neighbourIntPoints, horz_disp, vert_disp, dx, dy ) == false )
		return false;


	const int kCols = inImage.GetCol();
	const int kRows = inImage.GetRow();

	const int kNeighbours = 4;
	T pix[ kNeighbours ];				// in this buffer we store all 4 neighbours of a given point
											// These neighbours lie on the integer grid

	REQUIRE( neighbourIntPoints.size() == kNeighbours );

	for( int i = 0; i < kNeighbours; ++ i )
	{
		int col = (int)x( neighbourIntPoints[ i ] );
		int row = (int)y( neighbourIntPoints[ i ] );

		if( ! ( col < kCols && row < kRows && col >= 0 && row >= 0 ) )
			return false;				// out of range

		pix[ i ] = inImage.GetPixel( col, row );
	}


	pix[ 3 ] *= ( 1.0 - horz_disp ) * ( 1.0 - vert_disp );
	theInterpolatedVal = pix[ 3 ]; 

	pix[ 2 ] *= horz_disp * ( 1 - vert_disp );
	theInterpolatedVal += pix[ 2 ]; 

	pix[ 1 ] *= horz_disp * vert_disp;
	theInterpolatedVal += pix[ 1 ]; 

	pix[ 0 ] *= ( 1.0 - horz_disp ) * vert_disp;
	theInterpolatedVal += pix[ 0 ]; 

	return true;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////


// This class implements the bicubic interpolation scheme
// which for some cases gives smoother interpolation results.
template < class T = double >
class TBicubicPolyInterpolation : public TBilinearInterpolation< T >
{
	protected:

		TImageFor< T > fTmpImage;				// used in computations of the derivatives

		// Simoncelli horizontal (vertical) gradient is obtained in a twofold mode:
		// - first we do vertical (horizontal) smoothing with the prefilter
		// - then we apply the horizontal (vertical) derivative filter
		vector< T > fSimoncelliPrefilter;
		vector< T > fSimoncelliDerivative;

	protected:

		enum { kCoeffsCols = 4, kCoeffsRows = 4 };

		double fBicubicCoeffs[ kCoeffsRows ][ kCoeffsCols ];

		enum { kWeightsCols = 16, kWeightsRows = 16 };

		TImageFor< T > fWeightCoeffs;

		TImageFor< T > fHorzDerivImage;
		TImageFor< T > fVertDerivImage;
		TImageFor< T > fCrossDerivImage;

	protected:

		virtual void Compute_Horz_Derivative	( const TImageFor< T > & inImage );
		virtual void Compute_Vert_Derivative	( const TImageFor< T > & inImage );
		virtual void Compute_Cross_Derivative	( const TImageFor< T > & inImage );

		bool FillBicubicCoeffs( const TImageFor< T > & inImage );
			
		void FillWeightMatrix( void );
	
		bool ComputeBicubicCoeffs(	const TImageFor< T > & inImage, const Real_2D_Point & inCoord, double & horz_disp, double & vert_disp );

	
	public:
		
		TBicubicPolyInterpolation( const TImageFor< T > & inImage ) : fWeightCoeffs( kWeightsCols, kWeightsRows, 0.0 ),
			fHorzDerivImage( inImage.GetCol(), inImage.GetRow() ), fVertDerivImage( inImage.GetCol(), inImage.GetRow() ), fCrossDerivImage( inImage.GetCol(), inImage.GetRow() ),
			fTmpImage( inImage.GetCol(), inImage.GetRow() )
		{ 
			// Prepare the derivatives:
			// The three derivatives: df/dx, df/dy, and d2f/dxdy

			// Fill the Simoncelli filters
			fSimoncelliPrefilter.push_back( 0.22420981526374817 );
			fSimoncelliPrefilter.push_back( 0.5515803694725037 );
			fSimoncelliPrefilter.push_back( 0.22420981526374817 );

			fSimoncelliDerivative.push_back( -0.45527133345603943 );
			fSimoncelliDerivative.push_back( 0.0 );
			fSimoncelliDerivative.push_back( 0.45527133345603943 );
			
			// Compute the derivatives
			Compute_Horz_Derivative( inImage );
			Compute_Vert_Derivative( inImage );
			Compute_Cross_Derivative( inImage );

			// Prepare the weights:
			FillWeightMatrix(); 
		}

		virtual ~TBicubicPolyInterpolation() {}

	public:

		///////////////////////////////////////////////////////////
		// This function interpolates a pixel value
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			inImage - the image that serves pixel values
		//				for interpolation
		//			inCoord - a point to be interpolated (it
		//				can be further processed in derived classes
		//				e.g. for the inverse warping)
		//			theInterpolatedValue - the output interpolated
		//				value of a pixel
		//		
		// OUTPUT:
		//			true if operation successful,
		//			false otherwise
		//		
		// REMARKS:
		//		
		//		
		virtual bool operator()( const TImageFor< T > & inImage, const Real_2D_Point & inCoord, T & theInterpolatedVal );

};



//////////////////////////////////////////////////////////////////////////////////////////////////////


template < class T >
void TBicubicPolyInterpolation< T >::Compute_Horz_Derivative( const TImageFor< T > & inImage )
{
	Vert1DConvolve( inImage, fSimoncelliPrefilter, fTmpImage ); 
	Horz1DConvolve( fTmpImage, fSimoncelliDerivative, fHorzDerivImage ); 
}

template < class T >
void TBicubicPolyInterpolation< T >::Compute_Vert_Derivative( const TImageFor< T > & inImage )
{
	Horz1DConvolve( inImage, fSimoncelliPrefilter, fTmpImage ); 
	Vert1DConvolve( fTmpImage, fSimoncelliDerivative, fVertDerivImage );
}

template < class T >
void TBicubicPolyInterpolation< T >::Compute_Cross_Derivative( const TImageFor< T > & inImage )
{
	// It assumes that the fHorzDerivImage member has been already computed !!!
	Horz1DConvolve( fHorzDerivImage, fSimoncelliPrefilter, fTmpImage ); 
	Vert1DConvolve( fTmpImage, fSimoncelliDerivative, fCrossDerivImage ); 
}



template < class T >
void TBicubicPolyInterpolation< T >::FillWeightMatrix( void )
{

	// Set the specific values other than 0:
	fWeightCoeffs.SetPixel( 0, 0, 1.0 );
	fWeightCoeffs.SetPixel( 8, 1, 1.0 );

	fWeightCoeffs.SetPixel( 0, 2, -3.0 );
	fWeightCoeffs.SetPixel( 3, 2, 3.0 );
	fWeightCoeffs.SetPixel( 8, 2, -2.0 );
	fWeightCoeffs.SetPixel( 11, 2, -1.0 );

	fWeightCoeffs.SetPixel( 0, 3, 2.0 );
	fWeightCoeffs.SetPixel( 3, 3, -2.0 );
	fWeightCoeffs.SetPixel( 8, 3, 1.0 );
	fWeightCoeffs.SetPixel( 11, 3, 1.0 );

	fWeightCoeffs.SetPixel( 4, 4, 1.0 );

	fWeightCoeffs.SetPixel( 13, 5, 1.0 );

	fWeightCoeffs.SetPixel( 4, 6, -3.0 );
	fWeightCoeffs.SetPixel( 7, 6, 3.0 );
	fWeightCoeffs.SetPixel( 12, 6, -2.0 );
	fWeightCoeffs.SetPixel( 15, 6, -1.0 );

	fWeightCoeffs.SetPixel( 4, 7, 2.0 );
	fWeightCoeffs.SetPixel( 7, 7, -2.0 );
	fWeightCoeffs.SetPixel( 12, 7, 1.0 );
	fWeightCoeffs.SetPixel( 15, 7, 1.0 );

	fWeightCoeffs.SetPixel( 0, 8, -3.0 );
	fWeightCoeffs.SetPixel( 1, 8, 3.0 );
	fWeightCoeffs.SetPixel( 4, 8, -2.0 );
	fWeightCoeffs.SetPixel( 5, 8, -1.0 );

	fWeightCoeffs.SetPixel( 8, 9, -3.0 );
	fWeightCoeffs.SetPixel( 9, 9, 3.0 );
	fWeightCoeffs.SetPixel( 12, 9, -2.0 );
	fWeightCoeffs.SetPixel( 13, 9, -1.0 );

	fWeightCoeffs.SetPixel( 0, 10, 9.0 );
	fWeightCoeffs.SetPixel( 1, 10, -9.0 );
	fWeightCoeffs.SetPixel( 2, 10, 9.0 );
	fWeightCoeffs.SetPixel( 3, 10, -9.0 );
	fWeightCoeffs.SetPixel( 4, 10, 6.0 );
	fWeightCoeffs.SetPixel( 5, 10, 3.0 );
	fWeightCoeffs.SetPixel( 6, 10, -3.0 );
	fWeightCoeffs.SetPixel( 7, 10, -6.0 );
	fWeightCoeffs.SetPixel( 8, 10, 6.0 );
	fWeightCoeffs.SetPixel( 9, 10, -6.0 );
	fWeightCoeffs.SetPixel( 10, 10, -3.0 );
	fWeightCoeffs.SetPixel( 11, 10, 3.0 );
	fWeightCoeffs.SetPixel( 12, 10, 4.0 );
	fWeightCoeffs.SetPixel( 13, 10, 2.0 );
	fWeightCoeffs.SetPixel( 14, 10, 1.0 );
	fWeightCoeffs.SetPixel( 15, 10, 2.0 );

	fWeightCoeffs.SetPixel( 0,  11, -6.0 );
	fWeightCoeffs.SetPixel( 1,  11, 6.0 );
	fWeightCoeffs.SetPixel( 2,  11, -6.0 );
	fWeightCoeffs.SetPixel( 3,  11, 6.0 );
	fWeightCoeffs.SetPixel( 4,  11, -4.0 );
	fWeightCoeffs.SetPixel( 5,  11, -2.0 );
	fWeightCoeffs.SetPixel( 6,  11, 2.0 );
	fWeightCoeffs.SetPixel( 7,  11, 4.0 );
	fWeightCoeffs.SetPixel( 8,  11, -3.0 );
	fWeightCoeffs.SetPixel( 9,  11, 3.0 );
	fWeightCoeffs.SetPixel( 10, 11, 3.0 );
	fWeightCoeffs.SetPixel( 11, 11, -3.0 );
	fWeightCoeffs.SetPixel( 12, 11, -2.0 );
	fWeightCoeffs.SetPixel( 13, 11, -1.0 );
	fWeightCoeffs.SetPixel( 14, 11, -1.0 );
	fWeightCoeffs.SetPixel( 15, 11, -2.0 );

	fWeightCoeffs.SetPixel( 0,  12, 2.0 );
	fWeightCoeffs.SetPixel( 1,  12, -2.0 );
	fWeightCoeffs.SetPixel( 4,  12, 1.0 );
	fWeightCoeffs.SetPixel( 5,  12, 1.0 );

	fWeightCoeffs.SetPixel( 8,  13, 2.0 );
	fWeightCoeffs.SetPixel( 9,  13, -2.0 );
	fWeightCoeffs.SetPixel( 12,  13, 1.0 );
	fWeightCoeffs.SetPixel( 13,  13, 1.0 );

	fWeightCoeffs.SetPixel( 0,  14, -6.0 );
	fWeightCoeffs.SetPixel( 1,  14, 6.0 );
	fWeightCoeffs.SetPixel( 2,  14, -6.0 );
	fWeightCoeffs.SetPixel( 3,  14, 6.0 );
	fWeightCoeffs.SetPixel( 4,  14, -3.0 );
	fWeightCoeffs.SetPixel( 5,  14, -3.0 );
	fWeightCoeffs.SetPixel( 6,  14, 3.0 );
	fWeightCoeffs.SetPixel( 7,  14, 3.0 );
	fWeightCoeffs.SetPixel( 8,  14, -4.0 );
	fWeightCoeffs.SetPixel( 9,  14, 4.0 );
	fWeightCoeffs.SetPixel( 10, 14, 2.0 );
	fWeightCoeffs.SetPixel( 11, 14, -2.0 );
	fWeightCoeffs.SetPixel( 12, 14, -2.0 );
	fWeightCoeffs.SetPixel( 13, 14, -2.0 );
	fWeightCoeffs.SetPixel( 14, 14, -1.0 );
	fWeightCoeffs.SetPixel( 15, 14, -1.0 );

	fWeightCoeffs.SetPixel( 0,  15, 4.0 );
	fWeightCoeffs.SetPixel( 1,  15, -4.0 );
	fWeightCoeffs.SetPixel( 2,  15, 4.0 );
	fWeightCoeffs.SetPixel( 3,  15, -4.0 );
	fWeightCoeffs.SetPixel( 4,  15, 2.0 );
	fWeightCoeffs.SetPixel( 5,  15, 2.0 );
	fWeightCoeffs.SetPixel( 6,  15, -2.0 );
	fWeightCoeffs.SetPixel( 7,  15, -2.0 );
	fWeightCoeffs.SetPixel( 8,  15, 2.0 );
	fWeightCoeffs.SetPixel( 9,  15, -2.0 );
	fWeightCoeffs.SetPixel( 10, 15, -2.0 );
	fWeightCoeffs.SetPixel( 11, 15, 2.0 );
	fWeightCoeffs.SetPixel( 12, 15, 1.0 );
	fWeightCoeffs.SetPixel( 13, 15, 1.0 );
	fWeightCoeffs.SetPixel( 14, 15, 1.0 );
	fWeightCoeffs.SetPixel( 15, 15, 1.0 );
}



template < class T >
bool TBicubicPolyInterpolation< T >::ComputeBicubicCoeffs(	const TImageFor< T > & inImage, const Real_2D_Point & inCoord,
															double & horz_disp, double & vert_disp )
{
	// Fill values for the nearest four neighbours
	Real_2D_Point_Vec	neighbourIntPoints;

	double dx, dy;

	if( FindLocalDisplacements( inCoord, neighbourIntPoints, horz_disp, vert_disp, dx, dy ) == false )
		return false;

	REQUIRE( dx == 1.0 );
	REQUIRE( dy == 1.0 );

	REQUIRE( horz_disp >= 0.0 && horz_disp <= 1.0 );
	REQUIRE( vert_disp >= 0.0 && vert_disp <= 1.0 );

	const int kCols = inImage.GetCol();
	const int kRows = inImage.GetRow();

	const int kNeighbours = 4;
	T pix[ kNeighbours ];		// in this buffer we store all 4 neighbours of a given point - they lie on the integer grid
	T horz_deriv[ kNeighbours ];	
	T vert_deriv[ kNeighbours ];	
	T cross_deriv[ kNeighbours ];	

	REQUIRE( neighbourIntPoints.size() == kNeighbours );

	register int i, j;

	for( i = 0; i < kNeighbours; ++ i )
	{
		int col = (int)x( neighbourIntPoints[ i ] );
		int row = (int)y( neighbourIntPoints[ i ] );

		if( ! ( col < kCols && row < kRows && col >= 0 && row >= 0 ) )
			return false;				// out of range

		pix[ i ]			= inImage.GetPixel( col, row );
		horz_deriv[ i ]		= fHorzDerivImage.GetPixel( col, row );
		vert_deriv[ i ]		= fVertDerivImage.GetPixel( col, row );
		cross_deriv[ i ]	= fCrossDerivImage.GetPixel( col, row );
	}

	////\\////\\////\\////\\///\\////\\///\\////\\///\\////\\///\\////\\///

	double tmp, mult_dx_dy;
	double multi[ kWeightsCols ], mixed_package[ kWeightsCols ];

	mult_dx_dy = dx * dy;

	// Precompute the values
	for( i = 0; i < kNeighbours; ++ i )
	{
		mixed_package[ i ]		= pix[ i ];
		mixed_package[ i + 4 ]	= horz_deriv[ i ] * dx;
		mixed_package[ i + 8 ]	= vert_deriv[ i ] * dy;
		mixed_package[ i + 12 ] = cross_deriv[ i ] * mult_dx_dy;
	}

	// Multiply the matrices
	for( i = 0; i < kWeightsRows; ++ i )
	{
		tmp = 0.0;
		for( j = 0; j < kWeightsCols; ++ j )
		{
			tmp += fWeightCoeffs.GetPixel( j, i ) * mixed_package[ j ];
		}

		multi[ i ] = tmp;		// copy the result
	}

	int counter = 0;

	// Fill out the output matrix of coefficients
	for( i = 0; i < kCoeffsRows; ++ i )
		for( j = 0; j < kCoeffsCols; ++ j )
			fBicubicCoeffs[ i ][ j ] = multi[ counter ++ ];

	return true;
}



///////////////////////////////////////////////////////////
// This function interpolates a pixel value
///////////////////////////////////////////////////////////
//		
// INPUT:
//			inImage - the image that serves pixel values
//				for interpolation
//			inCoord - a point to be interpolated (it
//				can be further processed in derived classes
//				e.g. for the inverse warping)
//			theInterpolatedValue - the output interpolated
//				value of a pixel
//		
// OUTPUT:
//			true if operation successful,
//			false otherwise
//		
// REMARKS:
//		
//	
template < class T >
bool TBicubicPolyInterpolation< T >::operator ()( const TImageFor< T > & inImage, const Real_2D_Point & inCoord, T & theInterpolatedVal )
{
	const int kCols = inImage.GetCol();
	const int kRows = inImage.GetRow();

	double horz_disp, vert_disp;
	if( ComputeBicubicCoeffs( inImage, inCoord, horz_disp, vert_disp ) == false )
		return false;

	// After successful return of the ComputeBicubicCoeffs() we have filled bicubic coeffs 
	// in the fBicubicCoeffs member

	// Compute the interpolated value
	register int i, j;
	double tmp_horz_disp, tmp_vert_disp;
	double theSum = 0.0;
	double theOutVal = 0.0;


	tmp_horz_disp = 1.0;

	vert_disp = 1.0 - vert_disp;	// we must mirror reverse this displacement

	for( i = 0; i < kCoeffsRows; ++ i )
	{
		tmp_vert_disp = 1.0;
		theSum = 0.0;

		for( j = 0; j < kCoeffsCols; ++ j )
		{
			theSum += fBicubicCoeffs[ i ][ j ] * tmp_vert_disp;
			
			// Compute the consecutive power of the vert displacement
			tmp_vert_disp *= vert_disp;
		}

		theOutVal += tmp_horz_disp * theSum;

		// Compute the consecutive powers of the horz displacement
		tmp_horz_disp *= horz_disp;
	}

	theInterpolatedVal = theOutVal;


	return true;
}




//////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////
// This is a base class for image warping engine.
///////////////////////////////////////////////////////////
template < class T = double >
class TImageWarp
{
	public:
		
		typedef std::auto_ptr< TImageFor< T > > IMAGE_AUTO_PTR;


	protected:


		const TImageFor< T > & fInImage;

		TCoordTransformEngine & fCoordTransformEngine;

		TPixelInterpolation< T > & fPixelInterpolation;

	public:

		// ===================================================
		TImageWarp( const TImageFor< T > & inImage, TCoordTransformEngine & coordTransformEngine, TPixelInterpolation< T > & pixelInterpolation )
			: fInImage( inImage ), fCoordTransformEngine( coordTransformEngine ), fPixelInterpolation( pixelInterpolation ) {}
			
		virtual ~TImageWarp() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function performs the image warping.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			the warped image 
		//		
		// REMARKS:
		//		
		virtual IMAGE_AUTO_PTR operator () ( void ) = 0;


};




///////////////////////////////////////////////////////////
// This class implements the simple backward (or inverse)
// warping scheme: the images are mapped from the output
// grid into the input space (thus its name).
///////////////////////////////////////////////////////////
template < class T = double >
class TSimpleImageBackwardWarp : public TImageWarp< T >
{
	private:

		// Class inherent variables

	protected:


	public:

		// ===================================================
		TSimpleImageBackwardWarp( const TImageFor< T > & inImage, TCoordTransformEngine & coordTransformEngine, TPixelInterpolation< T > & pixelInterpolation = TBilinearInterpolation< T >::gkBilinearInterpolation )
			: TImageWarp< T >( inImage, coordTransformEngine, pixelInterpolation ) {}
		virtual ~TSimpleImageBackwardWarp() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function performs the image warping.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			the warped image 
		//		
		// REMARKS:
		//		
		//		
		virtual IMAGE_AUTO_PTR operator () ( void );

};




///////////////////////////////////////////////////////////
// This function performs the image warping.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			none
//		
// OUTPUT:
//			the warped image 
//		
// REMARKS:
//		
//		
template < class T >
typename TSimpleImageBackwardWarp< T >::IMAGE_AUTO_PTR TSimpleImageBackwardWarp< T >::operator () ( void )
{
	const int kCols = fInImage.GetCol();
	const int kRows = fInImage.GetRow();

	IMAGE_AUTO_PTR outImage( new TImageFor< T >( kCols, kRows, T() ) );		// will be intialized to 0 by a default constructor for built-in types
	
	register int i, j;

	// Traverse integer grid of the output image
	for( i = 0; i < kRows; ++ i )
	{
		for( j = 0; j < kCols; ++ j )
		{
			// The new coordinates are in the space of the input image
			Real_2D_Point inputSpaceCoords = fCoordTransformEngine( Real_2D_Point( j, i ) );

			// Find value of a pixel in new position (usually non integer) of the input image;
			// If the interpolation is not possible (e.g. the coordinates are out of range
			// of the input image) then skip
			T interpolPixelVal;
			if( ! ( inputSpaceCoords == TLinearTransformEngine::kPointAtInfinity ) && fPixelInterpolation( fInImage, inputSpaceCoords, interpolPixelVal ) == true )
				outImage->SetPixel( j, i, interpolPixelVal );
		}
	}

	return outImage;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////
// This class implements the simple backward scheme,
// exactly as the TSimpleImageBackwardWarp does.
// However, it allows output image to have different
// size than the input one.
///////////////////////////////////////////////////////////
template < class T = double >
class TImageBackwardWarp : public TSimpleImageBackwardWarp< T >
{
	private:

		// Class inherent variables

	protected:

		int fOutImage_Cols;			// desired number of columns in the output image
		int fOutImage_Rows;			// desired number of rows in the output image

	public:

		// ===================================================
		TImageBackwardWarp( const TImageFor< T > & inImage, const int outImage_Cols, const int outImage_Rows, TCoordTransformEngine & coordTransformEngine, TPixelInterpolation< T > & pixelInterpolation = TBilinearInterpolation< T >::gkBilinearInterpolation )
			: fOutImage_Cols( outImage_Cols ), fOutImage_Rows( outImage_Rows ), TSimpleImageBackwardWarp< T >( inImage, coordTransformEngine, pixelInterpolation ) {}
		TImageBackwardWarp( const TImageFor< T > & inImage, TCoordTransformEngine & coordTransformEngine, TPixelInterpolation< T > & pixelInterpolation = TBilinearInterpolation< T >::gkBilinearInterpolation )
			: fOutImage_Cols( inImage.GetCol() ), fOutImage_Rows( inImage.GetRow() ), TSimpleImageBackwardWarp< T >( inImage, coordTransformEngine, pixelInterpolation ) {}
		virtual ~TImageBackwardWarp() {}				// class virtual destructor
		// ===================================================


		///////////////////////////////////////////////////////////
		// This function performs the image warping.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			the warped image 
		//		
		// REMARKS:
		//		
		//		
		virtual IMAGE_AUTO_PTR operator () ( void );

};


///////////////////////////////////////////////////////////
// This function performs the image warping.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			none
//		
// OUTPUT:
//			the warped image 
//		
// REMARKS:
//		
//		
template < class T >
typename TImageBackwardWarp< T >::IMAGE_AUTO_PTR TImageBackwardWarp< T >::operator () ( void )
{
	REQUIRE( fOutImage_Cols < 10000 );		// sanity check
	REQUIRE( fOutImage_Rows < 10000 );		// sanity check

	IMAGE_AUTO_PTR outImage( new TImageFor< T >( fOutImage_Cols, fOutImage_Rows, T() ) );	// for the built-in types T() initializes them to 0 

	register int i, j;

	// Traverse integer grid of the output image
	for( i = 0; i < fOutImage_Rows; ++ i )
	{
		for( j = 0; j < fOutImage_Cols; ++ j )
		{
			// The new coordinates are in the space of the input image
			Real_2D_Point inputSpaceCoords = fCoordTransformEngine( Real_2D_Point( j, i ) );

			// Find value of a pixel in new position (usually non integer) of the input image;
			// If the interpolation is not possible (e.g. the coordinates are out of range
			// of the input image) then skip
			T interpolPixelVal;
			if( fPixelInterpolation( fInImage, inputSpaceCoords, interpolPixelVal ) == true )
				outImage->SetPixel( j, i, interpolPixelVal );
		}
	}

	return outImage;
}


//////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////


#endif BoxySoft_Warping_h




