
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: IPL2HIL_ConvertTest.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Format converters to/from the Intel's Open CV library.
//
// ==========================================================================

#pragma warning(disable:4786)




#include "IntelLIb_ImageConverters.h"

#if CONNECT_2_OPEN_CV

#include "HIL_BaseDefinitions.h"
#include "HIL_ArithmeticOperators.h"
#include "HIL_LogicalOperators.h"
#include "HIL_ComposedOperations.h"
#include "HIL_ConvolveOperators.h"
#include "HIL_FindValueInImageOperators.h"
#include "HIL_ImageFormatConvertOperators.h"
#include "HIL_MultiChannelImageFor.h"
#include "HIL_MultiChannelMixedCopy.h"
#include "HIL_ColorSpaceConvertOperators.h"

#include "ImageErrorMeasurementLibrary.h"
#include <time.h>
#include <fstream>



using namespace PHIL;


///////////////////////////////////////////////////////////
// Test function: conversion from IPL format, some operations
// output to the IPL format
///////////////////////////////////////////////////////////
//		
// INPUT:
//			fileName - name of a file to be converted 
//				It must be a monochrome and uncompressed format
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
void Mono_IPL_Test( const char * fileName )
{
	IplImage * image_1 = 0; 

	// Directly load a TIFF image to the IPL format
	// WARNING - the Intel's library accepts only non-compressed TIFF format !!!
	image_1 = cvLoadImage( fileName, 0 /*always grayscale*/ );
	if( image_1 == 0 )
		return;			// apparently cannot load that image


	// Create temporary fixed-pixel images
	ImageForFixed_16_16 temp_fixed_1( image_1->width, image_1->height );
	ImageForFixed_16_16 temp_fixed_2( image_1->width, image_1->height );


	// We can create and immediately execute each image operation, step by step.
	// Alternatively we can create a compound operation and then call it at once
	// - all registered operations will be executed in an order they were registered
	// (or in an another order if required). The latter approach is presented here.
	TVectorMultiImageOperation	convert_smooth_operator;		// this object can be automatic


	
	// At first we wish to register conversion from whatever format of an input image
	// to the format required in this function - in this case from the IPL to fixed.
	convert_smooth_operator.AddAdoptNewOperation( FormatConverter_AP( temp_fixed_1, * image_1 ) );


	// Let's prepare the smoothing kernel
	vector< FIXED_16_16 > theBinomialSmoothing;
	theBinomialSmoothing.push_back( 0.25 );
	theBinomialSmoothing.push_back( 0.5 );
	theBinomialSmoothing.push_back( 0.25 );


	// Then register vertical smoothing followed by a horizontal smoothing with the same mask
	convert_smooth_operator.AddAdoptNewOperation( Vert_Convolve_AP( temp_fixed_2, temp_fixed_1, theBinomialSmoothing ) );
	convert_smooth_operator.AddAdoptNewOperation( Horz_Convolve_AP( temp_fixed_1, temp_fixed_2, theBinomialSmoothing ) );


	// Now register copy-and-convert operation in opposite direction (fixed to IPL)
	convert_smooth_operator.AddAdoptNewOperation( FormatConverter_AP( * image_1, temp_fixed_1 ) );


	// All components already registered, so launch them all - they will be called linearly
	// because we are using the vector based composite operation (i.e. the TVectorMultiImageOperation)
	convert_smooth_operator();

	// Now try to save the result image using the Intel's library
	string str( fileName );	
	string::size_type period_place = str.find( '.' );
	if( period_place != string::npos )
		str.insert( period_place, string( "2" ) );	// modify name a little
	// Save the output result
	cvSaveImage( str.c_str(), image_1 );			

	cvReleaseImage( & image_1 );
}

///////////////////////////////////////////////////////////
// Test function for interlaced color images
///////////////////////////////////////////////////////////
//		
// INPUT:
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			none
//		
// REMARKS:
//		
//		
void Inter_Color_IPL_Test( const char * fileName )
{
	IplImage * image_1 = 0; 

	// Directly load a TIFF image to the IPL format
	// WARNING - the Intel's library accepts only non-compressed TIFF format !!!
	image_1 = cvLoadImage( fileName, -1 );
	if( image_1 == 0 )
		return;			// apparently cannot load that image

	// Create a temporardy object - a color (interlaced) image 
	ColorImage theColorImage( image_1->width, image_1->height );

	// Create and execute the conversion from the input IPL format
	( * FormatConverter_AP( theColorImage, * image_1 ) )();

	// Create a temporary IPL color image which is compatible with the input one
	IplImage * image_2 = cvCreateImage( cvSize( image_1->width, image_1->height ), image_1->depth, image_1->nChannels );
	if( image_2 == 0 )
		return;			// apparently cannot create that image
	
	// Do something with theColorImage...
	// ...

	// Convert the same image from RGB to HSI
	( * RGB_2_HSI_Converter_AP( theColorImage, theColorImage ) )();
	// Convert it back, now from HIS to RGB
	( * HSI_2_RGB_Converter_AP( theColorImage, theColorImage ) )();


	// Convert the same image from RGB to YCrCb
	ColorRealValuedImage theColorRealImage( image_1->width, image_1->height );
	MixedCopy( theColorRealImage, theColorImage );
	( * RGB_2_YCrCb_Converter_AP( theColorRealImage, theColorRealImage ) )();

	// -----------------------------------
	// Save YCrCb to visualize it:

	// Convert it back, now from YCrCb to RGB
	( * YCrCb_2_RGB_Converter_AP( theColorRealImage, theColorRealImage ) )();
	MixedCopy( theColorImage, theColorRealImage );


	// Create and execute the conversion to the temporary IPL image
	( * FormatConverter_AP( * image_2, theColorImage ) )();


	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

	ColorImage theColorImage2( theColorImage );
	IplImage * image_2_tmp = cvCreateImage( cvSize( image_1->width, image_1->height ), image_1->depth, image_1->nChannels );
	
	

	IplImage * ipl_test_float_1 = cvCreateImage( cvSize( image_1->width, image_1->height ), IPL_DEPTH_32F, 1 );
	IplImage * ipl_test_float_2 = cvCreateImage( cvSize( image_1->width, image_1->height ), IPL_DEPTH_32F, 1 );
	IplImage * ipl_test_float_3 = cvCreateImage( cvSize( image_1->width, image_1->height ), IPL_DEPTH_32F, 1 );

	TImageFor< float > hil_test_float_1( image_1->width, image_1->height );
	TImageFor< float > hil_test_float_2( image_1->width, image_1->height );
	TImageFor< float > hil_test_float_3( image_1->width, image_1->height );

	ImageForFixed_8_8 fixed_test_1( image_1->width, image_1->height );
	ImageForFixed_8_8 fixed_test_2( image_1->width, image_1->height );
	ImageForFixed_8_8 fixed_test_3( image_1->width, image_1->height );

	double start, finish, duration_1, duration_2;
	
	ofstream exec_time( "ExecTime.txt" );
	start = clock();

	////////////////////////////
	// Now for IPL
    for (int j=0; j<100;j++) {

		cvAdd( image_1, image_2, image_1 );
   }
	////////////////////////////

	duration_2 = clock() - start;
	exec_time << "Exec time for IPL: " << duration_2 << endl;

	start = clock();

	
	////////////////////////////
	for (int i=0; i<100;i++) {
           
			theColorImage += theColorImage2;
    }
	////////////////////////////
	
	duration_1 = clock() - start;
	exec_time << "Exec time for HIL: " << duration_1 << endl;

	cvReleaseImage( & image_2_tmp );

	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@



	// Save it back using Intel's procedures
	string str( fileName );	
	string::size_type period_place = str.find( '.' );
	if( period_place != string::npos )
		str.insert( period_place, string( "2" ) );	// modify an output file name a little
	cvSaveImage( str.c_str(), image_2 );

	cvReleaseImage( & image_1 );
	cvReleaseImage( & image_2 );

	cvReleaseImage( & ipl_test_float_1 );
	cvReleaseImage( & ipl_test_float_2 );
	cvReleaseImage( & ipl_test_float_3 );


}

///////////////////////////////////////////////////////////
// Test function for non-interlaced color images
///////////////////////////////////////////////////////////
//		
// INPUT:
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			none
//		
// REMARKS:
//		
//		
void NonInter_Color_IPL_Test( const char * fileName )
{
	IplImage * image_1 = 0; 

	// Directly load a TIFF image to the IPL format
	// WARNING - the Intel's library accepts only non-compressed TIFF format !!!
	image_1 = cvLoadImage( fileName, -1 );
	if( image_1 == 0 )
		return;			// apparently cannot load that image

	// Create a non-interlaced color image, fixed-point pixels, the same size
	// as the input IPL
	NI_Color_3xFi16_16_Image theColorImage( image_1->width, image_1->height, Color_3xFi16_16_Pixel() );

	// Create a temporardy object - a color (interlaced) image 
	( * FormatConverter_AP( theColorImage, * image_1 ) )();

	// Create a temporary IPL color image which is compatible with the input one
	IplImage * image_2 = cvCreateImage( cvSize( image_1->width, image_1->height ), image_1->depth, image_1->nChannels );
	if( image_2 == 0 )
		return;			// apparently cannot create that image

	// Do something with theColorImage...
	// ...


	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	{
		// Let's test copying from non-interlaced to interlaced color images
		// among HIL objects


	
		NI_Color_3x8_Image theMultiPixelColorImage( image_1->width, image_1->height );
		// Copy between two non-interlaced color images:
		( * FormatConverter_AP( theMultiPixelColorImage, theColorImage ) )();
		( * FormatConverter_AP( theColorImage, theMultiPixelColorImage ) )();


		ColorImage interlacedColorImage( image_1->width, image_1->height );
		// Copy between non-interlaced and interlaced color images:
		( * FormatConverter_AP( interlacedColorImage, theColorImage ) )();
		( * FormatConverter_AP( theColorImage, interlacedColorImage ) )();


	}
	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	


	// Create and execute the conversion to the temporary IPL image
	( * FormatConverter_AP( * image_2, theColorImage ) )();


	// Save it back using Intel's procedures
	string str( fileName );	
	string::size_type period_place = str.find( '.' );
	if( period_place != string::npos )
		str.insert( period_place, string( "3" ) );	// modify a name a little
	cvSaveImage( str.c_str(), image_2 );

	cvReleaseImage( & image_1 );
	cvReleaseImage( & image_2 );
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////
// Test function creates an interlaced image from a file.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			ptr to the created and ORPHANED ColorImage
//				object upon success;
//			0 if failure
//		
// REMARKS:
//			The caller is responsible of disposing 
//				the returned object.
//		
NI_Color_3x8_Image * Orphan_ColorImageFrom( const char * fileName )
{
	IplImage * image_1 = 0; 

	// Directly load a TIFF image to the IPL format
	// WARNING - the Intel's library accepts only non-compressed TIFF format !!!
	image_1 = cvLoadImage( fileName, -1 );
	if( image_1 == 0 )
		return 0;			// apparently cannot load that image

	NI_Color_3x8_Image * outImage = new NI_Color_3x8_Image( image_1->width, image_1->height );

	( * FormatConverter_AP( * outImage, * image_1 ) )();		// change format

	cvReleaseImage( & image_1 );

	return outImage;
}

///////////////////////////////////////////////////////////
// Test function saves a color image to a file.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			theImage - a reference to the image object
//				to be saved
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			true if success,
//			false otherwise
//		
// REMARKS:
//		
bool Save_ColorImageIn( const NI_Color_3x8_Image & theImage, const char * fileName )
{
	// Create a temporary IPL color image
	const int kPixelDepth = 8;		// 8 bits
	const int kChannels = 3;
	IplImage * image_1 = cvCreateImage( cvSize( theImage.GetCol(), theImage.GetRow() ), kPixelDepth, kChannels );
	if( image_1 == 0 )
		return false;			// apparently cannot create that image

	( * FormatConverter_AP( * image_1, theImage ) )();		// change format

	int retVal = cvSaveImage( fileName, image_1 );

	cvReleaseImage( & image_1 );

	return retVal == 0;
}


///////////////////////////////////////////////////////////
// Test function creates an non-interlaced image from a file.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			ptr to the created and ORPHANED ColorImage
//				object upon success;
//			0 if failure
//		
// REMARKS:
//			The caller is responsible of disposing 
//				the returned object.
//		
ColorImage * Orphan_NonInterColorImageFrom( const char * fileName )
{
	IplImage * image_1 = 0; 

	// Directly load a TIFF image to the IPL format
	// WARNING - the Intel's library accepts only non-compressed TIFF format !!!
	image_1 = cvLoadImage( fileName, -1 );
	if( image_1 == 0 )
		return 0;			// apparently cannot load that image

	ColorImage * outImage = new ColorImage( image_1->width, image_1->height );

	( * FormatConverter_AP( * outImage, * image_1 ) )();		// change format

	cvReleaseImage( & image_1 );

	return outImage;
}

///////////////////////////////////////////////////////////
// Test function saves a color image to a file.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			theImage - a reference to the image object
//				to be saved
//			fileName - name of a file to be converted 
//				It must be a color and uncompressed format
//		
// OUTPUT:
//			true if success,
//			false otherwise
//		
// REMARKS:
//		
bool Save_ColorImageIn( const ColorImage & theImage, const char * fileName )
{
	// Create a temporary IPL color image
	const int kPixelDepth = 8;		// 8 bits
	const int kChannels = 3;
	IplImage * image_1 = cvCreateImage( cvSize( theImage.GetCol(), theImage.GetRow() ), kPixelDepth, kChannels );
	if( image_1 == 0 )
		return false;			// apparently cannot create that image

	( * FormatConverter_AP( * image_1, theImage ) )();		// change format

	int retVal = cvSaveImage( fileName, image_1 );

	cvReleaseImage( & image_1 );

	return retVal == 0;
}



#endif CONNECT_2_OPEN_CV



























