
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_BooleanImage.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Specialization for the boolean images.
//
// ==========================================================================


#ifndef BoxySoft_BooleanImage
#define BoxySoft_BooleanImage



#include "HIL_ImageFor.h"


// ==========================================================================

namespace PHIL		
{


// ==========================================================================



////////////////////////////////////////////////////////////
// Specialization for binary images.
// Its internal structures are organized in such a way as to 
// save on space and access time.
////////////////////////////////////////////////////////////
template<> class TImageFor< bool >
{


protected:


	Dimension fRow;
	Dimension fCol;


	typedef unsigned long InternalAddressableData;

	enum { kBitsPerByte = 8 };
	enum { kBitsPerSingleWord = kBitsPerByte * sizeof( InternalAddressableData ) };

	InternalAddressableData * fData;

private:

	// Auxialiary data used to save computation time

	unsigned long fTotalWordCapacity;	// contains number of computer words devoted to store all bits
	unsigned long fTotalByteCapacity;	// contains number of computer words devoted to store all bits
										// The class invariant is as follows:
										//		fTotalByteCapacity = fTotalWordCapacity * sizeof( InternalAddressableData );	

public:

	virtual unsigned long GetOccupiedMemorySizeInBytes( void ) const { return fTotalByteCapacity; }		// contains number of computer words devoted to store all bits


protected:



	virtual void Create( Dimension col, Dimension row  ) 
	{
		fRow = row;
		fCol = col;

		int numOfElements = row * col;
		fTotalWordCapacity = numOfElements / kBitsPerSingleWord;
		if( fTotalWordCapacity * kBitsPerSingleWord < numOfElements )
			fTotalWordCapacity ++;

		fData = ( InternalAddressableData * ) new InternalAddressableData[ fTotalWordCapacity ];

		fTotalByteCapacity = fTotalWordCapacity * sizeof( InternalAddressableData );

		// Fill everything with 0 just to avoid undefined nibbles of data.
		memset( (unsigned char*)fData, 0, fTotalByteCapacity );
	}

protected:

	// Default protected constructor. 
	TImageFor( void ) 
	{
	}

public:

	//======================================
	TImageFor( Dimension col, Dimension row  ) 
	{
		Create( col, row );
	}

	TImageFor( Dimension col, Dimension row, bool init_val  ) 
	{
		Create( col, row );
		SetAll( init_val );
	}

	TImageFor( const Index index ) 
	{
		Create( index.fCol, index.fRow );
	}

	TImageFor( const Index index, bool init_val ) 
	{
		Create( index.fCol, index.fRow );
		SetAll( init_val );
	}

	// Template copy constructor (mixed copy)
	// If the input data is zero then set "false", otherwise set "true".
	template < typename U >
	TImageFor( const TImageFor< U > & ref, const U refThreshold = 0 )
	{
		fRow = ref.GetRow();
		fCol = ref.GetCol();

		int numOfElements = fRow * fCol;
		fTotalWordCapacity = numOfElements / kBitsPerSingleWord;
		if( fTotalWordCapacity * kBitsPerSingleWord < numOfElements )
			fTotalWordCapacity ++;

		fData = ( InternalAddressableData * ) new InternalAddressableData[ fTotalWordCapacity ];

		fTotalByteCapacity = fTotalWordCapacity * sizeof( InternalAddressableData );
		
		// Fill everything with 0 just to avoid undefined nibbles of data.
		memset( (unsigned char*)fData, 0, fTotalByteCapacity );

		// Copy data - if the input data is zero then set "false",
		// otherwise set "true".
 		register Dimension i, j;
 		for( i = 0; i < fRow; i ++ )
			for( j = 0; j < fCol; j ++ )
				SetPixel( j, i, ref.GetPixel( j, i ) <= refThreshold ? false : true );
	}

	// Copy constructor
	TImageFor( const TImageFor< bool > & ref )
	{
		fRow = ref.GetRow();
		fCol = ref.GetCol();
		fTotalWordCapacity = ref.fTotalWordCapacity;
		fTotalByteCapacity = ref.fTotalByteCapacity;

		fData = ( InternalAddressableData * ) new InternalAddressableData[ fTotalWordCapacity ];

		memcpy( (unsigned char*)fData, (unsigned char*)ref.fData, fTotalByteCapacity );
	}


	template < class U >
	TImageFor< bool > & operator = ( const TImageFor< U > & ref )
	{
		if( this != (TImageFor< bool > *) &ref )
		{
			fRow = ref.GetRow();
			fCol = ref.GetCol();
	
			delete [] fData;		// get rid of old data
			
			int numOfElements = fRow * fCol;
			fTotalWordCapacity = numOfElements / kBitsPerSingleWord;
			if( fTotalWordCapacity * kBitsPerSingleWord < numOfElements )
				fTotalWordCapacity ++;

			fData = ( InternalAddressableData * ) new InternalAddressableData[ fTotalWordCapacity ];

			fTotalByteCapacity = fTotalWordCapacity * sizeof( InternalAddressableData );

			// Fill everything with 0 just to avoid undefined nibbles of data.
			memset( (unsigned char*)fData, 0, fTotalByteCapacity );

			// Copy data - if the input data is zero then set "false",
			// otherwise set "true".
 			register Dimension i, j;
 			for( i = 0; i < fRow; i ++ )
				for( j = 0; j < fCol; j ++ )
					SetPixel( j, i, ref.GetPixel( j, i ) == 0 ? false : true );
		}

		return *this;
	}

	TImageFor< bool > & operator = ( const TImageFor< bool > & ref )
	{
		if( this != & ref )
		{
			fRow = ref.GetRow();
			fCol = ref.GetCol();
			fTotalWordCapacity = ref.fTotalWordCapacity;
			fTotalByteCapacity = ref.fTotalByteCapacity;

			delete [] fData;		// get rid of old data
			
			// Create new space for this object
			fData = ( InternalAddressableData * ) new InternalAddressableData[ fTotalWordCapacity ];
		
			// Copy data from the supplied object
			memcpy( (unsigned char*)fData, (unsigned char*)ref.fData, fTotalByteCapacity );
		}

		return *this;
	}


	virtual ~TImageFor() 
	{
		delete [] fData;
	}


	//======================================

	// Returns a pointer to the NEW image object which was built 
	// as a part of the "this" image. You are responsible for that 
	// object.
	// The new image is composed starting from (s_col,s_row) 
	// and extended to: width x height.
	//
	// INPUT:
	//			s_col - upper left coordinate of the new image (expressed
	//				in the mother coordinate system)
	//			s_row - top coordinate of the new image
	//			width - width of a new image
	//			height - height of a new image
	//			
	//
	// OUTPUT:
	//			pointer to the new image or 0 if creation not possible
	//			Responsibility of the object passed by this pointer is left
	//			to the adopting part.
	//
	TImageFor< bool > * OrphanBitBlit( Dimension s_col, Dimension s_row, Dimension width, Dimension height ) const
	{
		if( s_col >= fCol || s_row >= fRow || width == 0|| height == 0 )
			return 0;		// wrong input parameters

		// check whether the new window overlappes the whole image
		// - if not then exit 
		if( s_row + height > fRow || s_col + width > fCol )
			return 0;

		TImageFor< bool > * obj = new TImageFor< bool >( width, height );

		// copy data from the mother object
		register Dimension i, j;
		for( i = 0; i < height; i ++ )
		{
			for( j = 0; j < width; j ++ )
			{
				Dimension current_row = s_row + i;
				obj->SetPixel( j, i, GetPixel( s_col + j, current_row ) );
			}
		}

		return obj;
	}

	//======================================


	// Returns true if the two pictures are the same
	bool operator == ( const TImageFor< bool > & image ) const
	{
		// if different dimensions, then for sure not equal
		if( fRow != image.fRow || fCol != image.fCol )
			return false;
  		
		REQUIRE( fTotalWordCapacity == image.fTotalWordCapacity );
		REQUIRE( fTotalByteCapacity == image.fTotalByteCapacity );

		// Copy data from the supplied object
		return memcmp( fData, image.fData, fTotalByteCapacity ) == 0;
	}

	TImageFor< bool > operator | ( const TImageFor< bool > & image ) const
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		TImageFor< bool >	outImage( fCol, fRow );

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
			* ( outImage.fData + i ) = * ( fData + i ) | * ( image.fData + i );

		return outImage;
	}

	TImageFor< bool > & operator |= ( const TImageFor< bool > & image )
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
		{
			InternalAddressableData tmp = * ( fData + i );
			* ( fData + i ) = tmp | * ( image.fData + i );
		}

		return * this;
	}

	TImageFor< bool > operator & ( const TImageFor< bool > & image ) const
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		TImageFor< bool >	outImage( fCol, fRow );

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
			* ( outImage.fData + i ) = * ( fData + i ) & * ( image.fData + i );

		return outImage;
	}

	TImageFor< bool > & operator &= ( const TImageFor< bool > & image )
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
		{
			InternalAddressableData tmp = * ( fData + i );
			* ( fData + i ) = tmp & * ( image.fData + i );
		}

		return * this;
	}

	TImageFor< bool > operator ^ ( const TImageFor< bool > & image ) const
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		TImageFor< bool >	outImage( fCol, fRow );

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
			* ( outImage.fData + i ) = * ( fData + i ) ^ * ( image.fData + i );

		return outImage;
	}

	TImageFor< bool > & operator ^= ( const TImageFor< bool > & image )
	{
		REQUIRE( fCol == image.fCol && fRow == image.fRow );
		if( fCol != image.fCol || fRow != image.fRow )
			throw "";

		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
		{
			InternalAddressableData tmp = * ( fData + i );
			* ( fData + i ) = tmp ^ * ( image.fData + i );
		}

		return * this;
	}


	TImageFor< bool > & operator ~ ( void )
	{
		register Dimension i;
		for( i = 0; i < fTotalWordCapacity; ++ i )
		{
			InternalAddressableData tmp = * ( fData + i );
			* ( fData + i ) = ~ tmp;
		}

		return * this;
	}



	InternalAddressableData * GetImageData( void ) const { return fData; }



	// The xPixPosition should span from 0 to max_columns, while
	// the yPixPosition from 0 to max_rows.
	void SetPixel( Dimension xPixPosition, Dimension yPixPosition, bool value )
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		int in_matrix_bit_index = yPixPosition * fCol + xPixPosition;

		int word_offset			= in_matrix_bit_index / kBitsPerSingleWord;
		int bit_in_word_offset	= in_matrix_bit_index - word_offset * kBitsPerSingleWord;

		REQUIRE( bit_in_word_offset < kBitsPerByte * sizeof( InternalAddressableData ) );

		InternalAddressableData theWord = * ( fData + word_offset );
	
		const InternalAddressableData kOne = 1;
		InternalAddressableData theMask = kOne << bit_in_word_offset;

		* ( fData + word_offset ) = value == true ? theWord | theMask : theWord & ~theMask ;
	}

	// The xPixPosition should span from 0 to max_columns, while
	// the yPixPosition from 0 to max_rows.
	bool GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		int in_matrix_bit_index = yPixPosition * fCol + xPixPosition;

		int word_offset			= in_matrix_bit_index / kBitsPerSingleWord;
		int bit_in_word_offset	= in_matrix_bit_index - word_offset * kBitsPerSingleWord;
		
		REQUIRE( bit_in_word_offset < kBitsPerByte * sizeof( InternalAddressableData ) );

		InternalAddressableData theWord = * ( fData + word_offset );
		const InternalAddressableData kOne = 1;
		InternalAddressableData theMask = kOne << bit_in_word_offset;

		bool retBit = ( theWord & theMask ) == 0 ? false : true;

		return retBit;
	}

	virtual bool & operator [] ( Index index )
	{
		REQUIRE( false );		// this is not good function for this specialization
								// since in C++ we cannot have a reference to a signle bit
								// although we can return its value; left here to allow
								// instantializations of the MaskedBooleanImage
		static bool gTheDummyPixel = GetPixel( index.fCol, index.fRow );
		return gTheDummyPixel;
	}

	void SetAll( bool value )
	{
		// Fill everything with 0 or 1.
		memset( (unsigned char*)fData, value == true ? 0xFF : 0x00, fTotalByteCapacity );
	}


	Dimension GetRow( void ) const { return fRow; }
	Dimension GetCol( void ) const { return fCol; }


};


////////////////////////////////////////////////////////////////////////////////

typedef TImageFor< bool >			BinaryImage;

////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
// Auxiliary I/O

static istream & operator >> ( istream & i, bool & b )
{
	int tmp;
	i >> tmp;
	
	b = tmp == 0 ? false : true;

	return i;
}


////////////////////////////////////////////////////////////////////////////////



// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================


#endif BoxySoft_BooleanImage




