
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_ImageFor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Classes for representation of images.	
//
// ==========================================================================




#ifndef BoxySoft_HIL_ImageFor
#define BoxySoft_HIL_ImageFor

#pragma warning(disable:4786)


#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <xmemory>

#include "ImageDeclarations.h"
#include "HIL_PixelAccessTraits.h"



using namespace std;


// ==========================================================================

namespace PHIL		
{


// ==========================================================================



//////////////////////////////////////////////////////////////////////////////////	

const double kPi		= 3.1415926535897932384626433832795;
const double k2Pi		= 2.0 * kPi;
const double kPiHalf	= kPi / 2.0;

//////////////////////////////////////////////////////////////////////////////////	

typedef unsigned char			MonoPixel_Type;
typedef MonoPixel_Type *		MonoPixel_Ptr_Type;

typedef double					RealPixel_Type;
typedef RealPixel_Type *		RealPixel_Ptr_Type;

typedef long					Dimension;

//////////////////////////////////////////////////////////////////////////////////	

struct Index
{ 
	Dimension fCol; 
	Dimension fRow; 

	Index( void ) { fCol = 0; fRow = 0; }
	Index( Dimension _col, Dimension _row ) { fCol = _col; fRow = _row; }
};

//////////////////////////////////////////////////////////////////////////////////	



//////////////////////////////////////////////////////////////////////////////////	
// The basic structure for 2D images. Data is stored in rows like in the following
// diagram:
//
//			c o l u m n s
//
//			r	0   1   2
//			o	3	4	5
//			w	6	7	8
//			s
//
// memory alignement for the above is:
//
//	0 1 2 3 4 5 6 7 8 ...
//
//	or:
//
//	fData -> 0
//	fData + 1 -> 1
//	fData + 2 -> 2
//	..............
//
template< typename T, typename PAT = PixelAccess_Trait< T >, typename AL = allocator< T > >		// PAT = Pixel Access Trait
class TImageFor
{
public:

	typedef typename T PixelType;

	typedef typename PAT::PixelAccessType		PixelAccessType;
	typedef typename PAT::ConstPixelAccessType	ConstPixelAccessType;


	// Typedefs for compatibility with STL
	typedef typename AL									allocator_type;
	typedef typename AL::size_type						size_type;
	typedef typename AL::difference_type				difference_type;
	typedef typename AL::pointer						_Tptr;
	typedef typename AL::const_pointer					_Ctptr;
	typedef typename AL::reference						reference;
	typedef typename AL::const_reference				const_reference;
	typedef typename AL::value_type						value_type;


private:
protected:


	Dimension	fRow;		// contains number of rows of that image
	Dimension	fCol;		// contains number of columns of that image

	T *			fData;		// pointer to the compact data buffer


	Dimension	fElems;		// = fRow * fCol, redundant membe, used to spead up computations 


	allocator_type			fAllocator;	// the memory allocation object


protected:



	///////////////////////////////////////////////////////////
	// This function initializes an image
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			col - number of columns
	//			row - number of rows
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//			Memory for data is allocated but 
	//			data is NOT initialized.	
	//		
	void Create( Dimension col, Dimension row  ) 
	{
		REQUIRE( col > 0 );			// pre-condtions
		REQUIRE( row > 0 );

		fRow = row;
		fCol = col;

		fElems = row * col;
		fData = fAllocator.allocate( fElems, (void*)0 );


		REQUIRE( fData != 0 );		// post-condition

		T * ptr = fData;
		for( int i = 0; i < fElems; ++ i )
			fAllocator.construct( ptr ++, T() );
	}

public:

	// Default constructor. 
	TImageFor( const AL & allocator = AL() ) : fAllocator( allocator ), fRow( 0 ), fCol( 0 ), fData( 0 )
	{
	}

public:

	//======================================


	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			col - number of columns
	//			row - number of rows
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//			Memory for data is allocated but 
	//			data is NOT initialized.	
	//		
	TImageFor( Dimension col, Dimension row, const AL & allocator = AL() ) : fAllocator( allocator )
	{
		Create( col, row );
	}

	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			col - number of columns
	//			row - number of rows
	//			init_val - initial value for each pixel
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//		
	//		
	TImageFor( Dimension col, Dimension row, T init_val, const AL & allocator = AL() ) : fAllocator( allocator ) 
	{
		Create( col, row );
		SetAll( init_val );
	}

	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			index - number of columns and rows
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//			Memory for data is allocated but 
	//			data is NOT initialized.	
	//		
	explicit TImageFor( const Index index, const AL & allocator = AL() ) : fAllocator( allocator )	// explicit classifier
	{																								// to avoid unwanted implicit casts
		Create( index.fCol, index.fRow );
	}

	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			index - number of columns and rows
	//			init_val - initial value for each pixel
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//		
	//		
	TImageFor( const Index index, T init_val, const AL & allocator = AL() ) : fAllocator( allocator ) 
	{
		Create( index.fCol, index.fRow );
		SetAll( init_val );
	}


	///////////////////////////////////////////////////////////
	// Template copy constructor - mixed copy constructor 
	///////////////////////////////////////////////////////////
	template < typename U, typename P, typename A >			
	TImageFor( const TImageFor< U, P, A > & ref )			
		: fAllocator( AL() )
	{															

		fRow = ref.GetRow();
		fCol = ref.GetCol();

		fElems = fRow * fCol;

		fData = fAllocator.allocate( fElems, (void*)0 );

		REQUIRE( fData != 0 );		// post-condition

		U * source_ptr = ref.GetImageData();
		T * destin_ptr_start = fData;
		T * destin_ptr_end = fData + fElems;

		while( destin_ptr_start < destin_ptr_end )
			fAllocator.construct( destin_ptr_start ++, (T)* source_ptr ++ );

	}

	///////////////////////////////////////////////////////////
	// Copy constructor
	///////////////////////////////////////////////////////////
	TImageFor( const TImageFor< T, PAT, AL > & ref )
		: fAllocator( ref.fAllocator )
	{
		fRow = ref.GetRow();
		fCol = ref.GetCol();

		fElems = fRow * fCol;
		fData = fAllocator.allocate( fElems, (void*)0 );

		REQUIRE( fData != 0 );		// post-condition

		T * source_ptr = ref.fData;
		T * destin_ptr_start = fData;
		T * destin_ptr_end = fData + fElems;
		while( destin_ptr_start < destin_ptr_end )
			fAllocator.construct( destin_ptr_start ++, * source_ptr ++ );
	}

	///////////////////////////////////////////////////////////
	// Template assignement operator (mixed copy)
	///////////////////////////////////////////////////////////
	template < class U, typename P, typename A >
	TImageFor< T, PAT, AL > & operator = ( const TImageFor< U, P, A > & ref )
	{
		if( this != (TImageFor< T, PAT, AL > *) &ref )
		{
			T * ptr = fData;
			for( int i = 0; i < fElems; ++ i )
				fAllocator.destroy( ptr ++ );

			fAllocator.deallocate( fData, fElems );
			
			
			fRow = ref.GetRow();
			fCol = ref.GetCol();

			fElems = fRow * fCol;
		
			fData = fAllocator.allocate( fElems, (void*)0 );

			REQUIRE( fData != 0 );		// post-condition

			T * source_ptr = ref.fData;
			T * destin_ptr_start = fData;
			T * destin_ptr_end = fData + fElems;
			while( destin_ptr_start < destin_ptr_end )
				fAllocator.construct( destin_ptr_start ++, * source_ptr ++ );
		
		
		}

		return *this;
	}

	///////////////////////////////////////////////////////////
	// An assignement operator
	///////////////////////////////////////////////////////////
	TImageFor< T, PAT, AL > & operator = ( const TImageFor< T, PAT, AL > & ref )
	{
		if( this != (TImageFor< T, PAT, AL > *) &ref )
		{
			T * ptr = fData;
			for( int i = 0; i < fElems; ++ i )
				fAllocator.destroy( ptr ++ );

			fAllocator.deallocate( fData, fElems );

			
			
			fRow = ref.GetRow();
			fCol = ref.GetCol();

			fElems = fRow * fCol;

			fData = fAllocator.allocate( fElems, (void*)0 );

			REQUIRE( fData != 0 );		// post-condition

			T * source_ptr = ref.fData;
			T * destin_ptr_start = fData;
			T * destin_ptr_end = fData + fElems;
			while( destin_ptr_start < destin_ptr_end )
				fAllocator.construct( destin_ptr_start ++, * source_ptr ++ );
		}

		return *this;
	}


	///////////////////////////////////////////////////////////
	// Class virtual destructor.
	///////////////////////////////////////////////////////////
	virtual ~TImageFor() 
	{
		T * ptr = fData;
		for( int i = 0; i < fElems; ++ i )
			fAllocator.destroy( ptr ++ );

		fAllocator.deallocate( fData, fElems );
	}
	//======================================

	///////////////////////////////////////////////////////////
	// Returns a pointer to the NEW image object which was built 
	// as a part of the "this" image. You are responsible for that 
	// object.
	// The new image is composed starting from (s_col,s_row) 
	// and extended to: width x height.
	//
	// INPUT:
	//			s_col - upper left coordinate of the new image (expressed
	//				in the mother coordinate system)
	//			s_row - top coordinate of the new image
	//			width - width of a new image
	//			height - height of a new image
	//			
	//
	// OUTPUT:
	//			pointer to the new image or 0 if creation not possible
	//			Responsibility of the object passed by this pointer is left
	//			to the adopting part.
	//
	TImageFor< T, PAT, AL > * OrphanBitBlit( Dimension s_col, Dimension s_row, Dimension width, Dimension height ) const
	{
		if( s_col >= fCol || s_row >= fRow || width == 0|| height == 0 )
			return 0;		// wrong input parameters

		// check whether the new window overlappes the whole image
		// - if not then exit 
		if( s_row + height > fRow || s_col + width > fCol )
			return 0;

		TImageFor< T, PAT, AL > * obj = new TImageFor< T, PAT, AL >( width, height );

		// copy data from the mother object
		register Dimension i, j;
		for( i = 0; i < height; i ++ )
		{
			for( j = 0; j < width; j ++ )
			{
				Dimension current_row = s_row + i;

				obj->SetPixel( j, i, GetPixel( s_col + j, current_row ) );
			}
		}

		return obj;
	}

	//======================================

	///////////////////////////////////////////////////////////
	// Overloaded binary operators.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			image - a constant reference to the second image
	//		
	// OUTPUT:
	//			result image (by reference or a local copy,
	//				what should be avoided)
	//		
	// REMARKS:
	//
	//

	// Returns true if the two pictures are the same
	bool				operator == ( const TImageFor< T, PAT, AL > & image ) const; 

	TImageFor< T, PAT, AL >		operator + ( const TImageFor< T, PAT, AL > & image ) const; 
	
	TImageFor< T, PAT, AL > &	operator += ( const TImageFor< T, PAT, AL > & image );   

	TImageFor< T, PAT, AL >		operator - ( const TImageFor< T, PAT, AL > & image ) const;   

	TImageFor< T, PAT, AL > &	operator -= ( const TImageFor< T, PAT, AL > & image );   

	TImageFor< T, PAT, AL >		operator * ( const TImageFor< T, PAT, AL > & image ) const;  
	
	TImageFor< T, PAT, AL > &	operator *= ( const TImageFor< T, PAT, AL > & image );
	
	TImageFor< T, PAT, AL >		operator / ( const TImageFor< T, PAT, AL > & image ) const;
	
	TImageFor< T, PAT, AL > &	operator /= ( const TImageFor< T, PAT, AL > & image );   


	//======================================

	TImageFor< T, PAT, AL >		operator | ( const TImageFor< T, PAT, AL > & image ) const;   

	TImageFor< T, PAT, AL > &	operator |= ( const TImageFor< T, PAT, AL > & image );   

	TImageFor< T, PAT, AL >		operator & ( const TImageFor< T, PAT, AL > & image ) const;   

	TImageFor< T, PAT, AL > &	operator &= ( const TImageFor< T, PAT, AL > & image );   

	TImageFor< T, PAT, AL >		operator ^ ( const TImageFor< T, PAT, AL > & image ) const;   

	TImageFor< T, PAT, AL > &	operator ^= ( const TImageFor< T, PAT, AL > & image );   

	TImageFor< T, PAT, AL > & 	operator ~ ( void );

	//======================================

 	TImageFor< T, PAT, AL >		operator +  ( const T Value ) const;
 	TImageFor< T, PAT, AL > &	operator += ( const T Value );
	
 	TImageFor< T, PAT, AL >		operator - ( const T Value ) const;
 	TImageFor< T, PAT, AL > &	operator -= ( const T Value );
	
	TImageFor< T, PAT, AL >		operator * ( const T value ) const;   
	TImageFor< T, PAT, AL > &	operator *= ( const T value );   

	TImageFor< T, PAT, AL >		operator /  ( const T Value ) const;
	TImageFor< T, PAT, AL > &	operator /=  ( const T Value );
	//======================================

 	TImageFor< T, PAT, AL >		operator | ( const T Value ) const;
 	TImageFor< T, PAT, AL > &	operator |= ( const T Value );
	
 	TImageFor< T, PAT, AL >		operator & ( const T Value ) const;
 	TImageFor< T, PAT, AL > &	operator &= ( const T Value );

	TImageFor< T, PAT, AL >		operator ^  ( const T Value ) const;
	TImageFor< T, PAT, AL > &	operator ^=  ( const T Value );

	//======================================

	///////////////////////////////////////////////////////////
	// This function sets a pixel at position (x,y)or (col,row) 
	// of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//			value - a value to be set at pixel position
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//			From the OOP point of view this function should be virtual.
	//			However, to avoid run-time panalty it is not virtual,
	//			and the "value" parameter is passed by value. In derived
	//			classes this strategy can be overloaded.
	//		
	void SetPixel( Dimension xPixPosition, Dimension yPixPosition, /*const T*/ConstPixelAccessType value ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		* ( fData + yPixPosition * fCol + xPixPosition ) = value;
	}

	///////////////////////////////////////////////////////////
	// This function sets a pixel at position (x,y)or (col,row) 
	// of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//			value - a value to be set at pixel position, passed by REFERENCE!
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//			From the OOP point of view this function should be virtual.
	//			However, to avoid run-time panalty it is not virtual.
	//		
	void SetRefPixel( Dimension xPixPosition, Dimension yPixPosition, const T & value ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		* ( fData + yPixPosition * fCol + xPixPosition ) = value;
	}

	///////////////////////////////////////////////////////////
	// This function gets a VALUE of a pixel at position  
	// (x,y) or (col,row) of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a copy of a pixel, of type T, from the given position
	//		
	// REMARKS:
	//			The xPixPosition should span from 0 to max_columns-1, while
	//			the yPixPosition from 0 to max_rows-1.
	//		
	//			From the OOP point of view this function should be virtual.
	//			However, to avoid run-time panalty it is not virtual.
	//
	PixelAccessType GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return * ( fData + yPixPosition * fCol + xPixPosition );
	}

	///////////////////////////////////////////////////////////
	// This function gets a REFERENCE to a pixel at position  
	// (x,y) or (col,row) of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a reference to a pixel, of type T, from the given position
	//		
	// REMARKS:
	//			The xPixPosition should span from 0 to max_columns-1, while
	//			the yPixPosition from 0 to max_rows-1.
	//		
	//			From the OOP point of view this function should be virtual.
	//			However, to avoid run-time panalty it is not virtual.
	//
	T & GetRefPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return * ( fData + yPixPosition * fCol + xPixPosition );
	}

	///////////////////////////////////////////////////////////
	// This function sets a pixel at position (x,y)or (col,row) 
	// of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//			value - a value to be set at pixel position
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//			All positive input values are allowed of a pixel position.
	//		
	void SetPixel_Modulo( Dimension xPixPosition, Dimension yPixPosition, ConstPixelAccessType value ) const
	{
		* ( fData + ( yPixPosition % fRow ) * fCol + ( xPixPosition % fCol ) ) = value;
	}

	///////////////////////////////////////////////////////////
	// This function gets a VALUE of a pixel at position  
	// (x,y) or (col,row) of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a copy of a pixel, of type T, from the given position
	//		
	// REMARKS:
	//			All positive input values are allowed of a pixel position.
	//
	PixelAccessType GetPixel_Modulo( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		return * ( fData + ( yPixPosition % fRow ) * fCol + ( xPixPosition % fCol ) );
	}

	///////////////////////////////////////////////////////////
	// This function gets a REFERENCE to a pixel at position  
	// (x,y) or (col,row) of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a reference to a pixel, of type T, from the given position
	//		
	// REMARKS:
	//			All positive input values are allowed of a pixel position.
	//
	T & GetRefPixel_Modulo( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		return * ( fData + ( yPixPosition % fRow ) * fCol + ( xPixPosition % fCol ) );
	}
	///////////////


	///////////////////////////////////////////////////////////
	// This function sets all pixels to a given value.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			value - a value to be set upon all pixels
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//			A more robust function can be defined for 
	//			base types which uses memset()
	//		
	void SetAll( ConstPixelAccessType value ) const
	{
		register Dimension i, j;
 		for( i = 0; i < fRow; i ++ )
			for( j = 0; j < fCol; j ++ )
				SetPixel( j, i, value );
	}

	//======================================

	///////////////////////////////////////////////////////////
	// This operator returns a reference to a pixel at position.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			index - a position of a pixel;
	//				it MUST be in range of allowable
	//				dimensions of this image
	//		
	// OUTPUT:
	//			a reference to the pointed pixel
	//		
	// REMARKS:
	//		
	//		
	T & operator [] ( Index index ) const
	{
		REQUIRE( index.fCol < fCol );
		REQUIRE( index.fRow < fRow );
		return * ( fData + index.fRow * fCol + index.fCol );
	}

	
	//==============================================================================



	Dimension GetRow( void ) const { return fRow; }
	Dimension GetCol( void ) const { return fCol; }

	T * GetImageData( void ) const { return fData; }

	Dimension GetElems( void ) const { return fElems; }

	//==============================================================================

	///////////////////////////////////////////////////////////
	// This function returns number of computer bytes devoted
	// to store all data
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			none
	//		
	// OUTPUT:
	//			number of bytes allocated ONLY for image data,
	//			and NOT for this object alone!
	//		
	// REMARKS:
	//		
	//		
	virtual unsigned long GetOccupiedMemorySizeInBytes( void ) const { REQUIRE( fElems == fRow * fCol ); return fElems * sizeof( T ); }		

};



//////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
// This function loads an image from a binary file
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - a reference to an image to be loaded
//			fileName - a name of an input file
//		
// OUTPUT:
//			true on success,
//			false otherwise
//		
// REMARKS:
//		
//		
template< typename T, typename PAT, typename AL >	
bool LoadImage( TImageFor< T, PAT, AL > & image, const char * fileName )		// load 2D picture	
{
	ifstream inputFile( fileName, ios::binary | ios::in );

	if( ! inputFile.is_open() || inputFile.fail() )
		return false;

	inputFile.read( ( char *)image.GetImageData(), image.GetOccupiedMemorySizeInBytes() );

	if( inputFile.eof() )
		return false;

	inputFile.close();		
	return true;
}

///////////////////////////////////////////////////////////
// This function saves an image to a binary file
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - a reference to an image to be saved
//			fileName - a name of an output file
//		
// OUTPUT:
//			true on success,
//			false otherwise
//		
// REMARKS:
//		
//		
template< typename T, typename PAT, typename AL >	
bool SaveImage( const TImageFor< T, PAT, AL > & image, const char * fileName )		// store 2D picture
{
	ofstream outputFile( fileName, ios::binary  );

	if( ! outputFile.is_open() || outputFile.fail() )
		return false;

	outputFile.write( ( char *)image.GetImageData(), image.GetOccupiedMemorySizeInBytes() );
	outputFile.close();		
	return true;
}


///////////////////////////////////////////////////////////
// This function loads an image from an ASCII file
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - a reference to an image to be loaded
//			fileName - a name of an input file
//		
// OUTPUT:
//			true on success,
//			false otherwise
//		
// REMARKS:
//		
//		
template< typename T, typename PAT, typename AL >	
bool Load_ASCII_Image( TImageFor< T, PAT, AL > & image, const char * fileName )		// load 2D picture	
{
	// Open in text mode (default if no ios::binary)
	ifstream inputFile( fileName/*, ios::nocreate | ios::in*/ );

	if( ! inputFile.is_open() || inputFile.fail() )
		return false;

//	if( inputFile.setmode( filebuf::text ) == -1 )
//		return false;

	T val; 
	const Dimension kCol = image.GetCol();
	const Dimension kRow = image.GetRow();
	register Dimension i, j;

	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			if( inputFile.eof() )
				return false;

			inputFile >> val;
			image.SetPixel( j, i, val );
		}
	}

	inputFile.close();		
	return true;
}

///////////////////////////////////////////////////////////
// This function saves an image to an ASCII file
///////////////////////////////////////////////////////////
//		
// INPUT:
//			image - a reference to an image to be saved
//			fileName - a name of an output file
//		
// OUTPUT:
//			true on success,
//			false otherwise
//		
// REMARKS:
//		
//		
template< typename T, typename PAT, typename AL >	
bool Save_ASCII_Image( const TImageFor< T, PAT, AL > & image, const char * fileName, const char * separator = " " )		// store 2D picture
{
	REQUIRE( image.GetImageData() != 0 );

	// Open in text mode (default if no ios::binary)
	ofstream outputFile( fileName, ios::out );

	if( outputFile.is_open() == false )
		return false;			// ofstream error

	register Dimension i, j;
	const Dimension kCol = image.GetCol();
	const Dimension kRow = image.GetRow();
	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
			outputFile << image.GetPixel( j, i ) << separator;

		outputFile << endl;		// separate rows
	}

	outputFile.close();		
	return true;
}


//////////////////////////////////////////////////////////////////////////

// Returns true if the two pictures are the same
template< typename T, typename PAT, typename AL >	
bool TImageFor< T, PAT, AL >::operator == ( const TImageFor< T, PAT, AL > & image ) const
{
	// if different dimensions, then for sure not equal
	if( fRow != image.fRow || fCol != image.fCol )
		return false;

	return memcmp( fData, image.fData, GetOccupiedMemorySizeInBytes() ) == 0;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator + ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage += image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator += ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	// Comparing two addresses:

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;

	register T * one_after_last_src_data_ptr = src_data_ptr + fElems;

	while( src_data_ptr < one_after_last_src_data_ptr )
		* dst_data_ptr ++ += * src_data_ptr ++;

	return *this;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator -= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ -= * src_data_ptr ++;

	return *this;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator - ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage -= image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator ~ ( void )
{
	register T * src_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
	{
		* src_data_ptr = ~ * src_data_ptr;
		++ src_data_ptr;
	}

	return * this;
}


template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator ^= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ ^= * src_data_ptr ++;

	return *this;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator ^ ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage ^= image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator &= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ &= * src_data_ptr ++;

	return *this;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator & ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage &= image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator |= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ |= * src_data_ptr ++;

	return *this;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator | ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage |= image;
	return outImage;
}

// Pointwise multiplication
template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator * ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage *= image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator *= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	// Comparing two addresses:

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;

	register T * one_after_last_src_data_ptr = src_data_ptr + fElems;

	while( src_data_ptr < one_after_last_src_data_ptr )
		* dst_data_ptr ++ *= * src_data_ptr ++;

	return *this;
}

// Pointwise division
template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator / ( const TImageFor< T, PAT, AL > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	TImageFor< T, PAT, AL >	outImage( * this );
	outImage /= image;
	return outImage;
}

template< typename T, typename PAT, typename AL >	
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator /= ( const TImageFor< T, PAT, AL > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );

	register T * src_data_ptr = image.GetImageData();
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ /= * src_data_ptr ++;

	return *this;
}


template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator + ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage += Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator += ( const T Value )
{
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ += Value;

	return * this;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator - ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage -= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator -= ( const T Value )
{
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ -= Value;

   return * this;
}

// -----------------------------------------------------------
template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator / ( const T Value ) const
{
	REQUIRE( Value != (T)0 );
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage /= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator /= ( const T Value )
{
	REQUIRE( Value != (T)0 );
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ /= Value;

	return * this;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator * ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage *= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator *= ( const T Value )
{
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ *= Value;

	return * this;
}

// -----------------------------------------------------------

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator | ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage |= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator |= ( const T Value )
{
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ |= Value;

	return * this;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator & ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage &= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator &= ( const T Value )
{
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ &= Value;

	return * this;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > TImageFor< T, PAT, AL >::operator ^ ( const T Value ) const
{
	TImageFor< T, PAT, AL >	outImage( * this );
	outImage ^= Value;
	return outImage;
}

template< typename T, typename PAT, typename AL >
TImageFor< T, PAT, AL > & TImageFor< T, PAT, AL >::operator ^= ( const T Value )
{
	register Dimension i, j;
	register T * dst_data_ptr = fData;
	register int data_num = fElems;

	while( data_num -- != 0 )
		* dst_data_ptr ++ ^= Value;

	return * this;
}

// -----------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////


typedef TImageFor< unsigned char >	MonochromeImage;
typedef TImageFor< double >			TRealImage;
typedef TImageFor< unsigned long >	ULongImage;
typedef TImageFor<			long >	LongImage;


//////////////////////////////////////////////////////////////////////////////////

// Template-image definitions:

typedef TImageFor< MonochromeImage >	Monochrome_Template_Image;
typedef TImageFor< TRealImage >			Real_Template_Image;
typedef TImageFor< ULongImage >			ULong_Template_Image;
typedef TImageFor< LongImage >			Long_Template_Image;



//////////////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////////////	
// This derived class is an image where pixels are vectors (possibly of different
// lenghts).
//////////////////////////////////////////////////////////////////////////////////	
template< typename T, typename PAT  = /*PixelAccess_Trait< T >*/PixelAccess_Trait< vector< T > * > >
class TDanglingImageFor : public TImageFor< vector< T > *, PAT >
{
public:
	
	typedef TImageFor< vector< T > *, PAT >	BaseClass;

	typedef vector< T >			InternalStorage;
	typedef InternalStorage	*	InternalStoragePtr;

public:

	//======================================

	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			col - number of columns
	//			row - number of rows
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//		
	//		
	TDanglingImageFor( Dimension col, Dimension row  )
		: BaseClass( col, row )
	{
 		register Dimension i, j;
 		for( i = 0; i < fRow; i ++ )
			for( j = 0; j < fCol; j ++ )
				BaseClass::SetPixel( j, i, new vector< T > );
	}

	///////////////////////////////////////////////////////////
	// Class constructor 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			index - number of columns and rows
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//			Memory for data is allocated but 
	//			data is NOT initialized.	
	//		
	TDanglingImageFor( const Index index ) 
		: BaseClass( index )
	{
 		register Dimension i, j;
 		for( i = 0; i < fRow; i ++ )
			for( j = 0; j < fCol; j ++ )
				BaseClass::SetPixel( j, i, new vector< T > );
	}

	///////////////////////////////////////////////////////////
	// Class virtual destructor.
	///////////////////////////////////////////////////////////
	virtual ~TDanglingImageFor() 
	{
 		register Dimension i, j;
 		for( i = 0; i < fRow; i ++ )
			for( j = 0; j < fCol; j ++ )
				delete BaseClass::GetPixel( j, i );
	}
	//======================================

public:


	///////////////////////////////////////////////////////////
	// This function sets a pixel at position (x,y)or (col,row) 
	// of this image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//			value - a value to be set at pixel position
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//		
	void SetPixel( Dimension xPixPosition, Dimension yPixPosition, typename BaseClass::PixelAccessType/*T*//*BaseClass::PixelAccessType*//*PixelAccess_Trait< T >::PixelAccessType*/ value ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		// Attach the value to the dangling storage at the given place
		BaseClass::GetPixel( xPixPosition, yPixPosition )->push_back( value );
	}

	///////////////////////////////////////////////////////////
	// This function gets a VALUE of a pixel at position  
	// (x,y) or (col,row) of this image which is a pointer
	// to a vector< T >
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a pixel at position (x,y) or (col,row) of this image 
	//			which is a pointer to a vector< T >
	//		
	// REMARKS:
	//
	InternalStoragePtr GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return BaseClass::GetPixel( xPixPosition, yPixPosition );
	}

	///////////////////////////////////////////////////////////
	// This function returns size of a pixel - vector< T >
	// at position.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			size of a pixel - vector< T > at position.
	//		
	// REMARKS:
	//	
	int GetPixelDepth( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return BaseClass::GetPixel( xPixPosition, yPixPosition )->size();
	}

	void SetAll( const T value ) const
	{
		// do nothing for the dangling image
	}
};


//////////////////////////////////////////////////////////////////////////////////

typedef TDanglingImageFor< double >			RealDanglingImage;
typedef TDanglingImageFor< unsigned char >	MonochromeDanglingImage;


//////////////////////////////////////////////////////////////////////////////////
// Helper functions


// There must be a conversion operator defined from V to T.
// Returns false if cannot copy.
template< typename T, typename V, typename PAT1, typename PAT2, typename AL1, typename AL2 >
bool MixedCopy( TImageFor< T, PAT1, AL1 > & dest, const TImageFor< V, PAT2, AL2 > & source ) 
{
	register T * dst_data_ptr = dest.GetImageData();
	register V * src_data_ptr = source.GetImageData();
	register int data_num = source.GetElems();

	if( data_num != dest.GetElems() )
		return false;	// different sizes, cannot copy

	while( data_num -- != 0 )
		* dst_data_ptr ++ = /*(T)*/* src_data_ptr ++;
	
	return true;		// data copied successfully
}


//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////////////////////////
// The intention of the "proxy" class is to create a patch in other image,
// behaving like an ordinary image, with local coordinate system, etc.,
// but actually using storage of an underlying (mother) image. 
//////////////////////////////////////////////////////////////////////////////////
template < typename T, typename PAT  = PixelAccess_Trait< T >, typename AL = allocator< T > >
class TProxyImageFor : /*private*/public TImageFor< T, PAT, AL >
{
public:

	typedef TImageFor< T, PAT, AL >	BaseClass;

protected:

	TImageFor< T, PAT, AL > & fMotherImage;		// a reference to the substrate image
										// this proxy is created on

	Dimension	fMotherCol;				// num of columns and rows in a mother image
	Dimension	fMotherRow;

	Dimension	fAnchorCol;				// col/row position in a mother image this
	Dimension	fAnchorRow;				// proxy is anchored to

public:

	///////////////////////////////////////////////////////////
	// Class constructor.
	// The new proxy (i.e. a "patch" in a mother image) is composed starting from (s_col,s_row) 
	// up to but NOT included(!) (e_col,e_row) !!!
	///////////////////////////////////////////////////////////
	//
	// INPUT:
	//			image - a reference to the mother (underlying) image
	//			s_col - upper left coordinate of the new proxy (expressed
	//				in the mother coordinate system)
	//			s_row - top coordinate of the new proxy
	//			width - width of a proxy image
	//			height - height of a proxy image
	//
	// OUTPUT:
	//
	TProxyImageFor( TImageFor< T, PAT, AL > & image, Dimension s_col, Dimension s_row, Dimension width, Dimension height ) 
		:		fMotherImage( image )
	{
		fMotherCol = image.GetCol();
		fMotherRow = image.GetRow();

		if( s_col >= fMotherCol || s_row >= fMotherRow || width == 0|| height == 0 )
			throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );


		fCol = width;
		fRow = height;

		fAnchorCol = s_col;
		fAnchorRow = s_row;

		fData = 0;		// the key issue is that this object doesn't store any image data !

		// check whether the new window overlappes the whole image
		// - if not then exit 
		if( s_row + fRow > fMotherRow || s_col + fCol > fMotherCol )
			throw T_Standard_HIL_Exception( T_Standard_HIL_Exception::kImageDimensionsMismatch );
	}

	///////////////////////////////////////////////////////////
	// Copy constructor.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			image - a const reference to the proxy object
	//		
	// OUTPUT:
	//			this data are set according to the supplied proxy
	//			but no image data is copied!
	//		
	// REMARKS:
	//			Warning!!! SHALLOW copy only !!!
	//		
	TProxyImageFor( const TProxyImageFor< T, PAT, AL > & image ) 
		: fMotherImage( image.fMotherImage )
	{
		fMotherCol = image.fMotherCol;
		fMotherRow = image.fMotherRow;

		fCol = image.fCol;
		fRow = image.fRow;

		fAnchorCol = image.fAnchorCol;
		fAnchorRow = image.fAnchorRow;

		fData = image.fData;		
	}

	///////////////////////////////////////////////////////////
	// Assignement operator.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			image - a const reference to the proxy object
	//		
	// OUTPUT:
	//			this data are set according to the supplied proxy
	//			but no image data is copied!
	//		
	// REMARKS:
	//			Warning!!! SHALLOW copy only !!!
	//		
	TProxyImageFor & operator = ( const TProxyImageFor< T, PAT, AL > & image )
	{
		if( this != &image )
		{
			fMotherImage = image.fMotherImage;
			
			fMotherCol = image.fMotherCol;
			fMotherRow = image.fMotherRow;

			fCol = image.fCol;
			fRow = image.fRow;

			fAnchorCol = image.fAnchorCol;
			fAnchorRow = image.fAnchorRow;

			fData = image.fData;		
		}

		return *this;
	}

	///////////////////////////////////////////////////////////
	// Kind of an assignement operator with a deep copy.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			image - a const reference to another image
	//				which must be of the same size as this 
	//				proxy image
	//		
	// OUTPUT:
	//			true on success
	//		
	// REMARKS:
	//		
	bool Assign( const TImageFor< T, PAT, AL > & image )
	{
		// Only an image with exactly the same size can be assigned
		// i.e. copied into the mother place denoted by this proxy 
		if( &fMotherImage == &image || image.GetCol() != fCol || image.GetRow() != fRow )
			return false;


		register int i, j;

		for( i = 0; i < fCol; i ++ )
			for( j = 0; j < fRow; j ++ )
				SetPixel( i, j, image.GetPixel( i, j ) );	// actually sets mother pixels

		return true;
	}

	///////////////////////////////////////////////////////////
	// This function sets a pixel at position (x,y)or (col,row) 
	// of the mother image. 
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//			value - a value to be set at pixel position
	//		
	// OUTPUT:
	//			none
	//		
	// REMARKS:
	//		
	void SetPixel( Dimension xPixPosition, Dimension yPixPosition, typename BaseClass::ConstPixelAccessType value ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		fMotherImage.SetPixel( fAnchorCol + xPixPosition, fAnchorRow + yPixPosition, value );
	}

	void SetRefPixel( Dimension xPixPosition, Dimension yPixPosition, const T & value ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		fMotherImage.SetRefPixel( fAnchorCol + xPixPosition, fAnchorRow + yPixPosition, value );
	}

	///////////////////////////////////////////////////////////
	// This function gets a VALUE of a pixel at position  
	// (x,y) or (col,row) of the mother image
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a pixel at position (x,y) or (col,row) of the mother
	//			image
	// REMARKS:
	//
	typename BaseClass::PixelAccessType GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return fMotherImage.GetPixel( fAnchorCol + xPixPosition, fAnchorRow + yPixPosition );
	}

	///////////////////////////////////////////////////////////
	// This function gets a reference of a pixel at position  
	// (x,y) or (col,row) of the mother image
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the horizontal (or column) position of a pixel
	//			yPixPosition - the vertical (or row) position of a pixel
	//		
	// OUTPUT:
	//			a reference to a pixel at position (x,y) or (col,row) of 
	//			the mother image
	// REMARKS:
	//
	T & GetRefPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );

		return fMotherImage.GetRefPixel( fAnchorCol + xPixPosition, fAnchorRow + yPixPosition );
	}

	///////////////////////////////////////////////////////////
	// This operator returns a reference to a pixel at position.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			index - a position of a pixel;
	//				it MUST be in range of allowable
	//				dimensions of this proxy image
	//		
	// OUTPUT:
	//			a reference to the pointed pixel
	//		
	// REMARKS:
	//		
	//		
	T & operator [] ( Index index ) const
	{
		return fMotherImage.GetRefPixel( fAnchorCol + index.fCol, fAnchorRow + index.fRow );
	}


	// ==============================================================================

	// IO helpers
	friend bool LoadImage( TProxyImageFor< T, PAT > & proxyImage, const char * fileName );									// load 2D picture	
	friend bool SaveImage( const TProxyImageFor< T, PAT > & image, const char * fileName );									// store 2D picture
	friend bool Load_ASCII_Image( TProxyImageFor< T, PAT > & proxyImage, const char * fileName );							// load 2D picture	
	friend bool Save_ASCII_Image( const TProxyImageFor< T, PAT > & image, const char * fileName, const char * separator );	// store 2D picture

};



template< typename T, typename PAT, typename AL >	
bool LoadImage( TProxyImageFor< T, PAT, AL > & proxyImage, const char * fileName )		// load 2D picture	
{
	TImageFor< T, PAT, AL > temp( proxyImage.fCol, proxyImage.fRow );	// create temporary image

	if( temp.LoadImage( fileName ) == false )	// and get data into it from a file
		return false;

	return proxyImage.Assign( temp );			// now assign temp to this (mother based) proxy
}

template< typename T, typename PAT, typename AL >	
bool SaveImage( const TProxyImageFor< T, PAT, AL > & image, const char * fileName )		// store 2D picture
{
	// Create a buffer for the save operation -- it is an image extracted from the mother image
	TImageFor< T, PAT, AL > * temp = image.fMotherImage.OrphanBitBlit( image.fAnchorCol, image.fAnchorRow, image.fCol, image.fRow );

	REQUIRE( temp != 0 );

	bool retVal = SaveImage( * temp, fileName );
	delete temp;
	return retVal;
}


template< typename T, typename PAT, typename AL >	
bool Load_ASCII_Image( TProxyImageFor< T, PAT, AL > & proxyImage, const char * fileName )		// load 2D picture	
{
	TImageFor< T, PAT, AL > temp( proxyImage.fCol, proxyImage.fRow );	// create temporary image

	if( temp.Load_ASCII_Image( fileName ) == false )	// and get data into it from a file
		return false;

	return proxyImage.Assign( temp );			// now assign temp to this (mother based) proxy
}

template< typename T, typename PAT, typename AL >	
bool Save_ASCII_Image( const TProxyImageFor< T, PAT, AL > & image, const char * fileName, const char * separator )		// store 2D picture
{
	// Create a buffer for the save operation -- it is an image extracted from the mother image
	TImageFor< T, PAT, AL > * temp = image.fMotherImage.OrphanBitBlit( image.fAnchorCol, image.fAnchorRow, image.fCol, image.fRow );

	REQUIRE( temp != 0 );

	bool retVal = Save_ASCII_Image( * temp, fileName, separator );
	delete temp;
	return retVal;
}


// There must be a conversion operator defined from V to T.
// Returns false if cannot copy.
template< typename T, typename V, typename PAT1, typename PAT2, typename AL1, typename AL2 >
bool MixedCopy( TProxyImageFor< T, PAT1, AL1 > & dest, const TProxyImageFor< V, PAT2, AL2 > & source ) 
{
	if( dest.GetRow() != source.GetRow() || dest.GetCol() != source.GetCol() )
		return false;	// different sizes, cannot copy

	// Copy data 
 	register Dimension i, j;
	register Dimension row = dest.GetRow();
	register Dimension col = dest.GetCol();
 	for( i = 0; i < row; i ++ )
		for( j = 0; j < col; j ++ )
			dest.SetPixel( j, i, (T)source.GetPixel( j, i ) );

	return true;		// data copied successfully
}

template< typename T, typename V, typename PAT1, typename PAT2, typename AL1, typename AL2 >
bool MixedCopy( TImageFor< T, PAT1, AL1 > & dest, const TProxyImageFor< V, PAT2, AL2 > & source ) 
{
	if( dest.GetRow() != source.GetRow() || dest.GetCol() != source.GetCol() )
		return false;	// different sizes, cannot copy

	// Copy data 
 	register Dimension i, j;
	register Dimension row = dest.GetRow();
	register Dimension col = dest.GetCol();
 	for( i = 0; i < row; i ++ )
		for( j = 0; j < col; j ++ )
			dest.SetPixel( j, i, (T)source.GetPixel( j, i ) );

	return true;		// data copied successfully
}

//////////////////////////////////////////////////////////////////////////////

typedef TProxyImageFor< double >		RealImageProxy;
typedef TProxyImageFor< unsigned char >	MonochromeImageProxy;

//////////////////////////////////////////////////////////////////////////////////




// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================



#endif BoxySoft_HIL_ImageFor



