
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: AccumulationTraits.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the allocators.	
//
// ==========================================================================


#ifndef BoxySoft_InPlaceAllocator
#define BoxySoft_InPlaceAllocator

#pragma warning(disable:4786)





// ==========================================================================

namespace PHIL		
{


// ==========================================================================


///////////////////////////////////////////////////////////
// This class implements a special allocator which 
// allocates a container into a given buffer of memory.
// This buffer has to be created beforehand and must
// have a capacity suitable for a given container.
///////////////////////////////////////////////////////////
template < typename U >
class InPlaceAllocatorFor
{
	public:

		// Useful definitions
		typedef U				value_type;
		typedef U *				pointer;
		typedef const U *		const_pointer;
		typedef U &				reference;
		typedef const U &		const_reference;
		typedef size_t			size_type;
		typedef ptrdiff_t		difference_type;

	private:


		pointer		fStartDataPtr;		// This points at the beginning of the external memory location, provided in the constructor

	public:

		pointer		GetDataPtr( void ) const { return fStartDataPtr; }

	public:

		pointer address( reference value ) const 
		{
			return & value;
		}

		const_pointer address( const_reference value ) const 
		{
			return & value;
		}

		// ============
		// Constructors
		InPlaceAllocatorFor( void ) throw() : fStartDataPtr( 0 ) {}
		InPlaceAllocatorFor( pointer & ptr ) throw() : fStartDataPtr( ptr ) {}

		// Template copy constructor
		template < class T >
			InPlaceAllocatorFor( const InPlaceAllocatorFor< T > & r ) throw() : fStartDataPtr( (pointer)r.GetDataPtr() ) {}	

		~InPlaceAllocatorFor() throw() {}	// We do nothing with external memory
		// ============

		// Returns max number of elements that can be allocated with this class
		size_type max_size () const throw() 
		{
			return -1;			// return what is left
		}

		////////////////////////////////////////////////////////
		// Allocation of "num" elements U without initialization
		pointer allocate( size_type num, const void * = 0 ) 
		{
			return fStartDataPtr;		// just return the pointer to the buffer
		}

		// Initialization of the elements in the alreade allocated storage,
		// given by a pointer "p", with a value "value"
		void construct( pointer p, const U & value ) 
		{
			new ( (void*)p ) U( value );		// ... use the placement new
		}

		////////////////////////////////////////////////////////
		// Destroy elements of initialized storage given by "p"
		void destroy( pointer p ) 
		{
			p->~U();		// call destructor at address "p"
		}

		// Deallocate storage "p" of the already deleted elements
		void deallocate( pointer p, size_type num ) 
		{
			// Do nothing here - it is up to an external owner of the buffer to take care of it
		}
		////////////////////////////////////////////////////////


		// Rebinder for other types - do we need it, anyway?
		template < typename T >
		struct rebind 
		{
			typedef InPlaceAllocatorFor< T > other;
		};

};



////////////////////////////////////////////////////////////////////////////////////
// These express the idea that specializations of this allocator are interchangeable
template < typename P, typename Q >
bool operator == ( const InPlaceAllocatorFor< P > &, const InPlaceAllocatorFor< Q > & ) throw() 
{
	return true;
}


template < typename P, typename Q >
bool operator != ( const InPlaceAllocatorFor< P > &, const InPlaceAllocatorFor< Q > & ) throw() 
{
	return false;
}
////////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////
// The purpose of this class is to create a solid block
// of memory for a compact video structure.
///////////////////////////////////////////////////////////
template < typename U, typename AL = allocator< U > >
class SaveBufferAllocator
{
	public:

		// Useful definitions
		typedef U				value_type;
		typedef U *				pointer;
		typedef const U *		const_pointer;
		typedef U &				reference;
		typedef const U &		const_reference;
		typedef size_t			size_type;
		typedef ptrdiff_t		difference_type;

		typedef AL							buffer_allocator_type;
		typedef InPlaceAllocatorFor< U >	InPlaceAllocatorFor_U;


	private:


		pointer				fStartDataPtr;			// This points at the beginning of the external memory location, provided in the constructor
		pointer				fCurrentDataPtr;		// Points at the first free position in the buffer

		unsigned long		fTotalBufSize;			// Total size of the buffer in bytes
		unsigned long		fChunkSize;				// Size of a chunk of memory in each allocation (the same in each step)

		unsigned long		fNumOfChunks;			// Number of elements

	public:

		pointer		GetDataPtr( void ) const { return fStartDataPtr; }

	public:

		pointer address( reference value ) const 
		{
			return & value;
		}

		const_pointer address( const_reference value ) const 
		{
			return & value;
		}

		// ============
		// Constructors

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			chunk_size - size of a single data chung (such
		//				as a single video frame) expressed in bytes
		//			num_of_chunks - number of chunks
		//			alloc - optional allocator for the whole block
		//				If omitted then deafult allocator (with new)
		//				is used
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		SaveBufferAllocator( unsigned long chunk_size, unsigned long num_of_chunks, const AL & alloc = AL() ) throw() 
			: fStartDataPtr( 0 ), fCurrentDataPtr( 0 ), fChunkSize( chunk_size ), fNumOfChunks( num_of_chunks ), fBufferAllocator( alloc ) {}
		
		~SaveBufferAllocator() throw() 
		{
			fBufferAllocator.deallocate( fStartDataPtr, fTotalBufSize );
		}

		// ============

	protected:

		// Protected default constructor
		SaveBufferAllocator( const AL & allocator = AL() ) throw();

		// Protected copy constructor
		SaveBufferAllocator( const SaveBufferAllocator< U, AL > & r ) throw();	

	protected:

		buffer_allocator_type	fBufferAllocator;
		
		///////////////////////////////////////////////////////////
		// This function creates a buffer for the whole structure
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			true if ok
		//		
		// REMARKS:
		//			Parameters are class members
		//		
		bool CreateBuffer( void )
		{
			fTotalBufSize = fChunkSize * fNumOfChunks;
			fStartDataPtr = fCurrentDataPtr = fBufferAllocator.allocate( fTotalBufSize, (void*)0 );

			return fStartDataPtr == 0 ? false : true;
		}

	public:

		///////////////////////////////////////////////////////////
		// This function returns an in-place allocator for 
		// subsequent chunk of memory from a solid memory
		// buffer. The returned allocator should be passed
		// to an object created in a given chunk of memory
		// in the block.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			InPlaceAllocatorFor_U with next memory
		//				ptr; In case it was not possible then
		//				an allocator with 0 pointer is returned
		//		
		// REMARKS:
		//		
		//		
		InPlaceAllocatorFor_U GetNext_InPlaceAllocator( void ) 
		{
			pointer kZeroPtr = 0;
			// Lazy evaluation - check if there is already a buffer, if not then create one
			if( fStartDataPtr == 0 )
				if( CreateBuffer() == false )
					return InPlaceAllocatorFor_U( kZeroPtr );		// error - cannot create the buffer


			unsigned long size_left = fStartDataPtr + fTotalBufSize - fCurrentDataPtr;
			if( size_left >= fChunkSize )
			{
				pointer retPtr = fCurrentDataPtr;
				fCurrentDataPtr += fChunkSize;

				return InPlaceAllocatorFor_U( retPtr );
			}
			else
			{
				return InPlaceAllocatorFor_U( kZeroPtr );
			}


		}


		///////////////////////////////////////////////////////////
		// This function returns an in-place allocator for 
		// a chunk given by "index".
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			index - 0 based index of a chunk of memory 
		//		
		// OUTPUT:
		//			InPlaceAllocatorFor_U with next memory
		//				ptr; In case it was not possible then
		//				an allocator with 0 pointer is returned
		//		
		// REMARKS:
		//		
		//		
		InPlaceAllocatorFor_U Get_InPlaceAllocatorAt( unsigned long index )
		{
			if( index < fNumOfChunks )
			{
				return InPlaceAllocatorFor_U( fStartDataPtr + index * fChunkSize );
			}
			else
			{
				pointer kZeroPtr = 0;
				return InPlaceAllocatorFor_U( kZeroPtr );
			}
		}

};






// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================


#endif BoxySoft_InPlaceAllocator




