
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_MaskedImage.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the TMaskedImageFor - a special kind of an image
//				with a binary mask which tells object from the background.
//
// ==========================================================================


#ifndef BoxySoft_MaskedImage
#define BoxySoft_MaskedImage


#include "HIL_ImageFor.h"
#include "HIL_BooleanImage.h"






// ==========================================================================

namespace PHIL		
{


// ==========================================================================



// The idea behind the "masked" images is to associate a binary image
// to the classic TImageFor< ? >. This twofold structure operates
// in concordance. State of the binary image defines whether the pixel
// at the same position, but in the main image, is masked or not.

template < class T >
class TMaskedImageFor : public TImageFor< T >
{

protected:

	BinaryImage * fTheMask;		// if pixel in this mask is "true" then we assume
								// that a corresponding pixel in the main image
								// is set (i.e. "is of interest")

public:

	BinaryImage & GetMask( void ) { REQUIRE( fTheMask != 0 ); return * fTheMask; }


	///////////////////////////////////////////////////////////
	// This function safely assigns a new binary mask to
	// this masked image.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			newBinaryMask - new binary mask to be copied;
	//				its dimensions MUST be the same as dimensions
	//				of the old mask and the image.
	//		
	// OUTPUT:
	//			true - if assignement successful;
	//			false - otherwise (e.g. wrong dimensions)
	//		
	// REMARKS:
	//		
	//		
	bool AssignBinaryMask( const BinaryImage & newBinaryMask )
	{
		REQUIRE( fTheMask != 0 );		// since the mask should be always present
		REQUIRE( fCol == fTheMask->GetCol() &&  fRow == fTheMask->GetRow() );			// check compatibility of the old mask 
		REQUIRE( fCol == newBinaryMask.GetCol() &&  fRow == newBinaryMask.GetRow() );	// check compatibility of the new mask 
		if( fCol != newBinaryMask.GetCol() ||  fRow != newBinaryMask.GetRow() )
			return false;

		// Dimensions ok, so copy the mask
		* fTheMask = newBinaryMask;

		return true;
	}

public:


	virtual ~TMaskedImageFor() 
	{
		delete fTheMask;
	}

	//======================================
	TMaskedImageFor( Dimension col, Dimension row  )
		: TImageFor< T >( col, row )
	{
		fTheMask = new BinaryImage( col, row, true );	// all pixels set 
	}

	TMaskedImageFor( Dimension col, Dimension row, T init_val, bool mask_val = true ) 
		: TImageFor< T >( col, row, init_val  ) 
	{
		fTheMask = new BinaryImage( col, row, mask_val );	// all pixels pre set 
	}

	TMaskedImageFor( const Index index ) 
		: TImageFor< T >( const Index index ) 
	{
		fTheMask = new BinaryImage( index, true );	// all pixels set 
	}

	TMaskedImageFor( const Index index, T init_val, bool mask_val = true ) 
		: TImageFor< T >( const Index index, T init_val ) 
	{
		fTheMask = new BinaryImage( index, mask_val );	// all pixels pre set 
	}


	// Template copy constructor (mixed copy)
	template < typename U >
	TMaskedImageFor( const TImageFor< U > & ref, bool mask_val = true )
		: TImageFor< T >( ref )
	{
		fTheMask = new BinaryImage( fCol, fRow, mask_val );	// all pixels set 
	}

	template < class U >
	TMaskedImageFor< T > & operator = ( const TImageFor< U > & ref )
	{
		if( this != (TMaskedImageFor< T > *) & ref )
		{
			TImageFor< T >::operator = ( ref );	// call the base =

			delete fTheMask;
			fTheMask = new BinaryImage( fCol, fRow, true );	// all pixels set 
		}

		REQUIRE( fTheMask != 0 );
		REQUIRE( fCol == fTheMask->GetCol() );
		REQUIRE( fRow == fTheMask->GetRow() );

		return *this;
	}


	// Template copy constructor (mixed copy)
	template < typename U >
	TMaskedImageFor( TMaskedImageFor< U > & ref )
		: TImageFor< T >( ref )
	{
		fTheMask = new BinaryImage( ref.GetMask() );
	}

	// Copy constructor
	TMaskedImageFor( TMaskedImageFor< T > & ref )
		: TImageFor< T >( ref )
	{
		fTheMask = new BinaryImage( ref.GetMask() );
	}
	
	template < class U >
	TMaskedImageFor< T > & operator = ( TMaskedImageFor< U > & ref )
	{
		if( this != (TImageFor< T > *) &ref )
		{
			TImageFor< T >::operator = ( ref );	// call the base =

			delete fTheMask;
			fTheMask = new BinaryImage( ref.GetMask() );
		}

		REQUIRE( fTheMask != 0 );
		REQUIRE( fCol == fTheMask->GetCol() );
		REQUIRE( fRow == fTheMask->GetRow() );

		return *this;
	}



	///////////////////////////////////////////////////////////
	// This function operates on the mask image exclusively.
	///////////////////////////////////////////////////////////
	//		
	// INPUT:
	//			xPixPosition - the column index 
	//			yPixPosition - the row index
	//			theMaskVal - mask value to be set
	//		
	// OUTPUT:
	//		
	//		
	// REMARKS:
	//		
	//		
	void SetMaskPixel( Dimension xPixPosition, Dimension yPixPosition, bool theMaskVal = true )
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );
		REQUIRE( fTheMask != 0 );

		fTheMask->SetPixel( xPixPosition, yPixPosition, theMaskVal );
	}

	bool GetMaskPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );
		REQUIRE( fTheMask != 0 );

		return fTheMask->GetPixel( xPixPosition, yPixPosition );
	}


	// The xPixPosition should span from 0 to max_columns, while
	// the yPixPosition from 0 to max_rows.
	void SetPixel( Dimension xPixPosition, Dimension yPixPosition, T value, bool theMaskVal = true )
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );
		REQUIRE( fTheMask != 0 );

		fTheMask->SetPixel( xPixPosition, yPixPosition, theMaskVal );
		TImageFor< T >::SetPixel( xPixPosition, yPixPosition, value );
	}

	// The xPixPosition should span from 0 to max_columns, while
	// the yPixPosition from 0 to max_rows.
	T GetPixel( Dimension xPixPosition, Dimension yPixPosition, bool & theMaskVal ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );
		REQUIRE( fTheMask != 0 );

		theMaskVal = fTheMask->GetPixel( xPixPosition, yPixPosition );
		return TImageFor< T >::GetPixel( xPixPosition, yPixPosition );
	}

	T GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		return TImageFor< T >::GetPixel( xPixPosition, yPixPosition );	// delegate an action to the base class
	}

	T & GetRefPixel( Dimension xPixPosition, Dimension yPixPosition, bool & theMaskVal ) const
	{
		REQUIRE( xPixPosition < fCol );
		REQUIRE( yPixPosition < fRow );
		REQUIRE( fTheMask != 0 );

		theMaskVal = fTheMask->GetPixel( xPixPosition, yPixPosition );
		return TImageFor< T >::GetRefPixel( xPixPosition, yPixPosition );
	}

	T & GetRefPixel( Dimension xPixPosition, Dimension yPixPosition ) const
	{
		return TImageFor< T >::GetRefPixel( xPixPosition, yPixPosition );	// delegate an action to the base class
	}

	void SetAll( T value, bool theMaskVal )
	{
		REQUIRE( fTheMask != 0 );
		TImageFor< T >::SetAll( value );
		fTheMask->SetAll( theMaskVal );
	}

	//======================================

	virtual T & operator [] ( Index index )
	{
		return TImageFor< T >::operator [] ( index );		// call the base class
	}

	
	//==============================================================================

	// Returns true if the two pictures are the same
	bool operator == ( TMaskedImageFor< T > & image ) const
	{
		return TImageFor< T >::operator == ( image ) && * fTheMask == image.GetMask();
	}


	//======================================

	// Returns a pointer to the NEW image object which was built 
	// as a part of the "this" image. You are responsible for that 
	// object.
	// The new image is composed starting from (s_col,s_row) 
	// and extended to: width x height.
	//
	// INPUT:
	//			s_col - upper left coordinate of the new image (expressed
	//				in the mother coordinate system)
	//			s_row - top coordinate of the new image
	//			width - width of a new image
	//			height - height of a new image
	//			
	//
	// OUTPUT:
	//			pointer to the new image or 0 if creation not possible
	//			Responsibility of the object passed by this pointer is left
	//			to the adopting part.
	//
	TMaskedImageFor< T > * OrphanBitBlit( Dimension s_col, Dimension s_row, Dimension width, Dimension height ) const
	{
		if( s_col >= fCol || s_row >= fRow || width == 0|| height == 0 )
			return 0;		// wrong input parameters

		// check whether the new window overlappes the whole image
		// - if not then exit 
		if( s_row + height > fRow || s_col + width > fCol )
			return 0;

		TMaskedImageFor< T > * obj = new TMaskedImageFor< T >( width, height );

		// copy data from the mother object
		register Dimension i, j;
		for( i = 0; i < height; i ++ )
		{
			for( j = 0; j < width; j ++ )
			{
				Dimension current_row = s_row + i;

				bool mask;
				T pixel = GetPixel( s_col + j, current_row, mask );
				obj->SetPixel( j, i, pixel, mask );
			}
		}

		return obj;
	}

	//======================================


};

////////////////////////////////////////////////////

typedef TMaskedImageFor< unsigned char >	MaskedMonochromeImage;
typedef TMaskedImageFor< Color_3x8_Pixel >	MaskedColorImage;
typedef TMaskedImageFor< double >			MaskedRealImage;
typedef TMaskedImageFor< bool >				MaskedBooleanImage;

////////////////////////////////////////////////////



// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================



#endif BoxySoft_MaskedImage




