
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: MultiChannelImageFor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the multi-channel images, e.g. color images with
//				non-interlaced pixels.
//
// ==========================================================================


#ifndef BoxySoft_HIL_MultiChannelImage
#define BoxySoft_HIL_MultiChannelImage


#include "HIL_ImageFor.h"
#include "HIL_MultiPixelFor.h"




// ==========================================================================

namespace PHIL	
{





// ==========================================================================


//////////////////////////////////////////////////////////////////////////////////////
// This type of images can be used for multi-channel images with non-interlaced pixels.
// Implemented as a template-image pattern (i.e. image with pixels being images)
//////////////////////////////////////////////////////////////////////////////////////
template< class T, int Channels = kDefaultChannels >
class TMultiChannelImageFor : public TImageFor< TImageFor< T > >
{
	public:

		// For each instantiation of the enclosing class (TMultiChannelImageFor)
		// there will be its own type of MMultiPixelFor.

		enum { kChannels = Channels };

		typedef MMultiPixelFor< T, Channels >	MultiPixel;

		typedef TImageFor< T >					SingleChannelImage;

		typedef TImageFor< TImageFor< T > >		BaseClass;

	public:

		//======================================

		TMultiChannelImageFor( Dimension col, Dimension row )
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			SingleChannelImage initializingImage( col, row );
			for( int i = 0; i < Channels; ++ i )
				BaseClass::SetRefPixel( i, 0, initializingImage );
		}

		TMultiChannelImageFor( Dimension col, Dimension row, const MultiPixel & init_val ) 
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			SingleChannelImage initializingImage( col, row );
			for( int i = 0; i < Channels; ++ i )
			{
				BaseClass::SetRefPixel( i, 0, initializingImage );
				GetRefPixel( i, 0 ).SetAll( init_val.fData[ i ] );
			}
		}




		TMultiChannelImageFor( const Index index ) 
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			SingleChannelImage initializingImage( index.col, index.row );
			for( int i = 0; i < Channels; ++ i )
				BaseClass::SetRefPixel( i, 0, initializingImage );
		}

		TMultiChannelImageFor( const Index index, const T init_val ) 
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			SingleChannelImage initializingImage( index.col, index.row );
			for( int i = 0; i < Channels; ++ i )
			{
				BaseClass::SetRefPixel( i, 0, initializingImage );
				GetRefPixel( i, 0 ).SetAll( init_val.fData[ i ] );
			}
		}



		// Template copy constructor (mixed copy).
		// Pixel types can be different, so we can copy them and re-cast,
		// however, number of channels must be the same.
		template < class U, int Channels >
		TMultiChannelImageFor( const TMultiChannelImageFor< U, Channels > & ref )
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			for( int i = 0; i < Channels; ++ i )
				BaseClass::SetRefPixel( i, 0, (BaseClass::PixelType)ref.GetRefPixel( i, 0 ) );	// cast to (BaseClass::PixelType)
		}

		TMultiChannelImageFor( const TMultiChannelImageFor< T, Channels > & ref )
			: BaseClass( Channels, 1 )	// it is a single row with "Channels" entries, each entry is a separate image by itself
		{
			for( int i = 0; i < Channels; ++ i )
				BaseClass::SetRefPixel( i, 0, ref.GetRefPixel( i, 0 ) );	
		}



		// Pixel types can be different, so we can copy them and re-cast,
		// however, number of channels must be the same.
		template < class U, int Channels >
		TMultiChannelImageFor< T, Channels > & operator = ( const TMultiChannelImageFor< U, Channels > & ref )
		{
			if( this != (TMultiChannelImageFor< T, Channels > *) &ref )
			{
				REQUIRE( fRow == 1 );
				REQUIRE( fCol == ref.GetCol() );

				delete [] fData;		// get rid of old data

				// However, the channels can be quite different, so we have to recreate them
				BaseClass::Create( ref.GetCol(), ref.GetRow() );

				for( int i = 0; i < Channels; ++ i )
					BaseClass::SetRefPixel( i, 0, (BaseClass::PixelType)ref.GetRefPixel( i, 0 ) );	// cast to (BaseClass::PixelType)
			}

			return *this;
		}

		TMultiChannelImageFor< T, Channels > & operator = ( const TMultiChannelImageFor< T, Channels > & ref )
		{
			if( this != (TMultiChannelImageFor< T, Channels > *) &ref )
			{
				REQUIRE( fRow == 1 );
				REQUIRE( fCol == ref.fCol );

				delete [] fData;		// get rid of old data

				// However, the channels can be quite different, so we have to recreate them
				BaseClass::Create( ref.fCol, ref.fRow );

				for( int i = 0; i < Channels; ++ i )
					BaseClass::SetRefPixel( i, 0, ref.GetRefPixel( i, 0 ) );
			}

			return *this;
		}


		virtual ~TMultiChannelImageFor() {}

	///////
	public:

		///////////////////////////////////////////////////////////
		// This function sets a multi-pixel "multPixel" 
		// at position (x,y)or (col,row) of this multi-channel image. 
		// The multi-pixel is conveyed by reference thus there
		// is no "reference version" (i.e. SetRefPixel) of this function.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			xPixPosition - the horizontal (or column) position of a pixel
		//			yPixPosition - the vertical (or row) position of a pixel
		//			multPixel - a reference to the multi-pixel value to be set at pixel position
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			There is a question of passing a mutli-pixel argument either by reference
		//			of by value. In the base implementation of TImageFor<> the latter was chosen,
		//			and the second function SetRefPixel also supplied.
		//			In this derived class a passing by reference was chosen already for SetPixel
		//			and SetRefPixel is not supplied (also because GetRefPixel is not possible here).
		//			However, this can be changed (e.g. by creating a derived version) after 
		//			proper PROFILING and measuring performance in both cases.
		//		
		void SetPixel( Dimension xPixPosition, Dimension yPixPosition, const MultiPixel & multPixel ) const
		{
			REQUIRE( fCol == Channels );
			REQUIRE( fRow == 1 );
			for( int i = 0; i < Channels; ++ i )
				GetRefPixel( i, 0 ).SetPixel( xPixPosition, yPixPosition, multPixel.fData[ i ] );
		}

		///////////////////////////////////////////////////////////
		// This function gets a multi-pixel "multPixel" 
		// at position (x,y)or (col,row) of this multi-channel image. 
		// The multi-pixel is conveyed by value (a copy is created).
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			xPixPosition - the horizontal (or column) position of a pixel
		//			yPixPosition - the vertical (or row) position of a pixel
		//		
		// OUTPUT:
		//			multPixel - a multi-pixel value from pixel position
		//		
		// REMARKS:
		//			The xPixPosition should span from 0 to max_columns-1, while
		//			the yPixPosition from 0 to max_rows-1.
		//		
		//			It was tough to implement a resonable version of the GetRefPixel,
		//			since there is no single object we can refer to :(
		//			MMultiPixelFor & GetRefPixel( Dimension xPixPosition, Dimension yPixPosition ) const
		//			See also comment on SetPixel()
		//
		MultiPixel GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
		{
			REQUIRE( fCol == Channels );
			REQUIRE( fRow == 1 );

			MultiPixel retMulPixel;
			for( int i = 0; i < Channels; ++ i )
				retMulPixel.fData[ i ] = GetRefPixel( i, 0 ).GetPixel( xPixPosition, yPixPosition );

			return retMulPixel;
		}


	///////
	public:

		///////////////////////////////////////////////////////////
		// Overloaded binary operators.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			image - a constant reference to the second
		//				multi-channel image
		//		
		// OUTPUT:
		//			result image (by reference or a local copy,
		//				what should be avoided)
		//		
		// REMARKS:
		//			Each operation is performed on each channel
		//			separately. For each channel, action is delegated
		//			to the base implementation of the corresponding
		//			operator.
		//
		//			Additional operators can be added based
		//			on the supplied ones, either changing this
		//			class, or (presumably better) by deriving a new class.
		//
		
		// Returns true if the two pictures are the same
		bool									operator == ( const TMultiChannelImageFor< T, Channels > & image ) const; 

		TMultiChannelImageFor< T, Channels >	operator + ( const TMultiChannelImageFor< T, Channels > & image ) const; 
		TMultiChannelImageFor< T, Channels > &	operator += ( const TMultiChannelImageFor< T, Channels > & image );   
		
		TMultiChannelImageFor< T, Channels >	operator - ( const TMultiChannelImageFor< T, Channels > & image ) const;   
		TMultiChannelImageFor< T, Channels > &	operator -= ( const TMultiChannelImageFor< T, Channels > & image );   

		TMultiChannelImageFor< T, Channels >	operator * ( const TMultiChannelImageFor< T, Channels > & image ) const;  
		TMultiChannelImageFor< T, Channels > &	operator *= ( const TMultiChannelImageFor< T, Channels > & image );
		
		TMultiChannelImageFor< T, Channels >	operator / ( const TMultiChannelImageFor< T, Channels > & image ) const;
		TMultiChannelImageFor< T, Channels > &	operator /= ( const TMultiChannelImageFor< T, Channels > & image );   


};




// Returns true if the two pictures are the same
template< class T, int Channels >	
bool TMultiChannelImageFor< T, Channels >::operator == ( const TMultiChannelImageFor< T, Channels > & image ) const
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		if( ! ( GetRefPixel( i, 0 ) == image.GetRefPixel( i, 0 ) ) )
			return false;
  
	return true;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > TMultiChannelImageFor< T, Channels >::operator + ( const TMultiChannelImageFor< T, Channels > & image ) const
{
	TMultiChannelImageFor< T, Channels >	outImage( * this );		// copy this object at once

	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	// Add each channel separately
	for( int i = 0; i < Channels; ++ i )
		outImage.GetRefPixel( i, 0 ) += image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return outImage;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > & TMultiChannelImageFor< T, Channels >::operator += ( const TMultiChannelImageFor< T, Channels > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	// Add each channel separately
	for( int i = 0; i < Channels; ++ i )
		GetRefPixel( i, 0 ) += image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return * this;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > TMultiChannelImageFor< T, Channels >::operator - ( const TMultiChannelImageFor< T, Channels > & image ) const
{
	TMultiChannelImageFor< T, Channels >	outImage( * this );		// copy this object at once

	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		outImage.GetRefPixel( i, 0 ) -= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return outImage;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > & TMultiChannelImageFor< T, Channels >::operator -= ( const TMultiChannelImageFor< T, Channels > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		GetRefPixel( i, 0 ) -= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return * this;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > TMultiChannelImageFor< T, Channels >::operator * ( const TMultiChannelImageFor< T, Channels > & image ) const
{
	TMultiChannelImageFor< T, Channels >	outImage( * this );		// copy this object at once

	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		outImage.GetRefPixel( i, 0 ) *= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return outImage;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > & TMultiChannelImageFor< T, Channels >::operator *= ( const TMultiChannelImageFor< T, Channels > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		GetRefPixel( i, 0 ) *= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return * this;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > TMultiChannelImageFor< T, Channels >::operator / ( const TMultiChannelImageFor< T, Channels > & image ) const
{
	TMultiChannelImageFor< T, Channels >	outImage( * this );		// copy this object at once

	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		outImage.GetRefPixel( i, 0 ) /= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return outImage;
}

template< class T, int Channels >	
TMultiChannelImageFor< T, Channels > & TMultiChannelImageFor< T, Channels >::operator /= ( const TMultiChannelImageFor< T, Channels > & image )
{
	REQUIRE( fCol == image.GetCol() );
	REQUIRE( fRow == image.GetRow() );
	for( int i = 0; i < Channels; ++ i )
		GetRefPixel( i, 0 ) /= image.GetRefPixel( i, 0 );	// we only delegate an action here to each channel separately

	return * this;
}


// ==========================================================================

typedef	TMultiChannelImageFor< unsigned char >	NI_Color_3x8_Image;		// non-interlaced color image

// ==========================================================================


}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_MultiChannelImage




