
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_MultiChannelMixedCopy.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the mixed-copy functions for multi-channel
//				images.
//
// ==========================================================================



#ifndef BoxySoft_HIL_MultiChannelMixedCopy
#define BoxySoft_HIL_MultiChannelMixedCopy



#include "HIL_MultiChannelImageFor.h"


//////////////////////////////////////////////////////////////////////////////////////////


namespace PHIL		
{




///////////////////////////////////////////////////////////
// Copy from non-interlaced multi-channel to
// the interlaced image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. incompatible dimensions
//				of images)
//		
// REMARKS:
//		
//		Dimensions of images must match.
//	
template< typename P, int Channels, typename T >
bool MixedCopy( TImageFor< MMultiPixelFor< P, Channels > > & dest, const TMultiChannelImageFor< T, Channels > & source ) 
{
	register int i, j, k;

	const int kRow = dest.GetRow();
	const int kCol = dest.GetCol();

	if( source.GetRow() != 1 || source.GetCol() != Channels )
		return false;			// wrong size of the "source" image

	if( source.GetRefPixel( 0, 0 ).GetRow() != kRow || source.GetRefPixel( 0, 0 ).GetCol() != kCol )
		return false;			// cannot copy these images because different size of channels

 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			register MMultiPixelFor< P, Channels > thePixel;
			for( k = 0; k < Channels; k ++ )
				thePixel.fData[ k ] = source.GetRefPixel( k, 0 ).GetPixel( j, i );

			dest.SetPixel( j, i, thePixel );
		}
	}

	return true;
}

///////////////////////////////////////////////////////////
// Copy to non-interlaced multi-channel image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. incompatible dimensions
//				of images)
//		
// REMARKS:
//		
//		Dimensions of images must match.
//	
template< typename P, int Channels, typename T >
bool MixedCopy( TMultiChannelImageFor< T, Channels > & dest, const TImageFor< MMultiPixelFor< P, Channels > > & source ) 
{
	register int i, j, k;

	const int kRow = source.GetRow();
	const int kCol = source.GetCol();

	if( dest.GetRow() != 1 || dest.GetCol() != Channels )
		return false;			// wrong size of the "source" image

	if( dest.GetRefPixel( 0, 0 ).GetRow() != kRow || dest.GetRefPixel( 0, 0 ).GetCol() != kCol )
		return false;			// cannot copy these images because different size of channels

 	for( i = 0; i < kRow; i ++ )
	{
		for( j = 0; j < kCol; j ++ )
		{
			MMultiPixelFor< P, Channels > thePixel = source.GetPixel( j, i );
			for( k = 0; k < Channels; k ++ )
				dest.GetRefPixel( k, 0 ).SetPixel( j, i, thePixel.fData[ k ] );
		}
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Simple colour to scalar conversions:


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// Just a helper (kDefaultChannels == 3)
template < class PIX >
PIX RGB_2_Scalar( const MMultiPixelFor< PIX, kDefaultChannels > & in )
{
	// Standard coefficients for conversion from the RGB space into Y channel
	return (PIX)( 0.299 * (double)R(in) + 0.587 * (double)G(in) + 0.114 * (double)B(in) );
}

///////////////////////////////////////////////////////////
// Copy to non-interlaced multi-channel image
///////////////////////////////////////////////////////////
//		
// INPUT:
//			dest - reference to the destination image object
//			source - const reference to the source image object
//		
// OUTPUT:
//			true - if operation successful
//			false - otherwise (e.g. incompatible dimensions
//				of images)
//		
// REMARKS:
//		
//		Dimensions of images must match.
//	
static bool MixedCopy( MonochromeImage & dest, const ColorImage & source ) 
{
	register unsigned char * dst_data_ptr = dest.GetImageData();
	register Color_3x8_Pixel * src_data_ptr = source.GetImageData();
	register int data_num = source.GetElems();

	if( data_num != dest.GetElems() )
		return false;	// different sizes, cannot copy

	while( data_num -- != 0 )
		* dst_data_ptr ++ = RGB_2_Scalar( * src_data_ptr ++ );
	
	return true;		// data copied successfully
}



//////////////////////////////////////////////////////////////////////////////////////////

}	// end of the PHIL namespace


//////////////////////////////////////////////////////////////////////////////////////////




#endif BoxySoft_HIL_MultiChannelMixedCopy




