
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_MultiPixelFor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				MMultiPixelFor template class for all types of array-like pixels.	
//
// ==========================================================================


#ifndef BoxySoft_HIL_MultiPixelFor
#define BoxySoft_HIL_MultiPixelFor



#include "ImageDeclarations.h"			// for REQUIRE only
#include "HIL_PixelAccessTraits.h"


// ==========================================================================

namespace PHIL	
{





// ==========================================================================

const int kDefaultChannels = 3;

// ==========================================================================

// Multi-pixel models a pixel with "Channels" values of a type "P".
// It is used for interlaced color operations. 
template< typename P, int Channels = kDefaultChannels >
class MMultiPixelFor
{
	public:

		enum { kChannels = Channels };
		
		typedef P	PixelClusterArray[ Channels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { memset( (void*)fData, 0, Channels * sizeof( P ) ); }

		// Constructor 
		explicit MMultiPixelFor( const P * data ) { memcpy( (void*)fData, (void*)data, Channels * sizeof( P ) ); }
		
		explicit MMultiPixelFor( const P singleVal ) { for( register int i = 0; i < Channels; ++ i ) fData[ i ] = singleVal; }

		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] = (P)r.fData[ i ];
		}

		MMultiPixelFor( const MMultiPixelFor< P, Channels > & r ) { memcpy( (void*)fData, (void*)r.fData, Channels * sizeof( P ) ); }


		// Assignement
		template< class R >
		MMultiPixelFor & operator = ( const MMultiPixelFor< R, kChannels > & r ) 
		{ 
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] = (P)r.fData[ i ];
			return * this;
		}

		MMultiPixelFor & operator = ( const MMultiPixelFor< P, Channels > & r ) 
		{ 
			if( this != & r )
				memcpy( (void*)fData, (void*)r.fData, Channels * sizeof( P ) );
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, Channels * sizeof( P ) ) == 0;
		}

		// Just to introduce some order for the STL containers (is it good enough?)
		//bool operator < ( const MMultiPixelFor & r ) const
		//{
		//	return fData[ 0 ] < r.fData[ 0 ];
		//}

		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] += r.fData[ i ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] -= r.fData[ i ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] *= r.fData[ i ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] &= r.fData[ i ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] |= r.fData[ i ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] = ~fData[ i ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const P val )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] += (P)val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const P val )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] -= (P)val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const P val )
		{
			for( register int i = 0; i < Channels; ++ i ) 
				fData[ i ] *= (P)val;
			return * this;
		}

		MMultiPixelFor operator + ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		P & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < Channels );
			return fData[ index ];
		}

};

// ==========================================================================

#if MODERN_COMPILER

// Partial template specialization - for all types and exactly 2 channels
template< typename P >
class MMultiPixelFor< P, 2 >
{
	public:

		enum { kChannels = 2 };
		
		typedef P PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const P * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; }
		
		explicit MMultiPixelFor( const P singleVal ) { fData[ 0 ] = fData[ 1 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const P x0, const P x1 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; }


		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (P)r.fData[ 0 ]; fData[ 1 ] = (P)r.fData[ 1 ]; 
		}

		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; }



		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( P ) ) == 0;
		}

		// Just to introduce some order for the STL containers (is it good enough?)
		//bool operator < ( const MMultiPixelFor & r ) const
		//{
		//	return fData[ 0 ] < r.fData[ 0 ];
		//}

		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] &= r.fData[ 0 ];
			fData[ 1 ] &= r.fData[ 1 ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] |= r.fData[ 0 ];
			fData[ 1 ] |= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			fData[ 0 ] = ~fData[ 0 ];
			fData[ 1 ] = ~fData[ 1 ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}
		/////////////////////////////////////////////////

		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const P val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const P val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const P val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		P & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < Channels );
			return fData[ index ];
		}

};


// Partial template specialization - for all types and exactly 3 channels
template< typename P >
class MMultiPixelFor< P, 3 >
{
	public:

		enum { kChannels = 3 };
		
		typedef P PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; fData[ 2 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const P * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; fData[ 2 ] = * ++ data; }
		
		explicit MMultiPixelFor( const P singleVal ) { fData[ 0 ] = fData[ 1 ] = fData[ 2 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const P x0, const P x1, const P x2 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; fData[ 2 ] = x2; }


		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (P)r.fData[ 0 ]; fData[ 1 ] = (P)r.fData[ 1 ]; fData[ 2 ] = (P)r.fData[ 2 ];
		}

		//template<>		// let's make a specialization for its own type
		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; fData[ 2 ] = r.fData[ 2 ]; }



		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ], fData[ 2 ] = r.fData[ 2 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( P ) ) == 0;
		}

		// Just to introduce some order for the STL containers (is it good enough?)
		//bool operator < ( const MMultiPixelFor & r ) const
		//{
		//	return fData[ 0 ] < r.fData[ 0 ];
		//}

		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			fData[ 2 ] += r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			fData[ 2 ] -= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			fData[ 2 ] *= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] &= r.fData[ 0 ];
			fData[ 1 ] &= r.fData[ 1 ];
			fData[ 2 ] &= r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] |= r.fData[ 0 ];
			fData[ 1 ] |= r.fData[ 1 ];
			fData[ 2 ] |= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			fData[ 0 ] = ~fData[ 0 ];
			fData[ 1 ] = ~fData[ 1 ];
			fData[ 2 ] = ~fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}
		/////////////////////////////////////////////////
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const P val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			fData[ 2 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const P val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			fData[ 2 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const P val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			fData[ 2 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const P val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		P & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < Channels );
			return fData[ index ];
		}

};


#else // MODERN_COMPILER

// Unfortunately for older compilers (lack of partial-partial specializations) we have to cite all versions separately...
// The specializations are as follows:

// < int, 2 >
// < double, 2 >

// < unsigned char, 3 >
// < int, 3 >
// < double, 3 >

////////////////////////////////////////////
// For 2 members

template<>
class MMultiPixelFor< int, 2 >
{
	public:

		enum { kChannels = 2 };
		
		typedef int PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const int * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; }
		
		explicit MMultiPixelFor( const int singleVal ) { fData[ 0 ] = fData[ 1 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const int x0, const int x1 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; }


		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (int)r.fData[ 0 ]; fData[ 1 ] = (int)r.fData[ 1 ]; 
		}

		//template<>		// let's make a specialization for its own type
		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; }



		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( int ) ) == 0;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] &= r.fData[ 0 ];
			fData[ 1 ] &= r.fData[ 1 ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] |= r.fData[ 0 ];
			fData[ 1 ] |= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			fData[ 0 ] = ~fData[ 0 ];
			fData[ 1 ] = ~fData[ 1 ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}
		/////////////////////////////////////////////////
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const int val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const int val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const int val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		int & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < kChannels );
			return fData[ index ];
		}

};

template<>
class MMultiPixelFor< double, 2 >
{
	public:

		enum { kChannels = 2 };
		
		typedef double PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const double * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; }
		
		explicit MMultiPixelFor( const double singleVal ) { fData[ 0 ] = fData[ 1 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const double x0, const double x1 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; }


		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (double)r.fData[ 0 ]; fData[ 1 ] = (double)r.fData[ 1 ]; 
		}

		//template<>		// let's make a specialization for its own type
		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; }



		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const 
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( double ) ) == 0;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const double val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const double val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const double val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		double & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < kChannels );
			return fData[ index ];
		}

};



////////////////////////////////////////////
// For 3 members
template<>
class MMultiPixelFor< unsigned char, 3 >
{
	public:

		enum { kChannels = 3 };
		
		typedef unsigned char PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; fData[ 2 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const unsigned char * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; fData[ 2 ] = * ++ data; }
		
		explicit MMultiPixelFor( const unsigned char singleVal ) { fData[ 0 ] = fData[ 1 ] = fData[ 2 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const unsigned char x0, const unsigned char x1, const unsigned char x2 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; fData[ 2 ] = x2; }

		
		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (unsigned char)r.fData[ 0 ];
			fData[ 1 ] = (unsigned char)r.fData[ 1 ];
			fData[ 2 ] = (unsigned char)r.fData[ 2 ];
		}

		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; fData[ 2 ] = r.fData[ 2 ]; }

		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ], fData[ 2 ] = r.fData[ 2 ];
			return * this;
		}


	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( unsigned char ) ) == 0;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			fData[ 2 ] += r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			fData[ 2 ] -= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			fData[ 2 ] *= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] &= r.fData[ 0 ];
			fData[ 1 ] &= r.fData[ 1 ];
			fData[ 2 ] &= r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] |= r.fData[ 0 ];
			fData[ 1 ] |= r.fData[ 1 ];
			fData[ 2 ] |= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			fData[ 0 ] = ~fData[ 0 ];
			fData[ 1 ] = ~fData[ 1 ];
			fData[ 2 ] = ~fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}
		/////////////////////////////////////////////////
		
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const unsigned char val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			fData[ 2 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const unsigned char val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			fData[ 2 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const unsigned char val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			fData[ 2 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const unsigned char val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const unsigned char val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const unsigned char val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		unsigned char & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < kChannels );
			return fData[ index ];
		}

};

template<>
class MMultiPixelFor< int, 3 >
{
	public:

		enum { kChannels = 3 };
		
		typedef int PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; fData[ 2 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const int * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; fData[ 2 ] = * ++ data; }
		
		explicit MMultiPixelFor( const int singleVal ) { fData[ 0 ] = fData[ 1 ] = fData[ 2 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const int x0, const int x1, const int x2 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; fData[ 2 ] = x2; }

		
		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (int)r.fData[ 0 ];
			fData[ 1 ] = (int)r.fData[ 1 ];
			fData[ 2 ] = (int)r.fData[ 2 ];
		}


		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; fData[ 2 ] = r.fData[ 2 ]; }




		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ], fData[ 2 ] = r.fData[ 2 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( int ) ) == 0;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			fData[ 2 ] += r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			fData[ 2 ] -= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			fData[ 2 ] *= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}

		/////////////////////////////////////////////////
		// Logical operators
		MMultiPixelFor & operator &= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] &= r.fData[ 0 ];
			fData[ 1 ] &= r.fData[ 1 ];
			fData[ 2 ] &= r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator |= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] |= r.fData[ 0 ];
			fData[ 1 ] |= r.fData[ 1 ];
			fData[ 2 ] |= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator ~ ( void )
		{
			fData[ 0 ] = ~fData[ 0 ];
			fData[ 1 ] = ~fData[ 1 ];
			fData[ 2 ] = ~fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator & ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret &= r;
			return ret;
		}

		MMultiPixelFor operator | ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret |= r;
			return ret;
		}
		/////////////////////////////////////////////////
		
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const int val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			fData[ 2 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const int val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			fData[ 2 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const int val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			fData[ 2 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const int val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		int & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < kChannels );
			return fData[ index ];
		}

};

template<>
class MMultiPixelFor< double, 3 >
{
	public:

		enum { kChannels = 3 };
		
		typedef double PixelClusterArray[ kChannels ];

		PixelClusterArray fData;


	public:

		// A default constructor (all to 0)
		MMultiPixelFor( void ) { fData[ 0 ] = 0; fData[ 1 ] = 0; fData[ 2 ] = 0; }

		// Constructor 
		explicit MMultiPixelFor( const double * data ) { fData[ 0 ] = * data; fData[ 1 ] = * ++ data; fData[ 2 ] = * ++ data; }
		
		explicit MMultiPixelFor( const double singleVal ) { fData[ 0 ] = fData[ 1 ] = fData[ 2 ] = singleVal; }

		// Special constructor for exactly 3 channels
		MMultiPixelFor( const double x0, const double x1, const double x2 ) { fData[ 0 ] = x0; fData[ 1 ] = x1; fData[ 2 ] = x2; }

		
		// Copy constructors
		template< class R >
		MMultiPixelFor( const MMultiPixelFor< R, kChannels > & r )
		{ 
			fData[ 0 ] = (double)r.fData[ 0 ];
			fData[ 1 ] = (double)r.fData[ 1 ];
			fData[ 2 ] = (double)r.fData[ 2 ];
		}


		MMultiPixelFor( const MMultiPixelFor & r ) { fData[ 0 ] = r.fData[ 0 ]; fData[ 1 ] = r.fData[ 1 ]; fData[ 2 ] = r.fData[ 2 ]; }




		// Assignement
		MMultiPixelFor & operator = ( const MMultiPixelFor & r ) 
		{ 
			if( this != & r )
				fData[ 0 ] = r.fData[ 0 ], fData[ 1 ] = r.fData[ 1 ], fData[ 2 ] = r.fData[ 2 ];
			return * this;
		}

	public:

		bool operator == ( const MMultiPixelFor & r ) const
		{
			return memcmp( (void*)fData, (void*)r.fData, kChannels * sizeof( double ) ) == 0;
		}


		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const MMultiPixelFor & r )
		{
			fData[ 0 ] += r.fData[ 0 ];
			fData[ 1 ] += r.fData[ 1 ];
			fData[ 2 ] += r.fData[ 2 ];
			return * this;
		}


		MMultiPixelFor & operator -= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] -= r.fData[ 0 ];
			fData[ 1 ] -= r.fData[ 1 ];
			fData[ 2 ] -= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor & operator *= ( const MMultiPixelFor & r )
		{
			fData[ 0 ] *= r.fData[ 0 ];
			fData[ 1 ] *= r.fData[ 1 ];
			fData[ 2 ] *= r.fData[ 2 ];
			return * this;
		}

		MMultiPixelFor operator + ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret += r;
			return ret;
		}

		MMultiPixelFor operator - ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret -= r;
			return ret;
		}


		/////////////////////////////////////////////////
		
		// Point-wise multiplication
		MMultiPixelFor operator * ( const MMultiPixelFor & r ) const
		{
			MMultiPixelFor ret( * this );
			ret *= r;
			return ret;
		}
		/////////////////////////////////////////////////

		MMultiPixelFor & operator += ( const double val )
		{
			fData[ 0 ] += val;
			fData[ 1 ] += val;
			fData[ 2 ] += val;
			return * this;
		}

		MMultiPixelFor & operator -= ( const double val )
		{
			fData[ 0 ] -= val;
			fData[ 1 ] -= val;
			fData[ 2 ] -= val;
			return * this;
		}

		// Point-wise multiplication
		MMultiPixelFor & operator *= ( const double val )
		{
			fData[ 0 ] *= val;
			fData[ 1 ] *= val;
			fData[ 2 ] *= val;
			return * this;
		}

		MMultiPixelFor operator + ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret += val;
			return ret;
		}

		MMultiPixelFor operator - ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret -= val;
			return ret;
		}

		// Point-wise multiplication
		MMultiPixelFor operator * ( const double val ) const
		{
			MMultiPixelFor ret( * this );
			ret *= val;
			return ret;
		}

		// Overloaded subscript operator
		double & operator [] ( const int index )
		{
			REQUIRE( index >= 0 && index < kChannels );
			return fData[ index ];
		}

};


#endif // MODERN_COMPILER




// ==========================================================================

template< typename P, int Channels >
std::ostream & operator << ( std::ostream & o, const MMultiPixelFor< P, Channels > & f )
{
	for( register int i = 0; i < Channels; ++ i ) 
		o << f.fData[ i ] << ' ';
	o << endl;
	return o;
}

template< typename P, int Channels >
std::istream & operator >> ( std::istream & i, MMultiPixelFor< P, Channels > & f )
{
	for( register int c = 0; c < Channels; ++ c ) 
		i >> f.fData[ c ];
	return i;
}



// ==========================================================================

// This functor computes the inner product for multi-pixels.
template< class P, int Channels, class R, class PRODUCT = double >
class _InnerProduct
{
	public:

		PRODUCT operator () ( const MMultiPixelFor< P, Channels > & a, const MMultiPixelFor< R, Channels > & b ) 
		{
			PRODUCT retVal( 0 );		// init to zero
			for( register int i = 0; i < Channels; ++ i ) 
				retVal += (PRODUCT)a.fData[ i ] * (PRODUCT)b.fData[ i ];
			return retVal;
		}

};

// A helper function that returns the inner product always as "double".
// For other (maybe faster) versions - write your own function based on this one.
template< class P, int Channels, class R >
double InnerProduct( const MMultiPixelFor< P, Channels > & a, const MMultiPixelFor< R, Channels > & b )
{
	_InnerProduct< P, Channels, R > innerProduct;
	return innerProduct( a, b );
}


// ==========================================================================

typedef	MMultiPixelFor< unsigned char >			Color_3x8_Pixel;	
typedef	MMultiPixelFor< double >				Color_Real_Pixel;	

// ==========================================================================


// Specific traits are implemented as specializations.
// They are used by the TImageFor<> suite to access pixels
// of the types given in specifications

template <> class PixelAccess_Trait< Color_3x8_Pixel >
{
	public: 
		// For double access pixels by REFERENCE
		typedef			Color_3x8_Pixel &	PixelAccessType;
		typedef const	Color_3x8_Pixel &	ConstPixelAccessType;
};


template <> class PixelAccess_Trait< Color_Real_Pixel >
{
	public: 
		// For double access pixels by REFERENCE
		typedef			Color_Real_Pixel &	PixelAccessType;
		typedef const	Color_Real_Pixel &	ConstPixelAccessType;
};


// ==========================================================================

// interlaced color image ( pixel index 0 for RED, 1 for GREEN, 2 for BLUE )
typedef TImageFor< Color_3x8_Pixel >			ColorImage;	
typedef TImageFor< Color_Real_Pixel >			ColorRealValuedImage;	


// Helper macros

template< class P, int C >
inline P & R( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< class P, int C >
inline P R( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P & G( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P G( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P & B( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }
template< typename P, int C >
inline P B( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }


template< typename P, int C >
inline P & H( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P H( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P & S( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P S( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P & I( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }
template< typename P, int C >
inline P I( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }



template< typename P, int C >
inline P & Y( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P Y( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P & Cb( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P Cb( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P & Cr( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }
template< typename P, int C >
inline P Cr( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }


// ==========================================================================
// A point has almost the same semantics as a pixel, so maybe it is a good
// idea to think of them in the same way? If (in a future) we need some
// other semantics (e.g. exteriors, etc.) we can define a separate class
// for points (derived from the the MMultiPixelFor<> or containing a point
// object as a member?). However, at the other end is a run-time performance
// of such small, but usually many in action, objects. So, as for now, let us 
// stick to the idea of having pixels and points coming from the same template,
// and for sake of run-time performance, if necessary, instead of deriving
// some "fat" classes for points we can use some light helper functions
// or macros.


typedef MMultiPixelFor< int, 2 >	Int_2D_Point;
typedef MMultiPixelFor< double, 2 > Real_2D_Point;

// ==========================================================================

typedef MMultiPixelFor< int, 3 >	Int_3D_Point;
typedef MMultiPixelFor< double, 3 > Real_3D_Point;

// ==========================================================================
// Helper functions - instead of p.fData[0] we write x(p), etc.

template< typename P, int C >
inline P & x( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }
template< typename P, int C >
inline P x( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 0 ]; }

template< typename P, int C >
inline P & y( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }
template< typename P, int C >
inline P y( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 1 ]; }

template< typename P, int C >
inline P & z( MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }
template< typename P, int C >
inline P z( const MMultiPixelFor< P, C > & thePoint ) { return thePoint.fData[ 2 ]; }


// ==========================================================================
// Point containers:

typedef vector< Int_2D_Point >		Int_2D_Point_Vector;
typedef vector< Real_2D_Point >		Real_2D_Point_Vector;
typedef vector< Int_3D_Point >		Int_3D_Point_Vector;
typedef vector< Real_3D_Point >		Real_3D_Point_Vector;


// ==========================================================================
}	// end of the PHIL namespace
// ==========================================================================



#endif BoxySoft_HIL_MultiPixelFor




