
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: VideoFor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of the class for video stream processing.
//
// ==========================================================================



#ifndef BoxySoft_VideoFor
#define BoxySoft_VideoFor

#pragma warning(disable:4786)



#include "HIL_ImageFor.h"




// ==========================================================================

namespace PHIL	
{


// ==========================================================================



///////////////////////////////////////////////////////////
// This class implements an interface for video streams.
// A strict assumption is that the frames contain the
// same type of pixels. Dimensions of the frames can
// be different, though.
//
// For a default memory allocator the images are stored 
// rather as a list, i.e. they
// do not constitute a compact memory region.
// If other schemes are required the a proper 
// memory allocator should be passed.
//
///////////////////////////////////////////////////////////
template < typename PixType, typename PAT = PixelAccess_Trait< PixType >, typename AL = allocator< PixType > >
class TVideoFor
{
	public:

		typedef typename PixType												PixelType;
		typedef typename PixType *												PixelTypePtr;

		typedef typename TImageFor< PixType, PAT, AL >							FrameImage;
		typedef typename TImageFor< PixType, PAT, AL > *						FrameImagePtr;

		typedef typename TImageFor< PixType, PAT, AL >::PixelAccessType			FramePixelAccessType;
		typedef typename TImageFor< PixType, PAT, AL >::ConstPixelAccessType	FrameConstPixelAccessType;


	protected:


		typedef vector< FrameImagePtr >		VideoData;				// vector of pointers to frames

		VideoData							fVideoData;				// frame container


	public:

		// ===================================================
		TVideoFor( void );												// class default constructor
		virtual ~TVideoFor();											// class virtual destructor
		// ===================================================



		///////////////////////////////////////////////////////////
		// This function attaches a new and orphaned frame to the end
		// of the video object. The frame is owned, and in consequence
		// deleted, by this video object.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			newImage - pointer to the orphaned frame
		//		
		// OUTPUT:
		//			true if operation successful
		//			false otherwise
		//		
		// REMARKS:
		//		
		//		
		virtual bool AttachOrphanedFrame( FrameImagePtr newImage );


		///////////////////////////////////////////////////////////
		// This function inserts a new and orphaned frame at position
		// given by "index" to the collection. All other frames
		// are shifted.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			index - 0-based index 
		//			newImage - pointer to the orphaned frame
		//		
		// OUTPUT:
		//			true if operation successful
		//			false otherwise
		//		
		// REMARKS:
		//		
		//		
		virtual bool InsertOrphanedFrameAt( int index, FrameImagePtr newImage );


		///////////////////////////////////////////////////////////
		// This function removes and deletes frames from the video
		// stream in positions from "from" up to "to".
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			from - index to the first frame to be removed
		//			to - index to the last frame to be removed
		//		
		// OUTPUT:
		//			true if operation successful
		//			false otherwise
		//		
		// REMARKS:
		//			To remove a single frame "from" and "to"
		//			should be equal and point at a given frame.
		//		
		virtual bool RemoveFrames( int from, int to );


		///////////////////////////////////////////////////////////
		// This function returns a pointer to a frame at position
		// or 0 if the index is out of range.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			index - 0 based index of the frame to be returned
		//		
		// OUTPUT:
		//			ptr to the frame at index or
		//			0 if index is out of range
		//		
		// REMARKS:
		//		
		//		
		virtual FrameImagePtr GetFrameAt( int index );


		///////////////////////////////////////////////////////////
		// This function returns a pixel given its coordinates.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			cols - column position of a pixel
		//			rows - row position of a pixel
		//			frame - 0 based index of a frame in the 
		//				video stream from which a pixel is accessed
		//		
		// OUTPUT:
		//			pixel through defined access type (such as a value,
		//				or reference; defined by PAT)
		//		
		// REMARKS:
		//			This function is not save agains wrong indeces, so
		//			they should be checked before calling this function!
		//		
		FramePixelAccessType GetPixel( int cols, int rows, int frame );

		///////////////////////////////////////////////////////////
		// This function sets a value to a pixel.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			cols - column position of a pixel
		//			rows - row position of a pixel
		//			frame - 0 based index of a frame in the 
		//				video stream from which a pixel is accessed
		//			pixel - a value of a pixel, passed in accordance
		//				with a policy defined by the PAT template argument
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			This function is not save agains wrong indeces, so
		//			they should be checked before calling this function!
		//		
		void SetPixel( int cols, int rows, int frame, FrameConstPixelAccessType pixel );


		///////////////////////////////////////////////////////////
		// This function returns number of frames
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//		
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		int GetNumOfFrames( void ) const { return fVideoData.size(); }

};




// class constructor
template < typename PixType, typename PAT, typename AL >	
TVideoFor< PixType, PAT, AL >::TVideoFor( void )		
{
}


// class destructor
template < typename PixType, typename PAT, typename AL >	
TVideoFor< PixType, PAT, AL >::~TVideoFor()		
{
	for( int i = 0; i < fVideoData.size(); ++ i )
		delete fVideoData[ i ];
}



///////////////////////////////////////////////////////////
// This function attaches a new and orphaned frame to the end
// of the video object. The frame is owned, and in consequence
// deleted, by this video object.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			newImage - pointer to the orphaned frame
//		
// OUTPUT:
//			true if operation successful
//			false otherwise
//		
// REMARKS:
//		
//		
template < typename PixType, typename PAT, typename AL >	
bool TVideoFor< PixType, PAT, AL >::AttachOrphanedFrame( FrameImagePtr newImage )
{
	REQUIRE( newImage != 0 );
	REQUIRE( fVideoData.size() == 0 || ( fVideoData[ 0 ] != 0 && fVideoData[ 0 ]->GetCol() == newImage->GetCol() && fVideoData[ 0 ]->GetRow() == newImage->GetRow() ) );
	fVideoData.push_back( newImage );
	return true;
}


///////////////////////////////////////////////////////////
// This function inserts a new and orphaned frame at position
// given by "index" to the collection. All other frames
// are shifted.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			index - 0-based index 
//			newImage - pointer to the orphaned frame
//		
// OUTPUT:
//			true if operation successful
//			false otherwise
//		
// REMARKS:
//		
//		
template < typename PixType, typename PAT, typename AL >	
bool TVideoFor< PixType, PAT, AL >::InsertOrphanedFrameAt( int index, FrameImagePtr newImage )
{
	REQUIRE( newImage != 0 );
	REQUIRE( fVideoData.size() == 0 || ( fVideoData[ 0 ] != 0 && fVideoData[ 0 ]->GetCol() == newImage->GetCol() && fVideoData[ 0 ]->GetRow() == newImage->GetRow() ) );
	fVideoData.insert( fVideoData.begin() + index, newImage );	
	return true;
}


///////////////////////////////////////////////////////////
// This function removes and deletes frames from the video
// stream in positions from "from" up to "to".
///////////////////////////////////////////////////////////
//		
// INPUT:
//			from - index to the first frame to be removed
//			to - index to the last frame to be removed
//		
// OUTPUT:
//			true if operation successful
//			false otherwise
//		
// REMARKS:
//			To remove a single frame "from" and "to"
//			should be equal and point at a given frame.
//		
template < typename PixType, typename PAT, typename AL >	
bool TVideoFor< PixType, PAT, AL >::RemoveFrames( int from, int to )
{
	if( from > to || to >= fVideoData.size() )
		return false;					// wrong input range

	// Delete the frames
	for( int i = from; i <= to; ++ i )
		delete fVideoData[ i ];

	// Delete the pointers to the deleted frames
	fVideoData.erase( fVideoData.begin() + from, fVideoData.begin() + to + 1 );
	return true;
}


///////////////////////////////////////////////////////////
// This function returns a pointer to a frame at position
// or 0 if the index is out of range.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			index - 0 based index of the frame to be returned
//		
// OUTPUT:
//			ptr to the frame at index or
//			0 if index is out of range
//		
// REMARKS:
//		
//		
template < typename PixType, typename PAT, typename AL >	
typename TVideoFor< PixType, PAT, AL >::FrameImagePtr TVideoFor< PixType, PAT, AL >::GetFrameAt( int index )
{
	REQUIRE( index < fVideoData.size() );
	return fVideoData[ index ];
}


///////////////////////////////////////////////////////////
// This function returns a pixel given its coordinates.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			cols - column position of a pixel
//			rows - row position of a pixel
//			frame - 0 based index of a frame in the 
//				video stream from which a pixel is accessed
//		
// OUTPUT:
//			pixel through defined access type (such as a value,
//				or reference; defined by PAT)
//		
// REMARKS:
//			This function is not save agains wrong indeces, so
//			they should be checked before calling this function!
//		
template < typename PixType, typename PAT, typename AL >	
typename TVideoFor< PixType, PAT, AL >::FramePixelAccessType TVideoFor< PixType, PAT, AL >::GetPixel( int cols, int rows, int frame )
{
	REQUIRE( frame < fVideoData.size() );
	REQUIRE( fVideoData[ frame ] != 0 );
	return fVideoData[ frame ]->GetPixel( cols, rows );
}

///////////////////////////////////////////////////////////
// This function sets a value to a pixel.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			cols - column position of a pixel
//			rows - row position of a pixel
//			frame - 0 based index of a frame in the 
//				video stream from which a pixel is accessed
//			pixel - a value of a pixel, passed in accordance
//				with a policy defined by the PAT template argument
//		
// OUTPUT:
//			none
//		
// REMARKS:
//			This function is not save agains wrong indeces, so
//			they should be checked before calling this function!
//		
template < typename PixType, typename PAT, typename AL >	
void TVideoFor< PixType, PAT, AL >::SetPixel( int cols, int rows, int frame, FrameConstPixelAccessType pixel )
{
	REQUIRE( frame < fVideoData.size() );
	REQUIRE( fVideoData[ frame ] != 0 );
	fVideoData[ frame ]->SetPixel( cols, rows, pixel );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


typedef TVideoFor< MonochromeImage::PixelType >	MonochromeVideo;
typedef TVideoFor< ColorImage::PixelType >		ColorVideo;
typedef TVideoFor< TRealImage::PixelType >		RealVideo;


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper functions:


///////////////////////////////////////////////////////////
// This function creates a monochrome video object with 
// "frames" number of images of size cols x rows. 
// All pixels are set to 0.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			cols - number of columns in each frame
//			rows - number of rows in each frame
//			frames - number of frame
//		
// OUTPUT:
//		
//		
// REMARKS:
//			The standard memory allocator is applied here, 
//			i.e. it used new for each image, thus the images
//			not necessarily are in contiguous block of memory.
//		
inline MonochromeVideo * OrphanMonochromeVideo( int cols, int rows, int frames )
{
	MonochromeVideo * monoVid = new MonochromeVideo;

	for( int i = 0; i < frames; ++ i )
		monoVid->AttachOrphanedFrame( new MonochromeVideo::FrameImage( cols, rows, 0 ) );

	return monoVid;
}

///////////////////////////////////////////////////////////
// This function creates a monochrome video object with 
// "frames" number of images of size cols x rows. 
// All pixels are set to 0.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			cols - number of columns in each frame
//			rows - number of rows in each frame
//			frames - number of frame
//		
// OUTPUT:
//		
//		
// REMARKS:
//			The standard memory allocator is applied here, 
//			i.e. it used new for each image, thus the images
//			not necessarily are in contiguous block of memory.
//		
inline ColorVideo * OrphanColorVideo( int cols, int rows, int frames )
{
	ColorVideo * colorVid = new ColorVideo;

	for( int i = 0; i < frames; ++ i )
		colorVid->AttachOrphanedFrame( new ColorVideo::FrameImage( cols, rows, Color_3x8_Pixel( (unsigned char)0 ) ) );

	return colorVid;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

typedef std::auto_ptr< ColorVideo >				CVAP;		// for color video objects
typedef std::auto_ptr< MonochromeVideo >		MVAP;		// for scalar video objects


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================


#endif BoxySoft_VideoFor




