
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_ArithmeticOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Arithmetic operators for images.
//
// ==========================================================================


#ifndef BoxySoft_HIL_ArithmeticOperators
#define BoxySoft_HIL_ArithmeticOperators


#include "HIL_BaseOperations.h"


namespace PHIL		
{


// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// addition of two images
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Add_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Add_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Add_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which adds the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image addition for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 + fInImage2;	// each image is different object
				else
					fRetImage += fInImage1;		// ret and 2 are the same
			else
				fRetImage += fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};



// ==========================================================================


///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// subtraction of two images
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Sub_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Sub_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Sub_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which subtracts the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image subtraction for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 - fInImage2;	// each image is different object
				else									// cast to (InIm_Type1)
					fRetImage -= fInImage1;		// ret and 2 are the same
			else
				fRetImage -= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};


// ==========================================================================


///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// the Hadamard (pixel by pixel) image multiplication
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Mul_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Mul_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Mul_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which point-wise multiplies the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image multiplication for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 * fInImage2;	// each image is a different object
				else
					fRetImage *= fInImage1;		// ret and 2 are the same
			else
				fRetImage *= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};



// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// the Hadamard (pixel by pixel) image division.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Div_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Div_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Div_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which point-wise divides the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image divisiion for the given
		//			suite of input and output parameters.
		//
		//			WARNING: Can throw exception on divide by zero!!!
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 / fInImage2;	// each image is different object
				else
					fRetImage /= fInImage1;		// ret and 2 are the same
			else
				fRetImage /= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};

// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// image absolute value.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class Abs_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{


	public:

		///////////////////////////////////////////////////////////
		// Class constructor.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Abs_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			:	TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback )
				{}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Abs_OperationFor() {}


		///////////////////////////////////////////////////////////
		// The function operator which finds a pixel with maximal
		// value in an input image, as well as its position in 
		// that image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image;
		//		
		// REMARKS:
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImageRef != (void *)& fInImage1_Ref )
				fRetImageRef = fInImage1_Ref;	// if different then copy
			
			ImageAbsoluteValue( fRetImageRef );
			return & fRetImageRef;
		}

};



// ==========================================================================
// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Add_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Add( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Add_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Sub_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Sub( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Sub_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Mul_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Mul( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Mul_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Div_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Div( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Div_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1 >
Abs_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_Abs( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Abs_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}


// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Add_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Add_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Sub_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Sub_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Mul_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Mul_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Div_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Div_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Abs_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Abs_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}


// ==========================================================================


}	// end of the PHIL namespace

// ==========================================================================



#endif BoxySoft_HIL_ArithmeticOperators




