
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_BaseDefinitions.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A suite of basic definitions for image operators.
//
// ==========================================================================



#ifndef BoxySoft_HIL_BaseDefinitions
#define BoxySoft_HIL_BaseDefinitions



#include <memory>				// for auto_ptr
#include "HIL_Exceptions.h"
#include "FixedNumber.h"
#include "HIL_ImageForFixed.h"
#include "ThreadSecurity.h"




namespace PHIL		
{


// ==========================================================================

// This object is used as a default parameter. 
static TThreadSecurity kgThreadSecurity;

// ==========================================================================


///////////////////////////////////////////////////////////
// This is a root (base) class for the callback objects
// that can be (optionally) supplied to the image operator
// objects when it is created. The callback is called when
// the related image operation is finished.
//
// Derive your own class for more advanced callbacks.
//
///////////////////////////////////////////////////////////
class TOperationCompletionCallback
{
	public:

		virtual void * operator()( void ) 
		{
			return 0;
		}
};

// This object is used as a default parameter. 
static TOperationCompletionCallback kgOperationCompletionCallback;

// ==========================================================================


///////////////////////////////////////////////////////////
// This is a root (base) pure virtual class for ALL the image
// operations. All other operations are derived from this class.
//
// The general ideas for the image operations can be
// summarized as follows:
// 1.	The base class TImageOperation is a pure virtual
//		class, so all operators should be objects of the
//		derived classes.
// 2.	The base TImageOperation accepts and stores 
//		references to the two external objects:
//			- thread security object that in run-time
//			controls access to the processing resources;
//			- operation callback object which is called
//			upon completion of an operation.
// 3.	Each operation (i.e. the functional operator) should
//		be decorated with operation begin and operation end
//		sequences. To help this action an automatic variable 
//		of the inner MImageOperationRetinue class can
//		be defined at the beginning of each operator()
//		in a derived class. Then when this variable gets
//		out of scope of the operator() its destructor is
//		automatically invoked which, in turn, calls 
//		operator_end() member.
// 4.	Whereve it is possible all parameters are treated
//		as images. So, an image is a more ample notion
//		than a classical "visible" image. E.g. an image
//		can store in its pixels a value of a just found
//		maximum pixel in an another image, as well as 
//		x and y coordinates of that pixel, as its next
//		pixels. This is analogous to a matrix-processing
//		context where each value is treated as a matrix.
//
///////////////////////////////////////////////////////////
class TImageOperation
{
	protected:

		TThreadSecurity &				fResourceAccessPolicy;

		TOperationCompletionCallback &	fOperationCompletionCallback;



	public:

		///////////////////////////////////////////////////////////
		// Base class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TImageOperation( TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback ) : fResourceAccessPolicy( resourceAccessPolicy ), fOperationCompletionCallback( opCompCallback ) {}
	
		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		virtual ~TImageOperation() {}

	protected:

		///////////////////////////////////////////////////////////
		// This function should be called at the beginning
		// of each operator()
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			It calls resource access function of the 
		//			supplied thread security object
		//		
		virtual void operator_begin( void )
		{
			fResourceAccessPolicy.Enter_CriticalSection();		// enter the critical section
		}

		///////////////////////////////////////////////////////////
		// This function should be called at the end
		// of each operator()
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			It calls resource release function of the 
		//			supplied thread security object
		//			and the callback
		//		
		virtual void operator_end( void )
		{
			fResourceAccessPolicy.Exit_CriticalSection();		// ext the critical section 
			fOperationCompletionCallback();						// and spread the massage that opeartion is completed
		}

		//////////////////////////////////////////////////////////////////////////////////
		// A helper class...
		// Just create a local object of this class and the operator_begin() and
		// operator_end() will be called automatically due to auto object semantics of C++
		//////////////////////////////////////////////////////////////////////////////////
		class MImageOperationRetinue
		{
			private:

				TImageOperation &		fImageOperation;

				void * operator new ( size_t );		// make the operator new private to disable creation on the heap

			public:

				MImageOperationRetinue( TImageOperation & imOper ) : fImageOperation( imOper ) 
				{
					fImageOperation.operator_begin();	// resource acquisition is initialization
				}
				~MImageOperationRetinue()
				{
					fImageOperation.operator_end();		// destructor de-initializes automatically
				}
		};

		friend class MImageOperationRetinue;

	public:

		///////////////////////////////////////////////////////////
		// The function operator which - in a derived class -
		// defines an image operation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			user defined (in a derived class) void pointer
		//		
		// REMARKS:
		//			The necessary input parameters should be supplied
		//			to an appropriate CONSTRUCTOR of a derived class.
		//		
		virtual void * operator()( void ) = 0;

};

// ==========================================================================

typedef std::auto_ptr< TImageOperation >				ImageOperation_AutoPtr;
typedef std::auto_ptr< TOperationCompletionCallback >	OpCompletCallback_AutoPtr;

// ==========================================================================

}	// end of the PHIL namespace


// ==========================================================================

#endif BoxySoft_HIL_BaseDefinitions




