
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_BaseOperations.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A suite of base for image operators.
//
// ==========================================================================

#ifndef BoxySoft_HIL_BaseOperations
#define BoxySoft_HIL_BaseOperations


#include "HIL_BaseDefinitions.h"


namespace PHIL		
{




// ==========================================================================



///////////////////////////////////////////////////////////
//
// This is a base pure virtual class for all 
// UNARY IMAGE OPERATIONS, i.e. those that accept
// a single image and return a single image
// of any specific type, hoewver.
// This class is parameterized by two template
// types: for the return image and for the
// input image, respectively.
//
// It is derived from the base TImageOperation.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class TUnaryImageOperationFor : public TImageOperation
{
	public:

		enum { eNumOfParams = 1 };

		typedef RetIm_Type RetType;
		typedef InIm_Type1 InType_1;

	protected:

		RetType &				fRetImageRef;		// the reference to the output image
		const InType_1 const &	fInImage1_Ref;		// the constant reference to the constant input image
		

	public:

		///////////////////////////////////////////////////////////
		// Base class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TUnaryImageOperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: fRetImageRef( retImage ), fInImage1_Ref( inImage1 ), TImageOperation( resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~TUnaryImageOperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which - in a derived class -
		// defines an image operation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			user defined (in a derived class) void pointer
		//		
		// REMARKS:
		//			The input and output images should be already
		//			supplied to the class constructor.
		//		
		virtual void * operator()( void ) = 0;

};

// ==========================================================================

///////////////////////////////////////////////////////////
//
// This is a base pure virtual class for all 
// BINARY IMAGE OPERATIONS, i.e. those that accept
// two input images and return a single image
// of any specific type, hoewver.
// This class is parameterized by three template
// types: for the return image and for the
// two input images, respectively.
//
// It is derived from the base TImageOperation.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class TBinaryImageOperationFor : public TImageOperation
{
	public:

		enum { eNumOfParams = 2 };

		typedef RetIm_Type RetType;	
		typedef InIm_Type1 InType_1;
		typedef InIm_Type2 InType_2;

	protected:

		RetType &				fRetImage;		// the reference to the output image
		const InType_1 const &	fInImage1;		// the constant reference to the first constant input image
		const InType_2 const &	fInImage2;		// the constant reference to the second constant input image
		

	public:

		///////////////////////////////////////////////////////////
		// Base class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TBinaryImageOperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: fRetImage( retImage ), fInImage1( inImage1 ), fInImage2( inImage2 ), TImageOperation( resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~TBinaryImageOperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which - in a derived class -
		// defines an image operation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			user defined (in a derived class) void pointer
		//		
		// REMARKS:
		//			The input and output images should be already
		//			supplied to the class constructor.
		//		
		virtual void * operator()( void ) = 0;

};

// ==========================================================================

///////////////////////////////////////////////////////////
//
// This is a base pure virtual class for all 
// TEMPLATE-IMAGE IMAGE OPERATIONS, i.e. those that accept
// an input image and an input template-image,
// and return a single image of any specific type, hoewver.
// This class is parameterized by three template
// types: for the return image and for the
// input image and input template-image, respectively.
//
// It is derived from the base TImageOperation.
//
// Image templates are special images which for pixels have other images
// (kind of an image composites). They allow for unification
// of such image operations as convolutions, masks, windows, etc.
//
// Here we don't require that all images are of the same type or size,
// since image templates are images where pixels are also images.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class TImageTemplateOperationFor : public TImageOperation
{
	public:

		enum { eNumOfParams = 2 };

		typedef RetIm_Type RetType;
		typedef InIm_Type1 InType_1;
		typedef InIm_Type2 InType_2;

	protected:

		RetType &				fRetImage;			// the reference to the output image
		const InType_1 const &	fInImage1;			// the constant reference to the constant input image
		const InType_2 const &	fImageTemplate;		// the constant reference to the constant input template-image
		

	public:

		///////////////////////////////////////////////////////////
		// Base class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImageTemplate - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TImageTemplateOperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImageTemplate, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: fRetImage( retImage ), fInImage1( inImage1 ), fImageTemplate( inImageTemplate ), TImageOperation( resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~TImageTemplateOperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which - in a derived class -
		// defines an image operation.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			user defined (in a derived class) void pointer
		//		
		// REMARKS:
		//			The input and output images should be already
		//			supplied to the class constructor.
		//		
		virtual void * operator()( void ) = 0;


};

}	// end of the PHIL namespace

// ==========================================================================

#endif BoxySoft_HIL_BaseOperations




