
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_ComposedOperations.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the composite operations on images.
//
// ==========================================================================


#ifndef BoxySoft_HIL_ComposedOperations
#define BoxySoft_HIL_ComposedOperations


#include <list>
#include <vector>
#include "HIL_BaseOperations.h"


namespace PHIL		
{





// ==========================================================================

typedef vector< TImageOperation * >	DefaultContainerFor_Operation;
typedef list< TImageOperation * >	ListContainerFor_Operation;

// ==========================================================================


///////////////////////////////////////////////////////////
//
// This class implements a composite pattern.
// It allows for composition of many image operation
// into a complex image operation.
//
///////////////////////////////////////////////////////////
template < class Container = DefaultContainerFor_Operation >
class TComposedImageOperationFor : public TImageOperation
{

	protected:

		typedef Container ImageOperationContainer;
		
		ImageOperationContainer fImageOperationContainer;		// here we store the registered operations

	public:

		///////////////////////////////////////////////////////////
		// Class (default) constructor. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		TComposedImageOperationFor( TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
		  : TImageOperation( resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class (virtual) destructor
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			Disposes all registered (added) objects
		//			to this composite.
		//		
		virtual ~TComposedImageOperationFor() 
		{ 
			// delete all
			for( ImageOperationContainer::iterator it = fImageOperationContainer.begin(); it != fImageOperationContainer.end(); ++ it )
				delete * it;
		}

		///////////////////////////////////////////////////////////
		// This function calls function operator for each 
		// object added (with AddAdoptNewOperation) to this collection.
		// Order of execution is dependant on the semantics of
		// the container supplied as a template parameter (Container).
		// It is linear for vectors and lists.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			return value of the last operation
		//		
		// REMARKS:
		//			Execute all registered. Like for other image
		//			operations it is possible (but optional) to
		//			provide a resource access object and notify
		//			a callback object upon completion
		//		
		//		
		virtual void * operator()( void ) 
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			void * retVal = 0;
			for( ImageOperationContainer::iterator it = fImageOperationContainer.begin(); it != fImageOperationContainer.end(); ++ it )
			{
				REQUIRE( *it != 0 );
				retVal = (**it)();
			}
			return retVal;
		}

	public:


		///////////////////////////////////////////////////////////
		// This function adds new operation to the collection.
		// The added object is abandoned, i.e. it is left 
		// for exclusive use of the TComposedImageOperationFor
		// object.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			op - ptr to the TImageOperation object
		//				that is added to this collection
		//				The pointed object is ADOPTED by 
		//				this collection
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			The caller should "orphan" since
		//			this collection "adopts" the added
		//			operation. This object is AUTOMATICALLY
		//			destroyed by this collection.
		//		
		void AddAdoptNewOperation( TImageOperation * op ) 
		{ 
			fImageOperationContainer.push_back( op ); 
		}

		// Version with the auto pointer.
		void AddAdoptNewOperation( ImageOperation_AutoPtr op_AP ) 
		{ 
			fImageOperationContainer.push_back( op_AP.release() ); 
		}

};

// ==========================================================================

typedef TComposedImageOperationFor< DefaultContainerFor_Operation >	TVectorMultiImageOperation;
typedef TComposedImageOperationFor< ListContainerFor_Operation >	TListMultiImageOperation;


// ==========================================================================

}	// end of the PHIL namespace

// ==========================================================================



#endif BoxySoft_HIL_ComposedOperations




