
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: ConvolveOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				A suite of operators for image convolutions.
//
// ==========================================================================



#ifndef BoxySoft_HIL_ConvolveOperators
#define BoxySoft_HIL_ConvolveOperators



#include "HIL_BaseOperations.h"
#include "AccumulationTraits_ForFixed.h"
#include "GraphicLibrary.h"



namespace PHIL		
{





// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the two dimensional (full) convolution of 
// an image and template-image.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class _2D_Convolve_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). This
		//				is a 2D "mask" of the convolution.
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		_2D_Convolve_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~_2D_Convolve_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which 2D convolves an image with
		// the supplied mask (it is either a one row image or
		// a vector<>) and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//		
		virtual void * operator()( void )
		{
			REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			Convolve( fInImage1, fImageTemplate, fRetImage );				// action
			return & fRetImage;
		}
};

// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the one dimensional horizontal convolution of 
// an image and template-image.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class _Horz_Convolve_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). This
		//				is a 1D "mask" of the convolution (i.e.
		//				a one row image).
		//				This can be also a vector<> container.
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		_Horz_Convolve_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~_Horz_Convolve_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which horizontally convolves an image with
		// the supplied mask (it is either a one row image or
		// a vector<>) and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//		
		virtual void * operator()( void )
		{
			REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			Horz1DConvolve( fInImage1, fImageTemplate, fRetImage );
			return & fRetImage;
		}
};


// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the one dimensional vertical convolution of 
// an image and template-image.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class _Vert_Convolve_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). This
		//				is a 1D "mask" of the convolution (i.e.
		//				a one row image).
		//				This can be also a vector<> container.
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		_Vert_Convolve_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~_Vert_Convolve_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which vertically convolves an image with
		// the supplied mask (it is either a one row image or
		// a vector<>) and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//		
		virtual void * operator()( void )
		{
			REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			Vert1DConvolve( fInImage1, fImageTemplate, fRetImage );
			return & fRetImage;
		}
};

// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
_2D_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_2D_Convolve( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new _2D_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
_Horz_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Horz_Convolve( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new _Horz_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
_Vert_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Vert_Convolve( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new _Vert_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback );
}


// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr _2D_Convolve_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new _2D_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Horz_Convolve_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new _Horz_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Vert_Convolve_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new _Vert_Convolve_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) );
}



// ==========================================================================


}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_ConvolveOperators




