
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_FindValueInImageOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the operators for finding specific
//				values in images.
//
// ==========================================================================



#ifndef BoxySoft_HIL_FindValueInImageOperators
#define BoxySoft_HIL_FindValueInImageOperators


#include "HIL_BaseOperations.h"

namespace PHIL		
{





// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// it finds a pixel with a maximal value in an image,
// as well as a position of that pixel.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class FindMaxVal_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	protected:

		// We store index of an extremum but not its value since we don't know 
		// what is an actual type (and size!) of a pixel.

		Index fMaxValIndex;		// upon completion this member hold a position of a found pixel

	public:

		// Enums for return values
		enum { kRetRow_index = 0 };
		enum { kRet_val_index = 0, kRet_col_index = 1, kRet_row_index = 2 };

	public:

		///////////////////////////////////////////////////////////
		// This function returns a position of a just found pixel
		// with maximal value.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			index with the pixel coordinates
		//		
		// REMARKS:
		//		
		//		
		Index GetMaxValIndex( void ) const { return fMaxValIndex; }

		// Helpers
		Dimension				GetMaxVal_Column( void )	const { return fRetImageRef.GetPixel( kRet_col_index, kRetRow_index ); } 
		Dimension				GetMaxVal_Row( void )		const { return fRetImageRef.GetPixel( kRet_row_index, kRetRow_index ); } 

		typename RetIm_Type::PixelType	GetMaxVal( void )	const { return fRetImageRef.GetPixel( kRet_val_index, kRetRow_index ); } 



	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		FindMaxVal_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~FindMaxVal_OperationFor() {}


		///////////////////////////////////////////////////////////
		// The function operator which finds a pixel with maximal
		// value in an input image, as well as its position in 
		// that image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image;
		//			The output image has set result parameters as follows:
		//				   (c,r)
		//				At (kRet_val_index, kRetRow_index) = the maximal value
		//				At (kRet_col_index, kRetRow_index) = column (x coord.) of a pixel with max value
		//				At (kRet_row_index, kRetRow_index) = row (y coord.) of a pixel with max value
		//		
		// REMARKS:
		//			The return image should be
		//			able to receive three pixels: one
		//			with the found value and two with pixel
		//			position coordinates.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			REQUIRE( fRetImageRef.GetRow() >= 1 );		// so we have enough room for the return data
			REQUIRE( fRetImageRef.GetCol() >= 3 );		// so we have enough room for the return data
			//REQUIRE( (void*)&fRetImageRef != (void*)&fInImage1_Ref );	// otherwise, are you sure to obliterate the input image?
			fRetImageRef.SetPixel( kRet_val_index, kRetRow_index, GetImageMaxValue( fInImage1_Ref, fMaxValIndex ) );
			fRetImageRef.SetPixel( kRet_col_index, kRetRow_index, (RetIm_Type::PixelType)fMaxValIndex.fCol );
			fRetImageRef.SetPixel( kRet_row_index, kRetRow_index, (RetIm_Type::PixelType)fMaxValIndex.fRow );
			return & fRetImageRef;
		}

};


// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// it finds a pixel with a minimal value in an image,
// as well as a position of that pixel.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class FindMinVal_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	protected:

		// We store index of an extremum but not its value since we don't know 
		// what is an actual type (and size!) of a pixel.
		
		Index fMinValIndex; 	// upon completion this member hold a position of a found pixel

	public:

		// Enums for return values
		enum { kRetRow_index = 0 };
		enum { kRet_val_index = 0, kRet_col_index = 1, kRet_row_index = 2 };

	public:

		///////////////////////////////////////////////////////////
		// This function returns a position of a just found pixel
		// with minimal value.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			index with the pixel coordinates
		//		
		// REMARKS:
		//		
		//		
		Index GetMinValIndex( void ) const { return fMinValIndex; }

		// Helpers
		Dimension				GetMinVal_Column( void )	const { return fRetImageRef.GetPixel( kRet_col_index, kRetRow_index ); } 
		Dimension				GetMinVal_Row( void )		const { return fRetImageRef.GetPixel( kRet_row_index, kRetRow_index ); } 

		typename RetIm_Type::PixelType	GetMinVal( void )	const { return fRetImageRef.GetPixel( kRet_val_index, kRetRow_index ); } 

	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		FindMinVal_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~FindMinVal_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which finds a pixel with minimal
		// value in an input image, as well as its position in 
		// that image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//			The output image has set result parameters as follows:
		//				   (c,r)
		//				At (kRet_val_index, kRetRow_index) = the minimal value
		//				At (kRet_col_index, kRetRow_index) = column (x coord.) of a pixel with min value
		//				At (kRet_row_index, kRetRow_index) = row (y coord.) of a pixel with min value
		//		
		// REMARKS:
		//			The return image should be
		//			able to receive three pixels: one
		//			with the found value and two with pixel
		//			position coordinates.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			REQUIRE( fRetImageRef.GetCol() >= 3 );		// so we have enough room for the return data
			REQUIRE( fRetImageRef.GetRow() >= 1 );		// so we have enough room for the return data
			REQUIRE( (void*)&fRetImageRef != (void*)&fInImage1_Ref );	// otherwise, are you sure to obliterate the input iamge?
			fRetImageRef.SetPixel( kRet_val_index, kRetRow_index, GetImageMinValue( fInImage1_Ref, fMinValIndex ) );
			fRetImageRef.SetPixel( kRet_col_index, kRetRow_index, (RetIm_Type::PixelType)fMinValIndex.fCol );
			fRetImageRef.SetPixel( kRet_row_index, kRetRow_index, (RetIm_Type::PixelType)fMinValIndex.fRow );
			return & fRetImageRef;
		}

};

// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
FindMaxVal_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_FindMaxVal( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new FindMaxVal_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1 >
FindMinVal_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_FindMinVal( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new FindMinVal_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}


// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr FindMaxVal_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new FindMaxVal_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr FindMinVal_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new FindMinVal_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}



// ==========================================================================
}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_FindValueInImageOperators




