
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_LogicalOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definitions of the logical operations on images.
//
// ==========================================================================


#ifndef BoxySoft_HIL_LogicalOperators
#define BoxySoft_HIL_LogicalOperators

#include "HIL_BaseOperations.h"

namespace PHIL		
{





// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// the (pixel by pixel) logical AND on images.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class AND_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		AND_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~AND_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which point-wise logical AND the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image AND for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 & fInImage2;	// each image is different object
				else
					fRetImage &= fInImage1;		// ret and 2 are the same
			else
				fRetImage &= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};



// ==========================================================================


///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// the (pixel by pixel) logical OR on images.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class OR_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		OR_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~OR_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which point-wise logical OR the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image OR for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 | fInImage2;	// each image is different object
				else
					fRetImage |= fInImage1;		// ret and 2 are the same
			else
				fRetImage |= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};



// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements binary image operation:
// the (pixel by pixel) logical XOR on images.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class XOR_OperationFor : public TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inImage2 - constant reference to the second constant
		//				input image of type InIm_Type2 (specified  
		//				by the third template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		XOR_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TBinaryImageOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~XOR_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which point-wise logical XOR the two images
		// and puts a result to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			The function tries to pick the most efficient
		//			implementation of the image XOR for the given
		//			suite of input and output parameters.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImage != (void *)& fInImage1 )
				if( (void *)& fRetImage != (void *)& fInImage2 )
					fRetImage = fInImage1 * fInImage2;	// each image is a different object
				else
					fRetImage *= fInImage1;		// ret and 2 are the same
			else
				fRetImage *= fInImage2;			// ret and 1 are the same

			return & fRetImage;
		}

};

// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// bit negation NOT.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class NOT_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{

	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		NOT_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~NOT_OperationFor() {}


		///////////////////////////////////////////////////////////
		// The function negates bits of an input image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//		
		// REMARKS:
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			// Let's choose the optimal version
			if( (void *)& fRetImageRef != (void *)& fInImage1_Ref )
				fRetImageRef = fInImage1_Ref;	// if different then copy

			return & ~fRetImageRef;				// negate and return
		}

};

// ==========================================================================


// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
AND_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_AND( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new AND_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
OR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_OR( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new OR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
XOR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_XOR( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new XOR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1 >
NOT_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_NOT( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new NOT_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}

// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr AND_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new AND_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr OR_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new OR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr XOR_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new XOR_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr NOT_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new NOT_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}

// ==========================================================================

}	// end of the PHIL namespace

// ==========================================================================



#endif BoxySoft_HIL_LogicalOperators




