
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_NonParametricConvertOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of operators for non-parametric image conversions.
//
// ==========================================================================



#ifndef BoxySoft_HIL_NonParametricConvertOperators
#define BoxySoft_HIL_NonParametricConvertOperators


#include "HIL_BaseOperations.h"
#include "NonParametricTransformations.h"



namespace PHIL		
{





// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation for non
// parametric conversion of input images. The two commonly
// known non-parametric transformations are: Census and Rank.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class Census3x3_Convert_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). The input
		//				image will be converted-and-copied
		//				to this image.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//
		Census3x3_Convert_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Census3x3_Convert_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator converts and copies an input
		// image into an output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			There must be an appropriate conversion
		//			procedures defines for a given pair of images.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			Census_3x3 theNonParamProcessor( fInImage1_Ref );
			theNonParamProcessor.Convert( fRetImageRef );

			return & fRetImageRef;
		}

};


template< typename RetIm_Type, typename InIm_Type1 >
class Census5x5_Convert_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). The input
		//				image will be converted-and-copied
		//				to this image.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//
		Census5x5_Convert_OperationFor( RetType & retImage, const InType_1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Census5x5_Convert_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator converts and copies an input
		// image into an output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			There must be an appropriate conversion
		//			procedures defines for a given pair of images.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			Census_5x5 theNonParamProcessor( fInImage1_Ref );
			theNonParamProcessor.Convert( fRetImageRef );

			return & fRetImageRef;
		}

};



// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
Census3x3_Convert_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_Census3x3_Converter( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Census3x3_Convert_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}

// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Census3x3_Converter_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Census3x3_Convert_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}

// ==========================================================================

template< typename RetIm_Type, typename InIm_Type1 >
Census5x5_Convert_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_Census5x5_Converter( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Census5x5_Convert_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback );
}

// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Census5x5_Converter_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Census5x5_Convert_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) );
}


// ==========================================================================

}	// end of the PHIL namespace

// ==========================================================================



#endif BoxySoft_HIL_NonParametricConvertOperators




