
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_PyramidOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of operators for image pyramids.
//
// ==========================================================================



#ifndef BoxySoft_HIL_PyramidOperators
#define BoxySoft_HIL_PyramidOperators


#include "HIL_BaseOperations.h"
#include "NonParametricTransformations.h"
#include "GaussianImagePyramids.h"



namespace PHIL		
{





// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation for building
// image Gaussian pyramids.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class Gaussian_Pyramid_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	protected:
	
		// Some local storage for input parameters:
		int									fPyramidLevels;
		int									fMaskWidth;
		double								fSigma;
		const TRealLinearFilter_Factory &	fFilterFactory;

	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). The input
		//				image will be converted-and-copied
		//				to this image.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			pyramidLevels - number of requested 
		//				levels of the pyramid
		//			maskWidth - width in pixels of the smooting
		//				mask;
		//			sigma - the sigma parameter of the filter
		//			theFactory - ref to the filter factory;
		//				filter from this factory will be used to
		//				low-pass filtering of the input image
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//
		Gaussian_Pyramid_OperationFor( RetType & retImage, const InType_1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: fPyramidLevels( pyramidLevels ), fMaskWidth( maskWidth ), fSigma( sigma ), fFilterFactory( theFactory ), TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Gaussian_Pyramid_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator converts and copies an input
		// image into an output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			There must be an appropriate conversion
		//			procedures defines for a given pair of images.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			TGaussianImagePyramids theGaussianPyramid( fMaskWidth, fSigma, fFilterFactory );

			theGaussianPyramid( fInImage1_Ref, fRetImageRef, fPyramidLevels );

			return & fRetImageRef;
		}

};




// ==========================================================================


///////////////////////////////////////////////////////////
//
// This class implements unary image operation for building
// image Laplacian pyramids.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class Laplacian_Pyramid_OperationFor : public Gaussian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). The input
		//				image will be converted-and-copied
		//				to this image.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			pyramidLevels - number of requested 
		//				levels of the pyramid
		//			maskWidth - width in pixels of the smooting
		//				mask;
		//			sigma - the sigma parameter of the filter
		//			theFactory - ref to the filter factory;
		//				filter from this factory will be used to
		//				low-pass filtering of the input image
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//
		Laplacian_Pyramid_OperationFor( RetType & retImage, const InType_1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: Gaussian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, pyramidLevels, maskWidth, sigma, theFactory, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Laplacian_Pyramid_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator converts and copies an input
		// image into an output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			There must be an appropriate conversion
		//			procedures defines for a given pair of images.
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			TLaplacianImagePyramids theLaplacianPyramid( fMaskWidth, fSigma, fFilterFactory );

			theLaplacianPyramid( fInImage1_Ref, fRetImageRef, fPyramidLevels );


			return & fRetImageRef;
		}

};

// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
Gaussian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_Gaussian_Pyramid( RetIm_Type & retImage, const InIm_Type1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Gaussian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, pyramidLevels, maskWidth, sigma, theFactory, resourceAccessPolicy, opCompCallback );
}

// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Gaussian_Pyramid_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Gaussian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, pyramidLevels, maskWidth, sigma, theFactory, resourceAccessPolicy, opCompCallback ) );
}


// ==========================================================================

template< typename RetIm_Type, typename InIm_Type1 >
Laplacian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_Laplacian_Pyramid( RetIm_Type & retImage, const InIm_Type1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Laplacian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, pyramidLevels, maskWidth, sigma, theFactory, resourceAccessPolicy, opCompCallback );
}

// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Laplacian_Pyramid_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, int pyramidLevels, int maskWidth, double sigma, const TRealLinearFilter_Factory & theFactory = TGaussianFilter_Factory::gkGaussianFilter_Factory_Object, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Laplacian_Pyramid_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, pyramidLevels, maskWidth, sigma, theFactory, resourceAccessPolicy, opCompCallback ) );
}


// ==========================================================================

}	// end of the PHIL namespace

// ==========================================================================



#endif BoxySoft_HIL_PyramidOperators




