
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_RankFilterOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of operators for order filters.
//
// ==========================================================================



#ifndef BoxySoft_HIL_RankFilterOperators
#define BoxySoft_HIL_RankFilterOperators



#include "HIL_BaseOperations.h"
#include "AccumulationTraits_ForFixed.h"
#include "GraphicLibrary.h"
#include "Morphology.h"



namespace PHIL		
{


// ==========================================================================



///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the rank filtering of an image in pixel neighborhoods 
// defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Median_Filter_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Median_Filter_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Median_Filter_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator which median filters an image 
		// in local rectangular neighborhoods of pixels; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//	
		virtual void * operator()( void )
		{
			REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// We assume that the template-image (a parameter supplied to the constructor) 
			// is an image of type TImageFor< T > where T is a type that can be converted
			// to "Dimension". This image is a single row x two columns: the first element
			// denotes a horizontal span, while the second denotes the vertical.
			MedianFilter( fInImage1, fImageTemplate, fRetImage );	// action

			return & fRetImage;
		}


};

// ==========================================================================



#if MODERN_COMPILER

// Specialization for template-image given by "Index" - see member descriptions in the main class.
// Visual C++ 6.0 apparently cannot cope with this type of partial class template specialization :(
template< typename RetIm_Type, typename InIm_Type1/*, typename InIm_Type2*/ >
class Median_Filter_OperationFor< RetIm_Type, InIm_Type1, Index > : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, Index >
{
	public:


		Median_Filter_OperationFor( RetType & retImage, const InType_1 const & inImage1, const Index const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, Index >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		virtual ~Median_Filter_OperationFor() {}

		virtual void * operator()( void )
		{
			REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// We assume that the template-image (a parameter supplied to the constructor) 
			// is an image of type TImageFor< T > where T is a type that can be converted
			// to "Dimension". This image is a single row x two columns: the first element
			// denotes a horizontal span, while the second denotes the vertical.
			MedianFilter( fInImage1, fImageTemplate.fCol, fImageTemplate.fRow, fRetImage );	// action

			return & fRetImage;
		}
};



// ==========================================================================



// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
Median_Filter_OperationFor< RetIm_Type, InIm_Type1, Index > * Orphan_Median_Filter( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const Index const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Median_Filter_OperationFor< RetIm_Type, InIm_Type1, Index >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback );
}



// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr Median_Filter_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const Index const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Median_Filter_OperationFor< RetIm_Type, InIm_Type1, Index >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) );
}


// ==========================================================================



#else MODERN_COMPILER


#endif MODERN_COMPILER



// ==========================================================================
// ==========================================================================
// ==========================================================================






///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological erosion driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Erode_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Erode_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Erode_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological erosion of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data 
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.Erode( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};


///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological dilation driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Dilate_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Dilate_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Dilate_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological dilation of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data  
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.Dilate( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};


///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological opening driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Open_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Open_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Open_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological opening of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data 
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.Open( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};



///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological closing driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class Close_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Close_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Close_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological erosion of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data 
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.Close( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};



///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological gradient driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class MorphoGradient_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		MorphoGradient_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~MorphoGradient_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological erosion of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.Gradient( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};



///////////////////////////////////////////////////////////
//
// This class implements binary template image operation:
// the morphological top-hat driven by a structural 
// element which is defined by the supplied template-image.
//
// The class-template premises are as follows:
//   - RetIm_Type is a type of an output image (this
//     should be the same type as the input image given 
//     by InIm_Type1).
//   - InIm_Type1 is a type of an input image, i.e.
//     the one to be filtered.
//   - InIm_Type2 defines a template-image that defines
//     type of a neighborhood of pixels that are filtered.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
class TopHat_OperationFor : public TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >
{
	public:

		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter)
		//			inImage1 - constant reference to the first constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			inTemplateImage - constant reference to the constant
		//				input template-image of type InIm_Type2 (specified  
		//				by the third template parameter). In general this
		//				parameter defines the local neighborhood of pixels
		//				that are taken to determine median value which constitutes
		//				an output pixel in the retImage. 
		//
		//				In this class the simplest rectangular neighborhood
		//				is assumed (i.e. the Moore type). Thus we need only
		//				to supply two dimensions of the rectangle. They can 
		//				be supplied as a pair (h,w) where h - is a horizontal
		//				span, w - vertical. 
		//				
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		TopHat_OperationFor( RetType & retImage, const InType_1 const & inImage1, const InType_2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			: TImageTemplateOperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~TopHat_OperationFor() {}

		///////////////////////////////////////////////////////////
		// The function operator for morphological erosion of the  
		// input image in accordance with the structural element; 
		// Results go to the output image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image
		//		
		// REMARKS:
		//			It is required that the input image and the mask,
		//			are different from the output object.
		//			Otherwise data will be corrupted!
		//	
		virtual void * operator()( void )
		{
			//REQUIRE( (void*) & fInImage1 != (void*) & fRetImage );			// otherwise we destroy input data for convolution 
			//REQUIRE( (void*) & fImageTemplate != (void*) & fRetImage );		// the same is required for the template image 
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			
			// The StrictlyTypedMorphology is a type which fits to the types
			// of images, and their pixels, supplied to a given instantiation
			// of this class. Remember that usually InType_1 (through InIm_Type1)
			// is a type of an image whereas the MorphologyFor<> template class
			// requires a pixel-type template.
			typedef MorphologyFor< InType_1::PixelType > StrictlyTypedMorphology;

			StrictlyTypedMorphology theMorphologyObj;		// create the morphology object (very light) of the specific type
			StrictlyTypedMorphology::TheImageAutoPtr theErodedImage_AP( theMorphologyObj.TopHat( fInImage1, fImageTemplate ) );
			fRetImage = * theErodedImage_AP.release();		// take it over

			return & fRetImage;
		}


};


// ==========================================================================
// ==========================================================================





// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Median_Filter_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Median_Filter( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Median_Filter_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback );
}


// Morphology (binary and gray-valued)
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Erode_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Erode( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Erode_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Dilate_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Dilate( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Dilate_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Open_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Open( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Open_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
Close_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_Close( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Close_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
MorphoGradient_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_MorphoGradient( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new MorphoGradient_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
TopHat_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 > * Orphan_TopHat( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new TopHat_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback );
}






// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Median_Filter_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inTemplateImage, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Median_Filter_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inTemplateImage, resourceAccessPolicy, opCompCallback ) );
}


// Also for morphology (binary and gray-valued)
template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Erode_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Erode_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Dilate_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Dilate_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Open_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Open_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr Close_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Close_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr MorphoGradient_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new MorphoGradient_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}

template< typename RetIm_Type, typename InIm_Type1, typename InIm_Type2 >
ImageOperation_AutoPtr TopHat_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const InIm_Type2 const & inImage2, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new TopHat_OperationFor< RetIm_Type, InIm_Type1, InIm_Type2 >( retImage, inImage1, inImage2, resourceAccessPolicy, opCompCallback ) );
}




// ==========================================================================



}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_RankFilterOperators




