
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_RankFilterOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of operators for image thresholding.
//
// ==========================================================================



#ifndef BoxySoft_HIL_ThresholdingOperators
#define BoxySoft_HIL_ThresholdingOperators


#include "HIL_BaseOperations.h"

namespace PHIL		
{



// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// image global thresholding.
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class Threshold_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	protected:

		typedef typename InIm_Type1::PixelType Threshold_Type;
		typedef typename InIm_Type1::PixelType OutPixel_Type;



			OutPixel_Type	fOutValue_1;	// first output value

		Threshold_Type	fThresh_1;		// the first threshold value
		
			OutPixel_Type	fOutValue_2;	// second output value
		
		Threshold_Type	fThresh_2;		// the second threshold value

			OutPixel_Type	fOutValue_3;	// third output value


	public:

		///////////////////////////////////////////////////////////
		// Class constructor.
		// Thresholding is done by comparing each input pixel
		// with thresholds. Then the output value is set based
		// on this comparison, as follows:
		//
		// Computes the multi-graylevel thresholded output image.
		// The output gray value is one of the three ranges:
		//
		// Iout(x,y) = out_1 <==>             Iin( x, y ) < Thresh1
		// Iout(x,y) = out_2 <==> thresh_1 <= Iin( x, y ) < thresh_2
		// Iout(x,y) = out_3 <==>             Iin( x, y ) >= thresh_2
		//
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//			thresh_1 - first threshold value
		//			thresh_2 - second threshold value
		//			out_1 - output value 1
		//			out_2 - output value 2
		//			out_3 - output value 3
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		Threshold_OperationFor( RetType & retImage, const InType_1 const & inImage1, Threshold_Type thresh_1, Threshold_Type thresh_2, OutPixel_Type out_1, OutPixel_Type out_2, OutPixel_Type out_3, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			:	TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ),
				fThresh_1( thresh_1 ), fThresh_2( thresh_2 ), 
				fOutValue_1( out_1 ), fOutValue_2( out_2 ), fOutValue_3( out_3 ) {}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~Threshold_OperationFor() {}


		///////////////////////////////////////////////////////////
		// The function operator which finds a pixel with maximal
		// value in an input image, as well as its position in 
		// that image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image;
		//		
		// REMARKS:
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble
			MultiThresholding( fInImage1_Ref, fThresh_1, fThresh_2, fRetImageRef, fOutValue_1, fOutValue_2, fOutValue_3 );
			return & fRetImageRef;
		}

};



// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
Threshold_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_ImageThreshold( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const typename InIm_Type1::PixelType th1, const typename InIm_Type1::PixelType th2, const typename InIm_Type1::PixelType v1, const typename InIm_Type1::PixelType v2, const typename InIm_Type1::PixelType v3, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new Threshold_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, th2, th2, v1, v2, v3, resourceAccessPolicy, opCompCallback );
}


// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr ImageThreshold_AP( RetIm_Type & retImage, const InIm_Type1 const & inImage1, const typename InIm_Type1::PixelType th1, const typename InIm_Type1::PixelType th2, const typename InIm_Type1::PixelType v1, const typename InIm_Type1::PixelType v2, const typename InIm_Type1::PixelType v3, TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new Threshold_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, th2, th2, v1, v2, v3, resourceAccessPolicy, opCompCallback ) );
}



// ==========================================================================
}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_ThresholdingOperators




