
// ==========================================================================
//
// Software written by Boguslaw Cyganek (C) 2008
// Copyrights Pandora Inc. (C) 2008
//
// ==========================================================================
//
// This software is protected by the national and international copyrights. 
// The software is granted only for educational or academic/research purposes   
// and to accompany the book (credited to):
// "AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS"
// by Boguslaw Cyganek & J. Paul Siebert, Wiley, 2008.
// All other commercial and/or non-private applications require a written 
// permission.
// The software is supplied as it is without any guarantees nor responsibility  
// of its use in any application. 
//
// ==========================================================================
//
//  Background: AN INTRODUCTION TO 3D COMPUTER VISION TECHNIQUES AND ALGORITHMS
//  Version:    1.0
//
//  $Workfile: HIL_WarpingOperators.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-05-06 21:32:58
//
//
//  Description:
//				Definition of operators for image warping.
//
// ==========================================================================



#ifndef BoxySoft_HIL_WarpingOperators
#define BoxySoft_HIL_WarpingOperators


#include "HIL_BaseOperations.h"
#include "Warping.h"



namespace PHIL		
{



// ==========================================================================

///////////////////////////////////////////////////////////
//
// This class implements unary image operation:
// image warping
//
///////////////////////////////////////////////////////////
template< typename RetIm_Type, typename InIm_Type1 >
class AffineWarping_OperationFor : public TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >
{
	protected:

		double fRotationInDegs;					// contains a rotation angle (in degrees)
		Real_2D_Point fCenterOfRotation;		// stores a point in an input image which is a center of rotation

		Real_2D_Point fTranslationVector;		// horz, vert translation

		Real_2D_Point fScalingVector;			// defines horz, vert scaling

		TCompound_LinearTransformEngine fCoordTransformEngine;			// the compound affine transformation engine

	public:

		///////////////////////////////////////////////////////////
		// Class constructor.
		// Warpinging is done by comparing each input pixel
		// with thresholds. Then the output value is set based
		// on this comparison, as follows:
		//
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			retImage - reference to the output image
		//				of type RetIm_Type (specified by the 
		//				first template parameter). It should be
		//				able to receive three pixels: one
		//				with the found value and two with pixel
		//				position coordinates.
		//			inImage1 - constant reference to the constant
		//				input image of type InIm_Type1 (specified  
		//				by the second template parameter)
		//
		//			resourceAccessPolicy - optional reference to
		//				the thread security object (derivative
		//				of the TThreadSecurity class); by default
		//				the static kgThreadSecurity object is supplied
		//				which does nothing
		//			opCompCallback - optional reference to the callback
		//				object which is called upon completion of operation;
		//				by default the static kgOperationCompletionCallback
		//				object is supplied which does nothing
		//		
		AffineWarping_OperationFor( RetType & retImage, const InType_1 const & inImage1, 
									const double rotationInDegs, const Real_2D_Point & centerOfRotation,
									const Real_2D_Point & translationVector,
									const Real_2D_Point & scalingVector,
									TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
			:	TUnaryImageOperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, resourceAccessPolicy, opCompCallback ),
				fRotationInDegs( rotationInDegs ), fCenterOfRotation( centerOfRotation ), fTranslationVector( translationVector ), fScalingVector( scalingVector )
				{}

		///////////////////////////////////////////////////////////
		// Class virtual destructor
		///////////////////////////////////////////////////////////
		virtual ~AffineWarping_OperationFor() {}


		///////////////////////////////////////////////////////////
		// The function operator which finds a pixel with maximal
		// value in an input image, as well as its position in 
		// that image.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			none
		//		
		// OUTPUT:
		//			a pointer to the return image;
		//		
		// REMARKS:
		//		
		virtual void * operator()( void )
		{
			MImageOperationRetinue	theImageOperationRetinue( * this );		// create this variable for proper preamble

			TCompound_LinearTransformEngine coordTransformEngine;			// the compound affine transformation engine
		
			TCoordRotationEngine		rotation( fRotationInDegs / 180.0 * kPi, fCenterOfRotation );	// Create a rotation engine
			coordTransformEngine.AddNewTransformation( rotation );			// add rotation

			TCoordTranslationEngine		translation( fTranslationVector );	// Create a translation engine
			coordTransformEngine.AddNewTransformation( translation );		// add translation

			TCoordScalingEngine			scaling( fScalingVector );			// Create a scaling engine
			coordTransformEngine.AddNewTransformation( scaling );			// add scaling
	

			TSimpleImageBackwardWarp< InIm_Type1::PixelType >	simpleWarpEngine( fInImage1_Ref, coordTransformEngine );		// we use a backward warping scheme
	
			std::auto_ptr< TImageFor< InIm_Type1::PixelType > > outImage( simpleWarpEngine() );		// perform the warping transformation here
			
			MixedCopy( fRetImageRef, * outImage );													// and copy a warp result into an output image, whatever type it is...
	
			// Store coordinates for later
			fCoordTransformEngine = coordTransformEngine;

			return & fRetImageRef;
		}

		// Setters/Getters here...

	public:


		TCompound_LinearTransformEngine & GetCoordTransformEngine( void ) { return fCoordTransformEngine; }			// the compound affine transformation engine

};



// ==========================================================================

// Useful helpers:
template< typename RetIm_Type, typename InIm_Type1 >
AffineWarping_OperationFor< RetIm_Type, InIm_Type1 > * Orphan_ImageAffineWarping(	RetIm_Type & retImage, const InIm_Type1 const & inImage1, 
																					const double rotationInDegs, const Real_2D_Point & centerOfRotation,
																					const Real_2D_Point & translationVector,
																					const Real_2D_Point & scalingVector,
																					TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return new AffineWarping_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, rotationInDegs, centerOfRotation, translationVector, scalingVector, resourceAccessPolicy, opCompCallback );
}


// Yet another helpers:
template< typename RetIm_Type, typename InIm_Type1 >
ImageOperation_AutoPtr ImageAffineWarping_AP(	RetIm_Type & retImage, const InIm_Type1 const & inImage1, 
												const double rotationInDegs, const Real_2D_Point & centerOfRotation,
												const Real_2D_Point & translationVector,
												const Real_2D_Point & scalingVector,
												TThreadSecurity & resourceAccessPolicy = kgThreadSecurity, TOperationCompletionCallback & opCompCallback = kgOperationCompletionCallback )
{
	return ImageOperation_AutoPtr( new AffineWarping_OperationFor< RetIm_Type, InIm_Type1 >( retImage, inImage1, rotationInDegs, centerOfRotation, translationVector, scalingVector, resourceAccessPolicy, opCompCallback ) );
}



// ==========================================================================
}	// end of the PHIL namespace

// ==========================================================================


#endif BoxySoft_HIL_WarpingOperators




