
// ==========================================================================
//  Copyright  1998, 1999 The BoxySoft Inc.  All rights reserved.
// ==========================================================================
//
//  Background: Current Project
//  Version:    1.0
//
//  $Workfile: Tensor.cpp $
//  $Author  : Boguslaw Cyganek
//  $Created : 2007-03-24 23:51:48
//
//  $Revision: 114 $
//  $Author: Boguslaw Cyganek $
//  $Date: 2007-03-24 23:51:48 $
//
//  Description:
//
//
// ==========================================================================
//  History:
//
// ==========================================================================
//
//  $Log: Tensor.cpp $
//
// ==========================================================================




#include "Tensor.h"
#include "HIL_MultiPixelFor.h"
#include "HIL_VideoFor.h"




///////////////////////////////////////////////////////////
// This function does n-mode multiplication of a tensor
// with a matrix. Creates and returns the result n-mode
// tensor.
//
//			B = A  x    M
//                 (n)
//
// which is equivalent to
//
//			B    =  M   A
//          (n)         (n)
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//	
template < typename T >	
typename TFlatTensorFor< T >::TFlatTensor_AP n_mode_multiply( const TFlatTensorFor< T > & A, const typename TFlatTensorFor< T >::DataMatrix & M )
{
	// At first check dimensions and decide
	//REQUIRE( A.GetDataMatrix() != 0 );
	REQUIRE( A.GetImageData() != 0 );

	if( A.GetDataMatrix_Row() != M.GetCol() )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// wrong dimensions, cannot multiply, exiting...

	TFlatTensorFor< T >::TensorElem_IndexVector	index_B( A.GetIndexVector() );		// all the indices are the same
	index_B[ A.GetTensorMode() ] = M.GetRow();									// except this one

	// Create an output tensor
	TFlatTensorFor< T >::TFlatTensor_AP B( new TFlatTensorFor( index_B, A.GetTensorMode(), A.GetCyclicMode() ) );

	Mult_Matrix< TFlatTensorFor< T >::TensorElem_Type, TFlatTensorFor< T >::TensorElem_Type >( * B, M, A );

	REQUIRE( B->ClassInvariant() );

	return B;
}




///////////////////////////////////////////////////////////
// This function returns an orphaned copy of the input
// n-mode flat tensor, possibly in a different mode.
// Thus, it can be used for mode conversions of the flat
// tensors.
///////////////////////////////////////////////////////////
//		
// INPUT:
//			A - the input flat tensor 
//			k_mode - a mode of the output tensor
//		
// OUTPUT:
//			New tensor or
//			0 if error
//		
// REMARKS:
//		
//		
template < typename T >
typename TFlatTensorFor< T >::TFlatTensor_AP orphan_k_mode_tensor_from( const TFlatTensorFor< T > & A, int k_mode )
{
	// Check input data
	REQUIRE( k_mode < A.GetIndexVector().size() );
	if( k_mode >= A.GetIndexVector().size() )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// wrong dimensions, cannot proceed, exiting...

	// Create an output tensor
	TFlatTensorFor< T >::TFlatTensor_AP B( new TFlatTensorFor< T >( A.GetIndexVector(), k_mode ) );
	REQUIRE( B.get() != 0 );
	if( B.get() == 0 )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// problems when creating a new tensor (low memory?), exiting...


	// Now copy data
	const TFlatTensorFor< T >::TensorElem_IndexVector & elemIdx = A.GetIndexVector();
	int numOfDimensions = elemIdx.size();

	
	TFlatTensorFor< T >::TensorElem_IndexVector index( numOfDimensions, 0 );

	// Traverse each dimension and copy element-by-element
	bool all_traversed = false;
	while( all_traversed == false )
	{
		DEBUG_PrintVector( "mode_index.txt", index, true );


		// Here we copy a single element
		TFlatTensorFor< T >::TensorElem_Type theElement = A.GetElement( index );
		B->SetElement( index, theElement );


		// ----------------------------------------------
		int d = 0;
		for( d = 0; d < numOfDimensions; ++ d )
		{

			int left_index_val		= index[ d ];
			int left_index_val_max	= elemIdx[ d ] - 1;

			if( left_index_val < left_index_val_max )
			{
				index[ d ] = left_index_val + 1;		// turn this ratchet by one step
											
				break;									// and start counting from the least counter again
			}
			else
			{
				// This ratchet is full, so set it to 0 and advance the next one
				index[ d ] = 0;
			}

		}

		// Check STOP conditions
		if( d == numOfDimensions )
			all_traversed = true;
		// ----------------------------------------------


	}

	REQUIRE( B->ClassInvariant() );

	return B;
}

// ======================================================================================



// In future: the indices could be also saved to a file
template < typename T >
typename TFlatTensorFor< T >::TFlatTensor_AP orphan_k_mode_tensor_from( const char * fileName, const typename TFlatTensorFor< T >::TensorElem_IndexVector & tensorIndices, int k_mode )
{
	// Check input data
	REQUIRE( k_mode < tensorIndices.size() );
	if( k_mode >= tensorIndices.size() )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// wrong dimensions, cannot proceed, exiting...

	// Create an output tensor
	TFlatTensorFor< T >::TFlatTensor_AP B( new TFlatTensorFor< T >( tensorIndices, k_mode ) );
	REQUIRE( B.get() != 0 );
	if( B.get() == 0 )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// problems when creating a new tensor (low memory?), exiting...

	REQUIRE( B->GetImageData() != 0 );

	REQUIRE( B->ClassInvariant() );

	return Load_ASCII_Image( * B, fileName ) == true ? B : TFlatTensorFor< T >::TFlatTensor_AP( 0 );
}




template < typename T >
bool save_tensor_data( const char * fileName, const TFlatTensorFor< T > & tensor )
{
	return Save_ASCII_Image( tensor, fileName );
}


// ------------

template < typename T >
typename TFlatTensorFor< T >::TFlatTensor_AP orphan_k_mode_tensor_from( const MonochromeVideo & inMonoVideo )
{
	int numOfFrames = inMonoVideo.GetNumOfFrames();

	if( numOfFrames < 1 )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );


	const int kFrameCols = inMonoVideo.GetFrameAt( 0 )->GetCol();
	const int kFrameRows = inMonoVideo.GetFrameAt( 0 )->GetRow();

	TFlatTensorFor< T >::TensorElem_IndexVector tensorIndices;
	
	tensorIndices.push_back( kFrameRows );
	tensorIndices.push_back( numOfFrames );
	tensorIndices.push_back( kFrameCols );


	const int k_mode = 0;

	// Create an output tensor
	TFlatTensorFor< T >::TFlatTensor_AP B( new TFlatTensorFor< T >( tensorIndices, k_mode ) );
	REQUIRE( B.get() != 0 );
	if( B.get() == 0 )
		return TFlatTensorFor< T >::TFlatTensor_AP( 0 );		// problems when creating a new tensor (low memory?), exiting...

		
	// Copy the frames
	int anchor_x = 0;
	int anchor_y = 0;

	for( int f = 0; f < numOfFrames; ++ f )
	{
		RealImageProxy frameImage( * B, anchor_x, anchor_y, kFrameCols, kFrameRows );

		anchor_x += kFrameCols;

		MixedCopy( frameImage, * inMonoVideo.GetFrameAt( f ) );
	}


	return B;
}


template < typename U >
RealVideo * orphan_real_video_from( const TFlatTensorFor< U > & T )
{
	const TFlatTensorFor< U >::TensorElem_IndexVector & tensorIndices = T.GetIndexVector();

	int T_mode = T.GetTensorMode();

	// For this version we assume that T_mode must be 0.
	// If so, then 
	// index[ T_mode ] == frame_rows, 
	// index[ T_mode + 1 ] == num_of_frames
	// index[ T_mode + 2 ] == frame_cols

	REQUIRE( T_mode == 0 );
	REQUIRE( T_mode < tensorIndices.size() );
	REQUIRE( tensorIndices.size() == 3 );				// in this version, can be extended
	
	int mode = T_mode;

	int kFrameRows	= tensorIndices[ mode ];

	++ mode;
	if( mode >= tensorIndices.size() )
		mode = 0;
	int numOfFrames = tensorIndices[ mode ];

	++ mode;
	if( mode >= tensorIndices.size() )
		mode = 0;
	int kFrameCols	= tensorIndices[ mode ];

	// Based on this create an output mono video
	RealVideo * outVideo = new RealVideo( kFrameCols, kFrameRows, numOfFrames );				// class constructor
	REQUIRE( outVideo != 0 );


	// Now copy frames from the tensor matrix

	int anchor_x = 0;
	int anchor_y = 0;

	for( int f = 0; f < numOfFrames; ++ f )
	{
		RealImageProxy frameImage( (TFlatTensorFor< U > &)T, anchor_x, anchor_y, kFrameCols, kFrameRows );

		anchor_x += kFrameCols;

		MixedCopy( * outVideo->GetFrameAt( f ), frameImage );
	}

	return outVideo;
}



// ======================================================================================

template < typename T >
typename TFlatTensorFor< T >::TensorElem_IndexVector VectorFromTable( int * tab, int num_of_elems )
{
	TFlatTensorFor< T >::TensorElem_IndexVector retVector;
	copy( tab, tab + num_of_elems, back_inserter( retVector ) );
	return retVector;
}


void FlatTensorTest( void )
{
	int tensorDim[] = { 3, 4, 2 };

	int n_mode = 0;

	TFlatTensor theTensor( tensorDim, NUM_OF_ARRAY_ELEMENTS( tensorDim ), n_mode );


	// ---
	int elemIndices[] = { 2, 1, 0 };
	TFlatTensor::TensorElem_IndexVector elemIdx;
	copy( elemIndices, elemIndices + sizeof( elemIndices ) / sizeof( elemIndices[0] ), back_inserter( elemIdx ) );
	DEBUG_PrintVector( "elemIndices.txt", elemIdx );

	theTensor.SetElement( elemIdx, 13.31 );

	double retVal = theTensor.GetElement( elemIdx );


	// Add few more elements
	elemIdx[ 0 ] = 0; elemIdx[ 1 ] = 0; elemIdx[ 2 ] = 0;
	theTensor.SetElement( elemIdx, 11 );

	elemIdx[ 0 ] = 0; elemIdx[ 1 ] = 1; elemIdx[ 2 ] = 0;
	theTensor.SetElement( elemIdx, 12 );
	
	elemIdx[ 0 ] = 1; elemIdx[ 1 ] = 1; elemIdx[ 2 ] = 0;
	theTensor.SetElement( elemIdx, 13 );
	
	elemIdx[ 0 ] = 1; elemIdx[ 1 ] = 1; elemIdx[ 2 ] = 1;
	theTensor.SetElement( elemIdx, 14 );
	
	elemIdx[ 0 ] = 2; elemIdx[ 1 ] = 3; elemIdx[ 2 ] = 1;
	theTensor.SetElement( elemIdx, 15 );
	
	save_tensor_data( "theTensor_0.txt", theTensor );
	
	
	//VectorFromTable( elemIndices, NUM_OF_ARRAY_ELEMENTS( elemIndices ) );

	//
	TFlatTensor::TFlatTensor_AP tensor_2( orphan_k_mode_tensor_from( theTensor, 1 ) );
	save_tensor_data( "theTensor_1.txt", * tensor_2 );

	//
	TFlatTensor::TFlatTensor_AP tensor_3( orphan_k_mode_tensor_from( theTensor, 2 ) );
	save_tensor_data( "theTensor_2.txt", * tensor_3 );


	/////////////////////////////////////////////////////

	TFlatTensor::TensorElem_IndexVector tensor_2_104_1_Indices;
	tensor_2_104_1_Indices.push_back( 3 );
	tensor_2_104_1_Indices.push_back( 3 );
	tensor_2_104_1_Indices.push_back( 3 );

	TFlatTensor::TFlatTensor_AP tensor_2_104_1( orphan_k_mode_tensor_from< double >( "tensor_2_104_1.txt", tensor_2_104_1_Indices, 0 ) );

	if( tensor_2_104_1.get() == 0 )
		return;

	// Now save it in different modes

	//
	TFlatTensor::TFlatTensor_AP tensor_2_104_2( orphan_k_mode_tensor_from( * tensor_2_104_1, 1 ) );
	save_tensor_data( "tensor_2_104_2.txt", * tensor_2_104_2 );

	//
	TFlatTensor::TFlatTensor_AP tensor_2_104_3( orphan_k_mode_tensor_from( * tensor_2_104_1, 2 ) );
	save_tensor_data( "tensor_2_104_3.txt", * tensor_2_104_3 );

	/////////////////////////////////////////////////////



/*	TFlatTensorFor< T >::DataMatrix_Index inIndex( 1, 2 );
	TFlatTensorFor< T >::DataMatrix_Index outIndex( 0, 0 );

	int outTensorMode = n_mode; 

	theTensor.ConvertMatrixIndices( inIndex, outTensorMode, outIndex );
	REQUIRE( inIndex.row == outIndex.row );
	REQUIRE( inIndex.col == outIndex.col );
*/


	// 
	TFlatTensorProxy proxy_1( theTensor, 1 ), proxy_2( theTensor, 2 );

	save_tensor_data( "theTensor_1_proxy.txt", proxy_1 );
	save_tensor_data( "theTensor_2_proxy.txt", proxy_2 );

}


// ======================================================================================





