
// ==========================================================================
//  Copyright  1998, 1999 The BoxySoft Inc.  All rights reserved.
// ==========================================================================
//
//  Background: Current Project
//  Version:    1.0
//
//  $Workfile: Tensor.h $
//  $Author  : Boguslaw Cyganek
//  $Created : 2010-03-24 23:51:07
//
//  $Revision: 114 $
//  $Author: Boguslaw Cyganek $
//  $Date: 2010-03-24 23:51:07 $
//
//  Description:
//
//
// ==========================================================================
//  History:
//
// ==========================================================================
//
//  $Log: Tensor.h $
//
// ==========================================================================


#ifndef BoxySoft_Tensor
#define BoxySoft_Tensor



#include "HIL_ImageFor.h"
#include <vector>
#include <memory>



using namespace PHIL;



///////////////////////////////////////////////////////////
// This class implements a tensor in the matricized format.
///////////////////////////////////////////////////////////
template < typename T = double >
class TFlatTensorFor : public TImageFor< T >
{
	private:

		// Class inherent variables

	public:

		typedef TImageFor< T >	BaseClass;

		typedef double TensorElem_Type;


		typedef TensorElem_Type			ElemAccessType;		// by value, put TensorElem_Type & to set by reference
		typedef const ElemAccessType	ConstElemAccessType;


		typedef TImageFor< TensorElem_Type > DataMatrix;

		typedef int TensorElem_Index;
		typedef vector< TensorElem_Index > TensorElem_IndexVector;

		typedef Index DataMatrix_Index;


	protected:

		int				fTensorMode;	// selected index (in fIndexVector)


		TensorElem_IndexVector		fIndexVector;	// the vector of indices of the tensor

	public:

		int				GetDataMatrix_Row( void ) const { return fRow; }	
		int				GetDataMatrix_Col( void ) const { return fCol; }			

	public:

		int				GetTensorMode( void ) const { return fTensorMode; }			

		const TensorElem_IndexVector &	GetIndexVector( void ) const { return fIndexVector; }


	protected:


		bool Create( const TensorElem_IndexVector & indexVec, int n_mode );

		bool Create( const TensorElem_Index * indices, int numOfIndices, int n_mode );

	public:


		///////////////////////////////////////////////////////////
		// This function checks consistency of data members.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:	
		//			none
		//		
		// OUTPUT:
		//			true if ok
		//			false - serious internal coding error
		//		
		// REMARKS:
		//			
		//		
		bool ClassInvariant( void );


//	protected:


		bool ( TFlatTensorFor ::* fCyclicFunction )( const TensorElem_IndexVector & elementPosition, DataMatrix_Index & outOffset, int tensor_mode ) const;


		///////////////////////////////////////////////////////////
		// This function converts a multi-dimensional tensor
		// index (given by a vector) into an offset to the 
		// "flat" matricised representation. Used to access
		// elements of a tensor.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			elementPosition - tensor index of an element
		//				(its size must be the same as fIndexVector,
		//				and each value less than a corresponding
		//				in fIndexVector)
		//			outOffset - upon success contains an offset 
		//				of an element from the beginning of the
		//				data buffer in the fDataMatrix
		//		
		// OUTPUT:
		//			true on success
		//			false otherwise
		//		
		// REMARKS:
		//			Backward cycle starts from the selected index 
		//				(given by fTensorMode) and reaches indices in
		//				a LOWERING direction but cyclically
		//		
		bool ComputeElementOffset_BackwardCyclic( const TensorElem_IndexVector & elementPosition, DataMatrix_Index & outOffset, int tensor_mode ) const;

		// This is a forward version of the above - from fTensorMode this 
		// goes in a direction of INCREASING indices, also cyclically
		bool ComputeElementOffset_ForwardCyclic( const TensorElem_IndexVector & elementPosition, DataMatrix_Index & outOffset, int tensor_mode ) const;


	public:

		enum ECyclicMode { kBackward, kForward };

		ECyclicMode GetCyclicMode( void ) const { return fCyclicFunction == ComputeElementOffset_BackwardCyclic ? kBackward : kForward; }


	protected:

		// ===================================================

		///////////////////////////////////////////////////////////
		// PROTECTED class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			tensor_mode - mode of the tensor (0 by default)
		//				Tensor mode describes selected index of the
		//				tensor which becomes row counter of the 
		//				data matrix.
		//			cyclicMode - determins cyclic mode of the tensor
		//				which can be forward of backward (default)
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TFlatTensorFor( int tensor_mode = 0, ECyclicMode cyclicMode = kBackward );				// class default constructor


	public:


		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			indexVec - a vector with tensor dimensions
		//				Its size (i.e. number of entries) determins 
		//				dimensionality of a tensor. Eg. if indexVec
		//				is 4, 4, 5, 10 then this will be 4D tensor 
		//				with each dimension 4, 4, 5, 10, respectively.
		//			tensor_mode - mode of the tensor 
		//				Tensor mode describes selected index of the
		//				tensor which becomes row counter of the 
		//				data matrix.
		//			cyclicMode - determins cyclic mode of the tensor
		//				which can be forward of backward (default)
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TFlatTensorFor( const TensorElem_IndexVector & indexVec, int tensor_mode, ECyclicMode cyclicMode = kBackward );		// class constructor
		// As above but indices can be provided in standard C style (ptr to the table with indices and their number)
		TFlatTensorFor( const TensorElem_Index * indices, int numOfIndices, int tensor_mode, ECyclicMode cyclicMode = kBackward );				
		
		TFlatTensorFor( const TFlatTensorFor & );					// class copy constructor
		TFlatTensorFor & operator = ( const TFlatTensorFor & );		// class assignement operator
		
		virtual ~TFlatTensorFor();				// class virtual destructor
		// ===================================================

	public:


		///////////////////////////////////////////////////////////
		// This function returns a tensor's element at given index.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			elemIdx - a vector with an index of an element
		//				to read. Length of this vector must be 
		//				the same as a length of the indexVec provided
		//				during construction of the tensor. Moreover,
		//				each corresponding value should be positive 
		//				but less than that in the indexVec.
		//		
		// OUTPUT:
		//			tensor element at position
		//		
		// REMARKS:
		//		
		//		
		ElemAccessType	GetElement( const TensorElem_IndexVector & elemIdx ) const;

		///////////////////////////////////////////////////////////
		// This function sets a value of an element of the tensor
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			elemIdx - a vector with an index of an element
		//				to read. Length of this vector must be 
		//				the same as a length of the indexVec provided
		//				during construction of the tensor. Moreover,
		//				each corresponding value should be positive 
		//				but less than that in the indexVec.
		//			elemVal - value to be set
		//		
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		void			SetElement( const TensorElem_IndexVector & elemIdx, ElemAccessType elemVal );

	public:


		typedef std::auto_ptr< TFlatTensorFor< T > > TFlatTensor_AP;
};



template < typename T >
TFlatTensorFor< T >::TFlatTensorFor( int tensor_mode /*= 0*/, ECyclicMode cyclicMode /*= kBackward*/ )		// default constructor
	:	fTensorMode( tensor_mode ), 
		fCyclicFunction( cyclicMode == kBackward ? &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic : &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic )
{
	REQUIRE( fData == 0 );
}


template < typename T >
TFlatTensorFor< T >::TFlatTensorFor( const TensorElem_IndexVector & indexVec, int tensor_mode, ECyclicMode cyclicMode /*= kBackward*/ )		// class constructor
	:	fTensorMode( tensor_mode ), 
		fCyclicFunction( cyclicMode == kBackward ? &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic : &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic )
{
	bool createStatus = Create( indexVec, tensor_mode );
	REQUIRE( createStatus == true );			// what to do otherwise?
}

template < typename T >
TFlatTensorFor< T >::TFlatTensorFor( const TensorElem_Index * indices, int numOfIndices, int tensor_mode, ECyclicMode cyclicMode /*= kBackward*/ )		// class constructor
	:	fTensorMode( tensor_mode ), 
		fCyclicFunction( cyclicMode == kBackward ? &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic : &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic )
{
	bool createStatus = Create( indices, numOfIndices, tensor_mode );
	REQUIRE( createStatus == true );			// what to do otherwise?
}



///////////////////////////////////////////////////////////
// Class copy constructor
///////////////////////////////////////////////////////////
template < typename T >
TFlatTensorFor< T >::TFlatTensorFor( const TFlatTensorFor & ref )	
{
	fCyclicFunction = ref.GetCyclicMode() == kBackward ? &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic : &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic;

	bool createStatus = Create( ref.fIndexVector, ref.fTensorMode );
	REQUIRE( createStatus == true );			// what to do otherwise?
}


///////////////////////////////////////////////////////////
// Class assignement operator
///////////////////////////////////////////////////////////
template < typename T >
TFlatTensorFor< T > & TFlatTensorFor< T >::operator = ( const TFlatTensorFor< T > & ref )	
{
	if( this != & ref )
	{
		fCyclicFunction = ref.GetCyclicMode() == kBackward ? &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic : &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic;
		
		bool createStatus = Create( ref.fIndexVector, ref.fTensorMode );
		REQUIRE( createStatus == true );			// what to do otherwise?
	}

	return * this;
}


///////////////////////////////////////////////////////////
// Class destructor
///////////////////////////////////////////////////////////
template < typename T >
TFlatTensorFor< T >::~TFlatTensorFor()		
{
}


template < typename T >
bool TFlatTensorFor< T >::Create( const TensorElem_IndexVector & indexVec, int n_mode )
{
	REQUIRE( n_mode < indexVec.size() );
	if( n_mode >= indexVec.size() )
		return false;

	fTensorMode = n_mode;

	fIndexVector = indexVec;

	int row = indexVec[ n_mode ];

	int col = 1;
	for( int i = 0; i < fIndexVector.size(); ++ i )
		if( i != n_mode )
			col *= fIndexVector[ i ];

	delete BaseClass::fData;
	BaseClass::Create( col, row );
	BaseClass::SetAll( (TensorElem_Type)0 );

	REQUIRE( ClassInvariant() );

	return true;
}

template < typename T >
bool TFlatTensorFor< T >::Create( const TensorElem_Index * indices, int numOfIndices, int n_mode )
{
	REQUIRE( indices != 0 );
	REQUIRE( numOfIndices < 100 );		// sanity check
	if( indices == 0 )
		return false;

	copy( indices, indices + numOfIndices, back_inserter( fIndexVector ) );

	REQUIRE( n_mode < fIndexVector.size() );
	if( n_mode >= fIndexVector.size() )
		return false;

	fTensorMode = n_mode;

	int row = fIndexVector[ n_mode ];

	int col = 1;
	for( int i = 0; i < fIndexVector.size(); ++ i )
		if( i != n_mode )
			col *= fIndexVector[ i ];	// multiply all except that at fTensorMode (which is for rows)


	delete BaseClass::fData;
	BaseClass::Create( col, row );
	BaseClass::SetAll( (TensorElem_Type)0 );

	REQUIRE( ClassInvariant() );

	return true;
}

template < typename T >
bool TFlatTensorFor< T >::ClassInvariant( void )
{
	if( ! ( fCyclicFunction == &TFlatTensorFor<T>::ComputeElementOffset_BackwardCyclic || fCyclicFunction == &TFlatTensorFor<T>::ComputeElementOffset_ForwardCyclic ) )
		return false;		

	if( BaseClass::fData != 0 )
	{
		int cols = 1;
		for( int i = 0; i < fIndexVector.size(); ++ i )
			if( i != fTensorMode )
				cols *= fIndexVector[ i ];

		bool indexVector_cond = cols == fCol;
		bool mode_cond = fTensorMode < fIndexVector.size() && fRow == fIndexVector[ fTensorMode ];

		return indexVector_cond && mode_cond;
	}
	else
	{
		bool indexVector_cond = fIndexVector.size() == 0;
		bool mode_cond = fTensorMode == 0;

		return indexVector_cond && mode_cond;
	}
}

template < typename T >
bool TFlatTensorFor< T >::ComputeElementOffset_BackwardCyclic( const TensorElem_IndexVector & elementPosition, DataMatrix_Index & outOffset, int tensor_mode ) const
{
	REQUIRE( elementPosition.size() == fIndexVector.size() );

	REQUIRE( tensor_mode < elementPosition.size() );

	int r = elementPosition[ tensor_mode ];
	REQUIRE( r < GetDataMatrix_Row() );

	// Go the other direction
	register int c = 0;		// Must be zero here!

	// In backward mode go FORWARD, since we wish to access most
	// significant indices first for successive multiplications
	const int kNumOfDimensions = fIndexVector.size();
	for( register int i = 1, n = tensor_mode + 1; i < kNumOfDimensions; ++ i, ++ n )
	{											
		// This has to be cyclic
		if( n >= kNumOfDimensions )
			n = 0;

		REQUIRE( elementPosition[ n ] < fIndexVector[ n ] );		// all indices must be in the range

		REQUIRE( n >= 0 );
		REQUIRE( n < fIndexVector.size() );
		REQUIRE( n < elementPosition.size() );

		c *= fIndexVector	[ n ];		// first run will always result in 0
		c += elementPosition[ n ];

	}

	REQUIRE( c < GetDataMatrix_Col() );

	outOffset.fCol = c;
	outOffset.fRow = r;

	return true;
}


template < typename T >
bool TFlatTensorFor< T >::ComputeElementOffset_ForwardCyclic( const TensorElem_IndexVector & elementPosition, DataMatrix_Index & outOffset, int tensor_mode ) const
{
	REQUIRE( elementPosition.size() == fIndexVector.size() );

	REQUIRE( tensor_mode < elementPosition.size() );

	int r = elementPosition[ tensor_mode ];
	REQUIRE( r < GetDataMatrix_Row() );

	// Go the other direction
	register int c = 0;		// Must be zero here!

	// In forward mode go backward, since we wish to access the most
	// significant indices first for successive multiplications
	const int kNumOfDimensions = fIndexVector.size();
	for( register int i = 1, n = tensor_mode - 1; i < kNumOfDimensions; ++ i, -- n )
	{											
		// This has to be cyclic
		if( n < 0 )
			n = kNumOfDimensions - 1;

		REQUIRE( elementPosition[ n ] < fIndexVector[ n ] );		// all indices must be in the range

		REQUIRE( n >= 0 );
		REQUIRE( n < fIndexVector.size() );
		REQUIRE( n < elementPosition.size() );

		c *= fIndexVector	[ n ];		// first run will always result in 0
		c += elementPosition[ n ];

	}

	REQUIRE( c < GetDataMatrix_Col() );

	outOffset.fCol = c;
	outOffset.fRow = r;

	return true;
}


template < typename T >
typename TFlatTensorFor< T >::ElemAccessType	TFlatTensorFor< T >::GetElement( const TensorElem_IndexVector & elemIdx ) const
{
	DataMatrix_Index outOffset;
	bool indexStatus = ( this->*fCyclicFunction )( elemIdx, outOffset, fTensorMode );
	REQUIRE( indexStatus == true );

	return BaseClass::GetPixel( outOffset.fCol, outOffset.fRow );
}

template < typename T >
void TFlatTensorFor< T >::SetElement( const TensorElem_IndexVector & elemIdx, ElemAccessType elemVal )
{
	DataMatrix_Index outOffset;
	bool indexStatus = ( this->*fCyclicFunction )( elemIdx, outOffset, fTensorMode );
	REQUIRE( indexStatus == true );

	BaseClass::SetPixel( outOffset.fCol, outOffset.fRow, elemVal );
}


// ======================================================================================



// This class allows different modes of the base tensor
// without data copying, however. 
template < typename T >
class TFlatTensorProxyFor : public TFlatTensorFor< T >
{
	protected:

		typedef typename TFlatTensorFor< T > BaseClass;

	private:

		BaseClass & fMotherTensor;		// this must be a reference since a proxy cannot live without an external object

	public:

		// ===================================================

		
		///////////////////////////////////////////////////////////
		// Class constructor 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			motherTensor - a main object to which this one is
		//				a proxy.
		//			tensor_mode - mode of the proxy-tensor 
		//				The supplied tensor_mode can be DIFFERENT than 
		//				in the mother tensor !!
		//			cyclicMode - determins cyclic mode of the tensor
		//				which can be forward of backward (default)
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		TFlatTensorProxyFor( BaseClass & motherTensor, int tensor_mode, ECyclicMode cyclicMode = kBackward )	// class constructors
			: fMotherTensor( motherTensor ), TFlatTensorFor< T >( tensor_mode, cyclicMode )						// at first create an (almost) empty base
		{
			fIndexVector	= motherTensor.GetIndexVector(); 

			// We need to set up the matrix members
			REQUIRE( tensor_mode < fIndexVector.size() );
			fRow = fIndexVector[ tensor_mode ];		// contains number of rows of that image
			
			fCol = 1;								// contains number of columns of that image
			for( int i = 0; i < fIndexVector.size(); ++ i )
				if( i != tensor_mode )
					fCol *= fIndexVector[ i ];

			fData = motherTensor.GetImageData();	// anyway, the number of data will be the same, though 'in real' these would be shuffled 
	
			fElems = fRow * fCol;
			REQUIRE( fMotherTensor.GetElems() == fElems );

			if( fTensorMode != fMotherTensor.GetTensorMode() )
			{
				fGetPixel_FunPtr = & TFlatTensorProxyFor<T>::GetPixel_DiffModes;
				fSetPixel_FunPtr = & TFlatTensorProxyFor<T>::SetPixel_DiffModes;
			}
			else
			{
				fGetPixel_FunPtr = & TFlatTensorProxyFor<T>::GetPixel_TheSameModes;
				fSetPixel_FunPtr = & TFlatTensorProxyFor<T>::SetPixel_TheSameModes;
			}
		}
		
		virtual ~TFlatTensorProxyFor() 
		{
			fData = 0;		// detach asap, before the base destructor tries to get rid of the someone else's buffer
		}				

	protected:

		TFlatTensorProxyFor( const TFlatTensorProxyFor< T > & r );
		TFlatTensorProxyFor & operator = ( const TFlatTensorProxyFor< T > & r );
													
		// ===================================================

	public:


		///////////////////////////////////////////////////////////
		// This function returns a tensor's element at given index.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			elemIdx - a vector with an index of an element
		//				to read. Length of this vector must be 
		//				the same as a length of the indexVec provided
		//				during construction of the tensor. Moreover,
		//				each corresponding value should be positive 
		//				but less than that in the indexVec.
		//		
		// OUTPUT:
		//			tensor element at position
		//		
		// REMARKS:
		//		
		//		
		ElemAccessType	GetElement( const TensorElem_IndexVector & elemIdx ) const
		{
			return fMotherTensor.GetElement( elemIdx );		// simple action delegation
		}

		///////////////////////////////////////////////////////////
		// This function sets a value of an element of the tensor
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			elemIdx - a vector with an index of an element
		//				to read. Length of this vector must be 
		//				the same as a length of the indexVec provided
		//				during construction of the tensor. Moreover,
		//				each corresponding value should be positive 
		//				but less than that in the indexVec.
		//			elemVal - value to be set
		//		
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//		
		//		
		void SetElement( const TensorElem_IndexVector & elemIdx, ElemAccessType elemVal )
		{
			fMotherTensor.SetElement( elemIdx, elemVal );		// simple action delegation
		}

	public:


		///////////////////////////////////////////////////////////
		// This function sets a pixel at position (x,y)or (col,row) 
		// of this image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			xPixPosition - the horizontal (or column) position of a pixel
		//			yPixPosition - the vertical (or row) position of a pixel
		//			value - a value to be set at pixel position
		//		
		// OUTPUT:
		//			none
		//		
		// REMARKS:
		//			From the OOP point of view this function should be .
		//			However, to avoid run-time panalty it is not ,
		//			and the "value" parameter is passed by value. In derived
		//			classes this strategy can be overloaded.
		//		
		void SetPixel( Dimension xPixPosition, Dimension yPixPosition, const double value ) const
		{
			( this->*fSetPixel_FunPtr )( xPixPosition, yPixPosition, value );
		}


		///////////////////////////////////////////////////////////
		// This function gets a VALUE of a pixel at position  
		// (x,y) or (col,row) of this image. 
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//			xPixPosition - the horizontal (or column) position of a pixel
		//			yPixPosition - the vertical (or row) position of a pixel
		//		
		// OUTPUT:
		//			a copy of a pixel, of type T, from the given position
		//		
		// REMARKS:
		//			The xPixPosition should span from 0 to max_columns-1, while
		//			the yPixPosition from 0 to max_rows-1.
		//		
		//			From the OOP point of view this function should be .
		//			However, to avoid run-time panalty it is not .
		//
		double GetPixel( Dimension xPixPosition, Dimension yPixPosition ) const
		{
			return ( this->*fGetPixel_FunPtr )( xPixPosition, yPixPosition );
		}

	protected:



		// Vice versa...
		bool ComputeMatrixOffset_BackwardCyclic( const DataMatrix_Index & inOffset, TensorElem_IndexVector & outElementPosition, int tensorMode ) const;
		bool ComputeMatrixOffset_ForwardCyclic( const DataMatrix_Index & inOffset, TensorElem_IndexVector & outElementPosition, int tensor_mode ) const;


		///////////////////////////////////////////////////////////
		// This function converts a given matrix index into 
		// a corresponding one if the tensor were in a different
		// mode.
		///////////////////////////////////////////////////////////
		//		
		// INPUT:
		//		
		//		
		// OUTPUT:
		//		
		//		
		// REMARKS:
		//		
		//		
		void ConvertMatrixIndices( const	DataMatrix_Index & sourceTensorMatrixOffset, const int sourceTensorMode, 
											DataMatrix_Index & destTensorMatrixOffset, const int destTensorMode ) const;


	protected:



		T		( TFlatTensorProxyFor ::* fGetPixel_FunPtr )( Dimension xPixPosition, Dimension yPixPosition ) const;
		void	( TFlatTensorProxyFor ::* fSetPixel_FunPtr )( Dimension xPixPosition, Dimension yPixPosition, const double value ) const;

		// ---
		void SetPixel_DiffModes( Dimension xPixPosition, Dimension yPixPosition, const double value ) const
		{
			DataMatrix_Index sourceTensorMatrixOffset( xPixPosition, yPixPosition );
			DataMatrix_Index destTensorMatrixOffset( 0, 0 );

			// In the case of different modes we need to convert indices
			int destTensorMode = fMotherTensor.GetTensorMode();
			ConvertMatrixIndices(	sourceTensorMatrixOffset, fTensorMode /*set for THIS proxy*/, 
									destTensorMatrixOffset, destTensorMode );

			REQUIRE( destTensorMatrixOffset.fCol < fMotherTensor.GetDataMatrix_Col() );
			REQUIRE( destTensorMatrixOffset.fRow < fMotherTensor.GetDataMatrix_Row() );

			fMotherTensor.SetPixel( destTensorMatrixOffset.fCol, destTensorMatrixOffset.fRow, value  );
		}

		// ---
		void SetPixel_TheSameModes( Dimension xPixPosition, Dimension yPixPosition, const double value ) const
		{
			fMotherTensor.SetPixel( xPixPosition, yPixPosition, value  );
		}

		// ---
		double GetPixel_DiffModes( Dimension xPixPosition, Dimension yPixPosition ) const
		{
			DataMatrix_Index sourceTensorMatrixOffset( xPixPosition, yPixPosition );
			DataMatrix_Index destTensorMatrixOffset( 0, 0 );

			// In the case of different modes we need to convert indices
			int destTensorMode = fMotherTensor.GetTensorMode();
			ConvertMatrixIndices(	sourceTensorMatrixOffset, fTensorMode /*set for THIS proxy*/, 
									destTensorMatrixOffset, destTensorMode );

			REQUIRE( destTensorMatrixOffset.fCol < fMotherTensor.GetDataMatrix_Col() );
			REQUIRE( destTensorMatrixOffset.fRow < fMotherTensor.GetDataMatrix_Row() );

			return fMotherTensor.GetPixel( destTensorMatrixOffset.fCol, destTensorMatrixOffset.fRow );
		}

		// ---
		double GetPixel_TheSameModes( Dimension xPixPosition, Dimension yPixPosition ) const
		{
			return fMotherTensor.GetPixel( xPixPosition, yPixPosition );
		}


};


template < typename T >
bool TFlatTensorProxyFor< T >::ComputeMatrixOffset_BackwardCyclic( const DataMatrix_Index & inOffset, TensorElem_IndexVector & outElementPosition, int tensor_mode ) const
{
	REQUIRE( outElementPosition.size() == fIndexVector.size() );

	REQUIRE( tensor_mode < outElementPosition.size() );

	// This is the easiest to compute
	outElementPosition[ tensor_mode ] = inOffset.fRow;

	// Now goes the hard way...
	int c = inOffset.fCol;

	// Go vice versa
	const int kNumOfDimensions = fIndexVector.size();
	for( register int i = 1, n = tensor_mode - 1; i < kNumOfDimensions; ++ i, -- n )
	{											
		// This has to be cyclic
		if( n < 0 )
			n = kNumOfDimensions - 1;

		int ind = fIndexVector[ n ];
		int div = c / ind;
		int rst = c % ind;

		outElementPosition[ n ] = rst;
		c = div;
	}

	REQUIRE( c == 0 );

	return true;
}


template < typename T >
bool TFlatTensorProxyFor< T >::ComputeMatrixOffset_ForwardCyclic( const DataMatrix_Index & inOffset, TensorElem_IndexVector & outElementPosition, int tensor_mode ) const
{
	REQUIRE( outElementPosition.size() == fIndexVector.size() );

	REQUIRE( tensor_mode < outElementPosition.size() );

	// This is the easiest to compute
	outElementPosition[ tensor_mode ] = inOffset.fRow;

	// Now goes the hard way...
	int c = inOffset.col;

	// Go vice versa
	const int kNumOfDimensions = fIndexVector.size();
	for( register int i = 1, n = tensor_mode + 1; i < kNumOfDimensions; ++ i, ++ n )
	{											
		// This has to be cyclic
		if( n >= kNumOfDimensions )
			n = 0;

		int ind = fIndexVector[ n ];
		int div = c / ind;
		int rst = c % ind;

		outElementPosition[ n ] = rst;
		c = div;
	}

	REQUIRE( c == 0 );

	return true;
}



///////////////////////////////////////////////////////////
// This function converts a given matrix index into 
// a corresponding one if the tensor were in a different
// mode.
///////////////////////////////////////////////////////////
//		
// INPUT:
//		
//		
// OUTPUT:
//		
//		
// REMARKS:
//		
//		
template < typename T >
void TFlatTensorProxyFor< T >::ConvertMatrixIndices( const	DataMatrix_Index & sourceTensorMatrixOffset, const int sourceTensorMode, 
													DataMatrix_Index & destTensorMatrixOffset, const int destTensorMode ) const
{
	TensorElem_IndexVector tensorElementPosition( fIndexVector.size(), 0 );

	ComputeMatrixOffset_BackwardCyclic(		sourceTensorMatrixOffset,	tensorElementPosition,		sourceTensorMode );
//	DEBUG_PrintVector( "elemIndices.txt", tensorElementPosition );
	fMotherTensor.ComputeElementOffset_BackwardCyclic(	tensorElementPosition,		destTensorMatrixOffset,		destTensorMode );
}


// ======================================================================================


typedef TFlatTensorFor< double >		TFlatTensor;
typedef TFlatTensorProxyFor< double >	TFlatTensorProxy;


// ======================================================================================



#endif BoxySoft_Tensor




